with Errorout; use Errorout;
with Iirs_Utils; use Iirs_Utils;
with Types; use Types;
with Name_Table;
with Sem_Expr;
with Std_Names;
with GNAT.Table;
with Libraries;

package body Canon is
   --  Canonicalize a list of declarations.  LIST can be null.
   --  STMTS must be the concurrent statements list for LIST, or null_iir_list.
   procedure Canon_Declarations (Top : Iir_Design_Unit;
                                 List : Iir_List;
                                 Stmts : Iir_List);
   procedure Canon_Declaration (Top : Iir_Design_Unit;
                                Decl : Iir;
                                Stmts : Iir_List;
                                List : Iir_List);

   -- Compute the sensivity list of EXPR and add it to SENSIVITY_LIST.
   -- If IS_TARGET is true, the longuest static prefix of the signal name
   -- is not added to the sensitivity list, but other static prefix (such
   -- as indexes of an indexed name) are added.
   procedure Canon_Extract_Sensitivity
     (Expr: Iir; Sensitivity_List: Iir; Is_Target: Boolean);

   --  Canonicalize an association list.
   --  If ASSOCIATION_LIST is not null, then it is re-ordored and returned.
   --  If ASSOCIATION_LIST is null then:
   --    if INTERFACE_LIST is null then returns null.
   --    if INTERFACE_LIST is not null, a default list is created.
   function Canon_Association_List
     (Interface_List: Iir_List;
      Association_List: Iir_Association_List)
     return Iir_Association_List;

   procedure Canon_Block_Configuration (Top : Iir_Design_Unit;
                                        Conf : Iir_Block_Configuration);

   function Is_Signal_Object (Decl: Iir) return Boolean is
      Adecl: Iir;
   begin
      Adecl := Get_Base_Name (Decl);
      case Get_Kind (Adecl) is
         when Iir_Kind_Variable_Declaration
           | Iir_Kind_Variable_Interface_Declaration
           | Iir_Kind_Constant_Declaration
           | Iir_Kind_Constant_Interface_Declaration =>
            return False;
         when Iir_Kind_Signal_Declaration
           | Iir_Kind_Signal_Interface_Declaration =>
            return True;
         when others =>
            Error_Kind ("is_signal_object", Adecl);
      end case;
   end Is_Signal_Object;

   --  Extract the entity from ASPECT.
   --  Note: if ASPECT is a component declaration, returns ASPECT.
   function Get_Entity_From_Entity_Aspect (Aspect : Iir) return Iir
   is
      Inst : Iir;
   begin
      case Get_Kind (Aspect) is
         when Iir_Kind_Component_Declaration =>
            return Aspect;
         when Iir_Kind_Entity_Aspect_Entity =>
            return Get_Library_Unit (Get_Entity (Aspect));
         when Iir_Kind_Entity_Aspect_Configuration =>
            Inst := Get_Library_Unit (Get_Configuration (Aspect));
            return Get_Library_Unit (Get_Entity (Inst));
         when Iir_Kind_Entity_Aspect_Open =>
            return Null_Iir;
         when others =>
            Error_Kind ("get_entity_from_entity_aspect", Aspect);
      end case;
   end Get_Entity_From_Entity_Aspect;

   procedure Canon_Extract_Sensitivity_Aggregate
     (Aggr : Iir;
      Sensitivity_List : Iir_List;
      Is_Target : Boolean;
      Aggr_Type : Iir;
      Dim : Natural)
   is
      Assoc_List : Iir_List;
      El : Iir;
   begin
      Assoc_List := Get_Association_Choices_List (Aggr);
      if Get_Nbr_Elements (Get_Index_List (Aggr_Type)) = Dim then
         for I in Natural loop
            El := Get_Nth_Element (Assoc_List, I);
            exit when El = Null_Iir;
            Canon_Extract_Sensitivity
              (Get_Associated (El), Sensitivity_List, Is_Target);
         end loop;
      else
         Canon_Extract_Sensitivity_Aggregate
           (Get_Associated (El), Sensitivity_List, Is_Target, Aggr_Type,
            Dim + 1);
      end if;
   end Canon_Extract_Sensitivity_Aggregate;

   procedure Canon_Extract_Sensitivity
     (Expr: Iir; Sensitivity_List: Iir; Is_Target: Boolean)
   is
      El : Iir;
      List: Iir_List;
   begin
      if Get_Expr_Staticness (Expr) /= None then
         return;
      end if;

      case Get_Kind (Expr) is
         when Iir_Kind_Slice_Name =>
            if not Is_Target and then
              Get_Name_Staticness (Expr) >= Globally
            then
               if Is_Signal_Object (Expr) then
                  Add_Element (Sensitivity_List, Expr);
               end if;
            else
               Canon_Extract_Sensitivity (Get_Prefix (Expr),
                                          Sensitivity_List,
                                          Is_Target);
               Canon_Extract_Sensitivity (Get_Suffix (Expr),
                                          Sensitivity_List, False);
            end if;

         when Iir_Kind_Selected_Element =>
            if not Is_Target and then
              Get_Name_Staticness (Expr) >= Globally
            then
               if Is_Signal_Object (Expr) then
                  Add_Element (Sensitivity_List, Expr);
               end if;
            else
               Canon_Extract_Sensitivity (Get_Prefix (Expr),
                                          Sensitivity_List,
                                          Is_Target);
            end if;

         when Iir_Kind_Indexed_Name =>
            if not Is_Target
              and then Get_Name_Staticness (Expr) >= Globally
            then
               if Is_Signal_Object (Expr) then
                  Add_Element (Sensitivity_List, Expr);
               end if;
            else
               Canon_Extract_Sensitivity (Get_Prefix (Expr),
                                          Sensitivity_List,
                                          Is_Target);
               List := Get_Index_List (Expr);
               for I in Natural loop
                  El := Get_Nth_Element (List, I);
                  exit when El = Null_Iir;
                  Canon_Extract_Sensitivity (El, Sensitivity_List, False);
               end loop;
            end if;

         when Iir_Kind_Function_Call =>
            List := Get_Parameter_Association_List (Expr);
            if List /= Null_Iir_List then
               for I in Natural loop
                  El := Get_Nth_Element (List, I);
                  exit when El = Null_Iir;
                  case Get_Kind (El) is
                     when Iir_Kind_Association_Element_By_Expression =>
                        Canon_Extract_Sensitivity
                          (Get_Actual (El), Sensitivity_List, False);
                     when Iir_Kind_Association_Element_Open =>
                        null;
                     when others =>
                        Error_Kind ("canon_extract_sensitivity(call)", El);
                  end case;
               end loop;
            end if;

         when Iir_Kind_Qualified_Expression
           | Iir_Kind_Type_Conversion
           | Iir_Kind_Allocator_By_Expression =>
            Canon_Extract_Sensitivity
              (Get_Expression (Expr), Sensitivity_List, False);

         when Iir_Kind_Allocator_By_Subtype =>
            null;

         when Iir_Kinds_Monadic_Operator =>
            Canon_Extract_Sensitivity
              (Get_Operand (Expr), Sensitivity_List, False);
         when Iir_Kinds_Dyadic_Operator =>
            Canon_Extract_Sensitivity
              (Get_Left (Expr), Sensitivity_List, False);
            Canon_Extract_Sensitivity
              (Get_Right (Expr), Sensitivity_List, False);

         when Iir_Kind_Range_Expression =>
            Canon_Extract_Sensitivity
              (Get_Left_Limit (Expr), Sensitivity_List, False);
            Canon_Extract_Sensitivity
              (Get_Right_Limit (Expr), Sensitivity_List, False);

         when Iir_Kinds_Type_Attribute =>
            null;
         when Iir_Kind_Event_Attribute =>
            --  LRM 8.1
            --  An attribute name: [...], otherwise, apply this rule to the
            --  prefix of the attribute name.
            Canon_Extract_Sensitivity
              (Get_Prefix (Expr), Sensitivity_List, False);

         when Iir_Kind_Stable_Attribute =>
            declare
               Prefix, Static_Prefix: Iir;
               Attr: Iir;
            begin
               Prefix := Get_Prefix (Expr);
               Static_Prefix := Get_Longuest_Static_Prefix (Prefix);
               Canon_Extract_Sensitivity
                 (Get_Prefix (Expr), Sensitivity_List, True);
               if Static_Prefix = Prefix then
                  Add_Element (Sensitivity_List, Expr);
               else
                  Attr := Create_Iir (Get_Kind (Expr));
                  Location_Copy (Attr, Expr);
                  Set_Prefix (Attr, Static_Prefix);
                  Add_Element (Sensitivity_List, Attr);
               end if;
            end;

         when Iir_Kind_Last_Value_Attribute =>
            null;

         when Iir_Kind_Signal_Interface_Declaration
           | Iir_Kind_Signal_Declaration
           | Iir_Kind_Guard_Signal_Declaration =>
            if not Is_Target then
               Add_Element (Sensitivity_List, Expr);
            end if;

         when Iir_Kind_Alias_Declaration =>
            Canon_Extract_Sensitivity
              (Get_Name (Expr), Sensitivity_List, Is_Target);

         when Iir_Kind_Constant_Declaration
           | Iir_Kind_Constant_Interface_Declaration
           | Iir_Kind_Iterator_Declaration
           | Iir_Kind_Variable_Declaration
           | Iir_Kind_Variable_Interface_Declaration =>
            null;

         when Iir_Kind_Left_Array_Attribute
           | Iir_Kind_Right_Array_Attribute
           | Iir_Kind_Length_Array_Attribute
           | Iir_Kind_Low_Array_Attribute
           | Iir_Kind_High_Array_Attribute =>
            null;
            --Canon_Extract_Sensitivity
            --  (Get_Prefix (Expr), Sensitivity_List, Is_Target);

         when Iir_Kind_Aggregate =>
            declare
               Aggr_Type : Iir;
            begin
               Aggr_Type := Get_Base_Type (Get_Type (Expr));
               case Get_Kind (Aggr_Type) is
                  when Iir_Kind_Array_Type_Definition =>
                     Canon_Extract_Sensitivity_Aggregate
                       (Expr, Sensitivity_List, Is_Target, Aggr_Type, 1);
                  when Iir_Kind_Record_Type_Definition =>
                     List := Get_Association_Choices_List (Expr);
                     for I in Natural loop
                        El := Get_Nth_Element (List, I);
                        exit when El = Null_Iir;
                        Canon_Extract_Sensitivity
                          (Get_Associated (El), Sensitivity_List, Is_Target);
                     end loop;
                  when others =>
                     Error_Kind ("canon_extract_sensitivity(aggr)", Aggr_Type);
               end case;
            end;

         when others =>
            Error_Kind ("canon_extract_sensitivity", Expr);
      end case;
   end Canon_Extract_Sensitivity;

   function Make_Aggregate (Array_Type : Iir_Array_Type_Definition; El : Iir)
     return Iir_Aggregate
   is
      Res : Iir_Aggregate;
      Choice : Iir;
   begin
      Res := Create_Iir (Iir_Kind_Aggregate);
      Location_Copy (Res, El);
      Choice := Create_Iir (Iir_Kind_Association_Choice_By_None);
      Set_Associated (Choice, El);
      Append_Element (Get_Association_Choices_List (Res), Choice);

      --  will call sem_aggregate
      return Sem_Expr.Sem_Expression (Res, Array_Type);
   end Make_Aggregate;

   procedure Canon_Concatenation_Operator (Expr : Iir)
   is
      Array_Type : Iir_Array_Type_Definition;
      El_Type : Iir;
      Left, Right : Iir;
      Func_List : Iir_Implicit_Functions_List;
      Func : Iir_Implicit_Function_Declaration;
   begin
      Array_Type := Get_Type (Expr);
      El_Type := Get_Base_Type (Get_Element_Subtype (Array_Type));
      Left := Get_Left (Expr);
      if Get_Type (Left) = El_Type then
         Set_Left (Expr, Make_Aggregate (Array_Type, Left));
      end if;
      Right := Get_Right (Expr);
      if Get_Type (Right) = El_Type then
         Set_Right (Expr, Make_Aggregate (Array_Type, Right));
      end if;

      --  FIXME: must convert the implementation.
      --  Use implicit declaration list from the array_type ?
      Func_List := Get_Implicit_Functions_List
        (Get_Type_Declarator (Array_Type));
      for I in Natural loop
         Func := Get_Nth_Element (Func_List, I);
         if Get_Implicit_Definition (Func)
           = Iir_Predefined_Array_Array_Concat
         then
            Set_Implementation (Expr, Func);
            exit;
         end if;
      end loop;
   end Canon_Concatenation_Operator;

   -- canon on expressions, mainly for function calls.
   procedure Canon_Expression (Expr: Iir)
   is
      El : Iir;
      List: Iir;
   begin
      if Expr = Null_Iir then
         return;
      end if;
      case Get_Kind (Expr) is
         when Iir_Kind_Range_Expression =>
            Canon_Expression (Get_Left_Limit (Expr));
            Canon_Expression (Get_Right_Limit (Expr));

         when Iir_Kind_Slice_Name =>
            declare
               Suffix : Iir;
            begin
               Suffix := Get_Suffix (Expr);
               if Get_Kind (Suffix) not in Iir_Kinds_Discrete_Type_Definition
               then
                  Canon_Expression (Suffix);
               end if;
               Canon_Expression (Get_Prefix (Expr));
            end;

         when Iir_Kind_Indexed_Name =>
            Canon_Expression (Get_Prefix (Expr));
            List := Get_Index_List (Expr);
            for I in Natural loop
               El := Get_Nth_Element (List, I);
               exit when El = Null_Iir;
               Canon_Expression (El);
            end loop;

--          when Iir_Kind_Selected_Name =>
--             -- Use this order to allow tail recursion optimisation.
--             Canon_Expression (Get_Suffix (Expr));
--             Canon_Expression (Get_Prefix (Expr));
         when Iir_Kind_Selected_Element =>
            Canon_Expression (Get_Prefix (Expr));
         when Iir_Kind_Dereference
           | Iir_Kind_Implicit_Dereference =>
            Canon_Expression (Get_Prefix (Expr));

         when Iir_Kinds_Monadic_Operator =>
            Canon_Expression (Get_Operand (Expr));
         when Iir_Kinds_Dyadic_Operator =>
            Canon_Expression (Get_Left (Expr));
            Canon_Expression (Get_Right (Expr));
            if Get_Kind (Expr) = Iir_Kind_Concatenation_Operator
              and then Canon_Concatenation
              and then Get_Kind (Get_Implementation (Expr)) =
              Iir_Kind_Implicit_Function_Declaration
            then
               Canon_Concatenation_Operator (Expr);
            end if;

         when Iir_Kind_Function_Call =>
            declare
               Imp : Iir;
               Assoc_List : Iir_Association_List;
            begin
               Imp := Get_Implementation (Expr);
               if Get_Kind (Imp) /= Iir_Kind_Implicit_Function_Declaration then
                  Assoc_List := Get_Parameter_Association_List (Expr);
                  Assoc_List := Canon_Association_List
                     (Get_Interface_Declaration (Imp), Assoc_List);
                  Set_Parameter_Association_List (Expr, Assoc_List);
               else
                  -- FIXME:
                  -- should canon concatenation.
                  null;
               end if;
            end;
         when Iir_Kind_Type_Conversion
           | Iir_Kind_Qualified_Expression =>
            Canon_Expression (Get_Expression (Expr));
         when Iir_Kind_Aggregate =>
            --  FIXME
            null;
         when Iir_Kind_Allocator_By_Expression =>
            Canon_Expression (Get_Expression (Expr));
         when Iir_Kind_Allocator_By_Subtype =>
            null;

         when Iir_Kinds_Literal
           | Iir_Kind_Simple_Aggregate
           | Iir_Kind_Physical_Unit =>
            null;

         when Iir_Kinds_Array_Attribute =>
            -- No need to canon parameter, since it is a locally static
            -- expression.
            declare
               Prefix : Iir;
            begin
               Prefix := Get_Prefix (Expr);
               case Get_Kind (Prefix) is
                  when Iir_Kind_Type_Declaration
                    | Iir_Kind_Subtype_Declaration =>
                     null;
                  when others =>
                     Canon_Expression (Prefix);
               end case;
            end;

         when Iir_Kinds_Type_Attribute =>
            null;
         when Iir_Kind_Stable_Attribute
           | Iir_Kind_Quiet_Attribute
           | Iir_Kind_Delayed_Attribute =>
            --  FIXME: add the default parameter ?
            Canon_Expression (Get_Prefix (Expr));
         when Iir_Kind_Event_Attribute
           | Iir_Kind_Last_Value_Attribute
           | Iir_Kind_Active_Attribute
           | Iir_Kind_Last_Event_Attribute
           | Iir_Kind_Last_Active_Attribute =>
            Canon_Expression (Get_Prefix (Expr));

         when Iir_Kinds_Scalar_Type_Attribute
           | Iir_Kind_Image_Attribute
           | Iir_Kind_Value_Attribute =>
            Canon_Expression (Get_Parameter (Expr));

         when Iir_Kind_Signal_Interface_Declaration
           | Iir_Kind_Signal_Declaration
           | Iir_Kind_Guard_Signal_Declaration
           | Iir_Kind_Constant_Declaration
           | Iir_Kind_Constant_Interface_Declaration
           | Iir_Kind_Iterator_Declaration
           | Iir_Kind_Variable_Declaration
           | Iir_Kind_Variable_Interface_Declaration
           | Iir_Kind_File_Declaration
           | Iir_Kind_File_Interface_Declaration
           | Iir_Kind_Alias_Declaration =>
            null;

         when Iir_Kind_Enumeration_Literal =>
            null;

         when Iir_Kind_Element_Declaration =>
            null;

         when Iir_Kind_Attribute_Value =>
            null;

         when others =>
            Error_Kind ("canon_expression", Expr);
            null;
      end case;
   end Canon_Expression;

   procedure Canon_Waveform_List
     (Waveform: Iir_Waveform_List; Sensitivity_List: Iir_List)
   is
      We: Iir_Waveform_Element;
   begin
      for I in Natural loop
         We := Get_Nth_Element (Waveform, I);
         exit when We = Null_Iir;
         if Sensitivity_List /= Null_Iir_List then
            Canon_Extract_Sensitivity
              (Get_We_Value (We), Sensitivity_List, False);
         end if;
         Canon_Expression (Get_We_Value (We));
         if Get_Time (We) /= Null_Iir then
            Canon_Expression (Get_Time (We));
         end if;
      end loop;
   end Canon_Waveform_List;

   -- Names associations by position,
   -- reorder associations by name,
   -- create omitted association,
   function Canon_Association_List
     (Interface_List: Iir_List; Association_List: Iir_Association_List)
     return Iir_Association_List
   is
      -- The canon list of association.
      N_Assoc: Iir_Association_List;
      El, Assoc_El: Iir;

      Found : Boolean;
   begin
      --  No argument, so return now.
      if Interface_List = Null_Iir_List then
         if Association_List /= Null_Iir_List then
            raise Internal_Error;
         end if;
         return Null_Iir_List;
      end if;

      N_Assoc := Create_Iir (Iir_Kind_Association_List);

      -- Reorder list and add missing associations.
      for I in 0 .. Get_Nbr_Elements (Interface_List) - 1 loop
         El := Get_Nth_Element (Interface_List, I);

         --  Search associations with EL.
         if Association_List /= Null_Iir_List then
            Found := False;
            for J in Natural loop
               Assoc_El := Get_Nth_Element (Association_List, J);
               exit when Assoc_El = Null_Iir;
               if Get_Associated_Formal (Assoc_El) = El then
                  Append_Element (N_Assoc, Assoc_El);
                  case Get_Kind (Assoc_El) is
                     when Iir_Kind_Association_Element_Open =>
                        goto Done;
                     when Iir_Kind_Association_Element_By_Expression =>
                        if Get_Whole_Association_Flag (Assoc_El) then
                           goto Done;
                        end if;
                     when Iir_Kind_Association_Element_By_Individual =>
                        Found := True;
                     when others =>
                        Error_Kind ("canon_association_list", Assoc_El);
                  end case;
               end if;
            end loop;
            if Found then
               goto Done;
            end if;
         end if;

         -- No association, use default expr.
         Assoc_El := Create_Iir (Iir_Kind_Association_Element_Open);
         --  FIXME: association_list can be null_iir_list!
         --Location_Copy (Assoc_El, Association_List);
         Set_Formal (Assoc_El, El);
         Set_Associated_Formal (Assoc_El, El);
         Append_Element (N_Assoc, Assoc_El);
         << Done >> null;
      end loop;

      for I in 0 .. Get_Nbr_Elements (Interface_List) - 1 loop
         El := Get_Nth_Element (N_Assoc, I);
         if Get_Kind (El) = Iir_Kind_Association_Element_By_Expression then
            Canon_Expression (Get_Actual (El));
         end if;
      end loop;
      Free_Old_Iir (Association_List);
      return N_Assoc;
   end Canon_Association_List;

   --  Create a default association list for INTERFACE_LIST.
   --  The default is a list of interfaces associated with open.
   function Canon_Default_Association_List (Interface_List : Iir_List)
     return Iir_Association_List
   is
      Res : Iir_Association_List;
      Assoc, El : Iir;
   begin
      if Interface_List = Null_Iir_List then
         return Null_Iir_List;
      end if;
      Res := Create_Iir (Iir_Kind_Association_List);
      for I in Natural loop
         El := Get_Nth_Element (Interface_List, I);
         exit when El = Null_Iir;
         Assoc := Create_Iir (Iir_Kind_Association_Element_Open);
         Set_Formal (Assoc, El);
         Set_Associated_Formal (Assoc, El);
         Append_Element (Res, Assoc);
      end loop;
      return Res;
   end Canon_Default_Association_List;

   function Canon_Default_Map_Association_List
     (Formal_List, Actual_List : Iir_List; Loc : Location_Type)
     return Iir_Association_List
   is
      Res : Iir_Association_List;
      Formal, Actual : Iir;
      Assoc : Iir;
      Nbr_Assoc : Natural;
   begin
      --  formal is the entity port/generic.
      if Formal_List = Null_Iir_List then
         if Actual_List /= Null_Iir_List then
            raise Internal_Error;
         end if;
         return Null_Iir_List;
      end if;

      Res := Create_Iir (Iir_Kind_Association_List);
      Set_Location (Res, Loc);
      Nbr_Assoc := 0;
      for I in Natural loop
         Formal := Get_Nth_Element (Formal_List, I);
         exit when Formal = Null_Iir;
         Actual := Find_Name_In_List (Actual_List, Get_Identifier (Formal));
         if Actual /= Null_Iir then
            Assoc := Create_Iir (Iir_Kind_Association_Element_By_Expression);
            Set_Whole_Association_Flag (Assoc, True);
            Set_Actual (Assoc, Actual);
            Nbr_Assoc := Nbr_Assoc + 1;
         else
            Assoc := Create_Iir (Iir_Kind_Association_Element_Open);
         end if;
         Set_Location (Assoc, Loc);
         Set_Formal (Assoc, Formal);
         Set_Associated_Formal (Assoc, Formal);
         Append_Element (Res, Assoc);
      end loop;
      if Nbr_Assoc /= Get_Nbr_Elements (Actual_List) then
         --  There is non-associated actuals.
         raise Internal_Error;
      end if;
      return Res;
   end Canon_Default_Map_Association_List;

   --  Inner loop if any; used to canonicalize exit/next statement.
   Cur_Loop : Iir;

   procedure Canon_Sequential_Stmts (List: Iir_List) is
      Stmt: Iir;
      Expr: Iir;
      Prev_Loop : Iir;
      Label : Iir;
   begin
      if List = Null_Iir_List then
         return;
      end if;
      for I in Natural loop
         Stmt := Get_Nth_Element (List, I);
         exit when Stmt = Null_Iir;
         case Get_Kind (Stmt) is
            when Iir_Kind_If_Statement =>
               declare
                  Cond: Iir;
                  Clause: Iir := Stmt;
               begin
                  while Clause /= Null_Iir loop
                     Cond := Get_Condition (Clause);
                     if Cond /= Null_Iir then
                        Canon_Expression (Cond);
                     end if;
                     Canon_Sequential_Stmts
                       (Get_Then_Sequence_List (Clause));
                     Clause := Get_Else_Clause (Clause);
                  end loop;
               end;

            when Iir_Kind_Signal_Assignment_Statement =>
               Canon_Expression (Get_Target (Stmt));
               Canon_Waveform_List (Get_Waveform_List (Stmt), null);

            when Iir_Kind_Variable_Assignment_Statement =>
               Canon_Expression (Get_Target (Stmt));
               Canon_Expression (Get_Expression (Stmt));

            when Iir_Kind_Wait_Statement =>
               declare
                  Expr: Iir;
                  List: Iir_List;
               begin
                  Expr := Get_Timeout_Clause (Stmt);
                  if Expr /= Null_Iir then
                     Canon_Expression (Expr);
                  end if;
                  Expr := Get_Condition_Clause (Stmt);
                  if Expr /= Null_Iir then
                     Canon_Expression (Expr);
                  end if;
                  List := Get_Sensitivity_List (Stmt);
                  if List = Null_Iir_List and then Expr /= Null_Iir then
                     List := Create_Iir (Iir_Kind_Designator_List);
                     Canon_Extract_Sensitivity (Expr, List, False);
                     Set_Sensitivity_List (Stmt, List);
                  end if;
               end;

            when Iir_Kind_Case_Statement =>
               Canon_Expression (Get_Expression (Stmt));
               declare
                  Choices_List: Iir_Association_Choices_List;
                  Choice: Iir;
               begin
                  Choices_List := Get_Association_Choices_List (Stmt);
                  for I in Natural loop
                     Choice := Get_Nth_Element (Choices_List, I);
                     exit when Choice = Null_Iir;
                     -- FIXME: canon choice expr.
                     Canon_Sequential_Stmts (Get_Associated (Choice));
                  end loop;
               end;

            when Iir_Kind_Assertion_Statement
              | Iir_Kind_Report_Statement =>
               if Get_Kind (Stmt) = Iir_Kind_Assertion_Statement then
                  Canon_Expression (Get_Assertion_Condition (Stmt));
               end if;
               Expr := Get_Report_Expression (Stmt);
               if Expr /= Null_Iir then
                  Canon_Expression (Expr);
               end if;
               Expr := Get_Severity_Expression (Stmt);
               if Expr /= Null_Iir then
                  Canon_Expression (Expr);
               end if;

            when Iir_Kind_For_Loop_Statement =>
               -- FIXME: decl.
               Prev_Loop := Cur_Loop;
               Cur_Loop := Stmt;
               Canon_Sequential_Stmts (Get_Statement_List (Stmt));
               Cur_Loop := Prev_Loop;

            when Iir_Kind_While_Loop_Statement =>
               Expr := Get_Condition (Stmt);
               if Expr /= Null_Iir then
                  Canon_Expression (Expr);
               end if;
               Prev_Loop := Cur_Loop;
               Cur_Loop := Stmt;
               Canon_Sequential_Stmts (Get_Statement_List (Stmt));
               Cur_Loop := Prev_Loop;

            when Iir_Kind_Next_Statement
              | Iir_Kind_Exit_Statement =>
               Expr := Get_Condition (Stmt);
               if Expr /= Null_Iir then
                  Canon_Expression (Expr);
               end if;
               Label := Get_Loop (Stmt);
               if Label = Null_Iir then
                  Set_Loop (Stmt, Cur_Loop);
               end if;

            when Iir_Kind_Procedure_Call =>
               declare
                  Assoc_List : Iir_Association_List;
               begin
                  Assoc_List := Get_Parameter_Association_List (Stmt);
                  Assoc_List := Canon_Association_List
                    (Get_Interface_Declaration (Get_Implementation (Stmt)),
                     Assoc_List);
                  Set_Parameter_Association_List (Stmt, Assoc_List);
               end;

            when Iir_Kind_Null_Statement =>
               null;

            when Iir_Kind_Return_Statement =>
               Canon_Expression (Get_Expression (Stmt));

            when others =>
               Error_Kind ("canon_sequential_stmts", Stmt);
         end case;
      end loop;
   end Canon_Sequential_Stmts;

   procedure Add_Driver_For_Signal (Driver_List : Iir_List;
                                    Signal : Iir)
   is
      Choice_List : Iir_List;
      Choice : Iir;
   begin
      if Get_Kind (Signal) = Iir_Kind_Aggregate then
         Choice_List := Get_Association_Choices_List (Signal);
         for I in Natural loop
            Choice := Get_Nth_Element (Choice_List, I);
            exit when Choice = Null_Iir;
            Add_Driver_For_Signal (Driver_List, Get_Associated (Choice));
         end loop;
      else
         Add_Element (Driver_List, Get_Longuest_Static_Prefix (Signal));
      end if;
   end Add_Driver_For_Signal;

   -- Create a statement transform from a concurrent_signal_assignment
   -- statement (either selected or conditional).
   -- waveform transformation is not done.
   -- See LRM93 9.5
   procedure Canon_Concurrent_Signal_Assignment
     (Stmt: in out Iir;
      Proc: in out Iir_Sensitized_Process_Statement;
      List: out Iir_List)
   is
      If_Stmt: Iir;

   begin
      Proc := Create_Iir (Iir_Kind_Sensitized_Process_Statement);
      Location_Copy (Proc, Stmt);

      --  LRM93 9.5
      --  1. If a label appears on the concurrent signal assignment, then the
      --     same label appears on the process statement.
      Set_Label (Proc, Get_Label (Stmt));

      Set_Driver_List (Proc, Create_Iir (Iir_Kind_Driver_List));
      Add_Driver_For_Signal (Get_Driver_List (Proc), Get_Target (Stmt));

      Canon_Extract_Sensitivity
        (Get_Target (Stmt), Get_Sensitivity_List (Proc), True);
      Canon_Expression (Get_Target (Stmt));

      if Get_Guard (Stmt) /= Null_Iir then
         -- LRM93 9.1
         -- If the option guarded appears in the concurrent signal assignment
         -- statement, then the concurrent signal assignment is called a
         -- guarded assignment.
         -- If the concurrent signal assignement statement is a guarded
         -- assignment and the target of the concurrent signal assignment is
         -- a guarded target, then the statement transform is as follow:
         --   if GUARD then signal_transform else disconnect_statements end if;
         -- Otherwise, if the concurrent signal assignment statement is a
         -- guarded assignement, but the target if the concurrent signal
         -- assignment is not a guarded target, the then statement transform
         -- is as follows:
         --  if GUARD then signal_transform end if;
         If_Stmt := Create_Iir (Iir_Kind_If_Statement);
         Append_Element (Get_Process_Statement_List (Proc), If_Stmt);
         Location_Copy (If_Stmt, Stmt);
         Canon_Extract_Sensitivity (Get_Guard (Stmt),
                                    Get_Sensitivity_List (Proc),
                                    False);
         Set_Condition (If_Stmt, Get_Guard (Stmt));
         List := Get_Then_Sequence_List (If_Stmt);

         declare
            Target : Iir;
            Else_Clause : Iir_Elsif;
            Else_List : Iir_Sequential_Statement_List;
            Dis_Stmt : Iir_Signal_Assignment_Statement;
         begin
            Target := Get_Target (Stmt);
            if Get_Guarded_Target_State (Stmt) = True then
               --  The target is a guarded target.
               --  create the disconnection statement.
               Else_Clause := Create_Iir (Iir_Kind_Elsif);
               Location_Copy (Else_Clause, Stmt);
               Set_Else_Clause (If_Stmt, Else_Clause);
               Else_List := Get_Then_Sequence_List (Else_Clause);
               Dis_Stmt := Create_Iir (Iir_Kind_Signal_Assignment_Statement);
               Location_Copy (Dis_Stmt, Stmt);
               Set_Target (Dis_Stmt, Target);
               Append_Element (Else_List, Dis_Stmt);
               --  XX
               Set_Waveform_List (Dis_Stmt, Null_Iir);
            end if;
         end;
      else
         -- LRM93 9.1
         -- Finally, if the concurrent signal assignment statement is not a
         -- guarded assignment, and the traget of the concurrent signal
         -- assignment is not a guarded target,t hen the statement transform
         -- is as follows:
         --    signal_transform
         List := Get_Process_Statement_List (Proc);
      end if;
   end Canon_Concurrent_Signal_Assignment;

   function Canon_Concurrent_Procedure_Call (El : Iir)
     return Iir_Sensitized_Process_Statement
   is
      Proc : Iir_Sensitized_Process_Statement;
      Stmt : Iir_Procedure_Call;
      Assoc_List : Iir_List;
      Assoc : Iir;
      Imp : Iir;
      Driver_List : Iir_Driver_List;
      Interface : Iir;
   begin
      --  FIXME: canon to a process, *not* a sensitized process.
      Proc := Create_Iir (Iir_Kind_Sensitized_Process_Statement);
      Location_Copy (Proc, El);
      Set_Label (Proc, Get_Label (El));
      Set_Attribute_Value_List (Proc, Get_Attribute_Value_List (El));

      Stmt := Create_Iir (Iir_Kind_Procedure_Call);
      Append_Element (Get_Process_Statement_List (Proc), Stmt);
      Location_Copy (Stmt, El);
      Imp := Get_Implementation (El);
      Set_Implementation (Stmt, Imp);
      Assoc_List := Get_Parameter_Association_List (El);
      Assoc_List := Canon_Association_List
        (Get_Interface_Declaration (Imp), Assoc_List);
      Set_Parameter_Association_List (Stmt, Assoc_List);
      Driver_List := Null_Iir_List;
      if Assoc_List /= Null_Iir_List then
         for I in Natural loop
            Assoc := Get_Nth_Element (Assoc_List, I);
            exit when Assoc = Null_Iir;
            case Get_Kind (Assoc) is
               when Iir_Kind_Association_Element_By_Expression =>
                  Canon_Extract_Sensitivity
                    (Get_Actual (Assoc), Get_Sensitivity_List (Proc), False);
                  Interface := Get_Associated_Formal (Assoc);
                  if Get_Kind (Interface)
                    = Iir_Kind_Signal_Interface_Declaration
                    and then Get_Mode (Interface)
                    in Iir_Out_Mode .. Iir_Inout_Mode
                  then
                     if Driver_List = Null_Iir_List then
                        Driver_List := Create_Iir (Iir_Kind_Driver_List);
                        Set_Driver_List (Proc, Driver_List);
                     end if;
                     Add_Element
                       (Driver_List,
                        Get_Longuest_Static_Prefix (Get_Actual (Assoc)));
                  end if;
               when Iir_Kind_Association_Element_Open =>
                  null;
               when others =>
                  raise Internal_Error;
            end case;
         end loop;
      end if;
      return Proc;
   end Canon_Concurrent_Procedure_Call;

   function Canon_Wave_Transform (Orig_Stmt : Iir; Waveform : Iir; Proc : Iir)
     return Iir
   is
      Stmt : Iir;
   begin
      Stmt := Create_Iir (Iir_Kind_Signal_Assignment_Statement);
      Location_Copy (Stmt, Orig_Stmt);
      Set_Target (Stmt, Get_Target (Orig_Stmt));
      Canon_Waveform_List (Waveform, Get_Sensitivity_List (Proc));
      Set_Waveform_List (Stmt, Waveform);
      Set_Delay_Mechanism (Stmt, Get_Delay_Mechanism (Orig_Stmt));
      Set_Reject_Time_Expression
        (Stmt, Get_Reject_Time_Expression (Orig_Stmt));
      return Stmt;
   end Canon_Wave_Transform;

   --  Create signal_transform for a conditional concurrent signal assignment.
   procedure Canon_Conditional_Concurrent_Signal_Assigment
     (Conc_Stmt : Iir; Proc : Iir; Stmt_List : Iir_List)
   is
      List : Iir_List;
      El : Iir;
      Expr : Iir;
      Stmt : Iir;
      Res1 : Iir;
      Last_Res : Iir;
      Wf : Iir;
   begin
      List := Get_Conditional_Waveform_List (Conc_Stmt);
      Stmt := Null_Iir;
      for I in Natural loop
         El := Get_Nth_Element (List, I);
         exit when El = Null_Iir;
         Expr := Get_Condition (El);
         Wf := Canon_Wave_Transform (Conc_Stmt, Get_Waveform_List (El), Proc);
         if Expr = Null_Iir and I = 0 then
            Res1 := Wf;
         else
            if Expr /= Null_Iir then
               Canon_Expression (Expr);
               Canon_Extract_Sensitivity
                 (Expr, Get_Sensitivity_List (Proc), False);
            end if;
            if Stmt = Null_Iir then
               Res1 := Create_Iir (Iir_Kind_If_Statement);
            else
               Res1 := Create_Iir (Iir_Kind_Elsif);
            end if;
            Location_Copy (Res1, El);
            Set_Condition (Res1, Expr);
            Append_Element (Get_Then_Sequence_List (Res1), Wf);
         end if;
         if Stmt = Null_Iir then
            Stmt := Res1;
         else
            Set_Else_Clause (Last_Res, Res1);
         end if;
         Last_Res := Res1;
      end loop;
      Append_Element (Stmt_List, Stmt);
   end Canon_Conditional_Concurrent_Signal_Assigment;

   procedure Canon_Selected_Concurrent_Signal_Assignment
     (Conc_Stmt : Iir; Proc : Iir; Stmt_List : Iir_List)
   is
      Assoc_List : Iir_Association_Choices_List;
      Assoc: Iir;
      Case_Stmt: Iir_Case_Statement;
      Case_Assoc_List: Iir_Association_Choices_List;
      Expr : Iir;
      Stmt : Iir;
      Stmt_List1 : Iir_List;
   begin
      Case_Stmt := Create_Iir (Iir_Kind_Case_Statement);
      Append_Element (Stmt_List, Case_Stmt);
      Location_Copy (Case_Stmt, Conc_Stmt);
      Expr := Get_Expression (Conc_Stmt);
      Canon_Expression (Expr);
      Set_Expression (Case_Stmt, Expr);
      Canon_Extract_Sensitivity
        (Expr, Get_Sensitivity_List (Proc), False);

      Assoc_List := Get_Selected_Waveform_List (Conc_Stmt);
      Case_Assoc_List := Get_Association_Choices_List (Case_Stmt);
      for I in Natural loop
         Assoc := Get_Nth_Element (Assoc_List, I);
         exit when Assoc = Null_Iir;

         Stmt := Canon_Wave_Transform
           (Conc_Stmt, Get_Associated (Assoc), Proc);
         Stmt_List1 := Create_Iir (Iir_Kind_Sequential_Statement_List);
         Append_Element (Stmt_List1, Stmt);
         Set_Associated (Assoc, Stmt_List1);
         Append_Element (Case_Assoc_List, Assoc);
      end loop;
   end Canon_Selected_Concurrent_Signal_Assignment;

   procedure Canon_Concurrent_Stmts (Top : Iir_Design_Unit; List: Iir_List)
   is
      El: Iir;
      Proc: Iir;
      Stmt: Iir;
      Stmt_List: Iir_List;
      Expr: Iir;
   begin
      if List = Null_Iir_List then
         return;
      end if;
      for I in Natural loop
         El := Get_Nth_Element (List, I);
         exit when El = Null_Iir;

         --  Add a label if required.
         if Canon_Add_Labels and then Get_Label (El) = Null_Identifier then
            declare
               Str : String := Natural'Image (I);
            begin
               --  Note: the label starts with a capitalized letter, to avoid
               --  any clash with user's identifiers.
               Str (1) := 'P';
               Set_Label (El, Name_Table.Get_Identifier (Str));
            end;
         end if;

         case Get_Kind (El) is
            when Iir_Kind_Concurrent_Conditional_Signal_Assignment =>
               Canon_Concurrent_Signal_Assignment (El, Proc, Stmt_List);

               Canon_Conditional_Concurrent_Signal_Assigment
                 (El, Proc, Stmt_List);

               Replace_Nth_Element (List, I, Proc);
               Free_Iir (El);

            when Iir_Kind_Concurrent_Selected_Signal_Assignment =>
               Canon_Concurrent_Signal_Assignment (El, Proc, Stmt_List);

               Canon_Selected_Concurrent_Signal_Assignment
                 (El, Proc, Stmt_List);

               Replace_Nth_Element (List, I, Proc);
               Free_Iir (El);

            when Iir_Kind_Concurrent_Assertion_Statement =>
               -- Create a new entry.
               Proc := Create_Iir (Iir_Kind_Sensitized_Process_Statement);
               Location_Copy (Proc, El);
               Set_Label (Proc, Get_Label (El));
               Replace_Nth_Element (List, I, Proc);

               Stmt := Create_Iir (Iir_Kind_Assertion_Statement);
               Append_Element (Get_Process_Statement_List (Proc), Stmt);
               Location_Copy (Stmt, El);

               -- Expand the expression, fill the sensitivity list,
               Canon_Extract_Sensitivity
                 (Get_Assertion_Condition (El),
                  Get_Sensitivity_List (Proc), False);
               Canon_Expression (Get_Assertion_Condition (El));
               Set_Assertion_Condition
                 (Stmt, Get_Assertion_Condition (El));

               Expr := Get_Report_Expression (El);
               if Expr /= Null_Iir then
                  Canon_Expression (Expr);
               end if;
               Set_Report_Expression (Stmt, Expr);

               Expr := Get_Severity_Expression (El);
               if Expr /= Null_Iir then
                  Canon_Expression (Expr);
               end if;
               Set_Severity_Expression (Stmt, Expr);

            when Iir_Kind_Concurrent_Procedure_Call_Statement =>
               Replace_Nth_Element
                 (List, I, Canon_Concurrent_Procedure_Call (El));

            when Iir_Kind_Sensitized_Process_Statement =>
               Canon_Declarations
                 (Top, Get_Declaration_List (El), Null_Iir_List);
               Canon_Sequential_Stmts (Get_Process_Statement_List (El));

            when Iir_Kind_Process_Statement =>
               Canon_Declarations
                 (Top, Get_Declaration_List (El), Null_Iir_List);
               Canon_Sequential_Stmts (Get_Process_Statement_List (El));

            when Iir_Kind_Component_Instantiation_Statement =>
               declare
                  Inst : Iir;
                  Assoc_List : Iir_Association_List;
               begin
                  Inst := Get_Instantiated_Unit (El);
                  Inst := Get_Entity_From_Entity_Aspect (Inst);
                  Assoc_List := Get_Generic_Map_Aspect_List (El);
                  Assoc_List := Canon_Association_List
                    (Get_Generic_List (Inst), Assoc_List);
                  Set_Generic_Map_Aspect_List (El, Assoc_List);

                  Assoc_List := Get_Port_Map_Aspect_List (El);
                  Assoc_List := Canon_Association_List
                    (Get_Port_List (Inst), Assoc_List);
                  Set_Port_Map_Aspect_List (El, Assoc_List);
               end;

            when Iir_Kind_Block_Statement =>
               declare
                  Header : Iir_Block_Header;
                  List : Iir_List;
                  Block_Stmts : Iir_List;
                  Guard : Iir_Guard_Signal_Declaration;
               begin
                  Guard := Get_Guard_Decl (El);
                  if Guard /= Null_Iir then
                     Expr := Get_Guard_Expression (Guard);
                     Set_Guard_Sensitivity_List
                       (Guard, Create_Iir (Iir_Kind_Designator_List));
                     Canon_Extract_Sensitivity
                       (Expr, Get_Guard_Sensitivity_List (Guard), False);
                     Canon_Expression (Expr);
                  end if;
                  Header := Get_Block_Header (El);
                  if Header /= Null_Iir then
                     List := Get_Generic_Map_Aspect_List (Header);
                     if List /= Null_Iir_List then
                        List := Canon_Association_List
                          (Get_Generic_List (Header), List);
                     else
                        List := Canon_Default_Association_List
                          (Get_Generic_List (Header));
                     end if;
                     Set_Generic_Map_Aspect_List (Header, List);
                     List := Get_Port_Map_Aspect_List (Header);
                     if List /= Null_Iir_List then
                        List := Canon_Association_List
                          (Get_Port_List (Header), List);
                        Set_Port_Map_Aspect_List (Header, List);
                     end if;
                  end if;
                  Block_Stmts := Get_Concurrent_Statement_List (El);
                  Canon_Declarations
                    (Top, Get_Declaration_List (El), Block_Stmts);
                  Canon_Concurrent_Stmts (Top, Block_Stmts);
               end;

            when Iir_Kind_Generate_Statement =>
               declare
                  Scheme : Iir;
                  Block_Stmts : Iir_List;
               begin
                  Scheme := Get_Generation_Scheme (El);
                  if Get_Kind (Scheme) = Iir_Kind_Iterator_Declaration then
                     Canon_Declaration (Top, Scheme,
                                        Null_Iir_List, Null_Iir_List);
                  else
                     Canon_Expression (Scheme);
                  end if;
                  Block_Stmts := Get_Concurrent_Statement_List (El);
                  Canon_Declarations
                    (Top, Get_Declaration_List (El), Block_Stmts);
                  Canon_Concurrent_Stmts (Top, Block_Stmts);
               end;

            when others =>
               Error_Kind ("canon_concurrent_stmts", El);
         end case;
      end loop;
   end Canon_Concurrent_Stmts;

   procedure Canon_Binding_Indication
     (Component: Iir; Binding : Iir_Binding_Indication)
   is
      List : Iir_Association_List;
   begin
      if Binding = Null_Iir then
         return;
      end if;
      List := Get_Generic_Map_Aspect_List (Binding);
      List := Canon_Association_List (Get_Generic_List (Component), List);
      Set_Generic_Map_Aspect_List (Binding, List);
      List := Get_Port_Map_Aspect_List (Binding);
      List := Canon_Association_List (Get_Port_List (Component), List);
      Set_Port_Map_Aspect_List (Binding, List);
   end Canon_Binding_Indication;

   procedure Add_Binding_Indication_Dependence (Top : Iir_Design_Unit;
                                                Binding : Iir)
   is
      Aspect : Iir;
      Unit : Iir;
   begin
      if Binding = Null_Iir then
         return;
      end if;
      Aspect := Get_Entity_Aspect (Binding);
      if Aspect = Null_Iir then
         return;
      end if;
      case Get_Kind (Aspect) is
         when Iir_Kind_Entity_Aspect_Entity =>
            if Get_Architecture (Aspect) /= Null_Iir then
               Unit := Aspect;
            else
               Unit := Get_Entity (Aspect);
            end if;
         when Iir_Kind_Entity_Aspect_Configuration =>
            Unit := Get_Configuration (Aspect);
         when Iir_Kind_Entity_Aspect_Open =>
            Unit := Null_Iir;
         when others =>
            Error_Kind ("add_binding_indication_dependence", Aspect);
      end case;
      if Unit /= Null_Iir then
         Add_Dependence (Top, Unit);
      end if;
   end Add_Binding_Indication_Dependence;

   procedure Canon_Component_Configuration (Top : Iir_Design_Unit; Cfg : Iir)
   is
      Bind : Iir;
      Instances : Iir;
      Entity_Aspect : Iir;
      Block : Iir_Block_Configuration;
      Map_List : Iir_List;
      Entity : Iir;
   begin
      Bind := Get_Binding_Indication (Cfg);
      if Bind = Null_Iir then
         --  Add a default binding indication
         --  Extract a component instantiation
         Instances := Get_Instantiation_List (Cfg);
         case Get_Kind (Instances) is
            when Iir_Kind_Designator_List =>
               Bind := Get_Default_Binding_Indication
                 (Get_First_Element (Instances));
            when others =>
               --  designator_all and designator_others must have been replaced
               --  by a list during canon.
               raise Internal_Error;
         end case;
         if Bind = Null_Iir then
            --  Component is not bound.
            return;
         end if;
         Set_Binding_Indication (Cfg, Bind);
         Add_Binding_Indication_Dependence (Top, Bind);
         return;
      else
         Entity_Aspect := Get_Entity_Aspect (Bind);
         if Entity_Aspect = Null_Iir then
            Entity_Aspect := Get_Default_Entity_Aspect (Bind);
            Set_Entity_Aspect (Bind, Entity_Aspect);
         end if;
         if Entity_Aspect /= Null_Iir then
            Add_Binding_Indication_Dependence (Top, Bind);
            Entity := Get_Entity_From_Entity_Aspect (Entity_Aspect);
            Map_List := Get_Generic_Map_Aspect_List (Bind);
            if Map_List = Null_Iir_List then
               Map_List := Get_Default_Generic_Map_Aspect_List (Bind);
            else
               Map_List := Canon_Association_List (Get_Generic_List (Entity),
                                                   Map_List);
            end if;
            Set_Generic_Map_Aspect_List (Bind, Map_List);

            if Get_Port_Map_Aspect_List (Bind) = Null_Iir_List then
               Set_Port_Map_Aspect_List
                 (Bind, Get_Default_Port_Map_Aspect_List (Bind));
            end if;
            if Get_Kind (Cfg) = Iir_Kind_Component_Configuration then
               Block := Get_Block_Configuration (Cfg);
               if Block /= Null_Iir then
                  Canon_Block_Configuration (Top, Block);
               end if;
            end if;
         end if;
      end if;
   end Canon_Component_Configuration;

   procedure Canon_Component_Specification (Conf : Iir; Stmts : Iir_List)
   is
      Spec : Iir;
      List : Iir_Designator_List;
      El : Iir;
      Comp : Iir;
      Spec_Kind : Iir_Kind;
      Comp_Conf : Iir;
   begin
      Spec := Get_Instantiation_List (Conf);
      Spec_Kind := Get_Kind (Spec);

      if Spec_Kind = Iir_Kind_Designator_List then
         --  Already as a designator list.
         for I in Natural loop
            El := Get_Nth_Element (Spec, I);
            exit when El = Null_Iir;
            Comp_Conf := Get_Component_Configuration (El);
            if Comp_Conf /= Null_Iir and then Comp_Conf /= Conf then
               --  FIXME: handle merge.
               raise Internal_Error;
            else
               Set_Component_Configuration (El, Conf);
            end if;
         end loop;
         return;
      end if;

      Comp := Get_Component_Name (Conf);
      List := Create_Iir (Iir_Kind_Designator_List);
      Location_Copy (List, Spec);
      if Stmts /= Null_Iir_List then
         for I in Natural loop
            El := Get_Nth_Element (Stmts, I);
            exit when El = Null_Iir;
            if Get_Kind (El) = Iir_Kind_Component_Instantiation_Statement
              and then Get_Instantiated_Unit (El) = Comp
            then
               Comp_Conf := Get_Component_Configuration (El);
               if Comp_Conf = Null_Iir then
                  --  The component is not yet configured.
                  Append_Element (List, El);
                  Set_Component_Configuration (El, Conf);
               else
                  --  The component is already configured.
                  --  Handle incremental configuration.
                  if Get_Kind (Comp_Conf)
                    = Iir_Kind_Configuration_Specification
                    and then Spec_Kind = Iir_Kind_Designator_By_All
                  then
                     --  FIXME: handle incremental configuration.
                     raise Internal_Error;
                  end if;
                  case Spec_Kind is
                     when Iir_Kind_Designator_By_All =>
                        --  Several component configuration for an instance.
                        --  Must have been caught by sem.
                        raise Internal_Error;
                     when Iir_Kind_Designator_By_Others =>
                        null;
                     when others =>
                        raise Internal_Error;
                  end case;
               end if;
            end if;
         end loop;
      end if;
      Free_Iir (Spec);
      Set_Instantiation_List (Conf, Null_Iir);
      Set_Instantiation_List (Conf, List);
   end Canon_Component_Specification;

   procedure Canon_Disconnection_Specification
     (Dis : Iir_Disconnection_Specification; List : Iir_List)
   is
      Signal_List : Iir;
      Force : Boolean;
      El : Iir;
      N_List : Iir_Designator_List;
   begin
      Canon_Expression (Get_Expression (Dis));
      Signal_List := Get_Signal_List (Dis);
      case Get_Kind (Signal_List) is
         when Iir_Kind_Designator_List =>
            return;
         when Iir_Kind_Designator_By_Others =>
            Force := False;
         when Iir_Kind_Designator_By_All =>
            Force := True;
         when others =>
            Error_Kind ("canon_disconnection_specification", Signal_List);
      end case;
      N_List := Create_Iir (Iir_Kind_Designator_List);
      Location_Copy (N_List, Signal_List);
      Free_Iir (Signal_List);
      Set_Signal_List (Dis, N_List);
      for I in Natural loop
         El := Get_Nth_Element (List, I);
         exit when El = Null_Iir;
         if Get_Kind (El) = Iir_Kind_Signal_Declaration
           and then Get_Type (El) = Get_Type (Dis)
           and then Get_Signal_Kind (El) /= Iir_No_Signal_Kind
         then
            if not Get_Has_Disconnect_Flag (El) then
               Set_Has_Disconnect_Flag (El, True);
               Append_Element (N_List, El);
            else
               if Force then
                  raise Internal_Error;
               end if;
            end if;
         end if;
      end loop;
   end Canon_Disconnection_Specification;

   procedure Canon_Declaration (Top : Iir_Design_Unit;
                                Decl : Iir;
                                Stmts : Iir_List;
                                List : Iir_List)
   is
   begin
      case Get_Kind (Decl) is
         when Iir_Kind_Procedure_Spec_Body
           | Iir_Kind_Procedure_Body
           | Iir_Kind_Function_Spec_Body
           | Iir_Kind_Function_Body =>
            Canon_Declarations
              (Top, Get_Subprogram_Declaration (Decl), Null_Iir_List);
            Canon_Sequential_Stmts (Get_Subprogram_Statements (Decl));

         when Iir_Kind_Procedure_Specification
           | Iir_Kind_Function_Specification =>
            null;

         when Iir_Kind_Type_Declaration
           | Iir_Kind_Subtype_Declaration =>
            null;

         when Iir_Kind_Variable_Declaration
           | Iir_Kind_Signal_Declaration
           | Iir_Kind_Constant_Declaration =>
            Canon_Expression (Get_Default_Value (Decl));

         when Iir_Kind_Iterator_Declaration =>
            null;

         when Iir_Kind_Alias_Declaration =>
            null;

         when Iir_Kind_File_Declaration =>
            -- FIXME
            null;

         when Iir_Kind_Attribute_Declaration =>
            null;
         when Iir_Kind_Attribute_Specification =>
            Canon_Expression (Get_Expression (Decl));
         when Iir_Kind_Disconnection_Specification =>
            Canon_Disconnection_Specification (Decl, List);

         when Iir_Kind_Group_Template_Declaration =>
            null;
         when Iir_Kind_Group_Declaration =>
            null;

         when Iir_Kind_Use_Clause =>
            null;

         when Iir_Kind_Component_Declaration =>
            null;

         when Iir_Kind_Implicit_Procedure_Declaration =>
            null;

         when Iir_Kind_Configuration_Specification =>
            Canon_Component_Specification (Decl, Stmts);
            Canon_Component_Configuration (Top, Decl);
--             declare
--                List : Iir_List;
--                Binding : Iir_Binding_Indication;
--                Component : Iir_Component_Declaration;
--                Aspect : Iir;
--                Entity : Iir;
--             begin
--                Binding := Get_Binding_Indication (Decl);
--                Component := Get_Component_Name (Decl);
--                Aspect := Get_Entity_Aspect (Binding);
--                case Get_Kind (Aspect) is
--                   when Iir_Kind_Entity_Aspect_Entity =>
--                      Entity := Get_Entity (Aspect);
--                   when others =>
--                      Error_Kind ("configuration_specification", Aspect);
--                end case;
--                Entity := Get_Library_Unit (Entity);
--                List := Get_Generic_Map_Aspect_List (Binding);
--                if List = Null_Iir_List then
--                   Set_Generic_Map_Aspect_List
--                     (Binding,
--                      Canon_Default_Map_Association_List
--                    (Get_Generic_List (Entity), Get_Generic_List (Component),
--                       Get_Location (Decl)));
--                end if;
--                List := Get_Port_Map_Aspect_List (Binding);
--                if List = Null_Iir_List then
--                   Set_Port_Map_Aspect_List
--                     (Binding,
--                      Canon_Default_Map_Association_List
--                      (Get_Port_List (Entity), Get_Port_List (Component),
--                       Get_Location (Decl)));
--                end if;
--             end;

         when Iir_Kind_Implicit_Signal_Declaration =>
            null;
         when others =>
            Error_Kind ("canon_declaration", Decl);
      end case;
   end Canon_Declaration;

   procedure Canon_Declarations (Top : Iir_Design_Unit;
                                 List : Iir_List;
                                 Stmts : Iir_List)
   is
      Decl : Iir;
   begin
      if List = Null_Iir_List then
         return;
      end if;
      if Stmts /= Null_Iir then
         Clear_Instantiation_Configuration (Stmts);
      end if;
      for I in Natural loop
         Decl := Get_Nth_Element (List, I);
         exit when Decl = Null_Iir;
         Canon_Declaration (Top, Decl, Stmts, List);
      end loop;
   end Canon_Declarations;

   procedure Canon_Block_Configuration (Top : Iir_Design_Unit;
                                        Conf : Iir_Block_Configuration)
   is
      El : Iir;
      Spec : Iir;
      Stmts : Iir_List;
      Decls : Iir_List;
      Item_List : Iir_List;
      Blk : Iir;
   begin
      --  Note: the only allowed declarations are use clauses, which are not
      --  canonicalized.

      --  FIXME: handle indexed/sliced name?
      Spec := Get_Block_Specification (Conf);
      Blk := Get_Block_From_Block_Specification (Spec);
      Stmts := Get_Concurrent_Statement_List (Blk);
      Decls := Get_Declaration_List (Blk);
      Item_List := Get_Configuration_Item_List (Conf);
      Clear_Instantiation_Configuration (Stmts);

      --  1) Copy configuration specifications from the declarations.
      --  TODO: merge.
      if Decls /= Null_Iir_List then
         for I in Natural loop
            El := Get_Nth_Element (Decls, I);
            exit when El = Null_Iir;
            if Get_Kind (El) = Iir_Kind_Configuration_Specification then
               Append_Element (Item_List, El);
            end if;
         end loop;
      end if;

      --  2) Replaces component specification with the list of components.
      for I in Natural loop
         El := Get_Nth_Element (Item_List, I);
         exit when El = Null_Iir;
         case Get_Kind (El) is
            when Iir_Kind_Configuration_Specification =>
               --  Already canoncalized during canon of block declarations.
               --  But need to set configuration on instantiations.
               Canon_Component_Specification (El, Stmts);
            when Iir_Kind_Component_Configuration =>
               Canon_Component_Specification (El, Stmts);
            when Iir_Kind_Block_Configuration =>
               Blk := Get_Block_Specification (El);
               case Get_Kind (Blk) is
                  when Iir_Kind_Block_Statement =>
                     Set_Block_Block_Configuration (Blk, El);
                  when Iir_Kind_Indexed_Name
                    | Iir_Kind_Slice_Name =>
                     Blk := Get_Prefix (Blk);
                     Set_Prev_Block_Configuration
                       (El, Get_Generate_Block_Configuration (Blk));
                     Set_Generate_Block_Configuration (Blk, El);
                  when Iir_Kind_Generate_Statement =>
                     Set_Generate_Block_Configuration (Blk, El);
                  when others =>
                     Error_Kind ("canon_block_configuration(0)", Blk);
               end case;
            when others =>
               Error_Kind ("canon_block_configuration(1)", El);
         end case;
      end loop;

      --  3) Add default component configuration for unspecified component
      --     instantiation statements,
      --     Add default block configuration for unconfigured block statements.
      for I in Natural loop
         El := Get_Nth_Element (Stmts, I);
         exit when El = Null_Iir;
         case Get_Kind (El) is
            when Iir_Kind_Component_Instantiation_Statement =>
               if Get_Component_Configuration (El) = Null_Iir then
                  declare
                     Comp : Iir;
                     Res : Iir_Component_Configuration;
                     Designator : Iir_Designator_List;
                  begin
                     Comp := Get_Instantiated_Unit (El);
                     if Get_Kind (Comp) = Iir_Kind_Component_Declaration then
                        --  Create a component configuration.
                        --  FIXME: should merge all these default configuration
                        --    of the same component.
                        Res := Create_Iir (Iir_Kind_Component_Configuration);
                        Location_Copy (Res, El);
                        Set_Parent (Res, Conf);
                        Set_Component_Name (Res, Comp);
                        Designator := Create_Iir (Iir_Kind_Designator_List);
                        Append_Element (Designator, El);
                        Set_Instantiation_List (Res, Designator);
                        Append_Element (Item_List, Res);
                     end if;
                  end;
               end if;
            when Iir_Kind_Block_Statement =>
               declare
                  Res : Iir_Block_Configuration;
               begin
                  if Get_Block_Block_Configuration (El) = Null_Iir then
                     Res := Create_Iir (Iir_Kind_Block_Configuration);
                     Location_Copy (Res, El);
                     Set_Parent (Res, Conf);
                     Set_Block_Specification (Res, El);
                     Append_Element (Item_List, Res);
                  end if;
               end;
            when Iir_Kind_Generate_Statement =>
               declare
                  Res : Iir_Block_Configuration;
                  Scheme : Iir;
                  Blk_Config : Iir_Block_Configuration;
                  Blk_Spec : Iir;
               begin
                  Scheme := Get_Generation_Scheme (El);
                  Blk_Config := Get_Generate_Block_Configuration (El);
                  if Blk_Config = Null_Iir then
                     --  No block configuration for the (implicit) internal
                     --  block.  Create one.
                     Res := Create_Iir (Iir_Kind_Block_Configuration);
                     Location_Copy (Res, El);
                     Set_Parent (Res, Conf);
                     Set_Block_Specification (Res, El);
                     Append_Element (Item_List, Res);
                  elsif Get_Kind (Scheme) = Iir_Kind_Iterator_Declaration then
                     Blk_Spec := Get_Block_Specification (Blk_Config);
                     if Get_Kind (Blk_Spec) /= Iir_Kind_Generate_Statement then
                        --  There are partial configurations.
                        --  Create a default block configuration.
                        Res := Create_Iir (Iir_Kind_Block_Configuration);
                        Location_Copy (Res, El);
                        Set_Parent (Res, Conf);
                        Blk_Spec := Create_Iir (Iir_Kind_Selected_Name);
                        Location_Copy (Blk_Spec, Res);
                        Set_Identifier (Blk_Spec, Std_Names.Name_Others);
                        Set_Prefix (Blk_Spec, El);
                        Set_Block_Specification (Res, Blk_Spec);
                        Append_Element (Item_List, Res);
                     end if;
                  end if;
               end;
            when Iir_Kind_Sensitized_Process_Statement
              | Iir_Kind_Process_Statement =>
               null;
            when others =>
               Error_Kind ("canon_block_configuration(3)", El);
         end case;
      end loop;

      --  4) Canon component configuration and block configuration (recursion).
      for I in Natural loop
         El := Get_Nth_Element (Item_List, I);
         exit when El = Null_Iir;
         case Get_Kind (El) is
            when Iir_Kind_Block_Configuration =>
               Canon_Block_Configuration (Top, El);
            when Iir_Kind_Component_Configuration =>
               Canon_Component_Configuration (Top, El);
            when Iir_Kind_Configuration_Specification =>
               --  Already canonicalized.
               null;
            when others =>
               Error_Kind ("canon_block_configuration", El);
         end case;
      end loop;
   end Canon_Block_Configuration;

   procedure Canonicalize (Unit: Iir_Design_Unit) is
      El: Iir;
      List: Iir;
      Decls : Iir_List;
      Stmts : Iir_List;
   begin
      List := Get_Context_Items (Unit);
      for I in Natural loop
         El := Get_Nth_Element (List, I);
         exit when El = Null_Iir;
         case Get_Kind (El) is
            when Iir_Kind_Use_Clause =>
               null;
            when Iir_Kind_Library_Clause =>
               null;
            when others =>
               Error_Kind ("canonicalize1", El);
         end case;
      end loop;
      El := Get_Library_Unit (Unit);
      Decls := Get_Declaration_List (El);
      case Get_Kind (El) is
         when Iir_Kind_Entity_Declaration =>
            Stmts := Get_Concurrent_Statement_List (El);
            Canon_Declarations (Unit, Decls, Stmts);
            Canon_Concurrent_Stmts (Unit, Stmts);
         when Iir_Kind_Architecture_Declaration =>
            Stmts := Get_Concurrent_Statement_List (El);
            Canon_Declarations (Unit, Decls, Stmts);
            Canon_Concurrent_Stmts (Unit, Stmts);
         when Iir_Kind_Package_Declaration =>
            Canon_Declarations (Unit, Decls, Null_Iir_List);
         when Iir_Kind_Package_Body =>
            Canon_Declarations (Unit, Decls, Null_Iir_List);
         when Iir_Kind_Configuration_Declaration =>
            Canon_Declarations (Unit, Decls, Null_Iir_List);
            Canon_Block_Configuration (Unit, Get_Block_Configuration (El));
         when others =>
            Error_Kind ("canonicalize2", El);
      end case;
   end Canonicalize;

   --  Create a default component configuration for component instantiation
   --  statement INST.
   function Create_Default_Component_Configuration
     (Inst : Iir_Component_Instantiation_Statement;
      Parent : Iir;
      Config_Unit : Iir_Design_Unit)
     return Iir_Component_Configuration
   is
      Res : Iir_Component_Configuration;
      Designator : Iir;
      Comp : Iir_Component_Declaration;
      Bind : Iir;
      Aspect : Iir;
   begin
      Bind := Get_Default_Binding_Indication (Inst);

      if Bind = Null_Iir then
         --  Component is not bound.
         return Null_Iir;
      end if;

      Res := Create_Iir (Iir_Kind_Component_Configuration);
      Location_Copy (Res, Inst);
      Set_Parent (Res, Parent);
      Comp := Get_Instantiated_Unit (Inst);

      Set_Component_Name (Res, Comp);
      --  Create the instantiation list with only one element: INST.
      Designator := Create_Iir (Iir_Kind_Designator_List);
      Append_Element (Designator, Inst);
      Set_Instantiation_List (Res, Designator);

      Set_Binding_Indication (Res, Bind);
      Aspect := Get_Entity_Aspect (Bind);
      case Get_Kind (Aspect) is
         when Iir_Kind_Entity_Aspect_Entity =>
            Add_Dependence (Config_Unit, Get_Entity (Aspect));
            if Get_Architecture (Aspect) /= Null_Iir then
               raise Internal_Error;
            end if;
         when others =>
            Error_Kind ("Create_Default_Component_Configuration", Aspect);
      end case;

      return Res;
   end Create_Default_Component_Configuration;

   --  Create a default configuration declaration for architecture ARCH.
   function Create_Default_Configuration_Declaration
     (Arch : Iir_Architecture_Declaration)
     return Iir_Design_Unit
   is
      Loc : Location_Type;
      Config : Iir_Configuration_Declaration;
      Res : Iir_Design_Unit;
      Entity : Iir_Entity_Declaration;
      Blk_Cfg : Iir_Block_Configuration;
   begin
      Loc := Get_Location (Arch);
      Res := Create_Iir (Iir_Kind_Design_Unit);
      Set_Location (Res, Loc);
      Set_Parent (Res, Get_Parent (Get_Design_Unit (Arch)));
      Set_Loaded_Flag (Res, True);
      Set_Date (Res, Date_Analyzed);
      Config := Create_Iir (Iir_Kind_Configuration_Declaration);
      Set_Location (Config, Loc);
      Set_Library_Unit (Res, Config);
      Set_Design_Unit (Config, Res);
      Entity := Get_Entity (Arch);
      Set_Entity (Config, Get_Design_Unit (Entity));
      Add_Dependence (Res, Get_Design_Unit (Entity));
      Add_Dependence (Res, Get_Design_Unit (Arch));

      Blk_Cfg := Create_Iir (Iir_Kind_Block_Configuration);
      Set_Location (Blk_Cfg, Loc);
      Set_Parent (Blk_Cfg, Config);
      Set_Block_Specification (Blk_Cfg, Arch);
      Set_Block_Configuration (Config, Blk_Cfg);

      Canon_Block_Configuration (Res, Blk_Cfg);

      return Res;
   end Create_Default_Configuration_Declaration;

end Canon;
