with Libraries;
with Errorout; use Errorout;
with Std_Package;

package body Configuration is
   procedure Add_Design_Concurrent_Stmts (Stmts : Iir_List);
   procedure Add_Design_Block_Configuration (Blk : Iir_Block_Configuration);
   procedure Add_Design_Aspect (Aspect : Iir);

   procedure Add_Design (Unit : Iir_Design_Unit; From : Iir)
   is
      List : Iir_List;
      El : Iir;
      Lib_Unit : Iir;
   begin
      --  If already in the list, then nothing to do.
      if Get_Elab_Flag (Unit) then
         return;
      end if;

      Set_Elab_Flag (Unit, True);

      --  Add packages from depend list.
      List := Get_Dependence_List (Unit);
      for I in Natural loop
         El := Get_Nth_Element (List, I);
         exit when El = Null_Iir;
         if Get_Kind (El) = Iir_Kind_Design_Unit then
            Lib_Unit := Get_Library_Unit (El);
            if Get_Kind (Lib_Unit) = Iir_Kind_Package_Declaration then
               Add_Design (El, Unit);
            end if;
         end if;
      end loop;

      Lib_Unit := Get_Library_Unit (Unit);
      case Get_Kind (Lib_Unit) is
         when Iir_Kind_Package_Declaration =>
            null;
         when Iir_Kind_Configuration_Declaration =>
            --  Add architecture
            --  find all sub-configuration
            Libraries.Load_Design_Unit (Unit, From);
            Lib_Unit := Get_Library_Unit (Unit);
            Add_Design (Get_Entity (Lib_Unit), Unit);
            declare
               Blk : Iir_Block_Configuration;
               Arch : Iir;
            begin
               Blk := Get_Block_Configuration (Lib_Unit);
               Arch := Get_Block_Specification (Blk);
               Add_Design (Get_Design_Unit (Arch), Unit);
               Add_Design_Block_Configuration (Blk);
            end;
         when Iir_Kind_Architecture_Declaration =>
            --  Add entity
            --  find all entity/architecture/configuration instantiation
            Add_Design (Get_Design_Unit (Get_Entity (Lib_Unit)), Unit);
            Add_Design_Concurrent_Stmts
              (Get_Concurrent_Statement_List (Lib_Unit));
         when Iir_Kind_Entity_Declaration =>
            null;
         when Iir_Kind_Package_Body =>
            null;
         when others =>
            Error_Kind ("add_design", Lib_Unit);
      end case;

      Design_Units.Increment_Last;
      Design_Units.Table (Design_Units.Last) := Unit;

      if Get_Kind (Lib_Unit) = Iir_Kind_Package_Declaration then
         --  Add body (if any).
         declare
            Bod : Iir_Design_Unit;
         begin
            Bod := Libraries.Find_Secondary_Unit (Unit, Null_Identifier);
            if Bod = Null_Iir then
               if Get_Need_Body (Lib_Unit) then
                  Error_Msg_Elab
                    ("cannot find body of " & Disp_Node (Lib_Unit));
               end if;
            else
               Set_Package (Get_Library_Unit (Bod), Lib_Unit);
               Add_Design (Bod, Unit);
            end if;
         end;
      end if;
   end Add_Design;

   procedure Add_Design_Concurrent_Stmts (Stmts : Iir_List)
   is
      Stmt : Iir;
   begin
      if Stmts = Null_Iir_List then
         return;
      end if;
      for I in Natural loop
         Stmt := Get_Nth_Element (Stmts, I);
         exit when Stmt = Null_Iir;
         case Get_Kind (Stmt) is
            when Iir_Kind_Component_Instantiation_Statement =>
               declare
                  Unit : Iir;
               begin
                  Unit := Get_Instantiated_Unit (Stmt);
                  if Get_Kind (Unit) /= Iir_Kind_Component_Declaration then
                     Add_Design_Aspect (Unit);
                  end if;
               end;
            when Iir_Kind_Generate_Statement
              | Iir_Kind_Block_Statement =>
               Add_Design_Concurrent_Stmts
                 (Get_Concurrent_Statement_List (Stmt));
            when Iir_Kind_Process_Statement
              | Iir_Kind_Sensitized_Process_Statement =>
               null;
            when others =>
               Error_Kind ("add_design_concurrent_stmts(2)", Stmt);
         end case;
      end loop;
   end Add_Design_Concurrent_Stmts;

   procedure Add_Design_Aspect (Aspect : Iir)
   is
      use Libraries;

      Entity : Iir;
      Arch : Iir;
      Id : Name_Id;
      Entity_Lib : Iir;
   begin
      if Aspect = Null_Iir then
         return;
      end if;
      case Get_Kind (Aspect) is
         when Iir_Kind_Entity_Aspect_Entity =>
            Entity := Get_Entity (Aspect);
            Entity_Lib := Get_Library_Unit (Entity);
            Add_Design (Entity, Aspect);
            Arch := Get_Architecture (Aspect);
            if Arch /= Null_Iir then
               if Get_Kind (Arch) = Iir_Kind_Simple_Name then
                  Id := Get_Identifier (Arch);
                  Arch := Load_Secondary_Unit (Entity, Id, Aspect);
                  if Arch = Null_Iir then
                     Error_Msg_Elab
                       ("cannot find architecture " & Name_Table.Image (Id)
                        & " of " & Disp_Node (Entity_Lib));
                     return;
                  end if;
               end if;
            else
               Arch := Get_Latest_Architecture (Entity_Lib);
               if Arch = Null_Iir then
                  Error_Msg_Elab ("no architecture in library for "
                                  & Disp_Node (Entity_Lib), Aspect);
                  return;
               end if;
               Arch := Get_Design_Unit (Arch);
            end if;
            Load_Design_Unit (Arch, Aspect);
            Add_Design
              (Get_Default_Configuration_Declaration (Get_Library_Unit (Arch)),
               Aspect);
         when Iir_Kind_Entity_Aspect_Configuration =>
            Add_Design (Get_Configuration (Aspect), Aspect);
         when Iir_Kind_Entity_Aspect_Open =>
            null;
         when others =>
            Error_Kind ("add_design_binding_indication", Aspect);
      end case;
   end Add_Design_Aspect;

   procedure Add_Design_Binding_Indication (Bind : Iir_Binding_Indication) is
   begin
      if Bind = Null_Iir then
         return;
      end if;
      Add_Design_Aspect (Get_Entity_Aspect (Bind));
   end Add_Design_Binding_Indication;

   procedure Add_Design_Block_Configuration (Blk : Iir_Block_Configuration)
   is
      Item_List : Iir_List;
      Item : Iir;
   begin
      if Blk = Null_Iir then
         return;
      end if;
      Item_List := Get_Configuration_Item_List (Blk);
      if Item_List = Null_Iir_List then
         return;
      end if;
      for I in Natural loop
         Item := Get_Nth_Element (Item_List, I);
         exit when Item = Null_Iir;
         case Get_Kind (Item) is
            when Iir_Kind_Configuration_Specification =>
               Add_Design_Binding_Indication (Get_Binding_Indication (Item));
            when Iir_Kind_Component_Configuration =>
               Add_Design_Binding_Indication (Get_Binding_Indication (Item));
               Add_Design_Block_Configuration (Get_Block_Configuration (Item));
            when Iir_Kind_Block_Configuration =>
               Add_Design_Block_Configuration (Item);
            when others =>
               Error_Kind ("add_design_block_configuration", Item);
         end case;
      end loop;
   end Add_Design_Block_Configuration;

   --  elaboration of a design hierarchy:
   --  creates a list of design unit.
   --
   --  find top configuration (may be a default one), add it to the list.
   --  For each element of the list:
   --  add direct dependences (packages, entity, arch) if not in the list
   --  for architectures and configuration: find instantiations and add
   --  corresponding configurations
   function Configure (Primary_Id : Name_Id; Secondary_Id : Name_Id)
     return Iir
   is
      use Libraries;

      Unit : Iir_Design_Unit;
      Lib_Unit : Iir;
      Top : Iir;
   begin
      Unit := Find_Primary_Unit (Work_Library, Primary_Id);
      if Unit = Null_Iir then
         Error_Msg_Elab ("cannot find entity or configuration "
                         & Name_Table.Image (Primary_Id));
         return Null_Iir;
      end if;
      Lib_Unit := Get_Library_Unit (Unit);
      case Get_Kind (Lib_Unit) is
         when Iir_Kind_Entity_Declaration =>
            Load_Design_Unit_And_Dependences (Unit, Null_Iir);
            Lib_Unit := Get_Library_Unit (Unit);
            if Secondary_Id /= Null_Identifier then
               Unit := Find_Secondary_Unit (Unit, Secondary_Id);
               if Unit = Null_Iir then
                  Error_Msg_Elab
                    ("cannot find architecture "
                     & Name_Table.Image (Secondary_Id)
                     & " of " & Disp_Node (Lib_Unit));
                  return Null_Iir;
               end if;
            else
               declare
                  Arch_Unit : Iir_Architecture_Declaration;
               begin
                  Arch_Unit := Get_Latest_Architecture (Lib_Unit);
                  if Arch_Unit = Null_Iir then
                     Error_Msg_Elab
                       (Disp_Node (Lib_Unit)
                        & " has no architecture in library "
                        & Name_Table.Image (Get_Identifier (Work_Library)));
                     return Null_Iir;
                  end if;
                  Unit := Get_Design_Unit (Arch_Unit);
               end;
            end if;
            Load_Design_Unit_And_Dependences (Unit, Null_Iir);
            Lib_Unit := Get_Library_Unit (Unit);
            Top := Get_Default_Configuration_Declaration (Lib_Unit);
         when Iir_Kind_Configuration_Declaration =>
            Top := Unit;
         when others =>
            Error_Msg_Elab (Name_Table.Image (Primary_Id)
                            & " is neither an entity nor a configuration");
            return Null_Iir;
      end case;

      Set_Elab_Flag (Std_Package.Std_Standard_Unit, True);

      Add_Design (Top, Null_Iir);
      return Top;
   end Configure;

end Configuration;
