with Ada.Command_Line; use Ada.Command_Line;
with Ada.Text_IO; use Ada.Text_IO;
with Ada.Characters.Latin_1;
with GNAT.OS_Lib; use GNAT.OS_Lib;
with GNAT.Table;
with GNAT.Dynamic_Tables;
with Libraries;
with Name_Table; use Name_Table;
with Std_Names;
with Std_Package;
with Types; use Types;
with Iirs; use Iirs;
with Errorout; use Errorout;
with Iirs_Utils; use Iirs_Utils;
with Files_Map;
with Back_End;
with Flags;
--with Disp_Tree;
with Default_Pathes;
with Interfaces.C_Streams;
with System;
with Version;

package body Ghdldrv is
   --  Mode of GHDLDRV, ie the operation to perform.
   type Mode_Type is
     (Mode_Analyze,
      Mode_Elaborate,
      Mode_Dir,
      Mode_Find,
      Mode_Import,
      Mode_Make,
      Mode_Clean,
      Mode_Remove,
      Mode_Dispconfig,
      Mode_Version,
      Mode_Help,
      Mode_Options_Help);
   Mode : Mode_Type;

   --  Path prefix for libraries.
   Prefix_Path : String_Access := null;

   --  Suffix for asm files.
   Asm_Suffix : constant String := ".s";

   --  Suffix for post files.
   Post_Suffix : constant String := ".on";

   --  Suffix for list files.
   List_Suffix : constant String := ".lst";

   --  Prefix for elab files.
   Elab_Prefix : constant String := "e~";

   --  Name of the tools used.
   Compiler_Cmd : String_Access := null;
   Post_Processor_Cmd : String_Access := null;
   Assembler_Cmd : constant String := "as";
   Linker_Cmd : constant String := "gcc";

   --  Path of the tools.
   Compiler_Path : String_Access;
   Post_Processor_Path : String_Access;
   Assembler_Path : String_Access;
   Linker_Path : String_Access;

   Compile_Error : exception;

   --  "-o" string.
   Dash_O : String_Access;

   --  "-S" string.
   Dash_S : String_Access;

   --  "-quiet" option.
   Dash_Quiet : String_Access;

   --  "-v" flag, if set be verbose.
   Flag_Verbose : Boolean;

   --  If set, do not assmble
   Flag_Asm : Boolean;

   --  If true, executed commands are displayed.
   Flag_Disp_Commands : Boolean;

   --  Flag not quiet
   Flag_Not_Quiet : Boolean;

   --  Force recompilation of units in work library.
   Flag_Force : Boolean;

   --  If set, use a post-processor after compilation.
   Flag_Post : Boolean;

   --  Disp dependences during make.
   Flag_Depend_Unit : Boolean;
   Flag_Depend_File : Boolean;

   --  If true, use synopsys version of library ieee.  This just change pathes.
   type Ieee_Lib_Kind is (Lib_Standard, Lib_None, Lib_Synopsys);
   Flag_Ieee : Ieee_Lib_Kind;

   Has_Error : Boolean := False;

   --  Argument table for the tools.
   --  Each table low bound is 1 so that the length of a table is equal to
   --  the last bound.
   package Argument_Table_Pkg is new GNAT.Dynamic_Tables
     (Table_Component_Type => String_Access,
      Table_Index_Type => Integer,
      Table_Low_Bound => 1,
      Table_Initial => 4,
      Table_Increment => 100);
   use Argument_Table_Pkg;

   --  Arguments for tools.
   Compiler_Args : Argument_Table_Pkg.Instance;
   Postproc_Args : Argument_Table_Pkg.Instance;
   Assembler_Args : Argument_Table_Pkg.Instance;
   Linker_Args : Argument_Table_Pkg.Instance;

   function Get_Version_Path return String;

   --  Disp MSG on the standard output with the command name.
   procedure Error (Msg : String) is
   begin
      Put (Standard_Error, Command_Name);
      Put (Standard_Error, ": ");
      Put_Line (Standard_Error, Msg);
      Has_Error := True;
   end Error;

   --  Return FILENAME without the extension.
   function Get_Base_Name (Filename : String; Remove_Dir : Boolean := True)
                           return String
   is
      First : Natural;
      Last : Natural;
   begin
      First := Filename'First;
      Last := Filename'Last;
      for I in Filename'Range loop
         if Filename (I) = '.' then
            Last := I - 1;
         elsif Remove_Dir and then Filename (I) = Directory_Separator then
            First := I + 1;
            Last := Filename'Last;
         end if;
      end loop;
      return Filename (First .. Last);
   end Get_Base_Name;

   procedure My_Spawn (Program_Name : String; Args : Argument_List)
   is
      Success : Boolean;
   begin
      if Flag_Disp_Commands then
         Put (Program_Name);
         for I in Args'Range loop
            Put (' ');
            Put (Args (I).all);
         end loop;
         New_Line;
      end if;
      Spawn (Program_Name, Args, Success);
      if not Success then
         Error ("compilation error");
         raise Compile_Error;
      end if;
   end My_Spawn;

   function Append_Suffix (File : String; Suffix : String) return String_Access
   is
      Basename : String := Get_Base_Name (File);
   begin
      Image (Libraries.Work_Directory);
      Name_Buffer (Name_Length + 1 .. Name_Length + Basename'Length) :=
        Basename;
      Name_Length := Name_Length + Basename'Length;
      Name_Buffer (Name_Length + 1 .. Name_Length + Suffix'Length) := Suffix;
      Name_Length := Name_Length + Suffix'Length;
      return new String'(Name_Buffer (1 .. Name_Length));
   end Append_Suffix;

   --  Compile FILE with additional argument OPTS.
   procedure Do_Compile (Options : Argument_List; File : String)
   is
      Obj_File : String_Access;
      Asm_File : String_Access;
      Post_File : String_Access;
      Success : Boolean;
   begin
      --  Create the name for the assembly file.
      Asm_File := Append_Suffix (File, Asm_Suffix);

      if Flag_Post then
         Post_File := Append_Suffix (File, Post_Suffix);
      end if;

      --  Compile.
      declare
         P : Natural;
         Nbr_Args : Natural := Last (Compiler_Args) + Options'Length + 4;
         Args : Argument_List (1 .. Nbr_Args);
      begin
         P := 0;
         for I in First .. Last (Compiler_Args) loop
            P := P + 1;
            Args (P) := Compiler_Args.Table (I);
         end loop;
         for I in Options'Range loop
            P := P + 1;
            Args (P) := Options (I);
         end loop;

         --  Add -quiet.
         if not Flag_Not_Quiet then
            P := P + 1;
            Args (P) := Dash_Quiet;
         end if;

         Args (P + 1) := Dash_O;
         if Flag_Post then
            Args (P + 2) := Post_File;
         else
            Args (P + 2) := Asm_File;
         end if;
         Args (P + 3) := new String'(File);

         My_Spawn (Compiler_Path.all, Args (1 .. P + 3));
         Free (Args (P + 3));
      exception
         when Compile_Error =>
            --  Delete temporary file in case of error.
            Delete_File (Args (P + 2).all, Success);
            --  FIXME: delete object file too ?
            raise;
      end;

      --  Post-process.
      if Flag_Post then
         declare
            P : Natural;
            Nbr_Args : Natural := Last (Postproc_Args) + 4;
            Args : Argument_List (1 .. Nbr_Args);
         begin
            P := 0;
            for I in First .. Last (Postproc_Args) loop
               P := P + 1;
               Args (P) := Postproc_Args.Table (I);
            end loop;

            if not Flag_Not_Quiet then
               P := P + 1;
               Args (P) := Dash_Quiet;
            end if;

            Args (P + 1) := Dash_O;
            Args (P + 2) := Asm_File;
            Args (P + 3) := Post_File;
            My_Spawn (Post_Processor_Path.all, Args (1 .. P + 3));
         end;

         Free (Post_File);
      end if;

      --  Assemble.
      if not Flag_Asm then
         declare
            P : Natural;
            Nbr_Args : Natural := Last (Assembler_Args) + 4;
            Args : Argument_List (1 .. Nbr_Args);
            Success : Boolean;
         begin
            P := 0;
            for I in First .. Last (Assembler_Args) loop
               P := P + 1;
               Args (P) := Assembler_Args.Table (I);
            end loop;

            Obj_File := Append_Suffix (File, Get_Object_Suffix.all);

            Args (P + 1) := Dash_O;
            Args (P + 2) := Obj_File;
            Args (P + 3) := Asm_File;
            My_Spawn (Assembler_Path.all, Args (1 .. P + 3));
            Free (Obj_File);
            Delete_File (Asm_File.all, Success);
         end;
      end if;

      Free (Asm_File);
   end Do_Compile;

   package Files is new GNAT.Table
     (Table_Component_Type => String_Access,
      Table_Index_Type => Natural,
      Table_Low_Bound => 1,
      Table_Initial => 16,
      Table_Increment => 100);

   package Filelist is new GNAT.Table
     (Table_Component_Type => String_Access,
      Table_Index_Type => Natural,
      Table_Low_Bound => 1,
      Table_Initial => 16,
      Table_Increment => 100);

   procedure Add_File (File : Iir_Design_File)
   is
      use Files_Map;

      Name : Name_Id;
      Dir : Name_Id;
      Str : String_Access;
   begin
      Dir := Get_Library_Directory (Get_Parent (File));
      Name := Get_Design_File_Filename (File);
      Str := new String'(Image (Dir)
                         & Get_Base_Name (Image (Name))
                         & Get_Object_Suffix.all);
      Files.Increment_Last;
      Files.Table (Files.Last) := Str;
   end Add_File;

   --  Read a list of files from file FILENAME.
   --  Lines starting with a '#' are ignored (comments)
   --  Lines starting with a '>' are directory lines
   --  If first character of a line is a '@', it is replaced with
   --    the prefix_path.
   --  If TO_OBJ is true, then each file is converted to an object file name
   --   (suffix is replaced by the object file extension).
   procedure Add_File_List (Filename : String; To_Obj : Boolean)
   is
      use Interfaces.C_Streams;
      use System;
      use Ada.Characters.Latin_1;

      Dir : String (1 .. max_path_len);
      Dir_Len : Natural;
      Line : String (1 .. max_path_len);
      Stream : Interfaces.C_Streams.FILEs;
      Mode : constant String := "rt" & NUL;
      L : Natural;
      File : String_Access;
   begin
      Line (1 .. Filename'Length) := Filename;
      Line (Filename'Length + 1) := NUL;
      Stream := fopen (Line'Address, Mode'Address);
      if Stream = NULL_Stream then
         Error ("cannot open " & Filename);
         return;
      end if;
      Dir_Len := 0;
      loop
         exit when fgets (Line'Address, Line'Length, Stream) = NULL_Stream;
         if Line (1) /= '#' then
            --  Compute string length.
            L := 0;
            while Line (L + 1) /= NUL loop
               L := L + 1;
            end loop;

            --  Remove trailing NL.
            while L > 0 and then (Line (L) = LF or Line (L) = CR) loop
               L := L - 1;
            end loop;

            if Line (1) = '>' then
               Dir_Len := L - 1;
               Dir (1 .. Dir_Len) := Line (2 .. L);
            else
               if Line (1) = '@' then
                  File := new String'(Prefix_Path.all & Line (2 .. L));
               else
                  if To_Obj then
                     File := new String'(Dir (1 .. Dir_Len)
                                         & Get_Base_Name (Line (1 .. L))
                                         & Get_Object_Suffix.all);
                  else
                     File := new String'(Line (1 .. L));
                  end if;
               end if;

               Filelist.Increment_Last;
               Filelist.Table (Filelist.Last) := File;

               Dir_Len := 0;
            end if;
         end if;
      end loop;
      if fclose (Stream) /= 0 then
         Error ("cannot close " & Filename);
      end if;
   end Add_File_List;

   procedure Do_Bind (Options : Argument_List;
                      Prim_Name : String_Access; Sec_Name : String_Access)
   is
      Base_Name : String_Access;
      Elab_Name : String_Access;
      Filelist_Name : String_Access;
      Success : Boolean;
   begin
      --  create file name
      if Sec_Name = null then
         Base_Name := Prim_Name;
      else
         Base_Name := new String'(Prim_Name.all & '-' & Sec_Name.all);
      end if;

      Elab_Name := new String'(Elab_Prefix & Base_Name.all);
      Filelist_Name := new String'(Elab_Name.all & List_Suffix);

      --  call the compiler
      declare
         Comp_List : Argument_List (1 .. 6 + Options'Length);
         N : Natural;
      begin
         Comp_List (1) := new String'("--elab");
         Comp_List (2) := Prim_Name;
         if Sec_Name = null then
            N := 2;
         else
            Comp_List (3) := Sec_Name;
            N := 3;
         end if;
         Comp_List (N + 1) := new String'("-l");
         Comp_List (N + 2) := Filelist_Name;
         Comp_List (N + 3 .. N + 2 + Options'Length) := Options;
         Do_Compile (Comp_List (1 .. N + 2 + Options'Length), Elab_Name.all);
         Free (Comp_List (N + 1));
         Free (Comp_List (1));
      end;

      --  read files list
      Add_File_List (Filelist_Name.all, True);
      Add_File_List (Prefix_Path.all & "grt" & List_Suffix, False);

      --  call the linker
      declare
         P : Natural;
         Nbr_Args : Natural := Last (Linker_Args) + Filelist.Last + 4;
         Args : Argument_List (1 .. Nbr_Args);
         Obj_File : String_Access;
         Std_File : String_Access;
      begin
         Obj_File := Append_Suffix (Elab_Name.all, Get_Object_Suffix.all);
         Std_File := new
           String'(Prefix_Path.all & Get_Version_Path & Directory_Separator
                   & "std" & Directory_Separator
                   & "std_standard" & Get_Object_Suffix.all);
         P := 0;
         for I in First .. Last (Linker_Args) loop
            P := P + 1;
            Args (P) := Linker_Args.Table (I);
         end loop;
         Args (P + 1) := Dash_O;
         Args (P + 2) := Base_Name;
         Args (P + 3) := Obj_File;
         Args (P + 4) := Std_File;
         P := P + 4;
         for I in Filelist.First .. Filelist.Last loop
            P := P + 1;
            Args (P) := Filelist.Table (I);
         end loop;

         My_Spawn (Linker_Path.all, Args);

         Free (Obj_File);
         Free (Std_File);
      end;

      Delete_File (Filelist_Name.all, Success);

      for I in Filelist.First .. Filelist.Last loop
         Free (Filelist.Table (I));
      end loop;

      if Sec_Name /= null then
         Free (Base_Name);
      end if;
      Free (Elab_Name);
      Free (Filelist_Name);
   end Do_Bind;

   procedure Disp_Library_Unit (Unit : Iir) is
   begin
      case Get_Kind (Unit) is
         when Iir_Kind_Entity_Declaration =>
            Put ("entity " & Get_String (Unit));
         when Iir_Kind_Architecture_Declaration =>
            Put ("architecture " & Get_String (Unit)
                 & " of " & Get_String (Get_Entity (Unit)));
         when Iir_Kind_Configuration_Declaration =>
            Put ("configuration " & Get_String (Unit));
         when Iir_Kind_Package_Declaration =>
            Put ("package " & Get_String (Unit));
         when Iir_Kind_Package_Body =>
            Put ("package body " & Get_String (Unit));
         when others =>
            Put ("???");
      end case;
   end Disp_Library_Unit;

   procedure Disp_Library (Name : Name_Id) is
      Lib : Iir_Library_Declaration;
      File_List : Iir_List;
      File : Iir_Design_File;
      Unit_List : Iir_List;
      Unit : Iir;
   begin
      if Name = Std_Names.Name_Work then
         Lib := Libraries.Work_Library;
      elsif Name = Std_Names.Name_Std then
         Lib := Libraries.Std_Library;
      else
         Lib := Libraries.Get_Library (Name);
      end if;

      File_List := Get_Design_File_List (Lib);

      --  Disp contents of files.
      for I in Natural loop
         File := Get_Nth_Element (File_List, I);
         exit when File = Null_Iir;
         Unit_List := Get_Design_Unit_List (File);
         for J in Natural loop
            Unit := Get_Nth_Element (Unit_List, J);
            exit when Unit = Null_Iir;
            Disp_Library_Unit (Get_Library_Unit (Unit));
            New_Line;
         end loop;
      end loop;
   end Disp_Library;

   --  Return TRUE is UNIT can be at the apex of a design hierarchy.
   function Is_Top_Entity (Unit : Iir) return Boolean
   is
   begin
      if Get_Kind (Unit) /= Iir_Kind_Entity_Declaration then
         return False;
      end if;
      if Get_Port_List (Unit) /= Null_Iir_List then
         return False;
      end if;
      if Get_Generic_List (Unit) /= Null_Iir_List then
         return False;
      end if;
      return True;
   end Is_Top_Entity;

   --  Disp contents of libraries LIBS.
   procedure Do_Dir (Libs : Argument_List) is
      Id : Name_Id;
   begin
      --  Load word library.
      Std_Package.Create_Std_Standard_Package;
      --Libraries.Load_Std_Library;
      Libraries.Load_Work_Library;

      if Libs'Length = 0 then
         Disp_Library (Std_Names.Name_Work);
      else
         for L in Libs'Range loop
            Id := Get_Identifier (Libs (L).all);
            Disp_Library (Id);
         end loop;
      end if;
   end Do_Dir;

   --  Disp contents design files FILES.
   procedure Do_Find (Files : Argument_List)
   is
      Id : Name_Id;
      Design_File : Iir_Design_File;
      Unit_List : Iir_List;
      Unit : Iir;
      Lib : Iir;
      Flag_Add : Boolean := False;
   begin
      --  Load word library.
      Std_Package.Create_Std_Standard_Package;
      --Libraries.Load_Std_Library;
      Libraries.Load_Work_Library;

      for I in Files'Range loop
         Id := Get_Identifier (Files (I).all);
         Design_File := Libraries.Load_File (Id);
         if Design_File /= Null_Iir then
            Unit_List := Get_Design_Unit_List (Design_File);
            for J in Natural loop
               Unit := Get_Nth_Element (Unit_List, J);
               exit when Unit = Null_Iir;
               Lib := Get_Library_Unit (Unit);
               Disp_Library_Unit (Lib);
               if Is_Top_Entity (Lib) then
                  Put (" **");
               end if;
               New_Line;
               if Flag_Add then
                  Libraries.Add_Design_Unit_Into_Library (Unit);
               end if;
            end loop;
         end if;
      end loop;
      if Flag_Add then
         Libraries.Save_Work_Library;
      end if;
   end Do_Find;

   procedure Do_Import (Files : Argument_List)
   is
      Id : Name_Id;
      Design_File : Iir_Design_File;
      Unit_List : Iir_List;
      Unit : Iir;
      Lib : Iir;
      Files_List : Iir_List;
   begin
      --  Load word library.
      Std_Package.Create_Std_Standard_Package;
      --Libraries.Load_Std_Library;
      Libraries.Load_Work_Library;

      --  Parse all files.
      for I in Files'Range loop
         Id := Get_Identifier (Files (I).all);
         Design_File := Libraries.Load_File (Id);
         if Design_File /= Null_Iir then
            Unit_List := Get_Design_Unit_List (Design_File);
            for J in Natural loop
               Unit := Get_Nth_Element (Unit_List, J);
               exit when Unit = Null_Iir;
               if Flag_Verbose then
                  Lib := Get_Library_Unit (Unit);
                  Disp_Library_Unit (Lib);
                  if Is_Top_Entity (Lib) then
                     Put (" **");
                  end if;
                  New_Line;
               end if;
               Libraries.Add_Design_Unit_Into_Library (Unit);
            end loop;
         end if;
      end loop;

      --  Analyze all files.
      Files_List := Get_Design_File_List (Libraries.Work_Library);
      for I in Natural loop
         Design_File := Get_Nth_Element (Files_List, I);
         exit when Design_File = Null_Iir;
         Unit_List := Get_Design_Unit_List (Design_File);
         for J in Natural loop
            Unit := Get_Nth_Element (Unit_List, J);
            exit when Unit = Null_Iir;
            case Get_Date (Unit) is
               when Date_Valid
                 | Date_Analyzed =>
                  null;
               when Date_Parsed =>
                  Back_End.Finish_Compilation (Unit, False);
               when others =>
                  raise Internal_Error;
            end case;
         end loop;
      end loop;

      Libraries.Save_Work_Library;
   exception
      when Compilation_Error =>
         Error ("importation has failed due to compilation error");
   end Do_Import;

   package Mk_Dep is new GNAT.Table
     (Table_Component_Type => Iir,
      Table_Index_Type => Natural,
      Table_Low_Bound => 1,
      Table_Initial => 16,
      Table_Increment => 100);

   procedure Disp_Dependence (Unit : Iir_Design_Unit; Dep : Iir_Design_Unit) is
   begin
      Disp_Library_Unit (Get_Library_Unit (Dep));
      Put (" <- ");
      Disp_Library_Unit (Get_Library_Unit (Unit));
      New_Line;
   end Disp_Dependence;

   procedure Make_Depend (Unit : Iir_Design_Unit)
   is
      List : Iir_List;
      El : Iir;
      Lib_Unit : Iir;
      Sub : Iir_Design_Unit;
   begin
      if Get_Elab_Flag (Unit) then
         --  Unit is already in the dependence list.
         return;
      else
         Set_Elab_Flag (Unit, True);
      end if;
      List := Get_Dependence_List (Unit);
      if List /= Null_Iir_List then
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            Sub := Libraries.Find_Design_Unit (El);
            if Sub /= Null_Iir and then not Get_Elab_Flag (Sub) then
               --  FIXME: if sub = null_iir
               --  FIXME: replace el with sub.
               if Flag_Depend_Unit then
                  Disp_Dependence (Unit, Sub);
               end if;
               Make_Depend (Sub);
            end if;
         end loop;
      end if;
      Mk_Dep.Increment_Last;
      Mk_Dep.Table (Mk_Dep.Last) := Unit;

      --  Add package body if unit is a package declaration and if it needs a
      --  body.
      --  Note: lib_unit may be null_iir for a default configuration.
      Lib_Unit := Get_Library_Unit (Unit);
      if Lib_Unit /= Null_Iir
        and then Get_Kind (Lib_Unit) = Iir_Kind_Package_Declaration
        and then Get_Need_Body (Lib_Unit)
      then
         El := Libraries.Find_Secondary_Unit (Unit, Null_Identifier);
         if El = Null_Iir then
            Error ("package body " & Get_String (Lib_Unit)
                   & " was never analyzed");
         else
            if Flag_Depend_Unit then
               Disp_Dependence (Unit, El);
            end if;
            Make_Depend (El);
         end if;
      end if;
   end Make_Depend;

   procedure Add_Latest_Arch (Du : Iir_Design_Unit)
   is
      Arch : Iir;
      Unit : Iir;
   begin
      Unit := Get_Library_Unit (Du);
      if Get_Kind (Unit) = Iir_Kind_Entity_Declaration then
         Arch := Libraries.Get_Latest_Architecture (Unit);
         if Arch = Null_Iir then
            Error ("no architecture for " & Errorout.Disp_Node (Unit));
         else
            Arch := Get_Design_Unit (Arch);
            if not Get_Elab_Flag (Arch) then
               if Flag_Depend_Unit then
                  Disp_Dependence (Du, Arch);
               end if;
               Make_Depend (Arch);
            end if;
         end if;
      end if;
   end Add_Latest_Arch;

   procedure Make_File_Depend
   is
      New_File : Boolean;

      procedure Handle_Unit (Sub_Unit : Iir_Design_Unit)
      is
         Lib_Unit : Iir;
      begin
         if Get_Elab_Flag (Sub_Unit) then
            New_File := False;
         else
            if Flag_Depend_Unit then
               Lib_Unit := Get_Library_Unit (Sub_Unit);
               if Lib_Unit /= Null_Iir then
                  Disp_Library_Unit (Lib_Unit);
                  Put (" <- file");
                  New_Line;
               end if;
            end if;
            Make_Depend (Sub_Unit);
         end if;
      end Handle_Unit;

      Unit_List : Iir_List;
      File : Iir_Design_File;
      Unit : Iir_Design_Unit;
      Sub_Unit : Iir_Design_Unit;
      Lib_Unit : Iir;
      Cfg_Unit : Iir_Design_Unit;
   begin
      if Flag_Depend_Unit then
         Put_Line ("units required by linker but not used:");
      end if;

      --  Mark all elab flags of units.
      --  Clear all elab flags of file.
      for I in Mk_Dep.First .. Mk_Dep.Last loop
         Unit := Mk_Dep.Table (I);
         Set_Elab_Flag (Unit, True);
         File := Get_Design_File (Unit);
         Set_Elab_Flag (File, False);
      end loop;

      --  std.standard is already included.
      Set_Elab_Flag (Std_Package.Std_Standard_File, True);
      Set_Elab_Flag (Std_Package.Std_Standard_Unit, True);

      for I in Mk_Dep.First .. Natural'Last loop
         exit when I > Mk_Dep.Last;
         Unit := Mk_Dep.Table (I);
         File := Get_Design_File (Unit);
         if not Get_Elab_Flag (File) then
            Set_Elab_Flag (File, True);
            New_File := True;
            Unit_List := Get_Design_Unit_List (File);
            for J in Natural loop
               Sub_Unit := Get_Nth_Element (Unit_List, J);
               exit when Sub_Unit = Null_Iir;
               Handle_Unit (Sub_Unit);
               Lib_Unit := Get_Library_Unit (Sub_Unit);
               if Get_Kind (Lib_Unit) = Iir_Kind_Architecture_Declaration then
                  Cfg_Unit := Get_Default_Configuration_Declaration (Lib_Unit);
                  if Cfg_Unit /= Null_Iir then
                     Handle_Unit (Cfg_Unit);
                  end if;
               end if;
            end loop;
            if New_File then
               Add_File (File);
               --Put_Line ("file adding: " & Files.Table (Files.Last).all);
            end if;
         end if;
      end loop;
   end Make_File_Depend;

   function Is_Outdated (Unit : Iir_Design_Unit) return Boolean
   is
      use Files_Map;

      Design_File : Iir_Design_File;
      Library : Iir_Library_Declaration;
      File : Source_File_Entry;
      Obj_Path : String_Access;
      File_Id : Name_Id;
   begin
      --  Std.Standard is never outdated.
      if Unit = Std_Package.Std_Standard_Unit then
         return False;
      end if;

      Design_File := Get_Design_File (Unit);
      Library := Get_Parent (Design_File);
      File_Id := Get_Design_File_Filename (Design_File);
      File := Create_Source_File (Get_Design_File_Directory (Design_File),
                                  File_Id);
      if File = No_Source_File_Entry then
         Error ("source file " & Image (File_Id) & " not found");
         raise Compilation_Error;
      end if;

      --  Add file.
      if Get_Elab_Flag (Design_File) = False then
         Set_Elab_Flag (Design_File, True);
         Add_File (Design_File);
         Obj_Path := Files.Table (Files.Last);

         --  If the object file does not exist, recompile the file.
         if not Is_Regular_File (Obj_Path.all) then
            if Flag_Verbose then
               Put_Line ("object file for " & Image (Get_File_Name (File))
                         & " does not exist");
            end if;
            return True;
         end if;
      end if;

      --  1) not yet analyzed:
      if Get_Date (Unit) not in Date_Valid then
         if Flag_Verbose then
            Disp_Library_Unit (Get_Library_Unit (Unit));
            Put_Line (" was not analyzed");
         end if;
         return True;
      end if;

      --  2) file has been modified.
      if Get_File_Time_Stamp (File).all /=
        Get_File_Time_Stamp (Design_File).all
      then
         if Flag_Verbose then
            Put_Line ("file " & Image (Get_File_Name (File))
                 & " has been modified");
         end if;
         return True;
      end if;

      --  3) the object file does not exist.
      --  Already checked.

      --  4) one of the dependence is newer
      declare
         Depends : Iir_List;
         El : Iir;
         Dep : Iir_Design_Unit;
         Stamp : Time_Stamp_Acc;
         Dep_File : Iir_Design_File;
      begin
         Depends := Get_Dependence_List (Unit);
         Stamp := Get_Analysis_Time_Stamp (Design_File);
         if Depends /= Null_Iir_List then
            for I in Natural loop
               El := Get_Nth_Element (Depends, I);
               exit when El = Null_Iir;
               Dep := Libraries.Find_Design_Unit (El);
               if Dep = Null_Iir then
                  if Flag_Verbose then
                     Disp_Library_Unit (Unit);
                     Put (" depends on an unknown unit ");
                     Disp_Library_Unit (El);
                     New_Line;
                  end if;
                  return True;
               end if;
               Dep_File := Get_Design_File (Dep);
               if Dep /= Std_Package.Std_Standard_Unit
                 and then Stamp.all < Get_Analysis_Time_Stamp (Dep_File).all
               then
                  if Flag_Verbose then
                     Disp_Library_Unit (Get_Library_Unit (Unit));
                     Put (" depends on: ");
                     Disp_Library_Unit (Get_Library_Unit (Dep));
                     Put (" (more recently analyzed)");
                     New_Line;
                  end if;
                  return True;
               end if;
            end loop;
         end if;
      end;

      return False;
   end Is_Outdated;

   procedure Check_No_Elab_Flag (Lib : Iir_Library_Declaration)
   is
      File_List : Iir_List;
      Unit_List : Iir_List;
      File : Iir_Design_File;
      Unit : Iir_Design_Unit;
   begin
      File_List := Get_Design_File_List (Lib);
      for I in Natural loop
         File := Get_Nth_Element (File_List, I);
         exit when File = Null_Iir;
         Unit_List := Get_Design_Unit_List (File);
         for J in Natural loop
            Unit := Get_Nth_Element (Unit_List, J);
            exit when Unit = Null_Iir;
            if Get_Elab_Flag (Unit) then
               raise Internal_Error;
            end if;
         end loop;
      end loop;
   end Check_No_Elab_Flag;

   procedure Do_Make (Options : Argument_List;
                      Prim : String_Access; Sec : String_Access)
   is
      Top : Iir;
      Unit : Iir;
      File : Iir_Design_File;
      El : Iir;
      List : Iir_List;

      In_Work : Boolean;

      --  Time-stamp of the most recently compiled file.
      Last_Stamp : Time_Stamp_Acc;
      --  Most recently compiled file.
      Last_File : Iir_Design_File;

      Stamp : Time_Stamp_Acc;

      --  True if at least one compilation occured.
      Has_Compiled : Boolean;
   begin
      --Flag_Disp_Commands := True;

      Std_Package.Create_Std_Standard_Package;
      Libraries.Load_Work_Library;
      Check_No_Elab_Flag (Libraries.Work_Library);

      Top := Libraries.Find_Primary_Unit
        (Libraries.Work_Library, Get_Identifier (Prim.all));
      if Top = Null_Iir then
         Error ("cannot find primary unit " & Prim.all);
         raise Option_Error;
      end if;
      if Sec /= null then
         Top := Libraries.Find_Secondary_Unit (Top, Get_Identifier (Sec.all));
         if Top = Null_Iir then
            Error ("secondary unit """ & Sec.all & """ not found");
            raise Option_Error;
         end if;
      end if;

      --  Compute dependences.
      if Flag_Depend_Unit then
         Disp_Library_Unit (Get_Library_Unit (Top));
         Put_Line (" <- top unit");
      end if;
      Make_Depend (Top);

      --  Add an architecture (if necessary).
      if Get_Kind (Get_Library_Unit (Top)) = Iir_Kind_Entity_Declaration then
         Add_Latest_Arch (Top);
      end if;

      --  Add all latest architectures
      --  Note: the upper bound can be changed inside the loop!
      for I in Mk_Dep.First .. Natural'Last loop
         exit when I > Mk_Dep.Last;
         Unit := Mk_Dep.Table (I);
         if Unit /= Top then
            Add_Latest_Arch (Unit);
         end if;
      end loop;

      --  Disp dependences.
      if Flag_Depend_File then
         for I in Mk_Dep.First .. Mk_Dep.Last loop
            Disp_Library_Unit (Get_Library_Unit (Mk_Dep.Table (I)));
            New_Line;
         end loop;
      end if;

      --  Mark all files as not up to date.
      for I in Mk_Dep.First .. Mk_Dep.Last loop
         Set_Elab_Flag (Mk_Dep.Table (I), False);
      end loop;

      -- for each element of the mk_dep list:
      --   if outdated, recompile file, discard library, restart.
      Has_Compiled := False;
      Last_Stamp := null;
      for I in Mk_Dep.First .. Mk_Dep.Last loop
         Unit := Mk_Dep.Table (I);
         if not Get_Elab_Flag (Unit) then
            --  Unit has not just been compiled.

            File := Get_Design_File (Unit);
            Stamp := Get_Analysis_Time_Stamp (File);
            if Last_Stamp = null or else Last_Stamp.all < Stamp.all then
               --  FIXME: should not use analysis time stamp, since it is
               --  in local time, while file time are in GMT.
               Last_Stamp := Stamp;
               Last_File := File;
            end if;
            In_Work := Get_Parent (File) = Libraries.Work_Library;

            --  Due to side effect in IS_OUTDATED, always call it.
            if Is_Outdated (Unit) or else (Flag_Force and In_Work) then
               if not In_Work then
                  Error ("cannot build " & Disp_Node (Unit)
                         & " (not in work library)");
                  raise Compile_Error;
               end if;

               if not Flag_Verbose then
                  Put ("compile ");
                  Put (Image (Get_Design_File_Filename (File)));
                  --Disp_Library_Unit (Get_Library_Unit (Unit));
                  New_Line;
               end if;

               Has_Compiled := True;
               Do_Compile (Options, Image (Get_Design_File_Filename (File)));
               --  Set the analysus time stamp since the file has just been
               --  analyzed.
               Set_Analysis_Time_Stamp (File, Files_Map.Get_Os_Time_Stamp);

               --  Mark all units of FILE as up to date.
               List := Get_Design_Unit_List (File);
               for I in Natural loop
                  El := Get_Nth_Element (List, I);
                  exit when El = Null_Iir;
                  Set_Elab_Flag (El, True);
               end loop;

               --  Start again!
               --  FIXME: unload all libraries (except std).
               --  FIXME: again
            end if;
         end if;
      end loop;

      if not Flag_Verbose then
         Put ("elaborate and link");
         New_Line;
      end if;

      --  Add all files dependences.
      --  (Ex: If a file defines two architecture, the both entity are
      --   required to compile).
      Make_File_Depend;

      --  Elaboration.
      --  FIXME: secondary.
      --  if libgrt is more recent than the executable (FIXME).
      if Has_Compiled then
         if Flag_Verbose then
            Put_Line ("link due to a file compilation");
         end if;
         Do_Bind (Options, Prim, Sec);
      elsif not Is_Regular_File (Prim.all) then
         if Flag_Verbose then
            Put_Line ("link due to no binary file");
         end if;
         Do_Bind (Options, Prim, Sec);
      else
         declare
            Str : String := Prim.all & Character'Val (0);
         begin
            Stamp := Files_Map.Get_File_Time_Stamp (Str'Address);
         end;
         if Stamp.all < Last_Stamp.all then
            --  if a file is more recent than the executable.
            if Flag_Verbose then
               Put ("link due to outdated binary file: ");
               Put (Image (Get_Design_File_Filename (Last_File)));
               Put (" (");
               Put (Last_Stamp.all);
               Put (" > ");
               Put (Stamp.all);
               Put (")");
               New_Line;
            end if;
            Do_Bind (Options, Prim, Sec);
         end if;
      end if;
   end Do_Make;

   procedure Do_Clean (Remove_Library : Boolean)
   is
      Nul : constant Character := Character'Val (0);

      procedure Delete (Str : String)
      is
         Status : Boolean;
      begin
         Delete_File (Str'Address, Status);
         if Flag_Verbose and Status then
            Put_Line ("delete " & Str (Str'First .. Str'Last - 1));
         end if;
      end Delete;

      procedure Delete_Asm_Obj (Str : String) is
      begin
         Delete (Str & Get_Object_Suffix.all & Nul);
         Delete (Str & Asm_Suffix & Nul);
      end Delete_Asm_Obj;

      procedure Delete_Top_Unit (Str : String) is
      begin
         --  Delete elaboration file
         Delete_Asm_Obj (Image (Libraries.Work_Directory) & Elab_Prefix & Str);

         --  Delete file list.
         Delete (Image (Libraries.Work_Directory) & Str & List_Suffix & Nul);

         --  Delete executable.
         Delete (Str & Nul);
      end Delete_Top_Unit;

      File_List : Iir_List;
      File : Iir_Design_File;
      Design_Unit : Iir_Design_Unit;
      Lib_Unit : Iir;
      Ent_Unit : Iir;
      Design_Unit_List : Iir_List;
      Str : String_Access;
   begin
      Std_Package.Create_Std_Standard_Package;
      Libraries.Load_Work_Library;

      File_List := Get_Design_File_List (Libraries.Work_Library);
      for I in Natural loop
         File := Get_Nth_Element (File_List, I);
         exit when File = Null_Iir;

         --  Delete compiled file.
         Str := Append_Suffix (Image (Get_Design_File_Filename (File)), "");
         Delete_Asm_Obj (Str.all);
         Free (Str);

         Design_Unit_List := Get_Design_Unit_List (File);
         for J in Natural loop
            Design_Unit := Get_Nth_Element (Design_Unit_List, J);
            exit when Design_Unit = Null_Iir;
            Lib_Unit := Get_Library_Unit (Design_Unit);
            case Get_Kind (Lib_Unit) is
               when Iir_Kind_Entity_Declaration
                 | Iir_Kind_Configuration_Declaration =>
                  Delete_Top_Unit (Image (Get_Identifier (Lib_Unit)));
               when Iir_Kind_Architecture_Declaration =>
                  Ent_Unit := Get_Entity (Lib_Unit);
                  Delete_Top_Unit (Image (Get_Identifier (Ent_Unit))
                                   & '-'
                                   & Image (Get_Identifier (Lib_Unit)));
               when others =>
                  null;
            end case;
         end loop;
      end loop;

      if Remove_Library then
         Delete (Image (Libraries.Work_Directory)
                 & Back_End.Library_To_File_Name (Libraries.Work_Library)
                 & Nul);
      end if;
   end Do_Clean;

   procedure Add_Argument (Inst : in out Instance; Arg : String_Access)
   is
   begin
      Increment_Last (Inst);
      Inst.Table (Last (Inst)) := Arg;
   end Add_Argument;

   --  Convert option "-Wx,OPTIONS" to arguments for tool X.
   procedure Add_Arguments (Inst : in out Instance; Opt : String) is
   begin
      Add_Argument (Inst, new String'(Opt (Opt'First + 4 .. Opt'Last)));
   end Add_Arguments;

   function Get_Version_Path return String is
   begin
      case Flags.Vhdl_Std is
         when Vhdl_87 =>
            return "v87";
         when Vhdl_93c
           | Vhdl_93
           | Vhdl_98 =>
            return "v93";
      end case;
   end Get_Version_Path;

   function Get_Lib_Path (Suffix : String) return String_Access is
   begin
      return new String'
        ("-P" & Prefix_Path.all & Get_Version_Path & Directory_Separator
         & Suffix & Directory_Separator);
   end Get_Lib_Path;

   procedure Add_Library_Path (Name : String)
   is
      Str : String_Access;
   begin
      Str := Get_Lib_Path (Name);
      Add_Argument (Compiler_Args, Str);

      --  Also add in the current search list, but without -P prefix.
      Libraries.Add_Library_Path (Str.all (Str'First + 2 .. Str'Last));
   end Add_Library_Path;

   procedure Tool_Not_Found (Name : String) is
   begin
      Error ("installation problem: " & Name & " not found");
      raise Option_Error;
   end Tool_Not_Found;

   procedure Setup_Compiler
   is
   begin
      Compiler_Path := Locate_Exec_On_Path (Compiler_Cmd.all);
      if Compiler_Path = null then
         Tool_Not_Found (Compiler_Cmd.all);
      end if;
      if Flag_Post then
         Post_Processor_Path := Locate_Exec_On_Path (Post_Processor_Cmd.all);
         if Post_Processor_Path = null then
            Tool_Not_Found (Post_Processor_Cmd.all);
         end if;
      end if;
      Assembler_Path := Locate_Exec_On_Path (Assembler_Cmd);
      if Assembler_Path = null and not Flag_Asm then
         Tool_Not_Found (Assembler_Cmd);
      end if;
      Linker_Path := Locate_Exec_On_Path (Linker_Cmd);
      if Linker_Path = null then
         Tool_Not_Found (Linker_Cmd);
      end if;
      Dash_O := new String'("-o");
      Dash_S := new String'("-S");
      Dash_Quiet := new String'("-quiet");

      --  Add pathes for predefined libraries.
      if not Flags.Bootstrap then
         Add_Library_Path ("std");
         case Flag_Ieee is
            when Lib_Standard =>
               Add_Library_Path ("ieee");
            when Lib_Synopsys =>
               Add_Library_Path ("synopsys");
            when Lib_None =>
               null;
         end case;
      end if;
   end Setup_Compiler;

   procedure Do_Dispconfig is
   begin
      Put ("compiler command: ");
      Put_Line (Compiler_Cmd.all);
      if Flag_Post then
         Put ("post-processor command: ");
         Put_Line (Post_Processor_Cmd.all);
      end if;
      Put ("assembler command: ");
      Put_Line (Assembler_Cmd);
      Put ("linker command: ");
      Put_Line (Linker_Cmd);
      Put ("library directory: ");
      Put_Line (Prefix_Path.all);
      Setup_Compiler;
      Put ("compiler path: ");
      Put_Line (Compiler_Path.all);
      if Flag_Post then
         Put ("post-processor path: ");
         Put_Line (Post_Processor_Path.all);
      end if;
      Put ("assembler path: ");
      Put_Line (Assembler_Path.all);
      Put ("linker path: ");
      Put_Line (Linker_Path.all);
      Put_Line ("compiler options:");
      for I in First .. Last (Compiler_Args) loop
         Put (" ");
         Put_Line (Compiler_Args.Table (I).all);
      end loop;
   end Do_Dispconfig;

   procedure Do_Version is
   begin
      Put_Line (Version.Ghdl_Version);
      Put_Line ("Written by Tristan Gingold.");
      New_Line;
      Put_Line ("Copyright (C) 2003 Tristan Gingold.");
      Put_Line ("This is free software; see the source for copying conditions."
                & " There is NO");
      Put_Line ("warranty; not even for MERCHANTABILITY or FITNESS FOR A "
                & "PARTICULAR PURPOSE.");
   end Do_Version;

   --  Set the mode.
   procedure Set_Mode (Arg : String) is
   begin
      if Arg = "-a" then
         Mode := Mode_Analyze;
      elsif Arg = "-e" then
         Mode := Mode_Elaborate;
      elsif Arg = "-h" or else Arg = "--help" then
         Mode := Mode_Help;
      elsif Arg = "-d" then
         Mode := Mode_Dir;
      elsif Arg = "-f" then
         Mode := Mode_Find;
      elsif Arg = "-i" then
         Mode := Mode_Import;
      elsif Arg = "-m" then
         Mode := Mode_Make;
      elsif Arg = "--clean" then
         Mode := Mode_Clean;
      elsif Arg = "--remove" then
         Mode := Mode_Remove;
      elsif Arg = "--dispconfig" then
         Mode := Mode_Dispconfig;
      elsif Arg = "--version" then
         Mode := Mode_Version;
         Do_Version;
      elsif Arg = "--options-help" then
         Mode := Mode_Options_Help;
      else
         Error ("unknown action '" & Arg & "', try --help");
         raise Option_Error;
      end if;
   end Set_Mode;

   --  Disp help.
   procedure Do_Help
   is
      procedure P (S : String) renames Ada.Text_IO.Put_Line;
   begin
      P ("usage: " & Command_Name & " ACTION [OPTIONS] ...");
      P ("ACTION is one of:");
      P ("-a [OPTIONS] FILES...    analyze FILES");
      P ("-e [OPTIONS] UNIT_NAME [ARCH_NAME]    elaborate UNIT_NAME");
      P ("-d [LIBRARY_NAME...]     disp contents of LIBRARY_NAME");
      P ("-f FILES...              disp contents of FILES");
      P ("-i [OPTIONS] FILES...    import unit from FILES");
      P ("-m [OPTIONS] UNIT [ARCH] make UNIT");
      P ("--clean [OPTIONS]        remove object files");
      P ("--remove [OPTIONS]       as --clean and remove library file");
      P ("--dispconfig             disp configuration path");
      P ("-h or --help             disp this help");
      P ("--options-help           disp compiler options help");
      P ("--version                output version information and exit");
      New_Line;
      P ("Common options:");
      P (" --work=NAME           use NAME as work library");
      P (" --workdir=PATH        use PATH for the work library");
      P (" --std=XX              select VHDL standard (87,93c,93)");
      P (" --ieee=synopsys       use synopsys version of ieee library (bad)");
      P (" --ieee=standard       use standard version of ieee library (def)");
      P (" --ieee=none           no predefined ieee library");
      P (" -v                    verbose");
      P (" -Q                    do not add -quiet option to compiler");
      P (" -WX,OPTION            pass OPTION to X, where X is one of");
      P ("                        c(compiler) a(assembler) l(linker)");
      P (" --GHDL1=COMMAND       use COMMAND as compiler");
      P (" --PREFIX=DIR          use DIR as std default path");
      New_Line;
      P ("Additional options for -m:");
      P ("  -f                   force recompilation of work units");
      P ("  -Mu                  disp unit dependences (human format)");
      P ("  -Mf                  disp file dependences");
      New_Line;
      P ("Additional options for -a:");
      P ("  -S                   do not assemble");
      P ("  --expect-failure     expect analysis failure (for testsuite)");
      New_Line;
      P ("Report bugs to <ghdl@free.fr>.");
   end Do_Help;

   procedure Main is
      --  Index of the last argument.
      Argc : Natural := 0;
      --  Program arguments.
      Args : Argument_List (1 .. Argument_Count);

      Opts : Argument_List (1 .. Argument_Count);
      Optc : Natural := 0;
   begin
      Flag_Verbose := False;
      Flag_Not_Quiet := False;
      Flag_Disp_Commands := False;
      Flag_Force := False;
      Flag_Ieee := Lib_Standard;
      Flag_Asm := False;
      Flag_Post := False;

      Flag_Depend_Unit := False;
      Flag_Depend_File := False;

      --  The first line argument must be the action.
      if Argument_Count = 0 then
         Error ("no action selected, try --help");
         raise Option_Error;
      end if;

      Set_Mode (Argument (1));

      --  Do not decode options when --version.
      if Mode = Mode_Version then
         return;
      end if;

      --  Required for flag.parse_option.
      Std_Names.Std_Names_Initialize;
      Libraries.Init_Pathes;

      --  Initialize argument tables.
      Init (Compiler_Args);
      Init (Postproc_Args);
      Init (Assembler_Args);
      Init (Linker_Args);

      --  Decode options.
      for I in 2 .. Argument_Count loop
         declare
            Arg : String := Argument (I);
         begin
            if Arg (1) = '-' then
               --  Argument is an option.

               if Argc > 0 then
                  Error ("options after file");
                  raise Option_Error;
               end if;

               --  Option to pass to the compiler.
               if Arg = "-v" and then Flag_Verbose = False then
                  Flag_Verbose := True;
                  Flag_Disp_Commands := True;
               elsif Arg = "-Q" then
                  Flag_Not_Quiet := True;
               elsif Arg = "--post" then
                  Flag_Post := True;
               elsif Arg'Length > 8 and then Arg (1 .. 8) = "--GHDL1=" then
                  Compiler_Cmd := new String'(Arg (9 .. Arg'Last));
               elsif Arg'Length > 9 and then Arg (1 .. 9) = "--PREFIX=" then
                  Prefix_Path := new String'(Arg (10 .. Arg'Last));
               elsif Arg = "--ieee=synopsys" then
                  Flag_Ieee := Lib_Synopsys;
               elsif Arg = "--ieee=none" then
                  Flag_Ieee := Lib_None;
               elsif Arg = "--ieee=standard" then
                  Flag_Ieee := Lib_Standard;
               elsif Mode = Mode_Make and then Arg = "-f" then
                  Flag_Force := True;
               elsif Mode = Mode_Make and then Arg = "-Mu" then
                  Flag_Depend_Unit := True;
               elsif Mode = Mode_Make and then Arg = "-Mf" then
                  Flag_Depend_File := True;
               elsif Mode = Mode_Analyze and then Arg = "--expect-failure" then
                  Add_Argument (Compiler_Args, new String'(Arg));
               elsif Arg = "-S" then
                  Flag_Asm := True;
               elsif Arg'Length > 4
                 and then Arg (2) = 'W' and then Arg (4) = ','
               then
                  if Arg (3) = 'c' then
                     Add_Arguments (Compiler_Args, Arg);
                  elsif Arg (3) = 'a' then
                     Add_Arguments (Assembler_Args, Arg);
                  elsif Arg (3) = 'p' then
                     Add_Arguments (Postproc_Args, Arg);
                  elsif Arg (3) = 'l' then
                     Add_Arguments (Linker_Args, Arg);
                  else
                     Error
                       ("unknown tool name in '-W" & Arg (3) & ",' option");
                     raise Option_Error;
                  end if;
               else
                  if not Flags.Parse_Option (Arg) then
                     null;
                  end if;
                  --  Unknown option are saved.
                  Optc := Optc + 1;
                  Opts (Optc) := new String'(Arg);
               end if;
            else
               Argc := Argc + 1;
               Args (Argc) := new String'(Arg);
            end if;
         end;
      end loop;

      --  Set tools name.
      if Prefix_Path = null then
         Prefix_Path := new String'(Default_Pathes.Prefix);
      end if;
      if Compiler_Cmd = null then
         if Flag_Post then
            Compiler_Cmd := new String'(Default_Pathes.Compiler_Debug);
         else
            Compiler_Cmd := new String'(Default_Pathes.Compiler_Gcc);
         end if;
      end if;
      if Post_Processor_Cmd = null then
         Post_Processor_Cmd := new String'(Default_Pathes.Post_Processor);
      end if;

      if Mode /= Mode_Dispconfig then
         --  Find tools path.
         --  Since this can fail, this is not done for --dispconfig, because
         --  we want to display tools name even if they are not found.
         Setup_Compiler;
      end if;

      case Mode is
         when Mode_Analyze =>
            --  Compile each files.
            if Argc = 0 then
               Error ("no file to analyze");
               raise Option_Error;
            end if;

            for I in 1 .. Argc loop
               Do_Compile (Opts (1 .. Optc), Args (I).all);
            end loop;
         when Mode_Elaborate =>
            if Argc = 0 then
               Error ("no unit to elaborate");
               raise Option_Error;
            end if;
            if Argc = 1 then
               Do_Bind (Opts (1 .. Optc), Args (1), null);
            elsif Argc = 2 then
               Do_Bind (Opts (1 .. Optc), Args (1), Args (2));
            else
               Error ("too many units");
               raise Option_Error;
            end if;
         when Mode_Dir =>
            Do_Dir (Args (1 .. Argc));
         when Mode_Find =>
            Do_Find (Args (1 .. Argc));
         when Mode_Import =>
            Do_Import (Args (1 .. Argc));
         when Mode_Make =>
            if Argc = 1 then
               Do_Make (Opts (1 .. Optc), Args (1), null);
            elsif Argc = 2 then
               Do_Make (Opts (1 .. Optc), Args (1), Args (2));
            else
               Error ("bad option, try -h");
               raise Option_Error;
            end if;
         when Mode_Clean =>
            Do_Clean (False);
         when Mode_Remove =>
            Do_Clean (True);
         when Mode_Dispconfig =>
            Do_Dispconfig;
         when Mode_Help =>
            Do_Help;
         when Mode_Options_Help =>
            Flags.Disp_Options_Help;
         when Mode_Version =>
            null;
      end case;
      Set_Exit_Status (Success);
   exception
      when Option_Error
        | Compile_Error =>
         Set_Exit_Status (Failure);
   end Main;
end Ghdldrv;
