with System; use System;
with Ada.Unchecked_Conversion;
with Grt.Processes; use Grt.Processes;

package body Grt.Images is
   function To_Std_String_Basep is new Ada.Unchecked_Conversion
     (Source => System.Address, Target => Std_String_Basep);

   function To_Std_String_Boundp is new Ada.Unchecked_Conversion
     (Source => System.Address, Target => Std_String_Boundp);

   procedure Return_String (Res : Std_String_Ptr; Str : String)
   is
   begin
      Res.Base := To_Std_String_Basep (Ghdl_Stack2_Allocate (Str'Length));
      Res.Bounds := To_Std_String_Boundp
        (Ghdl_Stack2_Allocate (Std_String_Bound'Size / System.Storage_Unit));
      for I in 0 .. Str'Length - 1 loop
         Res.Base (Ghdl_Index_Type (I)) := Str (Str'First + I);
      end loop;
      Res.Bounds.Dim_1 := (Left => 1,
                           Right => Str'Length,
                           Dir => Dir_To,
                           Length => Str'Length);
   end Return_String;

   procedure Return_Str_Len (Res : Std_String_Ptr; Str : Ghdl_Str_Len_Type)
   is
      Char : String (1 .. 3);
   begin
      if Str.Str = null then
         Char (1) := ''';
         Char (2) := Character'Val (Str.Len);
         Char (3) := ''';
         Return_String (Res, Char);
      else
         Return_String (Res, Str.Str (1 .. Str.Len));
      end if;
   end Return_Str_Len;

   procedure Ghdl_Image_B2
     (Res : Std_String_Ptr; Val : Ghdl_B2; Type_Desc : Ghdl_Desc_Ptr)
   is
      Str : Ghdl_Str_Len_Type;
   begin
      if Val then
         Str := Type_Desc.B2.True_Name;
      else
         Str := Type_Desc.B2.False_Name;
      end if;
      Return_Str_Len (Res, Str);
   end Ghdl_Image_B2;

   procedure Ghdl_Image_E8
     (Res : Std_String_Ptr; Val : Ghdl_E8; Type_Desc : Ghdl_Desc_Ptr)
   is
   begin
      Return_Str_Len (Res, Type_Desc.E8.Values (Natural (Val)));
   end Ghdl_Image_E8;

   procedure Ghdl_Image_I32 (Res : Std_String_Ptr; Val : Ghdl_I32)
   is
      --  2**31 is ~ 2 * 10**9 -> 10 digits, with sign
      Str : String (1 .. 11);
      P : Natural;
      V : Ghdl_I32;
      D : Ghdl_I32;
   begin
      P := 11;
      V := Val;
      loop
         D := abs (V rem 10);
         V := V / 10;
         Str (P) := Character'Val (48 + D);
         exit when V = 0;
         P := P - 1;
      end loop;
      if Val < 0 then
         P := P - 1;
         Str (P) := '-';
      end if;
      Return_String (Res, Str (P .. Str'Last));
   end Ghdl_Image_I32;

   procedure Ghdl_Image_P64
     (Res : Std_String_Ptr; Val : Ghdl_P64; Type_Desc : Ghdl_Desc_Ptr)
   is
      --  2**63 is ~ 9 * 10**19 -> 20 digits, with sign
      Str : String (1 .. 21);
      P : Natural;
      V : Ghdl_P64;
      D : Ghdl_P64;
   begin
      P := Str'Last;
      if Val > 0 then
         V := -Val;
      else
         V := Val;
      end if;
      loop
         D := V rem 10;
         V := V / 10;
         Str (P) := Character'Val (48 - D);
         exit when V = 0;
         P := P - 1;
      end loop;
      if Val < 0 then
         P := P - 1;
         Str (P) := '-';
      end if;
      declare
         L : Natural := Str'Last + 1 - P;
         Str2 : String
           (1 .. L + 1 + Type_Desc.P64.Base_Name.Len);
      begin
         Str2 (1 .. L) := Str (P .. Str'Last);
         Str2 (L + 1) := ' ';
         Str2 (L + 2 .. Str2'Last) :=
           Type_Desc.P64.Base_Name.Str (1 .. Type_Desc.P64.Base_Name.Len);
         Return_String (Res, Str2);
      end;
   end Ghdl_Image_P64;

   procedure Ghdl_Image_F64 (Res : Std_String_Ptr; Val : Ghdl_F64)
   is
      --  Sign (1) + digit (1) + dot (1) + digits (15) + exp (1) + sign (1)
      --  + exp_digits (4) -> 24.
      Str : String (1 .. 25);
      P : Natural;
      V : Ghdl_F64;
      Vd : Ghdl_F64;
      Exp : Integer;
      D : Integer;
      B : Boolean;
   begin
      --  Handle sign.
      if Val < 0.0 then
         Str (1) := '-';
         P := 1;
         V := -Val;
      else
         P := 0;
         V := Val;
      end if;

      --  Compute the mantissa.
      --  FIXME: should do a dichotomy.
      if V  = 0.0 then
         Exp := 0;
      elsif V < 1.0 then
         Exp := -1;
         while V * (10.0 ** (-Exp)) < 1.0 loop
            Exp := Exp - 1;
         end loop;
      else
         Exp := 0;
         while V / (10.0 ** Exp) >= 10.0 loop
            Exp := Exp + 1;
         end loop;
      end if;

      --  Normalize VAL: in [0; 10[
      if Exp >= 0 then
         V := V / (10.0 ** Exp);
      else
         V := V * 10.0 ** (-Exp);
      end if;

      for I in 0 .. 15 loop
         Vd := Ghdl_F64'Floor (V);
         P := P + 1;
         Str (P) := Character'Val (48 + Integer (Vd));
         V := (V - Vd) * 10.0;

         if I = 0 then
            P := P + 1;
            Str (P) := '.';
         end if;
         exit when I > 0 and V < 10.0 ** (I + 1 - 15);
      end loop;

      if Exp /= 0 then
         --  LRM93 14.3
         --  if the exponent is present, the `e' is written as a lower case
         --  character.
         P := P + 1;
         Str (P) := 'e';

         if Exp < 0 then
            P := P + 1;
            Str (P) := '-';
            Exp := -Exp;
         end if;
         B := False;
         for I in 0 .. 4 loop
            D := (Exp / 10000) mod 10;
            if D /= 0 or B or I = 4 then
               P := P + 1;
               Str (P) := Character'Val (48 + D);
               B := True;
            end if;
            Exp := (Exp - D * 10000) * 10;
         end loop;
      end if;

      Return_String (Res, Str (1 .. P));
   end Ghdl_Image_F64;
end Grt.Images;
