with Ada.Command_Line; use Ada.Command_Line;
with Interfaces; use Interfaces;
with Grt.Errors; use Grt.Errors;
with Grt.Disp;
with Grt.Vcd;

package body Grt.Options is
   procedure Help
   is
      use Grt.Disp;
      procedure P (Str : String) renames Put_Line;
   begin
      P ("Usage: " & Command_Name & " [OPTIONS]");
      P ("Options are:");
      P (" --help, -h        disp this help");
      P (" --stop-time=X     stop the simulation at time X");
      P ("       X is expressed as a time value, without spaces: 1ns, ps...");
      P (" --stop-delta=X    stop the simulation cycle after X delta");
      P (" --assert-level=LEVEL   stop simulation if assert at LEVEL");
      P ("       LEVEL is note,warning,error,failure,none");
      P (" --disp-tree       disp the design hierarchy after elaboration");
      P (" --expect-failure  invert exit status");
      P (" --stack-size=X    set the stack size of non-sensitized processes");
      P (" --stack-max-size=X  set the maximum stack size");
      Grt.Vcd.Vcd_Help;
      P ("trace options:");
      P (" --trace-signals   disp signals after each cycle");
      P (" --trace-processes disp process name before each cycle");
      P ("debug options:");
      P (" --disp-order      disp signals order");
      P (" --disp-sources    disp sources while displaying signals");
      P (" --disp-sig-types  disp signal types");
      P (" --checks          do internal checks after each process");
   end Help;

   --  Extract from STR a number.
   --  First, all leading blanks are skipped.
   --  Then, all next digits are eaten.
   --  The position of the first non digit or one past the upper bound is
   --  returned into POS.
   --  If there is no digits, OK is set to false, else to true.
   procedure Extract_Integer
     (Str : String;
      Ok : out Boolean;
      Result : out Integer_64;
      Pos : out Natural)
   is
   begin
      Pos := Str'First;
      --  Skip blanks.
      while Pos <= Str'Last and then Str (Pos) = ' ' loop
         Pos := Pos + 1;
      end loop;
      Ok := False;
      Result := 0;
      loop
         exit when Pos > Str'Last or else Str (Pos) not in '0' .. '9';
         Ok := True;
         Result := Result * 10
           + (Character'Pos (Str (Pos)) - Character'Pos ('0'));
         Pos := Pos + 1;
      end loop;
   end Extract_Integer;

   function Extract_Size (Str : String; Option_Name : String) return Natural
   is
      Ok : Boolean;
      Val : Integer_64;
      Pos : Natural;
   begin
      Extract_Integer (Str, Ok, Val, Pos);
      if not Ok then
         Val := 1;
      end if;
      if Pos > Str'Last then
         --  No suffix.
         return Natural (Val);
      end if;
      if Pos = Str'Last
        or else (Pos + 1 = Str'Last
                 and then (Str (Pos + 1) = 'b' or Str (Pos + 1) = 'o'))
      then
         if Str (Pos) = 'k' or Str (Pos) = 'K' then
            return Natural (Val) * 1024;
         elsif Str (Pos) = 'm' or Str (Pos) = 'M' then
            return Natural (Val) * 1024 * 1024;
         end if;
      end if;
      Error ("bad memory unit for option " & Option_Name);
   end Extract_Size;

   function To_Lower (C : Character) return Character is
   begin
      if C in 'A' .. 'Z' then
         return Character'Val (Character'Pos (C) + 32);
      else
         return C;
      end if;
   end To_Lower;

   procedure Decode (Stop : out Boolean) is
   begin
      Stop := False;
      for I in 1 .. Argument_Count loop
         if Argument (I) = "--help" or else Argument (I) = "-h" then
            Help;
            Stop := True;
         elsif Argument (I) = "--disp-tree" then
            Disp_Design_Tree := True;
         elsif Argument (I) = "--trace-signals" then
            Trace_Signals := True;
         elsif Argument (I) = "--trace-processes" then
            Trace_Processes := True;
         elsif Argument (I) = "--disp-order" then
            Disp_Signals_Order := True;
         elsif Argument (I) = "--checks" then
            Checks := True;
         elsif Argument (I) = "--disp-sources" then
            Disp_Sources := True;
         elsif Argument (I) = "--disp-sig-types" then
            Disp_Sig_Types := True;
         elsif Argument (I)'Length >= 12
           and then Argument (I)(1 .. 12) = "--stop-time="
         then
            declare
               Ok : Boolean;
               Pos : Natural;
               Time : Integer_64;
               Unit : String (1 .. 3);
            begin
               Extract_Integer (Argument (I)(13 .. Argument (I)'Last),
                                Ok, Time, Pos);
               if not Ok then
                  Time := 1;
               end if;
               if (Argument (I)'Last - Pos + 1) not in 2 .. 3 then
                  Error ("bad unit for '" & Argument (I) & "'");
                  return;
               end if;
               Unit (1) := To_Lower (Argument (I)(Pos));
               Unit (2) := To_Lower (Argument (I)(Pos + 1));
               if Argument (I)'Last = Pos + 2 then
                  Unit (3) := To_Lower (Argument (I)(Pos + 2));
               else
                  Unit (3) := ' ';
               end if;
               if Unit = "fs " then
                  null;
               elsif Unit = "ps " then
                  Time := Time * (10 ** 3);
               elsif Unit = "ns " then
                  Time := Time * (10 ** 6);
               elsif Unit = "us " then
                  Time := Time * (10 ** 9);
               elsif Unit = "ms " then
                  Time := Time * (10 ** 12);
               elsif Unit = "sec" then
                  Time := Time * (10 ** 15);
               elsif Unit = "min" then
                  Time := Time * (10 ** 15) * 60;
               elsif Unit = "hr " then
                  Time := Time * (10 ** 15) * 3600;
               else
                  Error ("bad unit name for '" & Argument (I) & "'");
               end if;
               Stop_Time := Std_Time (Time);
            end;
         elsif Argument (I)'Length >= 13
           and then Argument (I)(1 .. 13) = "--stop-delta="
         then
            declare
               Ok : Boolean;
               Pos : Natural;
               Time : Integer_64;
            begin
               Extract_Integer (Argument (I)(13 .. Argument (I)'Last),
                                Ok, Time, Pos);
               if not Ok or else Pos <= Argument (I)'Last then
                  Error ("bad value in '" & Argument (I) & "'");
               else
                  if Time > Integer_64 (Integer'Last) then
                     Stop_Delta := Integer'Last;
                  else
                     Stop_Delta := Integer (Time);
                  end if;
               end if;
            end;
         elsif Argument (I)'Length > 15
           and then Argument (I)(1 .. 15) = "--assert-level="
         then
            if Argument (I)(16 .. Argument (I)'Last) = "note" then
               Severity_Level := Note_Severity;
            elsif Argument (I)(16 .. Argument (I)'Last) = "warning" then
               Severity_Level := Warning_Severity;
            elsif Argument (I)(16 .. Argument (I)'Last) = "error" then
               Severity_Level := Error_Severity;
            elsif Argument (I)(16 .. Argument (I)'Last) = "failure" then
               Severity_Level := Failure_Severity;
            elsif Argument (I)(16 .. Argument (I)'Last) = "none" then
               Severity_Level := 4;
            else
               Error ("bad argument for --assert-level= option, try --help");
            end if;
         elsif Argument (I) = "--expect-failure" then
            Expect_Failure := True;
         elsif Argument (I)'Length >= 13
           and then Argument (I)(1 .. 13) = "--stack-size="
         then
            Stack_Size := Extract_Size
              (Argument (I) (14 .. Argument (I)'Last), "--stack-size");
            if Stack_Size > Stack_Max_Size then
               Stack_Max_Size := Stack_Size;
            end if;
         elsif Argument (I)'Length >= 17
           and then Argument (I)(1 .. 17) = "--stack-max-size="
         then
            Stack_Max_Size := Extract_Size
              (Argument (I) (18 .. Argument (I)'Last), "--stack-size");
            if Stack_Size > Stack_Max_Size then
               Stack_Size := Stack_Max_Size;
            end if;
         elsif not Grt.Vcd.Vcd_Option (Argument (I)) then
               Error ("unknown option '" & Argument (I) & "', try --help");
         end if;
      end loop;
   end Decode;
end Grt.Options;
