with Types; use Types;
with Errorout; use Errorout;
with Std_Package; use Std_Package;
with Libraries;
with Std_Names;
with Sem_Name; use Sem_Name;
with Sem_Expr; use Sem_Expr;
with Iirs_Utils; use Iirs_Utils;
with Flags;
with Name_Table;
with Tokens;
with Evaluation; use Evaluation;

package body Sem is
   -- Forward declarations.
   procedure Sem_Concurrent_Statement_List
     (List : Iir_Concurrent_Statement_List; Is_Passive : Boolean);
   procedure Sem_Declaration_List (Decls: Iir_Declaration_List; Parent : Iir);
   procedure Sem_Specification_List
     (Decls: Iir_Declaration_List;
      Stmts: Iir_Concurrent_Statement_List);
   procedure Sem_Sequential_Statements_Internal
     (Stmts: Iir_Sequential_Statement_List);
   function Range_To_Subtype_Definition (A_Range: Iir) return Iir;
   procedure Sem_Context_Clauses (Design_Unit: Iir_Design_Unit);
   procedure Sem_Use_Clause (Clause: Iir_Use_Clause);
   procedure Sem_Block_Configuration
     (Block_Conf : Iir_Block_Configuration; Father: Iir);
   procedure Sem_Component_Specification
     (Stmts : Iir_Concurrent_Statement_List; Spec : Iir);
   procedure Sem_Binding_Indication (Bind : Iir_Binding_Indication;
                                     Comp : Iir_Component_Declaration;
                                     Parent : Iir);

   --  Create a default binding indication for component COMP which will be
   --  bound with entity ENTITY_UNIT.
   --  If ENTITY_UNIT is NULL_IIR, the component is not bound.
   --  If FORCE is True, a binding indication will be created even if the
   --   component is not bound (this is an open binding indication).
   --  PARENT is used to report error.
   function Sem_Create_Default_Binding_Indication
     (Comp : Iir_Component_Declaration;
      Entity_Unit : Iir_Design_Unit;
      Parent : Iir;
      Force : Boolean)
     return Iir_Binding_Indication;

   function Get_Visible_Entity_Declaration (Comp: Iir_Component_Declaration)
     return Iir_Design_Unit;
   --  Explain why there is no default binding for COMP.
   procedure Explain_No_Visible_Entity (Comp: Iir_Component_Declaration);

   --  Return TRUE iff LEFT and RIGHT are (in depth) equal.
   --  This corresponds to conformance rules, LRM93 2.7
   function Are_Trees_Equal (Left, Right : Iir) return Boolean;

   --  The kind of an inteface list.
   type Interface_Kind_Type is (Interface_Generic, Interface_Port,
                                Interface_Procedure, Interface_Function);

   -- Access to the current subprogram or process.
   Current_Subprogram: Iir := Null_Iir;

   -- Access to the current concurrent statement.
   -- Null_iir if no one.
   Current_Concurrent_Statement : Iir := Null_Iir;

   -- Semantic associed with signal mode.
   -- See 4.3.3
   type Boolean_Array_Of_Iir_Mode is array (Iir_Mode) of Boolean;
   Iir_Mode_Readable : constant Boolean_Array_Of_Iir_Mode :=
     (Iir_Unknown_Mode => False,
      Iir_In_Mode => True,
      Iir_Out_Mode => False,
      Iir_Inout_Mode => True,
      Iir_Buffer_Mode => True,
      Iir_Linkage_Mode => False);
   Iir_Mode_Writable : constant Boolean_Array_Of_Iir_Mode :=
     (Iir_Unknown_Mode => False,
      Iir_In_Mode => False,
      Iir_Out_Mode => True,
      Iir_Inout_Mode => True,
      Iir_Buffer_Mode => True,
      Iir_Linkage_Mode => False);

   --  Some signals are implicitly declared.  This is the case for signals
   --  declared by an attribute ('stable, 'quiet and 'transaction).
   --  Note: guard signals are also implicitly declared, but with a guard
   --   expression, which is located.
   --  Since these signals need resources and are not easily located (can be
   --  nearly in every expression), it is useful to add a node into a
   --  declaration list to declare them.
   --  However, only a few declaration_list can declare signals.  These
   --  declarations lists must register and unregister themselves with
   --  push_declarative_region_with_signals and
   --  pop_declarative_region_with_signals.
   type Implicit_Signal_Declaration_Type is record
      Declaration_List : Iir_Declaration_List;
      Implicit_Chain : Iir;
   end record;
   Current_Declarative_Region_With_Signals :
     Implicit_Signal_Declaration_Type := (null, Null_Iir);

   procedure Push_Declarative_Region_With_Signals
     (Cell: out Implicit_Signal_Declaration_Type;
      Decls : Iir_Declaration_List) is
   begin
      Cell := Current_Declarative_Region_With_Signals;
      Current_Declarative_Region_With_Signals := (Decls , Null_Iir);
   end Push_Declarative_Region_With_Signals;

   procedure Pop_Declarative_Region_With_Signals
     (Cell: in Implicit_Signal_Declaration_Type) is
   begin
      Current_Declarative_Region_With_Signals := Cell;
   end Pop_Declarative_Region_With_Signals;

   procedure Add_Declaration_For_Implicit_Signal (Sig : Iir) is
      Chain : Iir renames
        Current_Declarative_Region_With_Signals.Implicit_Chain;
   begin
      if Current_Declarative_Region_With_Signals.Declaration_List = null then
         raise Internal_Error;
      end if;
      if Chain = null then
         Chain := Create_Iir (Iir_Kind_Implicit_Signal_Declaration);
         Location_Copy
           (Chain,
            Current_Declarative_Region_With_Signals.Declaration_List);
         Append_Element
           (Current_Declarative_Region_With_Signals.Declaration_List, Chain);
      end if;
      Set_Implicit_Signal_Chain (Sig, Get_Implicit_Signal_Chain (Chain));
      Set_Implicit_Signal_Chain (Chain, Sig);
   end Add_Declaration_For_Implicit_Signal;

   function Get_Current_Subprogram return Iir is
   begin
      return Current_Subprogram;
   end Get_Current_Subprogram;

   procedure Add_Dependence (Unit : Iir) is
   begin
      Add_Dependence (Get_Current_Design_Unit, Unit);
   end Add_Dependence;

   -- Return TRUE iff FUNC is a resolution function for ATYPE.
   function Is_A_Resolution_Function (Func: Iir; Atype: Iir) return Boolean
   is
      Decl_List: Iir_List;
      Decl: Iir_List;
      Arg: Iir;
   begin
      -- LRM93 2.4
      --  A resolution function must be a [pure] function;
      if Get_Kind (Func) not in Iir_Kinds_Function_Specification then
         return False;
      end if;
      Decl_List := Get_Interface_Declaration (Func);
      -- LRM93 2.4
      --  moreover, it must have a single input parameter of class constant
      if Get_Nbr_Elements (Decl_List) /= 1 then
         return False;
      end if;
      Arg := Get_First_Element (Decl_List);
      if Get_Kind (Arg) /= Iir_Kind_Constant_Interface_Declaration then
         return False;
      end if;
      -- LRM93 2.4
      --  that is a one-dimensional, unconstrained array
      Decl := Get_Type (Arg);
      if Get_Kind (Decl) /= Iir_Kind_Array_Type_Definition then
         return False;
      end if;
      Decl_List := Get_Index_List (Decl);
      if Get_Nbr_Elements (Decl_List) /= 1 then
         return False;
      end if;
      -- LRM93 2.4
      --  whose element type is that of the resolved signal.
      Decl := Get_Element_Subtype (Decl);
      if Get_Base_Type (Decl) /= Get_Base_Type (Atype) then
         return False;
      end if;
      -- LRM93 2.4
      --  The type of the return value of the function must also be that of
      --  the signal.
      Decl := Get_Return_Type (Func);
      if Get_Base_Type (Decl) /= Get_Base_Type (Atype) then
         return False;
      end if;
      -- LRM93 2.4
      --  A resolution function must be a [pure] function;
      if Flags.Vhdl_Std >= Vhdl_93 and then Get_Pure_Flag (Func) = False then
         Error_Msg_Sem
           ("resolution " & Disp_Node (Func) & " must be pure", Atype);
         return False;
      end if;
      return True;
   end Is_A_Resolution_Function;

   procedure Sem_Resolution_Function (Decl: Iir) is
      use Sem_Name;
      Func: Iir;
      Name : Iir;
      Res: Iir;
      Inter: Name_Interpretation_Type;
   begin
      Name := Get_Resolution_Function (Decl);
      if Name = Null_Iir then
         -- This is not a resolved type.
         return;
      elsif Get_Kind (Name) in Iir_Kinds_Function_Declaration then
         --  The resolution function was already semantized.
         --  This can happen if comes from an unconstrained array subtype.
         return;
      end if;

      Res := Null_Iir;
      -- FIXME
      Inter := Get_Interpretation (Get_Identifier (Name));
      while Inter /= No_Name_Interpretation loop
         Func := Get_Declaration (Inter);
         if Is_A_Resolution_Function (Func, Decl) then
            if Res /= Null_Iir then
               Error_Msg_Sem
                 ("can't resolve overload for resolution function", Decl);
               return;
            end if;
            Res := Func;
            Set_Resolution_Function (Decl, Func);
            Set_Resolved_Flag (Decl, True);
         end if;
         Inter := Get_Next_Interpretation (Inter);
      end loop;
      if Res = Null_Iir then
         --  FIXME: better message.
         Error_Msg_Sem ("cannot find resolution function", Name);
      end if;
      Free_Name (Name);
   end Sem_Resolution_Function;

   --  Semantize array_subtype_definition DEF using BASE_TYPE as the base type
   --  of DEF.
   --  DEF must have an index list and may have a resolution function.
   --  Return DEF.
   function Sem_Array_Subtype_Indication (Base_Type : Iir; Def : Iir)
     return Iir
   is
      Type_Index, Subtype_Index: Iir;
      Staticness : Iir_Staticness;
      Error_Seen : Boolean;
      Type_Index_List : Iir_List;
      Subtype_Index_List : Iir_List;
   begin
      case Get_Kind (Base_Type) is
         when Iir_Kind_Array_Type_Definition
           | Iir_Kind_Unconstrained_Array_Subtype_Definition =>
            null;
         when others =>
            Error_Msg_Sem
              (Disp_Node (Base_Type) & " cannot be constrained", Def);
            --  Continue as if BASE_TYPE is really a base type, it is safe.
      end case;

      Staticness := Get_Type_Staticness (Get_Element_Subtype (Base_Type));
      Error_Seen := False;
      Type_Index_List := Get_Index_List (Base_Type);
      Subtype_Index_List := Get_Index_List (Def);
      for I in Natural loop
         Type_Index := Get_Nth_Element (Type_Index_List, I);
         Subtype_Index := Get_Nth_Element (Subtype_Index_List, I);
         exit when Type_Index = null and Subtype_Index = null;

         if Type_Index = Null_Iir then
            Error_Msg_Sem ("subtype has more indexes than "
                           & Disp_Node (Base_Type)
                           & " defined at " & Disp_Location (Base_Type),
                           Subtype_Index);
            --  Forget extra indexes.
            Set_Nbr_Elements (Subtype_Index_List, I);
            exit;
         end if;
         if Subtype_Index = Null_Iir then
            if not Error_Seen then
               Error_Msg_Sem ("subtype has less indexes than "
                              & Disp_Node (Base_Type)
                              & " defined at " & Disp_Location (Base_Type),
                              Def);
               Error_Seen := True;
            end if;
            --  Use type_index as a fake subtype
            --  FIXME: it is too fake.
            Append_Element (Subtype_Index_List, Type_Index);
            Staticness := None;
         else
            Subtype_Index := Sem_Discrete_Range_Expression
              (Subtype_Index, Type_Index);
            if Subtype_Index /= Null_Iir then
               Subtype_Index := Range_To_Subtype_Definition (Subtype_Index);
               Staticness := Min (Staticness,
                                  Get_Type_Staticness (Subtype_Index));
            end if;
            if Subtype_Index = Null_Iir then
               --  Create a fake subtype from type_index.
               --  FIXME: It is too fake.
               Subtype_Index := Type_Index;
               Staticness := None;
            end if;
            Replace_Nth_Element (Subtype_Index_List, I, Subtype_Index);
         end if;
      end loop;
      Set_Type_Staticness (Def, Staticness);
      Set_Element_Subtype (Def, Get_Element_Subtype (Base_Type));
      Sem_Resolution_Function (Def);
      if Get_Resolution_Function (Def) /= Null_Iir
        or else Get_Resolved_Flag (Get_Element_Subtype (Base_Type))
      then
         Set_Resolved_Flag (Def, True);
      else
         Set_Resolved_Flag (Def, False);
      end if;
      Set_Type_Mark (Def, Base_Type);
      Set_Signal_Type_Flag (Def, Get_Signal_Type_Flag (Base_Type));
      return Def;
   end Sem_Array_Subtype_Indication;

   --  Semantize a subtype indication.
   --  DEF can be either a name or an iir_subtype_definition.
   --  Return a new (an anonymous) subtype definition (with the correct kind),
   --  or an already defined type definition (if DEF is a name).
   function Sem_Subtype_Indication (Def: Iir; Incomplete : Boolean := False)
     return Iir
   is
      Type_Mark: Iir;
      Res: Iir;
      Decl_Kind : Decl_Kind_Type;
   begin
      if Incomplete then
         Decl_Kind := Decl_Incomplete_Type;
      else
         Decl_Kind := Decl_Type;
      end if;

      -- Simple case that correspond to no indication except a subtype
      -- identifier
      if Get_Kind (Def) in Iir_Kinds_Name then
         return Find_Declaration (Def, Decl_Kind);
      end if;

      --  Semantize the type mark.
      Type_Mark := Find_Declaration (Get_Base_Type (Def), Decl_Kind);
      if Type_Mark = Null_Iir then
         --  FIXME: handle inversion such as "subtype BASETYPE RESOLV", which
         --  should emit "resolution function must precede type name".
         return Null_Iir;
      end if;
      Set_Base_Type (Def, Type_Mark);

      case Get_Kind (Type_Mark) is
         when Iir_Kind_Array_Subtype_Definition
           | Iir_Kind_Array_Type_Definition
           | Iir_Kind_Unconstrained_Array_Subtype_Definition =>
            --  If the base type is an unconstrained array subtype, then get
            --  the *real* base type, and copy the resolution function (since
            --  a base type has no resolution function).
            if Get_Kind (Type_Mark) =
              Iir_Kind_Unconstrained_Array_Subtype_Definition
            then
               if Get_Resolution_Function (Def) = null then
                  if Get_Range_Constraint (Def) = null then
                     --  In this case, DEF must simply be a name.  There is
                     --  a parser internal error.
                     raise Internal_Error;
                  end if;
                  Set_Resolution_Function
                    (Def, Get_Resolution_Function (Type_Mark));
               end if;
               Type_Mark := Get_Base_Type (Type_Mark);
               Set_Base_Type (Def, Get_Base_Type (Type_Mark));
            end if;

            if Get_Kind (Type_Mark) = Iir_Kind_Array_Subtype_Definition then
               Set_Base_Type (Def, Get_Base_Type (Type_Mark));
            end if;

            if Get_Kind (Def) = Iir_Kind_Subtype_Definition then
               -- This is the case of "subtype new_array is [func] old_array".
               -- def must be a constrained array.
               if Get_Range_Constraint (Def) /= Null_Iir then
                  Error_Msg_Sem
                    ("cannot use a range constraint for an array", Def);
                  return Type_Mark;
               end if;
               if Get_Resolution_Function (Def) = null then
                  --  In this case, DEF must simply be a name.  There is
                  --  a parser internal error.
                  raise Internal_Error;
               end if;
               case Get_Kind (Type_Mark) is
                  when Iir_Kind_Array_Type_Definition =>
                     Res := Create_Iir
                       (Iir_Kind_Unconstrained_Array_Subtype_Definition);
                  when Iir_Kind_Array_Subtype_Definition =>
                     Res := Create_Iir (Iir_Kind_Array_Subtype_Definition);
                     Set_Element_Subtype
                       (Res, Get_Element_Subtype (Type_Mark));
                     Free_Old_Iir (Get_Index_List (Res));
                     Set_Index_List (Res, Get_Index_List (Type_Mark));
                  when others =>
                     Error_Kind ("sem_subtype_indication(array)", Type_Mark);
               end case;
               Location_Copy (Res, Def);
               Set_Base_Type (Res, Get_Base_Type (Type_Mark));
               Set_Resolution_Function (Res, Get_Resolution_Function (Def));
               Set_Type_Staticness (Res, Get_Type_Staticness (Type_Mark));
               Sem_Resolution_Function (Res);
               Set_Signal_Type_Flag (Res, Get_Signal_Type_Flag (Type_Mark));
               if Get_Resolution_Function (Res) /= Null_Iir
                 or else Get_Resolved_Flag (Get_Element_Subtype (Type_Mark))
               then
                  Set_Resolved_Flag (Res, True);
               else
                  Set_Resolved_Flag (Res, False);
               end if;
               Set_Type_Mark (Res, Type_Mark);
               Free_Name (Def);
               return Res;
            elsif Get_Kind (Def) = Iir_Kind_Array_Subtype_Definition then
               -- Case of a constraint for an array.
               -- Check each index constraint against array type.
               return Sem_Array_Subtype_Indication (Type_Mark, Def);
            else
               Error_Kind ("sem_subtype_indication(1)", Def);
               return Type_Mark;
            end if;

         when Iir_Kind_Integer_Subtype_Definition
           | Iir_Kind_Floating_Subtype_Definition
           | Iir_Kind_Enumeration_Subtype_Definition
           | Iir_Kind_Physical_Subtype_Definition =>
            if Get_Range_Constraint (Def) = Null_Iir
              and then Get_Resolution_Function (Def) = Null_Iir
            then
               --  This defines an alias, and must have been handled just
               --  before the case statment.
               raise Internal_Error;
            end if;
            declare
               A_Range : Iir;
            begin
               -- There are limits.  Create a new subtype.
               Res := Create_Iir (Get_Kind (Type_Mark));
               Location_Copy (Res, Def);
               Set_Base_Type (Res, Get_Base_Type (Type_Mark));
               Set_Type_Mark (Res, Type_Mark);
               Set_Resolution_Function (Res, Get_Resolution_Function (Def));
               A_Range := Get_Range_Constraint (Def);
               if A_Range = Null_Iir then
                  A_Range := Get_Range_Constraint (Type_Mark);
               else
                  A_Range := Sem_Discrete_Range_Expression
                    (A_Range, Type_Mark);
                  if A_Range = Null_Iir then
                     return Null_Iir;
                  end if;
               end if;
               Set_Range_Constraint (Res, A_Range);
               Set_Type_Staticness (Res, Get_Expr_Staticness (A_Range));
               Free_Name (Def);
               Sem_Resolution_Function (Res);
               Set_Resolved_Flag
                 (Res, Get_Resolution_Function (Res) /= Null_Iir);
               Set_Signal_Type_Flag (Res, Get_Signal_Type_Flag (Type_Mark));
               return Res;
            end;

         when Iir_Kind_Enumeration_Type_Definition =>
            if Get_Range_Constraint (Def) = Null_Iir and then
              Get_Resolution_Function (Def) = Null_Iir
            then
               raise Internal_Error;
            end if;

            declare
               Constraint : Iir_Range_Expression;
            begin
               -- There are limits.  Create a new subtype.
               Res := Create_Iir (Iir_Kind_Enumeration_Subtype_Definition);
               Location_Copy (Res, Def);
               Set_Base_Type (Res, Type_Mark);
               Set_Type_Mark (Res, Type_Mark);
               Set_Resolution_Function (Res, Get_Resolution_Function (Def));
               Constraint := Get_Range_Constraint (Def);
               if Constraint = Null_Iir then
                  Constraint := Get_Range_Constraint (Type_Mark);
               else
                  Constraint := Sem_Discrete_Range_Expression
                    (Constraint, Type_Mark);
                  -- FIXME: check bounds, check static
               end if;
               Set_Range_Constraint (Res, Constraint);
               Set_Type_Staticness (Res, Get_Expr_Staticness (Constraint));
            end;
            Free_Name (Def);
            Sem_Resolution_Function (Res);
            Set_Resolved_Flag
              (Res, Get_Resolution_Function (Res) /= Null_Iir);
            Set_Signal_Type_Flag (Res, True);
            return Res;

         when Iir_Kind_Record_Type_Definition =>
            declare
               Func: Iir;
            begin
               if Get_Kind (Def) /= Iir_Kind_Subtype_Definition then
                  Error_Kind ("sem_subtype_indication1", Def);
                  return null;
               end if;
               Func := Get_Resolution_Function (Def);
               if Func = Null_Iir then
                  --  This is an alias.
                  raise Internal_Error;
               end if;
               Res := Create_Iir (Iir_Kind_Record_Subtype_Definition);
               Location_Copy (Res, Def);
               Set_Base_Type (Res, Type_Mark);
               Set_Type_Staticness (Res, Get_Type_Staticness (Type_Mark));
               Set_Type_Mark (Res, Type_Mark);
               Set_Resolution_Function (Res, Func);
               Sem_Resolution_Function (Res);
               Set_Resolved_Flag (Res, Func /= Null_Iir);
               Set_Signal_Type_Flag (Res, Get_Signal_Type_Flag (Type_Mark));
               Free_Name (Def);
               return Res;
            end;

         when Iir_Kind_Access_Type_Definition =>
            --  LRM93 4.2
            --  A subtype indication denoting an access type [or a file type]
            --  may not contain a resolution function.
            if Get_Resolution_Function (Def) /= Null_Iir then
               Error_Msg_Sem
                 ("resolution function not allowed for an access type", Def);
            end if;

            case Get_Kind (Def) is
               when Iir_Kind_Subtype_Definition =>
                  Free_Name (Def);
                  return Type_Mark;
               when Iir_Kind_Array_Subtype_Definition =>
                  --  LRM93 3.3
                  --  The only form of constraint that is allowed after a name
                  --  of an access type in a subtype indication is an index
                  --  constraint.
                  declare
                     Sub_Type : Iir;
                     Base_Type : Iir;
                  begin
                     Base_Type := Get_Designated_Type (Type_Mark);
                     Sub_Type := Sem_Array_Subtype_Indication (Base_Type, Def);
                     Res := Create_Iir (Iir_Kind_Access_Subtype_Definition);
                     Location_Copy (Res, Def);
                     Set_Base_Type (Res, Type_Mark);
                     Set_Signal_Type_Flag (Res, False);
                     Free_Old_Iir (Def);
                     return Res;
                  end;
               when others =>
                  raise Internal_Error;
            end case;

         when Iir_Kind_File_Type_Definition =>
            if Get_Kind (Def) = Iir_Kind_Subtype_Definition then
               Free_Name (Def);
               return Type_Mark;
            else
               raise Internal_Error;
            end if;

         when others =>
            Error_Kind ("sem_subtype_indication", Type_Mark);
            return Def;
      end case;
   end Sem_Subtype_Indication;

   --  Sem a range expression.
   --  Both left and right bounds must be of the same type kind, ie
   --  integer types, or if INT_ONLY is false, real types.
   --  However, the two bounds need not have the same type.
   function Sem_Range_Expression (Expr : Iir; Int_Only : Boolean) return Iir
   is
      Left, Right: Iir;
      Bt_L_Kind, Bt_R_Kind : Iir_Kind;
   begin
      Left := Sem_Expression (Get_Left_Limit (Expr), Null_Iir);
      Right := Sem_Expression (Get_Right_Limit (Expr), Null_Iir);
      if Left = Null_Iir or Right = Null_Iir then
         return Null_Iir;
      end if;
      Set_Left_Limit (Expr, Left);
      Set_Right_Limit (Expr, Right);

      Set_Expr_Staticness (Expr, Min (Get_Expr_Staticness (Left),
                                      Get_Expr_Staticness (Right)));

      Bt_L_Kind := Get_Kind (Get_Base_Type (Get_Type (Left)));
      Bt_R_Kind := Get_Kind (Get_Base_Type (Get_Type (Right)));

      if Int_Only then
         if Bt_L_Kind /= Iir_Kind_Integer_Type_Definition
           and then Bt_R_Kind = Iir_Kind_Integer_Type_Definition
         then
            Error_Msg_Sem ("left bound must be an integer expression", Left);
            return Null_Iir;
         end if;
         if Bt_R_Kind /= Iir_Kind_Integer_Type_Definition
           and then Bt_L_Kind = Iir_Kind_Integer_Type_Definition
         then
            Error_Msg_Sem ("right bound must be an integer expression", Left);
            return Null_Iir;
         end if;
         if Bt_R_Kind /= Iir_Kind_Integer_Type_Definition
           and then Bt_L_Kind /= Iir_Kind_Integer_Type_Definition
         then
            Error_Msg_Sem ("each bound must be an integer expression", Expr);
            return Null_Iir;
         end if;
      else
         if Bt_L_Kind /= Bt_R_Kind then
            Error_Msg_Sem ("left and right bounds must be of the same type",
                           Expr);
            return Null_Iir;
         end if;
         case Bt_L_Kind is
            when Iir_Kind_Integer_Type_Definition
              | Iir_Kind_Floating_Type_Definition =>
               null;
         when others =>
            --  Enumeration range are not allowed to define a new type.
            Error_Msg_Sem
              ("bad range type, only integer or float is allowed", Expr);
            return Null_Iir;
         end case;
      end if;

      return Expr;
   end Sem_Range_Expression;

   -- Sem a physical type definition.  Create a subtype.
   function Sem_Physical_Type_Definition (Range_Expr: Iir; Decl : Iir)
      return Iir_Physical_Subtype_Definition
   is
      Unit: Iir_Physical_Unit;
      Def : Iir_Physical_Type_Definition;
      Sub_Type: Iir_Physical_Subtype_Definition;
      Range_Expr1: Iir;
      Unit_List : Iir_List;
      Val : Iir;
   begin
      Def := Get_Type (Range_Expr);

      --  LRM93 4.1
      --  The simple name declared by a type declaration denotes the
      --  declared type, unless the type declaration declares both a base
      --  type and a subtype of the base type, in which case the simple name
      --  denotes the subtype, and the base type is anonymous.
      Set_Type_Declarator (Def, Decl);
      Set_Base_Type (Def, Def);
      Set_Resolved_Flag (Def, False);
      Set_Type_Staticness (Def, Locally);
      Set_Signal_Type_Flag (Def, True);

      --  LRM93 3.1.3
      --  Each bound of a range constraint that is used in a physical type
      --  definition must be a locally static expression of some integer type
      --  but the two bounds need not have the same integer type.
      case Get_Kind (Range_Expr) is
         when Iir_Kind_Range_Expression =>
            Range_Expr1 := Sem_Range_Expression (Range_Expr, True);
         when others =>
            Error_Kind ("sem_physical_type_definition", Range_Expr);
      end case;
      if Range_Expr1 /= Null_Iir then
         if Get_Expr_Staticness (Range_Expr1) /= Locally then
            Error_Msg_Sem
              ("range constraint for a physical type must be static",
               Range_Expr1);
            Range_Expr1 := Null_Iir;
         else
            Range_Expr1 := Eval_Expr (Range_Expr1);
         end if;
      end if;

      --  Create the subtype.
      Sub_Type := Create_Iir (Iir_Kind_Physical_Subtype_Definition);
      Location_Copy (Sub_Type, Range_Expr);
      Set_Base_Type (Sub_Type, Def);
      Set_Signal_Type_Flag (Sub_Type, True);

      --  Sem primary units.
      Unit := Get_Primary_Unit (Def);
      Set_Value (Unit, 1);
      Sem_Name.Add_Name (Unit);
      Set_Type (Unit, Def);
      Set_Expr_Staticness (Unit, Locally);
      Sem_Name.Name_Visible (Unit);

      --  Sem secondary units.
      Unit_List := Get_Unit_List (Def);
      for I in Natural loop
         Unit := Get_Nth_Element (Unit_List, I);
         exit when Unit = null;
         --  Val := Sem_Physical_Literal (Get_Multiplier (Unit));
         Val := Sem_Expression (Get_Multiplier (Unit), Def);
         if Val /= Null_Iir then
            Val := Eval_Expr (Val);
            Set_Multiplier (Unit, Val);
            Set_Value (Unit, Get_Value (Val));

            --  LRM93 3.1
            --  The position number of unit names need not lie within the range
            --  specified by the range constraint.
            --  GHDL: this was not true in VHDL87.
            --  GHDL: This is not so simple if 1 is not included in the range.
            if False and then Flags.Vhdl_Std = Vhdl_87
              and then Range_Expr1 /= Null_Iir
            then
               if not Eval_Int_In_Range (Get_Value (Unit), Range_Expr1) then
                  Error_Msg_Sem
                    ("physical literal does not lie within the range", Unit);
               end if;
            end if;
         else
            --  Avoid errors storm.
            Set_Multiplier (Unit, Get_Primary_Unit (Def));
            Set_Value (Unit, 1);
         end if;

         Sem_Name.Add_Name (Unit);
         Set_Type (Unit, Def);
         Set_Expr_Staticness (Unit, Locally);
         Sem_Name.Name_Visible (Unit);
      end loop;

      if Range_Expr1 /= Null_Iir then
         declare
            --  Convert an integer literal to a physical literal.
            --  This is used to convert bounds.
            function Lit_To_Phys_Lit (Lim : Iir_Integer_Literal)
              return Iir_Physical_Int_Literal
            is
               Res : Iir_Physical_Int_Literal;
            begin
               Res := Create_Iir (Iir_Kind_Physical_Int_Literal);
               Location_Copy (Res, Lim);
               Set_Type (Res, Def);
               Set_Value (Res, Get_Value (Lim));
               Set_Name (Res, Get_Primary_Unit (Def));
               Set_Expr_Staticness (Res, Locally);
               Set_Literal_Origin (Res, Lim);
               return Res;
            end Lit_To_Phys_Lit;

            Phys_Range : Iir_Range_Expression;
         begin
            --  Create the physical range.
            Phys_Range := Create_Iir (Iir_Kind_Range_Expression);
            Location_Copy (Phys_Range, Range_Expr1);
            Set_Type (Phys_Range, Def);
            Set_Direction (Phys_Range, Get_Direction (Range_Expr1));
            Set_Left_Limit
              (Phys_Range, Lit_To_Phys_Lit (Get_Left_Limit (Range_Expr1)));
            Set_Right_Limit
              (Phys_Range, Lit_To_Phys_Lit (Get_Right_Limit (Range_Expr1)));
            Set_Expr_Staticness
              (Phys_Range, Get_Expr_Staticness (Range_Expr1));

            Set_Range_Constraint (Sub_Type, Phys_Range);
            --  This must be locally...
            Set_Type_Staticness (Sub_Type, Get_Expr_Staticness (Range_Expr1));
         end;
      end if;
      Set_Resolved_Flag (Sub_Type, False);

      return Sub_Type;
   end Sem_Physical_Type_Definition;

   function Create_Integer_Type (Loc : Iir; Constraint : Iir; Decl : Iir)
     return Iir
   is
      Ntype: Iir_Integer_Subtype_Definition;
      Ndef: Iir_Integer_Type_Definition;
   begin
      Ntype := Create_Iir (Iir_Kind_Integer_Subtype_Definition);
      Location_Copy (Ntype, Loc);
      Ndef := Create_Iir (Iir_Kind_Integer_Type_Definition);
      Location_Copy (Ndef, Loc);
      Set_Base_Type (Ndef, Ndef);
      Set_Type_Declarator (Ndef, Decl);
      Set_Type_Staticness (Ndef, Locally);
      Set_Signal_Type_Flag (Ndef, True);
      Set_Base_Type (Ntype, Ndef);
      Set_Type_Declarator (Ntype, Decl);
      Set_Range_Constraint (Ntype, Constraint);
      Set_Type_Staticness (Ntype, Get_Expr_Staticness (Constraint));
      Set_Resolved_Flag (Ntype, False);
      Set_Signal_Type_Flag (Ntype, True);
      if Get_Type_Staticness (Ntype) /= Locally then
         Error_Msg_Sem ("range constraint of type must be locally static",
                        Decl);
      end if;
      return Ntype;
   end Create_Integer_Type;

   function Range_Expr_To_Type_Definition (Expr : Iir; Decl: Iir)
     return Iir
   is
      Left, Right : Iir;
   begin
      if Sem_Range_Expression (Expr, False) = Null_Iir then
         return Null_Iir;
      end if;
      Left := Get_Left_Limit (Expr);
      Right := Get_Right_Limit (Expr);
      if Get_Expr_Staticness (Expr) = Locally then
         Left := Eval_Expr (Left);
         Set_Left_Limit (Expr, Left);
         Right := Eval_Expr (Right);
         Set_Right_Limit (Expr, Right);
      end if;

      case Get_Kind (Get_Base_Type (Get_Type (Left))) is
         when Iir_Kind_Integer_Type_Definition =>
            return Create_Integer_Type (Expr, Expr, Decl);
         when Iir_Kind_Floating_Type_Definition =>
            declare
               Ntype: Iir_Floating_Subtype_Definition;
               Ndef: Iir_Floating_Type_Definition;
            begin
               Ntype := Create_Iir (Iir_Kind_Floating_Subtype_Definition);
               Location_Copy (Ntype, Expr);
               Ndef := Create_Iir (Iir_Kind_Floating_Type_Definition);
               Location_Copy (Ndef, Expr);
               Set_Base_Type (Ndef, Ndef);
               Set_Type_Declarator (Ndef, Decl);
               Set_Type_Staticness (Ndef, Get_Expr_Staticness (Expr));
               Set_Signal_Type_Flag (Ndef, True);
               Set_Base_Type (Ntype, Ndef);
               Set_Type_Declarator (Ntype, Decl);
               Set_Range_Constraint (Ntype, Expr);
               Set_Resolved_Flag (Ntype, False);
               Set_Type_Staticness (Ntype, Get_Expr_Staticness (Expr));
               Set_Signal_Type_Flag (Ntype, True);
               return Ntype;
            end;
         when others =>
            --  sem_range_expression should catch such errors.
            raise Internal_Error;
      end case;
   end Range_Expr_To_Type_Definition;

   procedure Check_No_File_Type (El_Type : Iir; Loc : Iir) is
   begin
      case Get_Kind (El_Type) is
         when Iir_Kind_File_Type_Definition =>
            Error_Msg_Sem
              ("element of file type is not allowed in a composite type", Loc);
         when others =>
            null;
      end case;
   end Check_No_File_Type;

   --  Return true iff decl is std.textio.text
   function Is_Text_Type_Declaration (Decl : Iir_Type_Declaration)
     return Boolean
   is
      use Std_Names;
      P : Iir;
   begin
      if Get_Identifier (Decl) /= Name_Text then
         return False;
      end if;
      P := Get_Parent (Decl);
      if Get_Kind (P) /= Iir_Kind_Package_Declaration
        or else Get_Identifier (P) /= Name_Textio
      then
         return False;
      end if;
      --  design_unit, design_file, library_declaration.
      P := Get_Parent (Get_Design_File (Get_Design_Unit (P)));
      if Get_Kind (P) /= Iir_Kind_Library_Declaration then
         raise Internal_Error;
      end if;
      if Get_Identifier (P) /= Name_Std then
         return False;
      end if;
      return True;
   end Is_Text_Type_Declaration;

   --  Semantize the array_element type of DEF.
   --  Set type_staticness and resolved_flag of DEF.
   --  type_staticness of DEF (before calling this function) must be the
   --  staticness of the array indexes.
   procedure Sem_Array_Element (Def : Iir)
   is
      El_Type : Iir;
   begin
      El_Type := Get_Element_Subtype (Def);
      El_Type := Sem_Subtype_Indication (El_Type);
      if El_Type = Null_Iir then
         Set_Type_Staticness (Def, None);
         Set_Resolved_Flag (Def, False);
         --  FIXME: why not
         Set_Element_Subtype (Def, Integer_Type_Definition);
         return;
      end if;
      Set_Element_Subtype (Def, El_Type);
      Check_No_File_Type (El_Type, Def);
      Set_Signal_Type_Flag (Def, Get_Signal_Type_Flag (El_Type));

      --  LRM93 3.2.1.1
      --  The same requirement exists [must define a constrained
      --  array subtype] [...] for the element subtype indication
      --  of an array type definition, if the type of the array
      --  element is itself an array type.
      if not Sem_Is_Constrained (El_Type) then
         Error_Msg_Sem ("array element of unconstrained "
                        & Disp_Node (El_Type) & " is not allowed", Def);
      end if;
      Set_Type_Staticness (Def, Min (Get_Type_Staticness (El_Type),
                                     Get_Type_Staticness (Def)));
      Set_Resolved_Flag (Def, Get_Resolved_Flag (El_Type));
   end Sem_Array_Element;

   function Sem_Type_Definition (Def: Iir; Decl: Iir) return Iir
   is
   begin
      case Get_Kind (Def) is
         when Iir_Kind_Enumeration_Type_Definition =>
            Set_Base_Type (Def, Def);
            Set_Type_Staticness (Def, Locally);
            Set_Signal_Type_Flag (Def, True);

            Create_Range_Constraint_For_Enumeration_Type (Def);

            --  Makes all literal visible.
            declare
               El: Iir;
               Literal_List: Iir_List;
            begin
               Literal_List := Get_Enumeration_Literal_List (Def);
               for I in Natural loop
                  El := Get_Nth_Element (Literal_List, I);
                  exit when El = Null_Iir;
                  Set_Expr_Staticness (El, Locally);
                  Set_Type (El, Def);
                  Set_Enumeration_Decl (El, El);
                  Sem_Name.Add_Name (El);
                  Name_Visible (El);
               end loop;
            end;
            Set_Resolved_Flag (Def, False);
            return Def;

         when Iir_Kind_Range_Expression =>
            if Get_Type (Def) /= Null_Iir then
               return Sem_Physical_Type_Definition (Def, Decl);
            else
               return Range_Expr_To_Type_Definition (Def, Decl);
            end if;

         when Iir_Kind_Range_Array_Attribute
           | Iir_Kind_Attribute
           | Iir_Kind_Parenthesis_Name =>
            if Get_Type (Def) /= Null_Iir then
               return Sem_Physical_Type_Definition (Def, Decl);
            end if;
            --  Nb: the attribute is expected to be a 'range or
            --  a 'reverse_range attribute.
            declare
               Res : Iir;
            begin
               Res := Sem_Discrete_Range_Expression (Def, Null_Iir);
               if Res = Null_Iir then
                  return Null_Iir;
               end if;
               --  This cannot be a floating range.
               return Create_Integer_Type (Def, Res, Decl);
            end;

         when Iir_Kind_Array_Subtype_Definition =>
            declare
               Index_Type : Iir;
               Index_List : Iir_List;
               Staticness : Iir_Staticness;

               -- array_type_definition, which is the same as the subtype,
               -- but without any constraint in the indexes.
               Base_Type: Iir;
            begin
               -- FIXME: all indexes must be either constrained or
               -- unconstrained.
               -- If all indexes are unconstrained, this is really a type
               -- otherwise, this is a subtype.

               -- Create a definition for the base type of subtype DEF.
               Base_Type := Create_Iir (Iir_Kind_Array_Type_Definition);
               Location_Copy (Base_Type, Def);
               Set_Base_Type (Base_Type, Base_Type);
               Set_Type_Declarator (Base_Type, Decl);

               Staticness := Locally;
               Index_List := Get_Index_List (Def);
               for I in Natural loop
                  Index_Type := Get_Nth_Element (Index_List, I);
                  exit when Index_Type = Null_Iir;

                  Index_Type :=
                    Sem_Discrete_Range_Expression (Index_Type, Null_Iir);
                  if Index_Type /= Null_Iir then
                     Index_Type := Range_To_Subtype_Definition (Index_Type);
                     if Get_Base_Type (Index_Type) =
                       Universal_Integer_Type_Definition
                     then
                        -- FIXME: is it correct ???
                        Set_Base_Type (Index_Type, Integer_Type_Definition);
                     end if;

                     Replace_Nth_Element (Index_List, I, Index_Type);
                     Staticness :=
                       Min (Staticness, Get_Type_Staticness (Index_Type));

                     -- Set the index type in the array type.
                     -- must "unconstraint" the subtype.
                     Append_Element (Get_Index_List (Base_Type), Index_Type);
                  end if;
               end loop;
               Set_Type_Staticness (Def, Staticness);

               -- Element type.
               Sem_Array_Element (Def);

               Set_Element_Subtype (Base_Type, Get_Element_Subtype (Def));
               Set_Signal_Type_Flag (Base_Type, Get_Signal_Type_Flag (Def));
               --  According to LRM93 7.4.1, an unconstrained array type
               --  is not static.
               Set_Type_Staticness (Base_Type, None);
               Set_Type_Declarator (Base_Type, Decl);
               Set_Resolved_Flag (Base_Type, Get_Resolved_Flag (Def));

               Set_Base_Type (Def, Base_Type);
               return Def;
            end;

         when Iir_Kind_Array_Type_Definition =>
            declare
               Index_Type : Iir;
            begin
               Set_Base_Type (Def, Def);

               for I in Natural loop
                  Index_Type := Get_Nth_Element (Get_Index_List (Def), I);
                  exit when Index_Type = null;

                  Index_Type := Sem_Subtype_Indication (Index_Type);
                  if Index_Type /= null then
                     if Get_Kind (Index_Type) not in
                       Iir_Kinds_Discrete_Type_Definition
                     then
                        Error_Msg_Sem
                          ("index type of an array must be discrete",
                           Index_Type);
                     else
                        Replace_Nth_Element
                          (Get_Index_List (Def), I, Index_Type);
                     end if;
                  end if;
               end loop;
               --  According to LRM93 7.4.1, an unconstrained array type
               --  is not static.
               Set_Type_Staticness (Def, None);

               Sem_Array_Element (Def);
               return Def;
            end;

         when Iir_Kind_Record_Type_Definition =>
            declare
               --  Non semantized type of previous element.
               Last_El_Type : Iir;
               --  Semantized type of previous element
               Last_Type : Iir;

               El: Iir;
               El_Type : Iir;
               List: Iir;
               Resolved_Flag : Boolean;
               Staticness : Iir_Staticness;
            begin
               Sem_Name.Open_Declarative_Region;
               List := Get_Element_Declaration_List (Def);
               Resolved_Flag := True;
               Last_El_Type := Null_Iir;
               Last_Type := Null_Iir;
               Staticness := Locally;
               Set_Signal_Type_Flag (Def, True);
               for I in Natural loop
                  El := Get_Nth_Element (List, I);
                  exit when El = Null_Iir;
                  El_Type := Get_Type (El);
                  if El_Type /= Last_El_Type then
                     --  Be careful for a declaration list (r,g,b: integer).
                     Last_El_Type := El_Type;
                     El_Type := Sem_Subtype_Indication (El_Type);
                     Last_Type := El_Type;
                  else
                     El_Type := Last_Type;
                  end if;
                  if El_Type /= Null_Iir then
                     Set_Type (El, El_Type);
                     Check_No_File_Type (El_Type, El);
                     if not Get_Signal_Type_Flag (El_Type) then
                        Set_Signal_Type_Flag (Def, False);
                     end if;

                     --  LRM93 3.2.1.1
                     --  The same requirement [must define a constrained array
                     --  subtype] exits for the subtype indication of an
                     --  element declaration, if the type of the record
                     --  element is an array type.
                     if not Sem_Is_Constrained (El_Type) then
                        Error_Msg_Sem
                          ("element declaration of unconstrained "
                           & Disp_Node (El_Type) & " is not allowed", El);
                     end if;
                     Resolved_Flag :=
                       Resolved_Flag and Get_Resolved_Flag (El_Type);
                     Staticness := Min (Staticness,
                                        Get_Type_Staticness (El_Type));
                  else
                     Staticness := None;
                  end if;
                  Sem_Name.Add_Name (El);
                  Name_Visible (El);
               end loop;
               Sem_Name.Close_Declarative_Region;
               Set_Base_Type (Def, Def);
               Set_Resolved_Flag (Def, Resolved_Flag);
               Set_Type_Staticness (Def, Staticness);
               return Def;
            end;

         when Iir_Kind_Access_Type_Definition =>
            declare
               D_Type : Iir;
            begin
               D_Type := Sem_Subtype_Indication (Get_Designated_Type (Def),
                                                 True);
               if D_Type /= Null_Iir then
                  case Get_Kind (D_Type) is
                     when Iir_Kind_Incomplete_Type_Definition =>
                        Append_Element
                          (Get_Incomplete_Type_List (D_Type), Def);
                     when Iir_Kind_File_Type_Definition =>
                        --  LRM 3.3
                        --  The designated type must not be a file type.
                        Error_Msg_Sem
                          ("designated type must not be a file type", Def);
                     when others =>
                        null;
                  end case;
                  Set_Designated_Type (Def, D_Type);
               end if;
               Set_Base_Type (Def, Def);
               Set_Type_Staticness (Def, None);
               Set_Resolved_Flag (Def, False);
               Set_Signal_Type_Flag (Def, False);
               return Def;
            end;

         when Iir_Kind_File_Type_Definition =>
            declare
               Type_Mark : Iir;
            begin
               Type_Mark := Sem_Subtype_Indication (Get_Type_Mark (Def));
               Set_Type_Mark (Def, Type_Mark);
               if Type_Mark /= Null_Iir then
                  if Get_Signal_Type_Flag (Type_Mark) = False then
                     --  LRM 3.4
                     --  The base type of this subtype must not be a file type
                     --  or an access type.
                     --  If the base type is a composite type, it must not
                     --  contain a subelement of an access type.
                     Error_Msg_Sem
                       (Disp_Node (Type_Mark) & " cannot be a file type", Def);
                  elsif Get_Kind (Type_Mark) in Iir_Kinds_Array_Type_Definition
                  then
                     --  LRM 3.4
                     --  If the base type is an array type, it must be a one
                     --  dimensional array type.
                     if not Is_Unidim_Array_Type (Type_Mark) then
                        Error_Msg_Sem
                          ("multi-dimensional " & Disp_Node (Type_Mark)
                           & " cannot be a file type", Def);
                     end if;
                  end if;
               end if;
               Set_Base_Type (Def, Def);
               Set_Resolved_Flag (Def, False);
               Set_Text_File_Flag (Def, Is_Text_Type_Declaration (Decl));
               Set_Signal_Type_Flag (Def, False);
               return Def;
            end;

         when others =>
            Error_Kind ("sem_type_definition", Def);
            return Def;
      end case;
   end Sem_Type_Definition;

   procedure Sem_Type_Declaration (Decl: Iir)
   is
      Def: Iir;
      Inter : Name_Interpretation_Type;
      Old_Decl : Iir;
   begin
      --  Check if DECL complete a previous incomplete type declaration.
      Inter := Get_Interpretation (Get_Identifier (Decl));
      if Inter /= No_Name_Interpretation
        and then Is_In_Current_Declarative_Part (Inter)
      then
         Old_Decl := Get_Declaration (Inter);
         if Get_Kind (Old_Decl) /= Iir_Kind_Type_Declaration
           or else Get_Kind (Get_Type (Old_Decl)) /=
           Iir_Kind_Incomplete_Type_Definition
         then
            Old_Decl := Null_Iir;
         end if;
      end if;

      if Old_Decl = Null_Iir then
         Sem_Name.Add_Name (Decl);
      else
         Set_Visible_Flag (Old_Decl, False);
      end if;

      -- Check the definition of the type.
      Def := Get_Type (Decl);
      if Def = Null_Iir then
         --  Incomplete type declaration
         Def := Create_Iir (Iir_Kind_Incomplete_Type_Definition);
         Location_Copy (Def, Decl);
         Set_Type (Decl, Def);
         Set_Signal_Type_Flag (Def, True);
         Set_Type_Declarator (Def, Decl);
      else
         Def := Sem_Type_Definition (Def, Decl);
         if Def /= Null_Iir then
            Set_Type_Declarator (Def, Decl);
            Set_Type (Decl, Def);
            Create_Implicit_Operations (Decl, False);
            Sem_Name.Add_Implicit_Functions (Decl, False);
         end if;
      end if;

      if Old_Decl /= Null_Iir then
         --  Complete the type definition.
         declare
            List : Iir_List;
            El : Iir;
            Old_Def : Iir;
         begin
            Old_Def := Get_Type (Old_Decl);
            Set_Signal_Type_Flag (Old_Def, Get_Signal_Type_Flag (Def));
            List := Get_Incomplete_Type_List (Old_Def);
            for I in Natural loop
               El := Get_Nth_Element (List, I);
               exit when El = Null_Iir;
               Set_Designated_Type (El, Def);
            end loop;
            --  Complete the incomplete_type_definition node
            --  (set type_declarator and base_type).
            Set_Type_Declarator (Old_Def, Decl);
            Set_Base_Type (Old_Def, Get_Base_Type (Def));
            Replace_Name (Get_Identifier (Decl), Old_Decl, Decl);
         end;
      end if;
      Sem_Name.Name_Visible (Decl);
   end Sem_Type_Declaration;

   procedure Sem_Subtype_Declaration (Decl: Iir) is
      Def: Iir;
      Res: Iir;
   begin
      Sem_Name.Add_Name (Decl);

      -- Check the definition of the type.
      Def := Sem_Subtype_Indication (Get_Type (Decl));
      if Def = Null_Iir then
         return;
      end if;

      if not Is_Anonymous_Type_Definition (Def) then
         case Get_Kind (Def) is
            when Iir_Kind_Integer_Subtype_Definition
              | Iir_Kind_Floating_Subtype_Definition
              | Iir_Kind_Enumeration_Subtype_Definition
              | Iir_Kind_Physical_Subtype_Definition =>
               -- no limits, makes an alias.
               Res := Create_Iir (Get_Kind (Def));
               Location_Copy (Res, Decl);
               Set_Base_Type (Res, Get_Base_Type (Def));
               Set_Range_Constraint (Res, Get_Range_Constraint (Def));
               Set_Resolution_Function (Res, Get_Resolution_Function (Def));
            when Iir_Kind_Enumeration_Type_Definition =>
               --  makes an alias.
               Res := Create_Iir (Iir_Kind_Enumeration_Subtype_Definition);
               Location_Copy (Res, Decl);
               Set_Base_Type (Res, Def);
               Set_Type_Mark (Res, Def);
               Set_Range_Constraint (Res, Get_Range_Constraint (Def));
            when Iir_Kind_Access_Subtype_Definition
              | Iir_Kind_Access_Type_Definition =>
               --  Make an alias.
               Res := Create_Iir (Iir_Kind_Access_Subtype_Definition);
               Location_Copy (Res, Decl);
               Set_Base_Type (Res, Get_Base_Type (Def));
            when Iir_Kind_Array_Type_Definition
              | Iir_Kind_Unconstrained_Array_Subtype_Definition =>
               Res :=
                 Create_Iir (Iir_Kind_Unconstrained_Array_Subtype_Definition);
               Location_Copy (Res, Decl);
               Set_Base_Type (Res, Get_Base_Type (Def));
               Set_Type_Staticness (Res, Get_Type_Staticness (Def));
               if Get_Kind (Def) /= Iir_Kind_Array_Type_Definition then
                  Set_Resolution_Function
                    (Res, Get_Resolution_Function (Def));
               end if;
               Set_Resolved_Flag (Res, Get_Resolved_Flag (Def));
               Set_Type_Mark (Res, Def);
            when Iir_Kind_Record_Type_Definition
              | Iir_Kind_Record_Subtype_Definition =>
               Res := Create_Iir (Iir_Kind_Record_Subtype_Definition);
               Location_Copy (Res, Decl);
               Set_Base_Type (Res, Get_Base_Type (Def));
               Set_Type_Staticness (Res, Get_Type_Staticness (Def));
               if Get_Kind (Def) /= Iir_Kind_Record_Type_Definition then
                  Set_Resolution_Function
                    (Res, Get_Resolution_Function (Def));
               end if;
               Set_Resolved_Flag (Res, Get_Resolved_Flag (Def));
            when others =>
               --  FIXME: todo
               Error_Kind ("sem_subtype_declaration", Def);
         end case;
         Set_Type_Staticness (Res, Get_Type_Staticness (Def));
         Set_Signal_Type_Flag (Res, Get_Signal_Type_Flag (Def));
         Def := Res;
      end if;
      Set_Type (Decl, Def);
      Set_Type_Declarator (Def, Decl);
      Name_Visible (Decl);
   end Sem_Subtype_Declaration;

   function Sem_Is_Constrained (A_Type: Iir) return Boolean is
   begin
      case Get_Kind (A_Type) is
         when Iir_Kind_Array_Subtype_Definition =>
            return True;
         when Iir_Kind_Enumeration_Subtype_Definition
           | Iir_Kind_Enumeration_Type_Definition
           | Iir_Kind_Integer_Subtype_Definition
           | Iir_Kind_Integer_Type_Definition
           | Iir_Kind_Floating_Subtype_Definition
           | Iir_Kind_Floating_Type_Definition
           | Iir_Kind_Access_Type_Definition
           | Iir_Kind_Access_Subtype_Definition
           | Iir_Kind_Physical_Subtype_Definition
           | Iir_Kind_Record_Type_Definition
           | Iir_Kind_Record_Subtype_Definition
           | Iir_Kind_File_Type_Definition =>
            --| Iir_Kind_File_Subtype_Definition =>
            return True;
         when Iir_Kind_Array_Type_Definition
           | Iir_Kind_Unconstrained_Array_Subtype_Definition =>
            return False;
         when others =>
            Error_Kind ("sem_is_constrained", A_Type);
      end case;
   end Sem_Is_Constrained;

   --  If DECL is a constant declaration, and there is already a constant
   --  declaration in the current scope with the same name, then return it.
   --  Otherwise, return NULL.
   function Get_Deferred_Constant (Decl : Iir) return Iir
   is
      use Sem_Name;
      Deferred_Const : Iir;
      Interp : Name_Interpretation_Type;
   begin
      if Get_Kind (Decl) /= Iir_Kind_Constant_Declaration then
         return Null_Iir;
      end if;
      Interp := Get_Interpretation (Get_Identifier (Decl));
      if Interp = No_Name_Interpretation then
         return Null_Iir;
      end if;
      if not Is_In_Current_Declarative_Part (Interp) then
         return Null_Iir;
      end if;
      Deferred_Const := Get_Declaration (Interp);
      if Get_Kind (Deferred_Const) /= Iir_Kind_Constant_Declaration then
         return Null_Iir;
      end if;
      --  LRM93 4.3.1.1
      --  The corresponding full constant declaration, which defines the value
      --  of the constant, must appear in the body of the package.
      if Get_Kind (Get_Library_Unit (Get_Current_Design_Unit))
        /= Iir_Kind_Package_Body
      then
         Error_Msg_Sem
           ("full constant declaration must appear in package body", Decl);
      end if;
      return Deferred_Const;
   end Get_Deferred_Constant;

   procedure Sem_Object_Declaration (Decl: Iir; Parent : Iir)
   is
      Atype: Iir;
      Default_Value : Iir;
      Proxy : Iir;
      Deferred_Const : Iir;
      Staticness : Iir_Staticness;
   begin
      Deferred_Const := Get_Deferred_Constant (Decl);

      --  Semantize type and default value:
      Atype := Get_Type (Decl);
      if Get_Kind (Atype) /= Iir_Kind_Proxy then
         Atype := Sem_Subtype_Indication (Atype);
         if Atype = Null_Iir then
            Set_Type (Decl, Null_Iir);
            return;
         end if;
      end if;

      if Deferred_Const = Null_Iir then
         Sem_Name.Add_Name (Decl);
      end if;

      if Get_Kind (Atype) = Iir_Kind_Proxy then
         Proxy := Get_Proxy (Atype);
         Default_Value := Get_Default_Value (Proxy);
         Atype := Get_Type (Proxy);
         if Atype = Null_Iir then
            return;
         end if;
         Proxy := Get_Type (Decl);
         Free_Iir (Proxy);
      else
         Default_Value := Get_Default_Value (Decl);
         if Default_Value /= Null_Iir then
            declare
               Val : Iir;
            begin
               Val := Sem_Expression (Default_Value, Atype);
               if Val /= Null_Iir then
                  Default_Value := Val;
               end if;
            end;
         end if;
      end if;
      Set_Type (Decl, Atype);
      Set_Default_Value (Decl, Default_Value);
      Set_Base_Name (Decl, Decl);
      Set_Visible_Flag (Decl, True);

      --  LRM93 2.6
      --  The subtype indication given in the full declaration of the deferred
      --  constant must conform to that given in the deferred constant
      --  declaration.
      if Deferred_Const /= Null_Iir
        and then not Are_Trees_Equal (Get_Type (Decl),
                                      Get_Type (Deferred_Const))
      then
         Error_Msg_Sem
           ("subtype indication doesn't conform with the deferred constant",
            Decl);
      end if;

      --  LRM 4.3.1.3
      --  It is an error if a variable declaration declares a variable that is
      --  of a file type.
      --
      --  LRM 4.3.1.1
      --  It is an error if a constant declaration declares a constant that is
      --  of a file type, or an access type, or a composite type which has
      --  subelement that is a file type of an access type.
      --
      --  LRM 4.3.1.2
      --  It is an error if a signal declaration declares a signal that is of
      --  a file type [or an access type].
      case Get_Kind (Atype) is
         when Iir_Kind_File_Type_Definition =>
            Error_Msg_Sem (Disp_Node (Decl) & " cannot be of type file", Decl);
         when others =>
            if Get_Kind (Decl) /= Iir_Kind_Variable_Declaration
              and then Get_Signal_Type_Flag (Atype) = False
            then
               Error_Msg_Sem ("type of " & Disp_Node (Decl)
                              & " cannot be " & Disp_Node (Atype), Decl);
               Error_Msg_Sem ("(" & Disp_Node (Atype)
                              & " has an access subelement)", Decl);
            end if;
      end case;

      case Get_Kind (Decl) is
         when Iir_Kind_Constant_Declaration =>
            Set_Design_Unit (Decl, Get_Current_Design_Unit);
            --  LRM93 4.3.1.1
            --  If the assignment symbol ":=" followed by an expression is not
            --  present in a constant declaration, then the declaration
            --  declares a deferred constant.
            --  Such a constant declaration may only appear in a package
            --  declaration.
            if Deferred_Const /= Null_Iir then
               Set_Deferred_Declaration (Decl, Deferred_Const);
               Set_Deferred_Declaration (Deferred_Const, Decl);
            end if;
            if Default_Value = Null_Iir then
               if Deferred_Const /= Null_Iir then
                  Error_Msg_Sem
                    ("full constant declaration must have a default value",
                     Decl);
               else
                  Set_Deferred_Declaration_Flag (Decl, True);
               end if;
               if Get_Kind (Parent) /= Iir_Kind_Package_Declaration then
                  Error_Msg_Sem ("a constant must have a default value", Decl);
               end if;
               Set_Expr_Staticness (Decl, Globally);
            else
               --  LRM93 7.4.1: a locally static primary is defined:
               --  A constant (other than deferred constant) explicitly
               --  declared by a constant declaration and initialized
               --  with a locally static expression.
               --  Note: the staticness of the full declaration may be locally.
               if False and Deferred_Const /= Null_Iir then
                  --  This is a deferred constant.
                  Staticness := Globally;
               else
                  Staticness := Min (Get_Expr_Staticness (Default_Value),
                                     Get_Type_Staticness (Atype));
                  --  LRM 7.4.2 (Globally static primaries)
                  --  5. a constant
                  if Staticness < Globally then
                     Staticness := Globally;
                  end if;
               end if;
               Set_Expr_Staticness (Decl, Staticness);

               if Staticness = Locally then
                  Set_Default_Value
                    (Decl, Eval_Expr_Check (Default_Value, Atype));
               end if;
            end if;
         when Iir_Kind_Signal_Declaration =>
            --  LRM93 4.3.1.2
            --  It is also an error if a guarded signal of a
            --  scalar type is neither a resolved signal nor a
            --  subelement of a resolved signal.
            if Get_Signal_Kind (Decl) /= Iir_No_Signal_Kind
              and then not Get_Resolved_Flag (Atype)
            then
               Error_Msg_Sem
                 ("guarded " & Disp_Node (Decl) & " must be resolved", Decl);
            end if;
            Set_Expr_Staticness (Decl, None);
            Set_Has_Disconnect_Flag (Decl, False);
         when Iir_Kind_Variable_Declaration =>
            Set_Expr_Staticness (Decl, None);
         when others =>
            Error_Kind ("sem_object_declaration", Decl);
      end case;

      case Get_Kind (Decl) is
         when Iir_Kind_Constant_Declaration =>
            --  LRM93 3.2.1.1
            --  For a constant declared by an object declaration, the index
            --  ranges are defined by the initial value, if the subtype of the
            --  constant is unconstrained; otherwise they are defined by this
            --  subtype.
            --if Default_Value = Null_Iir
            --  and then not Sem_Is_Constrained (Atype)
            --then
            --   Error_Msg_Sem ("constant declaration of unconstrained "
            --                  & Disp_Node (Atype) & " is not allowed", Decl);
            --end if;
            null;
            --if Deferred_Const = Null_Iir then
            --   Name_Visible (Decl);
            --end if;

         when Iir_Kind_Variable_Declaration
           | Iir_Kind_Signal_Declaration =>
            --  LRM93 3.2.1.1
            --  For a variable or signal declared by an object declaration, the
            --  subtype indication of the corressponding object declaration
            --  must define a constrained array subtype.
            if not Sem_Is_Constrained (Atype) then
               Error_Msg_Sem
                 ("declaration of " & Disp_Node (Decl)
                  & " with unconstrained " & Disp_Node (Atype)
                  & " is not allowed", Decl);
               if Default_Value /= Null_Iir then
                  Error_Msg_Sem ("(even with a default value)", Decl);
               end if;
            end if;

         when others =>
            Error_Kind ("sem_object_declaration(2)", Decl);
      end case;
   end Sem_Object_Declaration;

   procedure Sem_File_Declaration (Decl: Iir_File_Declaration) is
      Atype: Iir;
      Logical_Name: Iir;
      Open_Kind : Iir;
   begin
      Sem_Name.Add_Name (Decl);
      Set_Expr_Staticness (Decl, None);
      Set_Base_Name (Decl, Decl);

      -- Try to find a type.
      Atype := Sem_Subtype_Indication (Get_Type (Decl));
      if Atype = Null_Iir then
         return;
      else
         Set_Type (Decl, Atype);
      end if;
      -- LRM93 4.3.1.4
      -- The subtype indication of a file declaration must define a file
      -- subtype.
      if Get_Kind (Atype) /= Iir_Kind_File_Type_Definition then
         Error_Msg_Sem ("file subtype expected for a file declaration", Decl);
         return;
      end if;

      Logical_Name := Get_File_Logical_Name (Decl);
      -- LRM93 4.3.1.4
      -- The file logical name must be an expression of predefined type
      -- STRING.
      if Logical_Name /= Null_Iir then
         Logical_Name :=
           Sem_Expression (Logical_Name, String_Type_Definition);
         if Logical_Name /= Null_Iir then
            Check_Read (Logical_Name);
            Set_File_Logical_Name (Decl, Logical_Name);
         end if;
      end if;

      Open_Kind := Get_File_Open_Kind (Decl);
      if Open_Kind /= Null_Iir then
         Open_Kind := Sem_Expression
           (Open_Kind, File_Open_Kind_Type_Definition);
         Check_Read (Open_Kind);
      else
         --  LRM93 4.3.1.4
         --  If a file open kind expression is not included in the file open
         --  information of a given file declaration, then the default value
         --  of READ_MODE is used during elaboration of the file declaration.
         if (Flags.Vhdl_Std >= Vhdl_93c and Get_Mode (Decl) = Iir_Unknown_Mode)
           and Logical_Name = Null_Iir
         then
            Open_Kind := File_Open_Kind_Read_Mode;
         end if;
      end if;
      if Open_Kind /= Null_Iir then
         Set_File_Open_Kind (Decl, Open_Kind);
      end if;
      Name_Visible (Decl);
   end Sem_File_Declaration;

   procedure Sem_Attribute_Declaration (Decl: Iir_Attribute_Declaration)
   is
      A_Type : Iir;
      Ident : Name_Id;
   begin
      --  LRM93 4.4
      --  The identifier is said to be the designator of the attribute.
      Ident := Get_Identifier (Decl);
      if Ident in Std_Names.Name_Id_Attributes then
         Warning_Msg_Sem ("predefined attribute """ & Name_Table.Image (Ident)
                          & """ overriden", Decl);
      end if;
      Sem_Name.Add_Name (Decl);

      A_Type := Sem_Subtype_Indication (Get_Type (Decl));
      if A_Type = Null_Iir then
         return;
      end if;
      --  LRM93 4.4
      --  The type mark must denote a subtype that is neither an access type
      --  nor a file type.
      --  The subtype need not be constrained.
      case Get_Kind (A_Type) is
         when Iir_Kind_Access_Type_Definition
           | Iir_Kind_Access_Subtype_Definition
           | Iir_Kind_File_Type_Definition =>
            --  FIXME: | Iir_Kind_File_Subtype_Definition =>
            Error_Msg_Sem
              ("attribute type can't be an access type nor a file type", Decl);
         when others =>
            null;
      end case;
      Set_Type (Decl, A_Type);
      Name_Visible (Decl);
   end Sem_Attribute_Declaration;

   procedure Sem_Group_Template_Declaration
     (Decl : Iir_Group_Template_Declaration)
   is
   begin
      Sem_Name.Add_Name (Decl);
      Sem_Name.Name_Visible (Decl);
   end Sem_Group_Template_Declaration;

   procedure Sem_Interface_List (Interface_List: Iir_List;
                                 Interface_Kind : Interface_Kind_Type)
   is
      El, A_Type: Iir;
      Proxy : Iir_Proxy;
      Default_Value: Iir;
   begin
      if Interface_List = Null_Iir_List then
         return;
      end if;
      for I in Natural loop
         El := Get_Nth_Element (Interface_List, I);
         exit when El = null;

         --  Avoid the reanalysed duplicated types.
         --  This is not an optimization, since the unanalysed type must have
         --  been freed.
         A_Type := Get_Type (El);
         if Get_Kind (A_Type) = Iir_Kind_Proxy then
            Proxy := A_Type;
            A_Type := Get_Type (Get_Proxy (Proxy));
            Default_Value := Get_Default_Value (Get_Proxy (Proxy));
            Free_Iir (Proxy);
         else
            A_Type := Sem_Subtype_Indication (A_Type);
            Default_Value := Get_Default_Value (El);
            if Default_Value /= Null_Iir and then A_Type /= Null_Iir then
               Deferred_Constant_Allowed := True;
               Default_Value := Sem_Expression (Default_Value, A_Type);
               Deferred_Constant_Allowed := False;
               Check_Read (Default_Value);
            end if;
         end if;

         if A_Type /= Null_Iir then
            Set_Type (El, A_Type);
            Set_Base_Name (El, El);

            if Get_Kind (El) = Iir_Kind_Signal_Interface_Declaration then
               case Get_Signal_Kind (El) is
                  when Iir_No_Signal_Kind =>
                     null;
                  when Iir_Bus_Kind =>
                     if Get_Kind (A_Type) not in Iir_Kinds_Subtype_Definition
                       or else Get_Resolution_Function (A_Type) = Null_Iir
                     then
                        Error_Msg_Sem
                          ("type `" & Disp_Node (A_Type)
                           & "' of guarded interface signal `" & Disp_Node (El)
                           & "' is not resolved", El);
                     end if;
                  when Iir_Register_Kind =>
                     Error_Msg_Sem
                       ("interface signal can't be of kind register", El);
               end case;
            end if;

            --  LRM93 4.3.2
            --  For an interface constant declaration or an interface signal
            --  declaration, the subtype indication must define a subtype
            --  that is neither a file type nor an access type.
            case Get_Kind (El) is
               when Iir_Kind_Constant_Interface_Declaration
                 | Iir_Kind_Signal_Interface_Declaration =>
                  case Get_Kind (Get_Base_Type (A_Type)) is
                     when Iir_Kind_File_Type_Definition =>
                        Error_Msg_Sem
                          ("constant or signal formal type can't "
                           & "be a file type", El);
                     when Iir_Kind_Access_Type_Definition =>
                        Error_Msg_Sem
                          ("constant or signal formal type can't "
                           & "be an access type", El);
                     when others =>
                        null;
                  end case;
               when Iir_Kind_Variable_Interface_Declaration =>
                  if Get_Kind (Get_Base_Type (A_Type))
                    = Iir_Kind_File_Type_Definition
                    and then Flags.Vhdl_Std >= Vhdl_93
                  then
                     Error_Msg_Sem
                       ("variable formal type can't be a file type (vhdl 93)",
                        El);
                  end if;
               when Iir_Kind_File_Interface_Declaration =>
                  if Get_Kind (Get_Base_Type (A_Type))
                    /= Iir_Kind_File_Type_Definition
                  then
                     Error_Msg_Sem
                       ("file formal type must be a file type", El);
                  end if;
               when others =>
                  --  El is not an interface.
                  raise Internal_Error;
            end case;

            if Default_Value /= Null_Iir then
               Set_Default_Value (El, Default_Value);
            end if;
         end if;

         Sem_Name.Add_Name (El);

         --  By default, interface are not static.
         --  This may be changed just below.
         Set_Expr_Staticness (El, None);

         case Interface_Kind is
            when Interface_Generic =>
               --  LRM93 1.1.1
               --  The generic list in the formal generic clause defines
               --  generic constants whose values may be determined by the
               --  environment.
               if Get_Kind (El) /= Iir_Kind_Constant_Interface_Declaration then
                  Error_Msg_Sem
                    ("generic " & Disp_Node (El) & " must be a constant",
                     El);
               else
                  --   LRM93 7.4.2 (Globally static primaries)
                  --   3. a generic constant.
                  Set_Expr_Staticness (El, Globally);
               end if;
            when Interface_Port =>
               if Get_Kind (El) /= Iir_Kind_Signal_Interface_Declaration then
                  Error_Msg_Sem
                    ("port " & Disp_Node (El) & " must be a signal", El);
               end if;
            when Interface_Procedure
              | Interface_Function =>
               if Get_Kind (El) = Iir_Kind_Variable_Interface_Declaration
                 and then Interface_Kind = Interface_Function
               then
                  Error_Msg_Sem ("variable interface parameter are not "
                                 & "allowed for a function (use a constant)",
                                 El);
               end if;
               case Get_Mode (El) is
                  when Iir_Unknown_Mode =>
                     raise Internal_Error;
                  when Iir_In_Mode =>
                     null;
                  when Iir_Inout_Mode
                    | Iir_Out_Mode =>
                     if Interface_Kind = Interface_Function then
                        Error_Msg_Sem ("inout or out is an allowed mode " &
                                       "for a function parameter", El);
                     end if;
                  when Iir_Buffer_Mode
                    | Iir_Linkage_Mode =>
                     Error_Msg_Sem ("buffer or linkage is not an allowed mode "
                                    & "for a subprogram parameter", El);
               end case;
         end case;
      end loop;

      --  Make all name visible.
      for I in Natural loop
         El := Get_Nth_Element (Interface_List, I);
         exit when El = null;
         Name_Visible (El);
      end loop;
   end Sem_Interface_List;

   procedure Sem_Component_Declaration (Component: Iir_Component_Declaration)
   is
   begin
      Sem_Name.Add_Name (Component);

      Sem_Name.Open_Declarative_Region;
      Sem_Interface_List (Get_Generic_List (Component), Interface_Generic);
      Sem_Interface_List (Get_Port_List (Component), Interface_Port);
      Sem_Name.Close_Declarative_Region;

      Name_Visible (Component);
   end Sem_Component_Declaration;

   -- Return the specification corresponding to a declaration DECL, or
   -- null_Iir if none.
   -- FIXME: respect rules of LRM93 2.7
   function Find_Subprogram_Specification (Decl: Iir) return Iir
   is
      use Sem_Name;
      Interpretation : Name_Interpretation_Type;
      Decl1: Iir;
   begin
      Interpretation := Get_Interpretation (Get_Identifier (Decl));
      while Interpretation /= No_Name_Interpretation loop
         if not Is_In_Current_Declarative_Part (Interpretation) then
            --  The declaration does not belong to the current declarative
            --  region, neither will the following one.  So, we do not found
            --  it.
            return Null_Iir;
         end if;
         Decl1 := Get_Declaration (Interpretation);
         --  Should be sure DECL1 and DECL belongs to the same declarative
         --  region, ie DECL1 was not made visible via a USE clause.
         if not Is_Potentially_Visible (Interpretation)
           and then
           ((Get_Kind (Decl) in Iir_Kinds_User_Function_Specification
             and then
             Get_Kind (Decl1) in Iir_Kinds_User_Function_Specification)
           or else
           (Get_Kind (Decl) in Iir_Kinds_User_Procedure_Specification
            and then
            Get_Kind (Decl1) in Iir_Kinds_User_Procedure_Specification))
           and then Check_Subprograms_Specification (Decl, Decl1)
         then
            return Decl1;
         end if;
         Interpretation := Get_Next_Interpretation (Interpretation);
      end loop;
      return Null_Iir;
   end Find_Subprogram_Specification;

   -- LRM93 8
   -- All statements may be labeled.
   -- Such labels are implicitly declared at the beginning of the declarative
   -- part of the innermost enclosing process statement of subprogram body.
   procedure Sem_Labels (List: Iir_List) is
      Stmt: Iir;
      Label: Name_Id;
   begin
      for I in Natural loop
         Stmt := Get_Nth_Element (List, I);
         exit when Stmt = null;
         Label := Get_Label (Stmt);
         if Label /= Null_Identifier then
            Sem_Name.Add_Name (Stmt);
            Name_Visible (Stmt);
         end if;

         --  Some statements have sub-lists of statements.
         case Get_Kind (Stmt) is
            when Iir_Kind_For_Loop_Statement
              | Iir_Kind_While_Loop_Statement =>
               Sem_Labels (Get_Statement_List (Stmt));
            when Iir_Kind_If_Statement =>
               declare
                  Clause : Iir;
               begin
                  Clause := Stmt;
                  while Clause /= Null_Iir loop
                     Sem_Labels (Get_Then_Sequence_List (Clause));
                     Clause := Get_Else_Clause (Clause);
                  end loop;
               end;
            when Iir_Kind_Case_Statement =>
               declare
                  Choice_List : Iir_Association_Choices_List;
                  El : Iir;
               begin
                  Choice_List := Get_Association_Choices_List (Stmt);
                  for I in Natural loop
                     El := Get_Nth_Element (Choice_List, I);
                     exit when El = Null_Iir;
                     Sem_Labels (Get_Associated (El));
                  end loop;
               end;
            when others =>
               null;
         end case;
      end loop;
   end Sem_Labels;

   procedure Sem_Labels_2 (List: Iir_List) is
      Stmt: Iir;
   begin
      for I in Natural loop
         Stmt := Get_Nth_Element (List, I);
         exit when Stmt = null;
         case Get_Kind (Stmt) is
            when Iir_Kind_Component_Instantiation_Statement =>
               declare
                  Comp : Iir;
                  Decl : Iir;
               begin
                  Comp := Get_Instantiated_Unit (Stmt);
                  if Get_Kind (Comp) in Iir_Kinds_Name then
                     Decl := Find_Declaration (Comp, Decl_Component);
                     Set_Instantiated_Unit (Stmt, Decl);
                  end if;
               end;
            when Iir_Kind_Concurrent_Procedure_Call_Statement =>
               --  Note: a statement such as
               --    label1: name;
               --  can be parsed as a procedure call statement or as a
               --  component instantiation statement.
               --  Check now and revert in case of error.
               if Get_Parameter_Association_List (Stmt) = Null_Iir_List then
                  declare
                     Decl : Iir;
                     Label : Name_Id;
                     N_Stmt : Iir_Component_Instantiation_Statement;
                     Imp : Iir;
                  begin
                     Imp := Get_Implementation (Stmt);
                     Sem_Names (Imp);
                     Decl := Get_Named_Entity (Imp);
                     if Get_Kind (Decl) = Iir_Kind_Component_Declaration then
                        N_Stmt := Create_Iir
                          (Iir_Kind_Component_Instantiation_Statement);
                        Label := Get_Label (Stmt);
                        Set_Label (N_Stmt, Label);
                        Set_Parent (N_Stmt, Get_Parent (Stmt));
                        Set_Instantiated_Unit (N_Stmt, Decl);
                        Replace_Nth_Element (List, I, N_Stmt);
                        Location_Copy (N_Stmt, Stmt);
                        if Label /= Null_Identifier then
                           --  A component instantiation statement must have
                           --  a label, this condition is checked during the
                           --  sem of the statement.
                           Sem_Name.Replace_Name (Label, Stmt, N_Stmt);
                        end if;
                        Free_Iir (Stmt);
                     end if;
                  end;
               end if;
            when others =>
               null;
         end case;
      end loop;
   end Sem_Labels_2;

   --  Semantize a declaration list DECL_LIST and a sequential statement
   --  list STMTS_LIST from declaration DECL.
   --  This is used by processes and subprograms semantization.
   procedure Sem_Sequential_Statements
     (Decl : Iir;
      Decl_List : Iir_Declaration_List;
      Stmts_List: Iir_Sequential_Statement_List)
   is
      --  Transition array for states.
      --  indexed as passive_array (current_state, callee_state).
      type Tri_State_Array is
        array (Tri_State_Type, Tri_State_Type) of Tri_State_Type;

      --  Passive or purity state:
      --  FALSE is the strongest (true < unknown < false).
      False_Array : constant Tri_State_Array :=
        (Unknown => (Unknown => Unknown, False => False, True => Unknown),
         False => (Unknown => False, False => False, True => False),
         True => (Unknown => Unknown, False => False, True => True));

      --  Wait state: TRUE is the strongest (false < unknown < true).
      True_Array : constant Tri_State_Array :=
        (Unknown => (Unknown => Unknown, False => Unknown, True => True),
         False => (Unknown => Unknown, False => False, True => True),
         True => (Unknown => True, False => True, True => True));

      Outer_Subprogram: Iir;
      Passive_State, Wait_State, Purity_State : Tri_State_Type;
      Callees_List : Iir_Callees_List;
      Callee: Iir;
   begin
      Outer_Subprogram := Current_Subprogram;
      Current_Subprogram := Decl;

      --  Set states.
      Set_Seen_Flag (Decl, False);
      Set_Passive_State (Decl, Unknown);
      Set_Wait_State (Decl, Unknown);
      Set_Purity_State (Decl, Unknown);

      -- Sem declarations
      Sem_Labels (Stmts_List);
      Sem_Declaration_List (Decl_List, Decl);
      Sem_Specification_List (Decl_List, null);

      -- Sem statements.
      Sem_Sequential_Statements_Internal (Stmts_List);

      --  Update states.
      Passive_State := Get_Passive_State (Decl);
      Wait_State := Get_Wait_State (Decl);
      Purity_State := Get_Purity_State (Decl);

      --  If a state in unknown, then initialize to the weakest value.
      if Passive_State = Unknown then
         Passive_State := True;
      end if;
      if Wait_State = Unknown then
         Wait_State := False;
      end if;
      if Purity_State = Unknown then
         Purity_State := True;
      end if;
      Callees_List := Get_Callees_List (Decl);
      if Callees_List /= null then
         for I in Natural loop
            Callee := Get_Nth_Element (Callees_List, I);
            exit when Callee = null;
            Passive_State := False_Array (Passive_State,
                                          Get_Passive_State (Callee));
            Wait_State := True_Array (Wait_State,
                                      Get_Wait_State (Callee));
            Purity_State := False_Array (Purity_State,
                                         Get_Purity_State (Callee));
         end loop;
      end if;
      Set_Passive_State (Decl, Passive_State);
      Set_Wait_State (Decl, Wait_State);
      Set_Purity_State (Decl, Purity_State);

      --  Check states.
      --  FIXME: this is not good enough, since it doesn't handle correctly
      --   forward declaration.  Avoid spurious warnings.
      if False then --and not Flags.Will_Elaborate then
         if Get_Kind (Decl) in Iir_Kinds_Process_Statement
           and then Get_Passive_Flag (Decl) = True
           and then Passive_State = Unknown
         then
            Warning_Msg_Sem
              ("can't assert that " & Disp_Node (Decl)
               & " is passive; will be checked at elaboration", Decl);
         end if;
         if Get_Kind (Decl) = Iir_Kind_Sensitized_Process_Statement
           and then Wait_State = Unknown
         then
            Warning_Msg_Sem
              ("can't assert that " & Disp_Node (Decl)
               & " has no wait statments; will be checked at elaboration",
               Decl);
         end if;
         if Get_Kind (Decl) in Iir_Kinds_Function_Specification
           and then Get_Pure_Flag (Decl) = True
           and then Purity_State = Unknown
         then
            Warning_Msg_Sem
              ("can't assert that " & Disp_Node (Decl)
               & " is pure; will be checked at elaboration", Decl);
         end if;
      end if;
      Current_Subprogram := Outer_Subprogram;
   end Sem_Sequential_Statements;

   --  Return TRUE iff LEFT and RIGHT are (in depth) equal.
   --  This corresponds to conformance rules, LRM93 2.7
   function Are_Trees_Equal (Left, Right : Iir) return Boolean
   is
      El_Left, El_Right : Iir;
   begin
      --  Short-cut to speed up.
      if Left = Right then
         return True;
      end if;

      --  Handle null_iir.
      if Left = Null_Iir or Right = Null_Iir then
         --  Note: LEFT *xor* RIGHT is null_iir.
         return False;
      end if;

      --  If nodes are not of the same kind, then they are not equals!
      --  Exept for procedure and function declarations.
      if Get_Kind (Left) /= Get_Kind (Right) then
         if (Get_Kind (Left) in Iir_Kinds_Procedure_Declaration
             and then Get_Kind (Right) in Iir_Kinds_Procedure_Declaration)
           or else
           (Get_Kind (Left) in Iir_Kinds_Function_Declaration
            and then Get_Kind (Right) in Iir_Kinds_Function_Declaration)
         then
            null;
         else
            return False;
         end if;
      end if;

      case Get_Kind (Left) is
         when Iir_Kinds_Procedure_Declaration =>
            return Are_Trees_Equal (Get_Interface_Declaration (Left),
                                    Get_Interface_Declaration (Right));
         when Iir_Kinds_Function_Declaration =>
            if not Are_Trees_Equal (Get_Interface_Declaration (Left),
                                    Get_Interface_Declaration (Right))
            then
               return False;
            end if;
            return Get_Return_Type (Left) = Get_Return_Type (Right);
         when Iir_Kind_Interface_List =>
            for I in Natural loop
               El_Left := Get_Nth_Element (Left, I);
               El_Right := Get_Nth_Element (Right, I);
               exit when El_Left = null and El_Right = null;
               if El_Left = null or El_Right = null then
                  return False;
               end if;
               if not Are_Trees_Equal (El_Left, El_Right) then
                  return False;
               end if;
            end loop;
            return True;
         when Iir_Kind_Constant_Interface_Declaration
           | Iir_Kind_Variable_Interface_Declaration
           | Iir_Kind_Signal_Interface_Declaration
           | Iir_Kind_File_Interface_Declaration =>
            if Get_Lexical_Layout (Left) /= Get_Lexical_Layout (Right)
              or else Get_Mode (Left) /= Get_Mode (Right)
            then
               return False;
            end if;
            if not Are_Trees_Equal (Get_Type (Left), Get_Type (Right)) then
               return False;
            end if;
            El_Left := Get_Default_Value (Left);
            El_Right := Get_Default_Value (Right);
            if ((El_Left = Null_Iir) xor (El_Right = Null_Iir)) = True then
               return False;
            end if;
            if El_Left /= Null_Iir
              and then Are_Trees_Equal (El_Left, El_Right) = False
            then
               return False;
            end if;
            return True;

         when Iir_Kind_Integer_Subtype_Definition
           | Iir_Kind_Enumeration_Subtype_Definition
           | Iir_Kind_Floating_Subtype_Definition
           | Iir_Kind_Physical_Subtype_Definition =>
            if Get_Base_Type (Left) /= Get_Base_Type (Right)
              or else Get_Resolution_Function (Left)
              /= Get_Resolution_Function (Right)
            then
               return False;
            end if;
            if Get_Type_Declarator (Left) /= Get_Type_Declarator (Right) then
               return False;
            end if;
            if Are_Trees_Equal (Get_Range_Constraint (Left),
                                Get_Range_Constraint (Right)) = False
            then
               return False;
            end if;
            return True;
         when Iir_Kind_Array_Subtype_Definition =>
            if Get_Base_Type (Left) /= Get_Base_Type (Right)
              or else (Get_Resolution_Function (Left)
                       /= Get_Resolution_Function (Right))
            then
               return False;
            end if;
            declare
               L_Left, L_Right : Iir_List;
            begin
               L_Left := Get_Index_List (Left);
               L_Right := Get_Index_List (Right);
               for I in Natural loop
                  El_Left := Get_Nth_Element (L_Left, I);
                  El_Right := Get_Nth_Element (L_Right, I);
                  exit when El_Left = Null_Iir;
                  if not Are_Trees_Equal (El_Left, El_Right) then
                     return False;
                  end if;
               end loop;
            end;
            return True;

         when Iir_Kind_Integer_Literal
           | Iir_Kind_Enumeration_Literal =>
            if Get_Value (Left) /= Get_Value (Right) then
               return False;
            end if;
            return Are_Trees_Equal (Get_Literal_Origin (Left),
                                    Get_Literal_Origin (Right));
         when Iir_Kind_Physical_Int_Literal =>
            if Get_Value (Left) /= Get_Value (Right)
              or else Get_Name (Left) /= Get_Name (Right)
            then
               return False;
            end if;
            return Are_Trees_Equal (Get_Literal_Origin (Left),
                                    Get_Literal_Origin (Right));
         when Iir_Kind_Floating_Point_Literal =>
            if Get_Fp_Value (Left) /= Get_Fp_Value (Right) then
               return False;
            end if;
            return Are_Trees_Equal (Get_Literal_Origin (Left),
                                    Get_Literal_Origin (Right));

         when Iir_Kinds_Dyadic_Operator =>
            return Are_Trees_Equal (Get_Left (Left), Get_Left (Right))
              and then Are_Trees_Equal (Get_Right (Left), Get_Right (Right));
         when Iir_Kinds_Monadic_Operator =>
            return Are_Trees_Equal (Get_Operand (Left), Get_Operand (Right));

         when Iir_Kind_Access_Type_Definition
           | Iir_Kind_Record_Type_Definition
           | Iir_Kind_Array_Type_Definition
           | Iir_Kind_Enumeration_Type_Definition
           | Iir_Kind_File_Type_Definition =>
            return Left = Right;

         when Iir_Kind_Range_Expression =>
            if Get_Type (Left) /= Get_Type (Right)
              or else Get_Direction (Left) /= Get_Direction (Right)
            then
               return False;
            end if;
            if not Are_Trees_Equal (Get_Left_Limit (Left),
                                    Get_Left_Limit (Right))
              or else not Are_Trees_Equal (Get_Right_Limit (Left),
                                           Get_Right_Limit (Right))
            then
               return False;
            end if;
            return True;

         when Iir_Kind_String_Literal =>
            return Get_Identifier (Left) = Get_Identifier (Right);

         when Iir_Kind_Aggregate =>
            if not Are_Trees_Equal (Get_Type (Left), Get_Type (Right)) then
               return False;
            end if;
            declare
               Assoc_L, Assoc_R : Iir_List;
               El_L, El_R : Iir;
            begin
               Assoc_L := Get_Association_Choices_List (Left);
               Assoc_R := Get_Association_Choices_List (Right);
               for I in Natural loop
                  El_L := Get_Nth_Element (Assoc_L, I);
                  El_R := Get_Nth_Element (Assoc_R, I);
                  exit when El_L = Null_Iir and El_R = Null_Iir;
                  if not Are_Trees_Equal (El_L, El_R) then
                     return False;
                  end if;
               end loop;
               return True;
            end;

         when Iir_Kind_Association_Choice_By_None =>
            return Are_Trees_Equal (Get_Associated (Left),
                                    Get_Associated (Right));
         when Iir_Kind_Association_Choice_By_Name =>
            if not Are_Trees_Equal (Get_Name (Left), Get_Name (Right)) then
               return False;
            end if;
            return Are_Trees_Equal (Get_Associated (Left),
                                    Get_Associated (Right));
         when Iir_Kind_Association_Choice_By_Expression
           | Iir_Kind_Association_Choice_By_Range =>
            if not Are_Trees_Equal (Get_Expression (Left),
                                    Get_Expression (Right)) then
               return False;
            end if;
            return Are_Trees_Equal (Get_Associated (Left),
                                    Get_Associated (Right));
         when others =>
            Error_Kind ("are_trees_equal", Left);
      end case;
   end Are_Trees_Equal;

   procedure Check_Conformance_Rules (Subprg, Spec: Iir) is
   begin
      if not Are_Trees_Equal (Subprg, Spec) then
         --  FIXME: should explain why it does not conform ?
         Error_Msg_Sem ("body does not conform with specification at "
                        & Disp_Location (Spec), Subprg);
      end if;
   end Check_Conformance_Rules;

   --  SUBPRG is either a _specification or a _spec_body.
   function Sem_Subprogram_Declaration (A_Subprg: Iir) return Iir
   is
      use Std_Names;
      Spec: Iir;
      Is_Body : Boolean;
      Subprg : Iir := A_Subprg;
      Interface_List : Iir_List;
   begin
      case Get_Kind (Subprg) is
         when Iir_Kind_Function_Spec_Body
           | Iir_Kind_Procedure_Spec_Body =>
            Is_Body := True;
         when Iir_Kind_Function_Specification
           | Iir_Kind_Procedure_Specification =>
            Is_Body := False;
         when others =>
            Error_Kind ("sem_subprogram_declaration", Subprg);
      end case;

      --  Sem interfaces
      Sem_Name.Open_Declarative_Region;
      Interface_List := Get_Interface_Declaration (Subprg);
      if Get_Kind (Subprg) in Iir_Kinds_Function_Specification then
         Sem_Interface_List (Interface_List, Interface_Function);
         Set_Return_Type
           (Subprg, Sem_Subtype_Indication (Get_Return_Type (Subprg)));
      else
         Sem_Interface_List (Interface_List, Interface_Procedure);
      end if;
      Sem_Name.Close_Declarative_Region;

      case Get_Identifier (Subprg) is
         when Name_Abs
           | Name_Not =>
            --  LRM93 2.3.1
            --  The subprogram specification of a unary operator must have a
            --  single parameter.
            if Get_Nbr_Elements (Interface_List) /= 1 then
               Error_Msg_Sem
                 ("spec of an unary operator must have a single parameter",
                  Subprg);
            end if;
         when Name_Logical_Operators
           | Name_Xnor
           | Name_Mod
           | Name_Rem
           | Name_Op_Mul
           | Name_Op_Div
           | Name_Relational_Operators
           | Name_Op_Concatenation
           | Name_Shift_Operators
           | Name_Op_Exp =>
            --  LRM93 2.3.1
            --  The subprogram specification of a binary operator must have
            --  two parameters.
            if Get_Nbr_Elements (Interface_List) /= 2 then
               Error_Msg_Sem
                 ("spec of an binary operator must have two parameters",
                  Subprg);
            end if;
         when Name_Op_Plus
           | Name_Op_Minus =>
            --  LRM93 2.3.1
            --  For each of the operators "+" and "-", overloading is allowed
            --  both as a unary operator and as a binary operator.
            if Get_Nbr_Elements (Interface_List) not in 1 .. 2 then
               Error_Msg_Sem
                 ("spec of ""+"" or ""-"" may only have 1 or 2 parameters",
                  Subprg);
            end if;
         when others =>
            null;
      end case;

      if Is_Body then
         Spec := Find_Subprogram_Specification (Subprg);
      end if;
--       if Spec /= Null_Iir and not Is_Body then
--          --  There is a previous specification with the same signature.
--          Error_Msg_Sem (Disp_Node (Subprg) & " already declared at "
--                         & Disp_Location (Spec), Subprg);
--       end if;

      if Is_Body and then Spec /= Null_Iir then
         -- SUBPRG is the body of the specification SPEC.
         Check_Conformance_Rules (Subprg, Spec);
         declare
            Old : Iir := Subprg;
         begin
            -- Free_Recursive (Get_Interface_Declaration (Old), True);
            case Get_Kind (Subprg) is
               when Iir_Kind_Function_Spec_Body =>
                  Subprg := Clone_Iir (Subprg, Iir_Kind_Function_Body);
               when Iir_Kind_Procedure_Spec_Body =>
                  Subprg := Clone_Iir (Subprg, Iir_Kind_Procedure_Body);
               when others =>
                  Error_Kind ("ssd1", Subprg);
            end case;
            Free_Iir (Old);
         end;
         Set_Subprogram_Specification (Subprg, Spec);
         Set_Subprogram_Body (Spec, Subprg);
      else
         Spec := Subprg;
         Sem_Name.Add_Name (Spec);
         Set_Visible_Flag (Spec, True);
      end if;

      if Is_Body then
         Sem_Name.Open_Declarative_Region;
         Set_Is_Within_Flag (Spec, True);

         -- Add the arguments name into the current declarative region.
         declare
            El: Iir;
            Interface_List: Iir;
         begin
            Interface_List := Get_Interface_Declaration (Spec);
            for I in Natural loop
               El := Get_Nth_Element (Interface_List, I);
               exit when El = null;
               Sem_Name.Add_Name (El);
            end loop;
         end;

         Sem_Sequential_Statements (Spec,
                                    Get_Subprogram_Declaration (Subprg),
                                    Get_Subprogram_Statements (Subprg));

         Set_Is_Within_Flag (Spec, False);
         Sem_Name.Close_Declarative_Region;
      end if;
      return Subprg;
   end Sem_Subprogram_Declaration;

   procedure Sem_Alias_Declaration (Alias: Iir_Alias_Declaration)
   is
      N_Type: Iir;
      N_Name: Iir;
      Name_Type : Iir;
   begin
      Sem_Name.Add_Name (Alias);

      --  FIXME: currently only object alias are handled.
      N_Name := Sem_Expression (Get_Name (Alias), Null_Iir);
      if N_Name = Null_Iir then
         return;
      end if;

      Name_Visible (Alias);

      Set_Name (Alias, N_Name);
      case Get_Kind (N_Name) is
         when Iir_Kinds_Object_Declaration
           | Iir_Kind_Slice_Name
           | Iir_Kind_Indexed_Name
           | Iir_Kind_Selected_Name
           | Iir_Kind_Selected_Element =>
            Set_Base_Name (Alias, Get_Base_Name (N_Name));
         when others =>
            Error_Msg_Sem ("can only alias named object", Alias);
            return;
      end case;

      --  LRM93 4.3.3.1 Object Aliases.
      --  1. A signature may not appear in a declaration of an object alias.
      -- FIXME: todo.
      --
      --  2. The name must be a static name that denotes an object.
      if Get_Name_Staticness (N_Name) < Globally then
         Error_Msg_Sem ("aliased name must be a static name", Alias);
      end if;

      --  LRM93 4.3.3.1
      --  The base type of the name specified in an alias declaration must be
      --  the same as the base type of the type mark in the subtype indication
      --  (if the subtype indication is present);
      Name_Type := Get_Type (N_Name);
      N_Type := Get_Type (Alias);
      if N_Type = Null_Iir then
         Set_Type (Alias, Name_Type);
         N_Type := Name_Type;
      else
         N_Type := Sem_Subtype_Indication (N_Type);
         if N_Type /= Null_Iir then
            Set_Type (Alias, N_Type);
            if Get_Base_Type (N_Type) /= Get_Base_Type (Name_Type) then
               Error_Msg_Sem ("base type of aliased name and name mismatch",
                              Alias);
            end if;
         end if;
      end if;

      --  LRM93 4.3.3.1
      --  This type must not be a multi-dimensional array type.
      if Get_Kind (N_Type) in Iir_Kinds_Array_Type_Definition then
         if Get_Nbr_Elements (Get_Index_List (N_Type)) > 1 then
            Error_Msg_Sem
              ("aliased name must not be a multi-dimensional array type",
               Alias);
         end if;
         if Get_Type_Staticness (N_Type) = Locally
           and then Get_Type_Staticness (Name_Type) = Locally
           and then Eval_Discrete_Type_Length
           (Get_Nth_Element (Get_Index_List (N_Type), 0))
           /= Eval_Discrete_Type_Length
           (Get_Nth_Element (Get_Index_List (Name_Type), 0))
         then
            Error_Msg_Sem ("number of elements not matching in type and name",
                           Alias);
         end if;
      end if;

      Set_Name_Staticness (Alias, Get_Name_Staticness (N_Name));
      Set_Expr_Staticness (Alias, Get_Expr_Staticness (N_Name));
   end Sem_Alias_Declaration;

   procedure Sem_Configuration_Specification
     (Decls : Iir_Concurrent_Statement_List;
      Conf : Iir_Configuration_Specification) is
   begin
      Sem_Component_Specification (Decls, Conf);
      --  Extend scope of component interface declaration.
      Sem_Name.Open_Scope_Extension;
      Sem_Name.Add_Component_Declarations (Get_Component_Name (Conf));
      Sem_Binding_Indication (Get_Binding_Indication (Conf),
                              Get_Component_Name (Conf), Conf);
      --  FIXME: check default port and generic association.
      Sem_Name.Close_Scope_Extension;
   end Sem_Configuration_Specification;

   function Get_Entity_Class_Kind (Decl : Iir) return Tokens.Token_Type
   is
      use Tokens;
   begin
      case Get_Kind (Decl) is
         when Iir_Kind_Design_Unit =>
            case Get_Kind (Get_Library_Unit (Decl)) is
               when Iir_Kind_Entity_Declaration =>
                  return Tok_Entity;
               when Iir_Kind_Architecture_Declaration =>
                  return Tok_Architecture;
               when Iir_Kind_Configuration_Declaration =>
                  return Tok_Configuration;
               when Iir_Kind_Package_Declaration =>
                  return Tok_Package;
               when others =>
                  Error_Kind ("get_entity_class_kind(unit)", Decl);
            end case;
         when Iir_Kind_Procedure_Specification
           | Iir_Kind_Procedure_Spec_Body
           | Iir_Kind_Implicit_Procedure_Declaration =>
            return Tok_Procedure;
         when Iir_Kind_Function_Specification
           | Iir_Kind_Function_Spec_Body
           | Iir_Kind_Implicit_Function_Declaration =>
            return Tok_Function;
         when Iir_Kind_Type_Declaration =>
            return Tok_Type;
         when Iir_Kind_Subtype_Declaration =>
            return Tok_Subtype;
         when Iir_Kind_Constant_Declaration
           | Iir_Kind_Constant_Interface_Declaration =>
            return Tok_Constant;
         when Iir_Kind_Signal_Declaration
           | Iir_Kind_Signal_Interface_Declaration
           | Iir_Kind_Implicit_Signal_Declaration =>
            return Tok_Signal;
         when Iir_Kind_Variable_Declaration
           | Iir_Kind_Variable_Interface_Declaration =>
            return Tok_Variable;
         when Iir_Kind_Component_Declaration =>
            return Tok_Component;
         when Iir_Kind_Concurrent_Conditional_Signal_Assignment
           | Iir_Kind_Concurrent_Selected_Signal_Assignment
           | Iir_Kind_Sensitized_Process_Statement
           | Iir_Kind_Process_Statement
           | Iir_Kind_Concurrent_Assertion_Statement
           | Iir_Kind_Component_Instantiation_Statement
           | Iir_Kind_Block_Statement
           | Iir_Kind_If_Statement
           | Iir_Kind_For_Loop_Statement
           | Iir_Kind_While_Loop_Statement
           | Iir_Kind_Next_Statement
           | Iir_Kind_Exit_Statement
           | Iir_Kind_Signal_Assignment_Statement
           | Iir_Kind_Variable_Assignment_Statement
           | Iir_Kind_Assertion_Statement
           | Iir_Kind_Wait_Statement
           | Iir_Kind_Return_Statement
           | Iir_Kind_Case_Statement
           | Iir_Kind_Procedure_Call
           | Iir_Kind_Concurrent_Procedure_Call_Statement
           | Iir_Kind_Null_Statement =>
            return Tok_Label;
         when Iir_Kind_Enumeration_Literal =>
            return Tok_Literal;
         when Iir_Kind_Physical_Unit =>
            return Tok_Units;
         when Iir_Kind_Group_Declaration =>
            return Tok_Group;
         when Iir_Kind_File_Declaration
           | Iir_Kind_File_Interface_Declaration =>
            return Tok_File;
         when Iir_Kind_Attribute_Declaration =>
            --  Even if an attribute can't have a attribute...
            --  Because an attribute declaration can appear in a declaration
            --  region.
            return Tok_Attribute;
         when others =>
            Error_Kind ("get_entity_class_kind", Decl);
      end case;
      return Tok_Invalid;
   end Get_Entity_Class_Kind;

   procedure Sem_Group_Declaration (Group : Iir_Group_Declaration)
   is
      use Tokens;

      Template : Iir_Group_Template_Declaration;
      Entity_Class_List : Iir_Entity_Class_Entry_List;
      Is_Box_Reached : Boolean;
      Constituent_List : Iir_Group_Constituent_List;
      Class : Token_Type;
      El : Iir;
      El_Entity : Iir_Entity_Class;
   begin
      Sem_Name.Add_Name (Group);
      Template := Find_Declaration (Get_Name (Group), Decl_Group_Template);
      if Template = null then
         return;
      end if;
      Set_Name (Group, Template);
      Entity_Class_List := Get_Entity_Class_Entry_List (Template);
      Constituent_List := Get_Group_Constituent_List (Group);
      Is_Box_Reached := False;
      for I in Natural loop
         El := Get_Nth_Element (Constituent_List, I);
         if El = null then
            --  Several constituents seen and allowed.
            exit when Is_Box_Reached;
            --  End of entity_class list reached.
            El_Entity := Get_Nth_Element (Entity_Class_List, I);
            exit when El_Entity = null;
            --  Several constituents allowed, one seen.
            exit when Get_Entity_Kind (El_Entity) = Tok_Box;
            El_Entity := Get_Nth_Element (Entity_Class_List, I + 1);
            --  Several constituents allowed, zero seen.
            exit when El_Entity /= null
              and then Get_Entity_Kind (El_Entity) = Tok_Box;
            Error_Msg_Sem ("not enough elements in group constituent list",
                           Group);
            exit;
         end if;
         if not Is_Box_Reached then
            El_Entity := Get_Nth_Element (Entity_Class_List, I);
            if El_Entity = null then
               Error_Msg_Sem ("to many elements in group constituent list",
                              Group);
               exit;
            end if;
            Class := Get_Entity_Kind (El_Entity);
            if Class = Tok_Box then
               --  LRM93 4.6
               --  An entity class entry that includes a box (<>) allows zero
               --  or more group constituents to appear in this position in the
               --  corresponding group declaration.
               Is_Box_Reached := True;
               Class := Get_Entity_Kind
                 (Get_Nth_Element (Entity_Class_List, I - 1));
            end if;
         end if;
         El := Find_Declaration (El, Decl_Any);
         if El /= null then
            --  LRM93 4.7
            --  It is an error if the class of any group constituent in the
            --  group constituent list is not the same as the class specified
            --  by the corresponding entity class entry in the entity class
            --  entry list of the group template.
            if Get_Entity_Class_Kind (El) /= Class then
               Error_Msg_Sem
                 ("constituent no of class `" & Tokens.Image (Class) & ''',
                  El);
            end if;
            Replace_Nth_Element (Constituent_List, I, El);
         end if;
      end loop;
   end Sem_Group_Declaration;

   --  Decorate DECL with attribute ATTR.
   --  If CHECK_CLASS is true, class of DECL must be class of ATTR, otherwise
   --   returns silently.
   --  If CHECK_DEFINED is true, DECL must not have bee, decorated, otherwise
   --   returns silently.
   procedure Attribute_A_Decl
     (Decl : Iir;
      Attr : Iir_Attribute_Specification;
      Check_Class : Boolean;
      Check_Defined : Boolean)
   is
      use Tokens;
      Decl1 : Iir;
      List : Iir_Attribute_Value_List;
      El : Iir_Attribute_Value;
   begin
      --  LRM93 5.1
      --  An entity designator that denotes an alias of an object is required
      --  to denote the entire object, and not a subelement or slice thereof.
      if Get_Kind (Decl) = Iir_Kind_Alias_Declaration then
         Decl1 := Get_Object_Prefix (Get_Name (Decl));
      else
         Decl1 := Decl;
      end if;

      --  LRM93 5.1
      --  It is an error if the class of those names is not the same as that
      --  denoted by the entity class.
      if Get_Entity_Class_Kind (Decl1) /= Get_Entity_Kind (Attr) then
         if Check_Class then
            Error_Msg_Sem (Disp_Node (Decl1) & " is not of class `"
                           & Tokens.Image (Get_Entity_Kind (Attr)) & ''',
                           Attr);
         end if;
         return;
      end if;

      --  LRM93 5.1
      --  An attribute specification for an attribute of a design unit
      --  (ie an entity interface, an architecture, a configuration, or a
      --  package) must appear immediately within the declarative part of
      --  that design unit.
      case Get_Kind (Decl1) is
         when Iir_Kind_Entity_Declaration
           | Iir_Kind_Architecture_Declaration
           | Iir_Kind_Configuration_Declaration
           | Iir_Kind_Package_Declaration =>
            if Decl1 /= Get_Library_Unit (Get_Current_Design_Unit) then
               Error_Msg_Sem (Disp_Node (Attr) & " must appear immediatly "
                              & "within " & Disp_Node (Decl1), Attr);
               return;
            end if;
         when others =>
            null;
      end case;

      List := Get_Attribute_Value_List (Decl1);
      if List = Null_Iir then
         List := Create_Iir (Iir_Kind_Attribute_Value_List);
         Set_Attribute_Value_List (Decl1, List);
      end if;

      --  LRM93 5.1
      --  It is an error if a given attribute is associated more than once with
      --  a given named entity.
      for I in Natural loop
         El := Get_Nth_Element (List, I);
         exit when El = null;
         if Get_Attribute_Designator (Get_Attribute_Specification (El))
           = Get_Attribute_Designator (Attr)
         then
            if Get_Attribute_Specification (El) = Attr then
               --  Was already specified with the same attribute value.
               --  This is possible only in one case:
               --
               --    signal S1       : real;
               --    alias  S1_too   : real is S1;
               --    attribute ATTR : T1;
               --    attribute ATTR of ALL : signal is '1';
               return;
            end if;
            if Check_Defined then
               Error_Msg_Sem
                 (Disp_Node (Decl) & " has already " & Disp_Node (Attr), Attr);
               Error_Msg_Sem ("previous attribute specification at "
                              & Disp_Location (El), Attr);
            end if;
            return;
         end if;
      end loop;
      El := Create_Iir (Iir_Kind_Attribute_Value);
      Location_Copy (El, Attr);
      Set_Name_Staticness (El, None);
      Set_Attribute_Specification (El, Attr);
      Set_Expr_Staticness (El, Get_Expr_Staticness (Get_Expression (Attr)));
      Set_Designated_Entity (El, Decl);
      Set_Type (El, Get_Type (Get_Attribute_Designator (Attr)));
      Set_Base_Name (El, El);
      Append_Element (List, El);
      Append_Element (Get_Attribute_Value_List (Attr), El);
   end Attribute_A_Decl;

   --  Return TRUE if an entity was attributed.
   function Sem_Named_Entities
     (Scope : Iir;
      Name : Name_Id;
      Attr : Iir_Attribute_Specification;
      Check_Class : Boolean;
      Check_Defined : Boolean)
     return Boolean
   is
      Res : Boolean;

      procedure Sem_Named_Entity (Ent : Iir) is
      begin
         case Get_Kind (Ent) is
            when Iir_Kind_Design_Unit
              | Iir_Kinds_Concurrent_Statement
              | Iir_Kinds_Function_Specification
              | Iir_Kinds_Procedure_Specification
              | Iir_Kinds_Sequential_Statement
              | Iir_Kinds_Object_Declaration
              | Iir_Kind_Type_Declaration
              | Iir_Kind_Subtype_Declaration
              | Iir_Kind_Component_Declaration =>
               if Name = Null_Identifier or else Get_Identifier (Ent) = Name
               then
                  if Get_Visible_Flag (Ent) = False then
                     Error_Msg_Sem
                       (Disp_Node (Ent) & " is not yet visible", Attr);
                  else
                     Attribute_A_Decl (Ent, Attr, Check_Class, Check_Defined);
                     Res := True;
                  end if;
               end if;
            when Iir_Kind_Attribute_Declaration
              | Iir_Kind_Attribute_Specification
              | Iir_Kind_Configuration_Specification
              | Iir_Kind_Use_Clause =>
               null;
            when others =>
               Error_Kind ("sem_named_entity", Ent);
         end case;
      end Sem_Named_Entity;

      procedure Sem_Named_Entity_List (List : Iir_List)
      is
         El : Iir;
      begin
         if List = Null_Iir then
            return;
         end if;
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            Sem_Named_Entity (El);
         end loop;
      end Sem_Named_Entity_List;

      First : Iir;
   begin
      Res := False;
      case Get_Kind (Scope) is
         when Iir_Kind_Architecture_Declaration =>
            Sem_Named_Entity (Get_Design_Unit (Scope));
            Sem_Named_Entity_List (Get_Declaration_List (Scope));
            Sem_Named_Entity_List (Get_Concurrent_Statement_List (Scope));
         when Iir_Kind_Process_Statement =>
            Sem_Named_Entity (Scope);
            Sem_Named_Entity_List (Get_Declaration_List (Scope));
            Sem_Named_Entity_List (Get_Process_Statement_List (Scope));
         when Iir_Kind_Entity_Declaration =>
            Sem_Named_Entity (Get_Design_Unit (Scope));
            Sem_Named_Entity_List (Get_Generic_List (Scope));
            Sem_Named_Entity_List (Get_Port_List (Scope));
            Sem_Named_Entity_List (Get_Declaration_List (Scope));
            Sem_Named_Entity_List (Get_Concurrent_Statement_List (Scope));
         when Iir_Kind_Block_Statement =>
            Sem_Named_Entity (Scope);
            First := Get_Block_Header (Scope);
            if First /= Null_Iir then
               Sem_Named_Entity_List (Get_Generic_List (First));
               Sem_Named_Entity_List (Get_Port_List (First));
            end if;
            Sem_Named_Entity_List (Get_Declaration_List (Scope));
            Sem_Named_Entity_List (Get_Concurrent_Statement_List (Scope));
         when Iir_Kind_Package_Declaration =>
            Sem_Named_Entity (Get_Design_Unit (Scope));
            Sem_Named_Entity_List (Get_Declaration_List (Scope));
         when Iir_Kind_Configuration_Declaration =>
            Sem_Named_Entity (Get_Design_Unit (Scope));
         when Iir_Kind_Function_Spec_Body
           | Iir_Kind_Function_Body
           | Iir_Kind_Procedure_Spec_Body
           | Iir_Kind_Procedure_Body =>
            First := Get_Subprogram_Spec (Scope);
            Sem_Named_Entity (First);
            Sem_Named_Entity_List (Get_Interface_Declaration (First));
            Sem_Named_Entity_List (Get_Subprogram_Declaration (Scope));
            Sem_Named_Entity_List (Get_Subprogram_Statements (Scope));
         when others =>
            Error_Kind ("sem_named_entities", Scope);
      end case;
      return Res;
   end Sem_Named_Entities;

   procedure Sem_Attribute_Specification
     (Spec : Iir_Attribute_Specification;
      Scope : Iir)
   is
      Name : Iir_Attribute_Declaration;
      List : Iir_List;
      Expr : Iir;
      Res : Boolean;
   begin
      --  LRM93 5.1
      --  The attribute designator must denote an attribute.
      Name := Find_Declaration (Get_Attribute_Designator (Spec),
                                Decl_Attribute);
      if Name /= Null_Iir then
         Set_Attribute_Designator (Spec, Name);

         --  LRM93 5.1
         --  The type of the expression in the attribute specification must be
         --  the same as (or implicitly convertible to) the type mark in the
         --  corresponding attribute declaration.
         Expr := Sem_Expression (Get_Expression (Spec), Get_Type (Name));
         if Expr /= Null_Iir then
            Check_Read (Expr);
            Set_Expression (Spec, Eval_Expr_If_Static (Expr));
         end if;
      end if;

      --  LRM93 5.1
      --  The entity name list identifies those named entities, both
      --  implicitly and explicitly defined, that inherit the attribute, as
      --  defined below:
      List := Get_Entity_Name_List (Spec);
      case Get_Kind (List) is
         when Iir_Kind_Designator_List =>
            --  o If a list of entity designators is supplied, then the
            --  attribute specification applies to the named entities denoted
            --  by those designators.
            declare
               use Sem_Name;
               El : Iir;
            begin
               for I in Natural loop
                  El := Get_Nth_Element (List, I);
                  exit when El = null;
                  if not Sem_Named_Entities (Scope, Get_Identifier (El), Spec,
                                             True, True)
                  then
                     Error_Msg_Sem ("no named entities `" & Get_String (El)
                                    & "' in current scope", El);
                  end if;
               end loop;
            end;
         when Iir_Kind_Designator_By_All
           | Iir_Kind_Designator_By_Others =>
            --  o If the reserved word OTHERS is supplied, then the attribute
            --  specification applies to named entities of the specified class
            --  that are declared in the immediately enclosing declarative
            --  part, provided that each such entity is not explicitly named
            --  in the entity name list of a previous attribute specification
            --  for the given attribute.
            if Get_Kind (List) = Iir_Kind_Designator_By_All then
               Res := Sem_Named_Entities (Scope, Null_Identifier, Spec,
                                          False, True);
            else
               Res := Sem_Named_Entities (Scope, Null_Identifier, Spec,
                                          False, False);
            end if;
         when others =>
            raise Internal_Error;
      end case;
   end Sem_Attribute_Specification;

   --  Compare ATYPE and TYPE_MARK.
   --  ATYPE is a type definition, which can be anonymous.
   --  TYPE_MARK is a subtype definition, established from a type mark.
   --   Therefore, it is the name of a type or a subtype.
   --  Return TRUE iff the type mark of ATYPE is TYPE_MARK.
   function Is_Same_Type_Mark (Atype : Iir; Type_Mark : Iir)
     return Boolean is
   begin
      if Get_Kind (Atype) in Iir_Kinds_Subtype_Definition
        and then Is_Anonymous_Type_Definition (Atype)
      then
         --  FIXME: to be removed; used to catch uninitialized type_mark.
         if Get_Type_Mark (Atype) = null then
            raise Internal_Error;
         end if;
         return Get_Type_Mark (Atype) = Type_Mark;
      else
         return Atype = Type_Mark;
      end if;
   end Is_Same_Type_Mark;

   procedure Sem_Disconnect_Specification
     (Dis : Iir_Disconnection_Specification)
   is
      Atype : Iir;
      Time_Expr : Iir;
      List : Iir;

      procedure Sem_A_Signal_Name (El : Iir)
      is
         Prefix : Iir;
      begin
         --  FIXME: errors are based on EL, which may not point on the current
         --  line.

         Prefix := Get_Base_Name (El);
         --  LRM93 5.3
         --  Each signal name in a signal list in a guarded signal
         --  specification must be a locally static name that
         --  denotes a guarded signal.
         case Get_Kind (Prefix) is
            when Iir_Kind_Signal_Declaration
              | Iir_Kind_Signal_Interface_Declaration =>
               null;
            when others =>
               Error_Msg_Sem ("object must be a signal", El);
               return;
         end case;
         if Get_Name_Staticness (El) /= Locally then
            Error_Msg_Sem ("signal name must be locally static", El);
         end if;
         if Get_Signal_Kind (Prefix) = Iir_No_Signal_Kind then
            Error_Msg_Sem ("signal must be a guarded signal", El);
         end if;
         Set_Has_Disconnect_Flag (Prefix, True);

         --  LRM93 5.3
         --  If the guarded signal is a declared signal or a slice of thereof,
         --  the type mark must be the same as the type mark indicated in the
         --  guarded sugnal specification.
         --  If the guarded signal is an array element of an explicitly
         --  declared signal, the type mark must be the same as the element
         --  subtype indication in the (explicit or implicit) array type
         --  declaration that declares the base type of the explicitly
         --  declared signal.
         --  If the guarded signal is a record element of an explicitly
         --  declared signal, then the type mark must be the same as the type
         --  mark in the element subtype definition of the record type
         --  declaration that declares the type of the explicitly declared
         --  signal.
         -- FIXME: to be checked: the expression type (as set by
         --  sem_expression) may be a base type instead of a type mark.
         if not Is_Same_Type_Mark (Get_Type (El), Atype) then
            Error_Msg_Sem ("type mark and signal type mismatch", El);
         end if;

         --  LRM93 5.3
         --  Each signal must be declared in the declarative part enclosing
         --  the disconnection specification.
         --  FIXME: todo.
      end Sem_A_Signal_Name;
   begin
      --  Sem type mark.
      Atype := Get_Type (Dis);
      Atype := Sem_Subtype_Indication (Atype);
      if Atype /= Null_Iir then
         Set_Type (Dis, Atype);
      end if;

      --  LRM93 5.3
      --  The time expression in a disconnection specification must be static
      --  and must evaluate to a non-negative value.
      Time_Expr := Sem_Expression
        (Get_Expression (Dis), Time_Subtype_Definition);
      if Time_Expr /= Null_Iir then
         Check_Read (Time_Expr);
         Set_Expression (Dis, Time_Expr);
         if Get_Expr_Staticness (Time_Expr) < Globally then
            Error_Msg_Sem ("time expression must be static", Time_Expr);
         end if;
      end if;

      List := Get_Signal_List (Dis);
      case Get_Kind (List) is
         when Iir_Kind_Designator_List =>
            declare
               El : Iir;
            begin
               for I in Natural loop
                  El := Get_Nth_Element (List, I);
                  exit when El = null;
                  El := Sem_Expression (El, null);
                  if El /= null then
                     Sem_A_Signal_Name (El);
                     Replace_Nth_Element (List, I, El);
                  end if;
               end loop;
            end;
         when Iir_Kind_Designator_By_All
           | Iir_Kind_Designator_By_Others =>
            --  FIXME: checks todo
            null;
         when others =>
            raise Internal_Error;
      end case;
   end Sem_Disconnect_Specification;

   -- Check all declarations of LIST are complete
   -- This is used for subprograms in any declaration_list or deferred
   -- constants in a package declaration.
   -- DECL is the declaration that contains the declaration_list LIST.
   procedure Check_Full_Declaration (List: Iir_Declaration_List; Decl: Iir)
   is
      El: Iir;
   begin
      for I in Natural loop
         El := Get_Nth_Element (List, I);
         exit when El = null;
         case Get_Kind (El) is
            when Iir_Kind_Constant_Declaration =>
               if Get_Deferred_Declaration_Flag (El) then
                  if Get_Deferred_Declaration (El) = Null_Iir then
                     Error_Msg_Sem ("missing value for constant declared at "
                                    & Disp_Location (El), Decl);
                  else
                     --  Remove from visibility the full declaration of the
                     --  constant.
                     Set_Deferred_Declaration (El, Null_Iir);
                  end if;
               end if;
            when Iir_Kind_Function_Specification
              | Iir_Kind_Procedure_Specification =>
               if Iirs.Get_Subprogram_Body (El) = null then
                  Error_Msg_Sem ("missing body for `"
                                 & Name_Table.Image (Get_Identifier (El))
                                 & "' declared at "
                                 & Disp_Location (El), Decl);
               end if;
            when Iir_Kind_Type_Declaration =>
               declare
                  Def : Iir;
               begin
                  Def := Get_Type (El);
                  if Get_Kind (Def) = Iir_Kind_Incomplete_Type_Definition
                    and then Get_Type_Declarator (Def) = El
                  then
                     Error_Msg_Sem ("missing full type declaration for "
                                    & Disp_Node (El), El);
                  end if;
               end;
            when others =>
               null;
         end case;
      end loop;
   end Check_Full_Declaration;

   --  Semantize every declaration of DECLS.
   --  STMTS is the concurrent statement list associated with DECLS if any,
   --   or null.  This is used for specification.
   procedure Sem_Declaration_List (Decls: Iir_Declaration_List; Parent : Iir)
   is
      Decl: Iir;
   begin
      -- Du to implicit declarations, the list can grow during sem.
      for I in Natural loop
         Decl := Get_Nth_Element (Decls, I);
         exit when Decl = Null_Iir;
         case Get_Kind (Decl) is
            when Iir_Kind_Type_Declaration =>
               Sem_Type_Declaration (Decl);
            when Iir_Kind_Subtype_Declaration =>
               Sem_Subtype_Declaration (Decl);
            when Iir_Kind_Signal_Declaration =>
               Sem_Object_Declaration (Decl, Parent);
            when Iir_Kind_Constant_Declaration =>
               Sem_Object_Declaration (Decl, Parent);
            when Iir_Kind_Variable_Declaration =>
               Sem_Object_Declaration (Decl, Parent);
            when Iir_Kind_Attribute_Declaration =>
               Sem_Attribute_Declaration (Decl);
            when Iir_Kind_Attribute_Specification =>
               Sem_Attribute_Specification (Decl, Parent);
            when Iir_Kind_Component_Declaration =>
               Sem_Component_Declaration (Decl);
            when Iir_Kinds_Function_Declaration
              | Iir_Kinds_Procedure_Declaration =>
               Replace_Nth_Element
                 (Decls, I, Sem_Subprogram_Declaration (Decl));
            when Iir_Kind_Alias_Declaration =>
               Sem_Alias_Declaration (Decl);
            when Iir_Kind_File_Declaration =>
               Sem_File_Declaration (Decl);
            when Iir_Kind_Use_Clause =>
               Sem_Use_Clause (Decl);
            when Iir_Kind_Configuration_Specification =>
               null;
            when Iir_Kind_Disconnection_Specification =>
               Sem_Disconnect_Specification (Decl);
            when Iir_Kind_Group_Template_Declaration =>
               Sem_Group_Template_Declaration (Decl);
            when Iir_Kind_Group_Declaration =>
               Sem_Group_Declaration (Decl);
            when Iir_Kind_Implicit_Signal_Declaration =>
               --  Added by sem, so nothing to do.
               null;
            when others =>
               Error_Kind ("sem_declaration_list", Decl);
         end case;
      end  loop;

      case Get_Kind (Parent) is
         when Iir_Kind_Package_Declaration
           | Iir_Kind_Entity_Declaration =>
            null;
         when others =>
            Check_Full_Declaration (Decls, Parent);
      end case;
   end Sem_Declaration_List;

   procedure Sem_Specification_List
     (Decls: Iir_Declaration_List;
      Stmts: Iir_Concurrent_Statement_List)
   is
      Decl: Iir;
   begin
      for I in Natural loop
         Decl := Get_Nth_Element (Decls, I);
         exit when Decl = null;
         case Get_Kind (Decl) is
            when Iir_Kind_Configuration_Specification =>
               Sem_Configuration_Specification (Stmts, Decl);
            when others =>
               null;
         end case;
      end loop;
   end Sem_Specification_List;

   procedure Sem_Entity_Declaration (Entity: Iir_Entity_Declaration)
   is
      Unit : Iir_Design_Unit;
      Port_List: Iir_Port_List;
      Generic_List: Iir_Generic_List;
      Decls : Iir_Declaration_List;
      Implicit : Implicit_Signal_Declaration_Type;
   begin
      Unit := Get_Design_Unit (Entity);
      Sem_Name.Add_Name (Unit);
      Set_Visible_Flag (Unit, True);

      Set_Is_Within_Flag (Entity, True);

      --  LRM93 10.1
      --  1.  An entity declaration, together with a corresponding architecture
      --      body.
      Sem_Name.Open_Declarative_Region;

      -- Check generic list.
      Generic_List := Get_Generic_List (Entity);
      if Generic_List /= null then
         Sem_Interface_List (Generic_List, Interface_Generic);
      end if;

      -- Check port list.
      Port_List := Get_Port_List (Entity);
      if Port_List /= null then
         Sem_Interface_List (Port_List, Interface_Port);
      end if;

      -- entity declarative part.
      Decls := Get_Declaration_List (Entity);

      Push_Declarative_Region_With_Signals (Implicit, Decls);
      Sem_Declaration_List (Decls, Entity);
      Sem_Specification_List (Decls, null);

      -- statements.
      Sem_Concurrent_Statement_List
        (Get_Concurrent_Statement_List (Entity), True);
      Pop_Declarative_Region_With_Signals (Implicit);
      Sem_Name.Close_Declarative_Region;
      Set_Is_Within_Flag (Entity, False);
   end Sem_Entity_Declaration;

   -- Semantize a waveform_list WAVEFORM_LIST that is assigned via statement
   -- ASSIGN_STMT to a subelement or a slice of a signal SIGNAL_DECL.
   procedure Sem_Waveform_List
     (Assign_Stmt: Iir;
      Waveform_List: Iir_Waveform_List;
      Waveform_Type : in out Iir)
   is
      Expr: Iir;
      We: Iir_Waveform_Element;
      Time, Last_Time : Iir_Int64;
   begin
      if Waveform_List = null then
         Error_Msg_Sem ("sem_waveform_list: no waveform", Assign_Stmt);
         return;
      end if;

      --  Start with -1 to allow after 0 ns.
      Last_Time := -1;
      for I in Natural loop
         We := Get_Nth_Element (Waveform_List, I);
         exit when We = Null_Iir;

         Expr := Get_We_Value (We);
         if Get_Kind (Expr) = Iir_Kind_Null_Literal then
            --  GHDL: allowed only if target is guarded; this is checked by
            --  sem_check_waveform_list.
            null;
         else
            if Get_Kind (Expr) = Iir_Kind_Aggregate
              and then Waveform_Type = Null_Iir
            then
               Error_Msg_Sem
                 ("type of waveform is unknown, (use type qualifier)", Expr);
            else
               Expr := Sem_Expression (Expr, Waveform_Type);
               if Expr /= Null_Iir then
                  Check_Read (Expr);
                  Set_We_Value (We, Eval_Expr_If_Static (Expr));
                  if Waveform_Type = Null_Iir then
                     Waveform_Type := Get_Type (Expr);
                  end if;
               end if;
            end if;
         end if;

         if Get_Time (We) /= Null_Iir then
            Expr := Sem_Expression (Get_Time (We), Time_Type_Definition);
            if Expr /= Null_Iir then
               Check_Read (Expr);
               if Get_Expr_Staticness (Expr) = Locally then
                  --  LRM93 8.4
                  --  It is an error if the time expression in a waveform
                  --  element evaluates to a negative value.
                  --  [...]
                  --  It is an error if the sequence of new transactions is not
                  --  in ascending order with repect to time.
                  -- GHDL: this must be checked at run-time, but this is also
                  --  checked now for static expressions.
                  Expr := Eval_Expr (Expr);
                  Time := Get_Value (Expr);
                  if Time < 0 then
                     Error_Msg_Sem
                       ("waveform time expression must be >= 0", Expr);
                  elsif Time <= Last_Time then
                     Error_Msg_Sem ("time must be greather than previous one",
                                    Expr);
                  else
                     Last_Time := Time;
                  end if;
               end if;
               Set_Time (We, Expr);
            end if;
         else
            if I /= 0 then
               --  Time expression must be in ascending order.
               Error_Msg_Sem ("time expression required here", We);
            end if;
         end if;
      end loop;
      return;
   end Sem_Waveform_List;

   -- Semantize a waveform_list WAVEFORM_LIST that is assigned via statement
   -- ASSIGN_STMT to a subelement or a slice of a signal SIGNAL_DECL.
   procedure Sem_Check_Waveform_List
     (Assign_Stmt: Iir; Waveform_List: Iir_Waveform_List)
   is
      We: Iir_Waveform_Element;
      Expr : Iir;
   begin
      if Waveform_List = null then
         return;
      end if;

      for I in Natural loop
         We := Get_Nth_Element (Waveform_List, I);
         exit when We = null;

         Expr := Get_We_Value (We);
         if Get_Kind (Expr) = Iir_Kind_Null_Literal then
            --  This is a null waveform element.
            --  LRM93 8.4.1
            --  It is an error if the target of a signal assignment statement
            --  containing a null waveform is not a guarded signal or an
            --  aggregate of guarded signals.
            if Get_Guarded_Target_State (Assign_Stmt) = False then
               Error_Msg_Sem
                 ("null transactions can be assigned only to guarded signals",
                  Assign_Stmt);
            end if;
         end if;
      end loop;
   end Sem_Check_Waveform_List;

   --  Add a driver for SIG.
   --  STMT is used in case of error (it is the statement that creates the
   --   driver).
   --  Do nothing if:
   --    The current statement list does not belong to a process,
   --    SIG is a formal signal interface.
   procedure Sem_Add_Driver (Sig : Iir; Stmt : Iir)
   is
      Sig_Object : Iir;
      Sig_Object_Type : Iir;
      Parent : Iir;
      Driver_List : Iir_List;
      Driver : Iir;
   begin
      if Sig = Null_Iir then
         return;
      end if;
      Sig_Object := Get_Base_Name (Sig);
      Sig_Object_Type := Get_Type (Sig_Object);

      --  Check for multiple driver for a unresolved signal declaration.
      --  Do this only if the object is a non-composite signal declaration.
      if Get_Kind (Sig_Object) = Iir_Kind_Signal_Declaration
        and then Get_Kind (Sig_Object_Type)
        not in Iir_Kinds_Composite_Type_Definition
        and then not Get_Resolved_Flag (Sig_Object_Type)
      then
         if Get_Signal_Driver (Sig_Object) /= Null_Iir and then
           Get_Signal_Driver (Sig_Object) /= Current_Concurrent_Statement
         then
            Error_Msg_Sem ("unresolved " & Disp_Node (Sig_Object)
                           & " has already a driver at "
                           & Disp_Location (Get_Signal_Driver (Sig_Object)),
                           Stmt);
         else
            Set_Signal_Driver (Sig_Object, Current_Concurrent_Statement);
         end if;
      end if;

      --  The driver is attached to the current process (if any), or to
      --  the current subprogram (if any) or to nothing.
      if Current_Concurrent_Statement /= Null_Iir
        and then (Get_Kind (Current_Concurrent_Statement)
                  in Iir_Kinds_Process_Statement)
      then
         Driver := Current_Concurrent_Statement;
      elsif Current_Subprogram /= Null_Iir then
         Driver := Current_Subprogram;
      else
         return;
      end if;

      case Get_Kind (Sig_Object) is
         when Iir_Kind_Signal_Interface_Declaration =>
            Parent := Get_Parent (Sig_Object);
            case Get_Kind (Parent) is
               when Iir_Kind_Block_Statement
                 | Iir_Kind_Entity_Declaration
                 | Iir_Kind_Block_Header =>
                  null;
               when Iir_Kinds_User_Procedure_Specification =>
                  return;
               when others =>
                  Error_Kind ("sem_add_driver", Parent);
            end case;
         when Iir_Kind_Signal_Declaration =>
            null;
         when others =>
            Error_Kind ("sem_add_driver(2)", Sig_Object);
      end case;

      Driver_List := Get_Driver_List (Driver);
      if Driver_List = Null_Iir_List then
         Driver_List := Create_Iir (Iir_Kind_Driver_List);
         Set_Driver_List (Driver, Driver_List);
      end if;

      Add_Element (Driver_List, Get_Longuest_Static_Prefix (Sig));
   end Sem_Add_Driver;

   procedure Fill_Array_From_Aggregate_Associated
     (List : Iir_List; Nbr : in out Natural; Arr : Iir_Array_Acc)
   is
      El : Iir;
      Ass : Iir;
   begin
      for I in Natural loop
         El := Get_Nth_Element (List, I);
         exit when El = Null_Iir;
         Ass := Get_Associated (El);
         if Get_Kind (Ass) = Iir_Kind_Aggregate then
            Fill_Array_From_Aggregate_Associated
              (Get_Association_Choices_List (Ass), Nbr, Arr);
         else
            if Arr /= null then
               Arr (Nbr) := Ass;
            end if;
            Nbr := Nbr + 1;
         end if;
      end loop;
   end Fill_Array_From_Aggregate_Associated;

   --  Return TRUE iff there is no common elements designed by N1 and N2.
   --  N1 and N2 are static names.
   --  FIXME:  The current implementation is completly wrong; should check from
   --   prefix to suffix.
   function Is_Disjoint (N1, N2: Iir) return Boolean
   is
      List1, List2 : Iir_List;
      El1, El2 : Iir;
   begin
      if N1 = N2 then
         return False;
      end if;
      if Get_Kind (N1) = Iir_Kind_Indexed_Name
        and then Get_Kind (N2) = Iir_Kind_Indexed_Name
      then
         if Is_Disjoint (Get_Prefix (N1), Get_Prefix (N2)) then
            return True;
         end if;
         --  Check indexes.
         List1 := Get_Index_List (N1);
         List2 := Get_Index_List (N2);
         for I in Natural loop
            El1 := Get_Nth_Element (List1, I);
            El2 := Get_Nth_Element (List2, I);
            exit when El1 = Null_Iir;
            El1 := Eval_Expr (El1);
            Replace_Nth_Element (List1, I, El1);
            El2 := Eval_Expr (El2);
            Replace_Nth_Element (List2, I, El2);
            --  EL are of discrete type.
            if Get_Value (El1) /= Get_Value (El2) then
               return True;
            end if;
         end loop;
         return False;
      end if;
      return True;
   end Is_Disjoint;

   procedure Check_Uniq_Aggregate_Associated
     (Aggr : Iir_Aggregate; Nbr : Natural)
   is
      Index : Natural;
      Arr : Iir_Array_Acc;
      List : Iir_List;
   begin
      List := Get_Association_Choices_List (Aggr);
      --  Count number of associated values, and create the array.
      --  Already done: use nbr.
      -- Fill_Array_From_Aggregate_Associated (List, Nbr, null);
      Arr := new Iir_Array (0 .. Nbr - 1);
      --  Fill the array.
      Index := 0;
      Fill_Array_From_Aggregate_Associated (List, Index, Arr);
      if Index /= Nbr then
         --  Should be the same.
         raise Internal_Error;
      end if;
      --  Check each element is uniq.
      for I in Arr.all'Range loop
         if Get_Name_Staticness (Arr (I)) = Locally then
            for J in 0 .. I - 1 loop
               if Get_Name_Staticness (Arr (J)) = Locally
                 and then not Is_Disjoint (Arr (I), Arr (J))
               then
                  Error_Msg_Sem ("target is assigned more than once", Arr (I));
                  Error_Msg_Sem ("previous assignment: "
                                 & Disp_Location (Arr (J)), Arr (I));
                  Free (Arr);
                  return;
               end if;
            end loop;
         end if;
      end loop;
      Free (Arr);
      return;
   end Check_Uniq_Aggregate_Associated;

   --  Do checks for the target of an assignment.
   procedure Check_Simple_Signal_Target
     (Stmt : Iir; Target : Iir; Staticness : Iir_Staticness);
   --  STMT is used to localize the error (if any).
   procedure Check_Simple_Variable_Target
     (Stmt : Iir; Target : Iir; Staticness : Iir_Staticness);

   procedure Check_Aggregate_Target
     (Stmt : Iir; Target : Iir; Nbr : in out Natural)
   is
      Choice_List : Iir_List;
      Choice : Iir;
      Ass : Iir;
   begin
      Choice_List := Get_Association_Choices_List (Target);
      for I in Natural loop
         Choice := Get_Nth_Element (Choice_List, I);
         exit when Choice = Null_Iir;
         case Get_Kind (Choice) is
            when Iir_Kind_Association_Choice_By_Range =>
               --  LRM93 8.4
               --  It is an error if an element association in such an
               --  aggregate contains an OTHERS choice or a choice that is
               --  a discrete range.
               Error_Msg_Sem ("discrete range choice not allowed for target",
                              Choice);
            when Iir_Kind_Association_Choice_By_Others =>
               --  LRM93 8.4
               --  It is an error if an element association in such an
               --  aggregate contains an OTHERS choice or a choice that is
               --  a discrete range.
               Error_Msg_Sem ("others choice not allowed for target", Choice);
            when Iir_Kind_Association_Choice_By_Expression
              | Iir_Kind_Association_Choice_By_Name
              | Iir_Kind_Association_Choice_By_List
              | Iir_Kind_Association_Choice_By_None =>
               --  LRM93 9.4
               --  Such a target may not only contain locally static signal
               --  names [...]
               Ass := Get_Associated (Choice);
               if Get_Kind (Ass) = Iir_Kind_Aggregate then
                  Check_Aggregate_Target (Stmt, Ass, Nbr);
               else
                  if Get_Kind (Stmt) = Iir_Kind_Variable_Assignment_Statement
                  then
                     Check_Simple_Variable_Target (Stmt, Ass, Locally);
                  else
                     Check_Simple_Signal_Target (Stmt, Ass, Locally);
                  end if;
                  Nbr := Nbr + 1;
               end if;
            when others =>
               Error_Kind ("check_aggregate_target", Choice);
         end case;
      end loop;
   end Check_Aggregate_Target;

   procedure Check_Simple_Signal_Target
     (Stmt : Iir; Target : Iir; Staticness : Iir_Staticness)
   is
      Target_Object : Iir;
      Guarded_Target : Tri_State_Type;
   begin
      if not Is_Object_Name (Target) then
         Error_Msg_Sem ("target is not a signal name", Target);
         return;
      end if;

      Target_Object := Get_Base_Name (Target);

      case Get_Kind (Target_Object) is
         when Iir_Kind_Signal_Interface_Declaration =>
            if not Iir_Mode_Writable (Get_Mode (Target_Object)) then
               Error_Msg_Sem ("interface signal can't be assigned", Target);
            else
               Sem_Add_Driver (Target, Stmt);
            end if;
            --  LRM93 2.1.1.2
            --  A formal signal parameter is a guarded signal if and only if
            --  it is associated with an actual signal that is a guarded
            --  signal.
            --  GHDL: a formal signal interface of a subprogram has no static
            --   kind.  This is determined at run-time, according to the actual
            --   associated with the formal.
            if Get_Kind (Get_Parent (Target_Object)) in
              Iir_Kinds_Subprogram_Declaration
            then
               Guarded_Target := Unknown;
            end if;
         when Iir_Kind_Signal_Declaration =>
            Sem_Add_Driver (Target, Stmt);
         when Iir_Kind_Guard_Signal_Declaration =>
            Error_Msg_Sem ("implicit GUARD signal cannot be assigned", Stmt);
            return;
         when others =>
            Error_Msg_Sem ("target is not a signal", Stmt);
            return;
      end case;
      if Get_Name_Staticness (Target) < Staticness then
         Error_Msg_Sem ("signal name must be static", Stmt);
      end if;

      --  LRM93 2.1.1.2
      --  A formal signal parameter is a guarded signal if and only if
      --  it is associated with an actual signal that is a guarded
      --  signal.
      --  GHDL: a formal signal interface of a subprogram has no static
      --   kind.  This is determined at run-time, according to the actual
      --   associated with the formal.
      if Get_Kind (Target_Object) = Iir_Kind_Signal_Interface_Declaration
        and then Get_Kind (Get_Parent (Target_Object)) in
        Iir_Kinds_Subprogram_Declaration
      then
         Guarded_Target := Unknown;
      else
         if Get_Signal_Kind (Target_Object) /= Iir_No_Signal_Kind then
            Guarded_Target := True;
         else
            Guarded_Target := False;
         end if;
      end if;

      case Get_Guarded_Target_State (Stmt) is
         when Unknown =>
            Set_Guarded_Target_State (Stmt, Guarded_Target);
         when True
           | False =>
            if Get_Guarded_Target_State (Stmt) /= Guarded_Target then
               --  LRM93 9.5
               --  It is an error if the target of a concurrent signal
               --  assignment is neither a guarded target nor an
               --  unguarded target.
               Error_Msg_Sem ("guarded and unguarded target", Target);
            end if;
      end case;
   end Check_Simple_Signal_Target;

   procedure Check_Simple_Variable_Target
     (Stmt : Iir; Target : Iir; Staticness : Iir_Staticness)
   is
      Target_Prefix : Iir;
   begin
      if not Is_Object_Name (Target) then
         Error_Msg_Sem ("target is not a variable name", Stmt);
         return;
      end if;
      Target_Prefix := Get_Base_Name (Target);
      case Get_Kind (Target_Prefix) is
         when Iir_Kind_Variable_Interface_Declaration =>
            if not Iir_Mode_Writable (Get_Mode (Target_Prefix)) then
               Error_Msg_Sem (Disp_Node (Target_Prefix)
                              & " cannot be written (bad mode)", Target);
               return;
            end if;
         when Iir_Kind_Variable_Declaration =>
            null;
         when Iir_Kind_Implicit_Dereference
           | Iir_Kind_Dereference  =>
            --  LRM 3.3
            --  An object designated by an access type is always an object of
            --  class variable.
            null;
         when others =>
            Error_Msg_Sem (Disp_Node (Target_Prefix)
                           & " is not a variable to be assigned", Stmt);
            return;
      end case;
      if Get_Name_Staticness (Target) < Staticness then
         Error_Msg_Sem ("variable element of aggregate must be a static name",
                        Target);
      end if;
   end Check_Simple_Variable_Target;

   procedure Check_Target (Stmt : Iir; Target : Iir)
   is
      Nbr : Natural;
   begin
      if Get_Kind (Target) = Iir_Kind_Aggregate then
         Nbr := 0;
         Check_Aggregate_Target (Stmt, Target, Nbr);
         Check_Uniq_Aggregate_Associated (Target, Nbr);
      else
         if Get_Kind (Stmt) = Iir_Kind_Variable_Assignment_Statement then
            Check_Simple_Variable_Target (Stmt, Target, None);
         else
            Check_Simple_Signal_Target (Stmt, Target, None);
         end if;
      end if;
   end Check_Target;

   --  Return FALSE in case of error.
   function Sem_Signal_Assignment_Target_And_Option (Stmt: Iir; Sig_Type : Iir)
     return Boolean
   is
      --  The target of the assignment.
      Target: Iir;
      --  The value that will be assigned.
      Expr: Iir;
      Ok : Boolean;
   begin
      Ok := True;
      -- Find the signal.
      Target := Get_Target (Stmt);
      Target := Sem_Expression (Target, Sig_Type);
      if Target /= Null_Iir then
         Set_Target (Stmt, Target);
         Check_Target (Stmt, Target);
      else
         Ok := False;
      end if;

      Expr := Get_Reject_Time_Expression (Stmt);
      if Expr /= Null_Iir then
         Expr := Sem_Expression (Expr, Time_Type_Definition);
         if Expr /= Null_Iir then
            Check_Read (Expr);
            Set_Reject_Time_Expression (Stmt, Expr);
         else
            Ok := False;
         end if;
      end if;
      return Ok;
   end Sem_Signal_Assignment_Target_And_Option;

   procedure Sem_Signal_Assignment (Stmt: Iir)
   is
      Target : Iir;
      Waveform_Type : Iir;
   begin
      Target := Get_Target (Stmt);
      if Get_Kind (Target) /= Iir_Kind_Aggregate then
         if not Sem_Signal_Assignment_Target_And_Option (Stmt, null) then
            return;
         end if;

         -- check the expression.
         Waveform_Type := Get_Type (Get_Target (Stmt));
         if Waveform_Type /= Null_Iir then
            Sem_Waveform_List (Stmt, Get_Waveform_List (Stmt), Waveform_Type);
            Sem_Check_Waveform_List (Stmt, Get_Waveform_List (Stmt));
         end if;
      else
         Sem_Waveform_List (Stmt, Get_Waveform_List (Stmt), Waveform_Type);
         if Waveform_Type = Null_Iir
           or else
           not Sem_Signal_Assignment_Target_And_Option (Stmt, Waveform_Type)
         then
            return;
         end if;
         Sem_Check_Waveform_List (Stmt, Get_Waveform_List (Stmt));
      end if;
   end Sem_Signal_Assignment;

   procedure Sem_Guard (Stmt: Iir)
   is
      use Sem_Name;
      Guard: Iir;
      Guard_Interpretation : Name_Interpretation_Type;
   begin
      Guard := Get_Guard (Stmt);
      if Guard = Null_Iir then
         --  This assignment is not guarded.

         --  LRM93 9.5
         --  It is an error if a concurrent signal assignment is not a guarded
         --  assignment, and the target of the concurrent signal assignment
         --  is a guarded target.
         if Get_Guarded_Target_State (Stmt) = True then
            Error_Msg_Sem
              ("not a guarded assignment for a guarded target", Stmt);
         end if;
         return;
      end if;
      if Guard /= Stmt then
         -- if set, guard must be equal to stmt here.
         raise Internal_Error;
      end if;
      Guard_Interpretation := Get_Interpretation (Std_Names.Name_Guard);
      if Guard_Interpretation = No_Name_Interpretation then
         Error_Msg_Sem ("no guard signals for this guarded assignment",
                        Stmt);
         return;
      end if;

      Guard := Get_Declaration (Guard_Interpretation);
      -- LRM93 9.5:
      -- The signal GUARD [...] an explicitly declared signal of type
      -- BOOLEAN that is visible at the point of the concurrent signal
      -- assignment statement
      -- FIXME.
      case Get_Kind (Guard) is
         when Iir_Kind_Signal_Declaration
           | Iir_Kind_Signal_Interface_Declaration
           | Iir_Kind_Guard_Signal_Declaration =>
            null;
         when others =>
            Error_Msg_Sem ("visible GUARD object is not a signal", Stmt);
            Error_Msg_Sem ("GUARD object is " & Disp_Node (Guard), Stmt);
            return;
      end case;

      if Get_Type (Guard) /= Boolean_Type_Definition then
         Error_Msg_Sem ("GUARD is not of boolean type", Guard);
      end if;
      Set_Guard (Stmt, Guard);
   end Sem_Guard;

   procedure Sem_Concurrent_Conditional_Signal_Assignment (Stmt: Iir)
   is
      List : Iir_List;
      El : Iir;
      Expr : Iir;
      Wf_List : Iir_List;
      Target_Type : Iir;
      Target : Iir;
   begin
      List := Get_Conditional_Waveform_List (Stmt);
      Target := Get_Target (Stmt);
      if Get_Kind (Target) /= Iir_Kind_Aggregate then
         if not Sem_Signal_Assignment_Target_And_Option (Stmt, null) then
            return;
         end if;
         Target := Get_Target (Stmt);
         Target_Type := Get_Type (Target);
      else
         Target_Type := Null_Iir;
      end if;

      for I in Natural loop
         El := Get_Nth_Element (List, I);
         exit when El = Null_Iir;
         Wf_List := Get_Waveform_List (El);
         Sem_Waveform_List (Stmt, Wf_List, Target_Type);
         Sem_Check_Waveform_List (Stmt, Wf_List);
         Expr := Get_Condition (El);
         if Expr /= Null_Iir then
            Expr := Sem_Expression (Expr, Boolean_Type_Definition);
            if Expr /= Null_Iir then
               Check_Read (Expr);
               Set_Condition (El, Expr);
            end if;
         end if;
      end loop;
      Sem_Guard (Stmt);
      if Get_Kind (Target) = Iir_Kind_Aggregate then
         if not Sem_Signal_Assignment_Target_And_Option (Stmt, Target_Type)
         then
            return;
         end if;
      end if;
   end Sem_Concurrent_Conditional_Signal_Assignment;

   --  Semantize a list of case choice LIST, and check for correct CHOICE type.
   procedure Sem_Case_Choices
     (Choice : Iir; List : Iir_List; Loc : Location_Type)
   is
      Choice_Type : Iir;
      Low, High : Iir;
   begin
      Choice_Type := Get_Type (Choice);
      case Get_Kind (Choice_Type) is
         when Iir_Kinds_Discrete_Type_Definition =>
            Sem_Choices_Range (List, Choice_Type, False, Loc, Low, High);
         when Iir_Kind_Array_Subtype_Definition =>
            Sem_String_Choices_Range (List, Choice);
         when others =>
            Error_Msg_Sem ("type of expression must be discrete", Choice);
      end case;
   end Sem_Case_Choices;

   procedure Sem_Concurrent_Selected_Signal_Assignment (Stmt: Iir)
   is
      Expr: Iir;
      List: Iir_Association_Choices_List;
      El: Iir;
      Waveform_Type : Iir;
   begin
      if not Sem_Signal_Assignment_Target_And_Option (Stmt, Null_Iir) then
         return;
      end if;
      Waveform_Type := Get_Type (Get_Target (Stmt));

      Expr := Sem_Expression (Get_Expression (Stmt), Null_Iir);
      if Expr = Null_Iir then
         return;
      end if;
      Check_Read (Expr);
      Set_Expression (Stmt, Expr);
      List := Get_Selected_Waveform_List (Stmt);
      Sem_Case_Choices (Expr, List, Get_Location (Stmt));

      -- Sem on associated.
      if Waveform_Type /= Null_Iir then
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = null;
            Sem_Waveform_List (Stmt, Get_Associated (El), Waveform_Type);
            Sem_Check_Waveform_List (Stmt, Get_Associated (El));
         end loop;
      end if;
      Sem_Guard (Stmt);
   end Sem_Concurrent_Selected_Signal_Assignment;

   procedure Sem_Variable_Assignment (Stmt: Iir) is
      Target: Iir;
      Expr: Iir;
      Target_Type : Iir;
   begin
      -- Find the variable.
      Target := Get_Target (Stmt);
      Expr := Get_Expression (Stmt);
      if Get_Kind (Target) = Iir_Kind_Aggregate then
         if Get_Kind (Expr) = Iir_Kind_Aggregate then
            Error_Msg_Sem ("can't determine type, use type qualifier", Expr);
            return;
         end if;
         Expr := Sem_Expression (Get_Expression (Stmt), Null_Iir);
         if Expr = Null_Iir then
            return;
         end if;
         Check_Read (Expr);
         Set_Expression (Stmt, Expr);
         Target_Type := Get_Type (Expr);
      else
         Target_Type := Null_Iir;
      end if;

      Target := Sem_Expression (Target, Target_Type);
      if Target = Null_Iir then
         return;
      end if;
      Set_Target (Stmt, Target);

      Check_Target (Stmt, Target);

      if Get_Kind (Target) /= Iir_Kind_Aggregate then
         Expr := Sem_Expression (Expr, Get_Type (Target));
         if Expr /= Null_Iir then
            Check_Read (Expr);
            Set_Expression (Stmt, Eval_Expr_If_Static (Expr));
         end if;
      end if;
   end Sem_Variable_Assignment;

--    procedure Check_Parent_For_Return (Parent : Iir; Stmt : Iir)
--    is
--       P : Iir;
--    begin
--       P := Parent;
--       loop
--          case Get_Kind (P) is
--             when Iir_Kind_Process_Statement
--               | Iir_Kind_Sensitized_Process_Statement =>
--                Error_Msg_Parse ("return not allowed in a process");
--                return;
--             when Iir_Kind_Procedure_Spec_Body =>
--                if Get_Expression (Expr) /= Null_Iir then
--                   Error_Msg_Parse ("return with an expression not "
--                                    & "allowed in a procedure");
--                end if;
--                return;
--             when Iir_Kind_Function_Spec_Body =>
--                return;
--             when Iir_Kind_If_Statement =>
--                null;
--             when others =>
--                Error_Kind ("check_parent_for_return", Parent);
--          end case;
--          P := Get_Parent_Statement (P);
--       end loop;
--    end Check_Parent_For_Return;

   procedure Sem_Return_Statement (Stmt: Iir_Return_Statement) is
      Expr: Iir;
   begin
      if Current_Subprogram = null then
         Error_Msg_Sem ("return statement not in a subprogram body", Stmt);
         return;
      end if;
      Expr := Get_Expression (Stmt);
      case Get_Kind (Current_Subprogram) is
         when Iir_Kinds_Procedure_Specification =>
            if Expr /= Null_Iir then
               Error_Msg_Sem
                 ("return in a procedure can't have an expression", Stmt);
            end if;
            return;
         when Iir_Kinds_Function_Specification =>
            if Expr = Null_Iir then
               Error_Msg_Sem
                 ("return in a function must have an expression", Stmt);
               return;
            end if;
         when Iir_Kinds_Process_Statement =>
            Error_Msg_Sem ("return statement not allowed in a process", Stmt);
            return;
         when others =>
            Error_Kind ("sem_return_statement", Stmt);
      end case;
      Set_Type (Stmt, Get_Return_Type (Current_Subprogram));
      Expr := Sem_Expression (Expr, Get_Return_Type (Current_Subprogram));
      if Expr /= Null_Iir then
         Check_Read (Expr);
         Set_Expression (Stmt, Eval_Expr_If_Static (Expr));
      end if;
   end Sem_Return_Statement;

   --  Convert a range expression to a subtype definition whose constraint is
   --  A_RANGE.
   --  This function extract the type of the range expression.
   function Range_To_Subtype_Definition (A_Range: Iir) return Iir
   is
      Sub_Type: Iir;
      Range_Type : Iir;
   begin
      case Get_Kind (A_Range) is
         when Iir_Kind_Range_Expression
           | Iir_Kind_Range_Array_Attribute
           | Iir_Kind_Reverse_Range_Array_Attribute =>
            --  Create a sub type.
            Range_Type := Get_Type (A_Range);
         when Iir_Kinds_Discrete_Type_Definition =>
            --  A_RANGE is already a subtype definition.
            return A_Range;
         when others =>
            Error_Kind ("range_to_subtype_definition", A_Range);
            return Null_Iir;
      end case;

      case Get_Kind (Range_Type) is
         when Iir_Kind_Enumeration_Type_Definition
           | Iir_Kind_Enumeration_Subtype_Definition =>
            Sub_Type := Create_Iir (Iir_Kind_Enumeration_Subtype_Definition);
         when Iir_Kind_Integer_Type_Definition
           | Iir_Kind_Integer_Subtype_Definition =>
            Sub_Type := Create_Iir (Iir_Kind_Integer_Subtype_Definition);
         when others =>
            raise Internal_Error;
      end case;
      Location_Copy (Sub_Type, A_Range);
      Set_Range_Constraint (Sub_Type, A_Range);
      Set_Base_Type (Sub_Type, Get_Base_Type (Range_Type));
      Set_Type_Staticness (Sub_Type, Get_Expr_Staticness (A_Range));
      Set_Signal_Type_Flag (Sub_Type, True);
      return Sub_Type;
   end Range_To_Subtype_Definition;

   procedure Sem_Iterator (Iterator : Iir_Iterator_Declaration;
                           Staticness : Iir_Staticness)
   is
      It_Type: Iir;
      A_Range: Iir;
      Range_Type : Iir;
   begin
      It_Type := Get_Type (Iterator);
      A_Range := Sem_Discrete_Range_Expression (It_Type, null);
      if A_Range = Null_Iir then
         Set_Type (Iterator, Error_Iir);
         return;
      end if;
      if Get_Kind (A_Range) in Iir_Kinds_Type_And_Subtype_Definition then
         Range_Type := A_Range;
      else
         Range_Type := Get_Type (A_Range);
      end if;
      case Get_Kind (Range_Type) is
         when Iir_Kinds_Discrete_Type_Definition =>
            null;
         when others =>
            Error_Msg_Sem ("iterator is not of discrete type", A_Range);
            Set_Type (Iterator, Null_Iir);
            return;
      end case;

      Set_Type (Iterator, Range_To_Subtype_Definition (A_Range));
      Set_Base_Name (Iterator, Iterator);
      Set_Expr_Staticness (Iterator, Staticness);
   end Sem_Iterator;

   -- Sem for concurrent and sequential assertion statements.
   procedure Sem_Report_Statement (Stmt : Iir)
   is
      Expr : Iir;
   begin
      Expr := Get_Report_Expression (Stmt);
      if Expr /= Null_Iir then
         Expr := Sem_Expression (Expr, String_Type_Definition);
         Check_Read (Expr);
         Set_Report_Expression (Stmt, Expr);
      end if;

      Expr := Get_Severity_Expression (Stmt);
      if Expr /= Null_Iir then
         Expr := Sem_Expression (Expr, Severity_Level_Type_Definition);
         Check_Read (Expr);
         Set_Severity_Expression (Stmt, Expr);
      end if;
   end Sem_Report_Statement;

   procedure Sem_Assertion_Statement (Stmt: Iir)
   is
      Expr : Iir;
   begin
      Expr := Get_Assertion_Condition (Stmt);
      Expr := Sem_Expression (Expr, Boolean_Type_Definition);
      Check_Read (Expr);
      Expr := Eval_Expr_If_Static (Expr);
      Set_Assertion_Condition (Stmt, Expr);

      Sem_Report_Statement (Stmt);
   end Sem_Assertion_Statement;

   procedure Sem_Case_Statement (Stmt: Iir_Case_Statement)
   is
      Expr: Iir;
      List: Iir_Association_Choices_List;
      El: Iir;
      Loc : Location_Type;
   begin
      Expr := Get_Expression (Stmt);
      Loc := Get_Location (Expr);
      -- FIXME: overload.
      Expr := Sem_Expression (Expr, null);
      if Expr = Null_Iir then
         return;
      end if;
      Check_Read (Expr);
      Set_Expression (Stmt, Expr);
      List := Get_Association_Choices_List (Stmt);
      Sem_Case_Choices (Expr, List, Get_Location (Stmt));
      -- Sem on associated.
      for I in Natural loop
         El := Get_Nth_Element (List, I);
         exit when El = null;
         Sem_Sequential_Statements_Internal (Get_Associated (El));
      end loop;
   end Sem_Case_Statement;

   procedure Sem_Sensitivity_List (List: Iir_Designator_List) is
      El: Iir;
      Res: Iir;
      Loc : Location_Type;
      Prefix : Iir;
   begin
      for I in Natural loop
         -- El is an iir_identifier.
         El := Get_Nth_Element (List, I);
         exit when El = null;
         if Get_Kind (El) not in Iir_Kinds_Name then
            Error_Msg_Sem ("a sensitivity element must be a signal name", El);
         else
            Loc := Get_Location (El);
            Res := Sem_Expression (El, Null_Iir);

            -- Dummy loop, just to avoid goto.
            loop
               exit when Res = Null_Iir;
               Prefix := Get_Base_Name (Res);
               case Get_Kind (Prefix) is
                  when Iir_Kind_Signal_Declaration
                    | Iir_Kind_Guard_Signal_Declaration =>
                     null;
                  when Iir_Kind_Signal_Interface_Declaration =>
                     if not Iir_Mode_Readable (Get_Mode (Prefix)) then
                        Error_Msg_Sem
                          (Disp_Node (Res) & " of mode out"
                           & " can't be in a sensivity list", Loc);
                     end if;
                  when others =>
                     Error_Msg_Sem (Disp_Node (Res)
                                    & " is neither a signal nor a port", Loc);
                     exit;
               end case;
               if Get_Name_Staticness (Res) < Globally then
                  Error_Msg_Sem ("name is not static", Loc);
               end if;
               Replace_Nth_Element (List, I, Res);
               exit;
            end loop;
         end if;
      end loop;
   end Sem_Sensitivity_List;

   procedure Sem_Wait_Statement (Stmt: Iir_Wait_Statement)
   is
      Expr: Iir;
   begin
      --  Check validity.
      case Get_Kind (Current_Subprogram) is
         when Iir_Kind_Process_Statement =>
            null;
         when Iir_Kinds_Function_Declaration =>
            --  LRM93 8.2
            --  It is an error if a wait statement appears in a function
            --  subprogram [...]
            Error_Msg_Sem
              ("wait statement not allowed in a function subprogram", Stmt);
            return;
         when Iir_Kinds_Procedure_Declaration =>
            --  LRM93 8.2
            --  [It is an error ...] or in a procedure that has a parent that
            --  is a function subprogram.
            --  LRM93 8.2
            --  [...] or in a procedure that has a parent that is such a
            --  process statement.
            -- GHDL: this is checked during execution.
            null;
         when Iir_Kind_Sensitized_Process_Statement =>
            --  LRM93 8.2
            --  Furthermore, it is an error if a wait statement appears in an
            --  explicit process statement that includes a sensitivity list,
            --  [...]
            Error_Msg_Sem
              ("wait statement not allowed in a sensitized process", Stmt);
            return;
         when others =>
            raise Internal_Error;
      end case;

      Expr := Get_Sensitivity_List (Stmt);
      if Expr /= null then
         Sem_Sensitivity_List (Expr);
      end if;
      Expr := Get_Condition_Clause (Stmt);
      if Expr /= Null_Iir then
         Expr := Sem_Expression (Expr, Boolean_Type_Definition);
         Check_Read (Expr);
         Set_Condition_Clause (Stmt, Expr);
      end if;
      Expr := Get_Timeout_Clause (Stmt);
      if Expr /= Null_Iir then
         Expr := Sem_Expression (Expr, Time_Type_Definition);
         if Expr /= Null_Iir then
            Check_Read (Expr);
            Expr := Eval_Expr_If_Static (Expr);
            Set_Timeout_Clause (Stmt, Expr);
            if Get_Expr_Staticness (Expr) = Locally
              and then Get_Value (Expr) < 0
            then
               Error_Msg_Sem ("timeout value must be positive", Stmt);
            end if;
         end if;
      end if;
      Set_Wait_State (Current_Subprogram, True);
   end Sem_Wait_Statement;

   procedure Sem_Exit_Next_Statement (Stmt : Iir)
   is
      Cond: Iir;
      Label: Iir;
      P : Iir;
   begin
      Cond := Get_Condition (Stmt);
      if Cond /= Null_Iir then
         Cond := Sem_Expression (Cond, Boolean_Type_Definition);
         Check_Read (Cond);
         Set_Condition (Stmt, Cond);
      end if;
      Label := Get_Loop (Stmt);
      if Label /= Null_Iir then
         Label := Find_Declaration (Label, Decl_Label);
      end if;
      if Label /= Null_Iir then
         case Get_Kind (Label) is
            when Iir_Kind_While_Loop_Statement
              | Iir_Kind_For_Loop_Statement =>
               Set_Loop (Stmt, Label);
            when others =>
               Error_Msg_Sem ("loop label expected", Stmt);
               Label := Null_Iir;
         end case;
      end if;
      --  Check the current statement is inside the labeled loop.
      P := Stmt;
      loop
         P := Get_Parent (P);
         case Get_Kind (P) is
            when Iir_Kind_While_Loop_Statement
              | Iir_Kind_For_Loop_Statement =>
               if Label = Null_Iir or else Label = P then
                  exit;
               end if;
            when Iir_Kind_If_Statement
              | Iir_Kind_Elsif
              | Iir_Kind_Case_Statement =>
               null;
            when others =>
               --  FIXME: should emit a message for label mismatch.
               Error_Msg_Sem ("exit/next must be inside a loop", Stmt);
               exit;
         end case;
      end loop;
   end Sem_Exit_Next_Statement;

   -- Process is the scope, this is also the process for which drivers can
   -- be created.
   procedure Sem_Sequential_Statements_Internal
     (Stmts: Iir_Sequential_Statement_List)
   is
      Stmt: Iir;
   begin
      for I in Natural loop
         Stmt := Get_Nth_Element (Stmts, I);
         exit when Stmt = Null_Iir;
         case Get_Kind (Stmt) is
            when Iir_Kind_Null_Statement =>
               null;
            when Iir_Kind_If_Statement =>
               declare
                  Clause: Iir := Stmt;
                  Cond: Iir;
               begin
                  while Clause /= null loop
                     Cond := Get_Condition (Clause);
                     if Cond /= null then
                        Cond := Sem_Expression (Cond, Boolean_Type_Definition);
                        Check_Read (Cond);
                        Set_Condition (Clause, Cond);
                     end if;
                     Sem_Sequential_Statements_Internal
                       (Get_Then_Sequence_List (Clause));
                     Clause := Get_Else_Clause (Clause);
                  end loop;
               end;
            when Iir_Kind_For_Loop_Statement =>
               declare
                  Iterator: Iir;
               begin
                  Sem_Name.Open_Declarative_Region;
                  Set_Is_Within_Flag (Stmt, True);
                  Iterator := Get_Iterator_Scheme (Stmt);
                  Sem_Name.Add_Name (Iterator);
                  Sem_Iterator (Iterator, None);
                  Set_Visible_Flag (Iterator, True);
                  Sem_Sequential_Statements_Internal
                    (Get_Statement_List (Stmt));
                  Set_Is_Within_Flag (Stmt, False);
                  Sem_Name.Close_Declarative_Region;
               end;
            when Iir_Kind_While_Loop_Statement =>
               declare
                  Cond: Iir;
               begin
                  Cond := Get_Condition (Stmt);
                  if Cond /= Null_Iir then
                     Cond := Sem_Expression (Cond, Boolean_Type_Definition);
                     Check_Read (Cond);
                     Set_Condition (Stmt, Cond);
                  end if;
                  Sem_Sequential_Statements_Internal
                    (Get_Statement_List (Stmt));
               end;
            when Iir_Kind_Signal_Assignment_Statement =>
               Sem_Signal_Assignment (Stmt);
               if Current_Concurrent_Statement /= Null_Iir and then
                 Get_Kind (Current_Concurrent_Statement)
                 in Iir_Kinds_Process_Statement
                 and then Get_Passive_Flag (Current_Concurrent_Statement)
               then
                  Error_Msg_Sem
                    ("signal statement forbidden in passive process", Stmt);
               end if;
               Set_Passive_State (Current_Subprogram, False);
            when Iir_Kind_Variable_Assignment_Statement =>
               Sem_Variable_Assignment (Stmt);
            when Iir_Kind_Return_Statement =>
               Sem_Return_Statement (Stmt);
            when Iir_Kind_Assertion_Statement =>
               Sem_Assertion_Statement (Stmt);
            when Iir_Kind_Report_Statement =>
               Sem_Report_Statement (Stmt);
            when Iir_Kind_Case_Statement =>
               Sem_Case_Statement (Stmt);
            when Iir_Kind_Wait_Statement =>
               Sem_Wait_Statement (Stmt);
            when Iir_Kind_Procedure_Call =>
               Sem_Procedure_Call (Stmt);
            when Iir_Kind_Next_Statement
              | Iir_Kind_Exit_Statement =>
               Sem_Exit_Next_Statement (Stmt);
            when others =>
               Error_Kind ("sem_sequential_statements_Internal", Stmt);
         end case;
      end loop;
   end Sem_Sequential_Statements_Internal;

   procedure Sem_Process_Statement (Proc: Iir) is
   begin
      Set_Is_Within_Flag (Proc, True);

      -- LRM93 10.1
      -- 8. A process statement
      Sem_Name.Open_Declarative_Region;

      -- Sem declarations
      Sem_Sequential_Statements (Proc,
                                 Get_Declaration_List (Proc),
                                 Get_Process_Statement_List (Proc));

      Sem_Name.Close_Declarative_Region;

      Set_Is_Within_Flag (Proc, False);
   end Sem_Process_Statement;

   procedure Sem_Sensitized_Process_Statement
     (Proc: Iir_Sensitized_Process_Statement) is
   begin
      Sem_Sensitivity_List (Get_Sensitivity_List (Proc));
      Sem_Process_Statement (Proc);
   end Sem_Sensitized_Process_Statement;

   --  Check for restrictions in 1.1.1.2
   --  Return FALSE in case of error.
   function Check_Port_Association_Restriction
     (Formal : Iir_Signal_Interface_Declaration;
      Actual : Iir_Signal_Interface_Declaration;
      Assoc : Iir)
     return Boolean
   is
   begin
      case Get_Mode (Formal) is
         when Iir_In_Mode =>
            case Get_Mode (Actual) is
               when Iir_In_Mode
                 | Iir_Inout_Mode
                 | Iir_Buffer_Mode =>
                  return True;
               when others =>
                  Error_Msg_Sem
                    ("formal in " & Disp_Node (Formal)
                     & " may be only associated with an actual port of "
                     & "mode `inout', `in' or `buffer",
                     Assoc);
                  return False;
            end case;
         when Iir_Out_Mode =>
            case Get_Mode (Actual) is
               when Iir_Out_Mode
                 | Iir_Inout_Mode =>
                  return True;
               when others =>
                  Error_Msg_Sem
                    ("formal out " & Disp_Node (Formal)
                     & " may be only associated "
                     & "with an actual port of mode `out' or `inout'",
                     Assoc);
                  return False;
            end case;
         when Iir_Inout_Mode =>
            if Get_Mode (Actual) /= Iir_Inout_Mode then
               Error_Msg_Sem
                 ("formal inout " & Disp_Node (Formal)
                  & " may be only associated with an actual port "
                  & "of mode `inout'",
                  Assoc);
               return False;
            else
               return True;
            end if;
         when Iir_Buffer_Mode =>
            if Get_Mode (Actual) /= Iir_Buffer_Mode then
               Error_Msg_Sem
                 ("formal buffer " & Disp_Node (Formal)
                  & " may be only associated "
                  & "with an actual port of mode `buffer'",
                  Assoc);
               return False;
            else
               return True;
            end if;
         when Iir_Linkage_Mode =>
            return True;
         when others =>
            raise Internal_Error;
      end case;
   end Check_Port_Association_Restriction;

   procedure Sem_Generic_Port_Association_List
     (Formal_Generic : Iir_Generic_List;
      Generic_Associations : Iir_Association_List;
      Formal_Port : Iir_Port_List;
      Port_Associations : Iir_Association_List;
      Parent : Iir)
   is
      El : Iir;
      Actual : Iir;
      Prefix : Iir;
      Check_Match : Boolean;
   begin
      --  Note: CHECK_MATCH argument of sem_subprogram_arguments must be
      --   true if parent is a component instantiation.
      case Get_Kind (Parent) is
         when Iir_Kind_Component_Instantiation_Statement =>
            Check_Match := True;
         when Iir_Kind_Binding_Indication =>
            Check_Match := False;
         when Iir_Kind_Block_Header =>
            Check_Match := True;
         when others =>
            Error_Kind ("sem_generic_port_association_list", Parent);
      end case;

      if Sem_Association_List (Generic_Associations) then
         if Sem_Subprogram_Arguments (Formal_Generic, Generic_Associations,
                                      True, Check_Match, True, Parent) /= Full
         then
            null;
         end if;
      end if;
      if not Sem_Association_List (Port_Associations) then
         return;
      end if;
      if Sem_Subprogram_Arguments (Formal_Port, Port_Associations, True,
                                   Check_Match, True, Parent) /= Full
      then
         return;
      end if;
      --  LRM93 1.1.1.2
      --  The actual, if a port or signal, must be denoted by a static name.
      --  The actual, if an expression, must be a globally static expression.
      if Port_Associations /= null then
         for I in Natural loop
            El := Get_Nth_Element (Port_Associations, I);
            exit when El = null;
            if Get_Kind (El) = Iir_Kind_Association_Element_By_Expression then
               Actual := Get_Actual (El);
               --  There has been an error, exit from the loop.
               exit when Actual = Null_Iir;
               Prefix := Get_Base_Name (Actual);
               case Get_Kind (Prefix) is
                  when Iir_Kind_Signal_Declaration
                    | Iir_Kind_Signal_Interface_Declaration =>
                     --  Port or signal.
                     if Get_Name_Staticness (Actual) < Globally then
                        Error_Msg_Sem ("actual must be a static name", Actual);
                     end if;
                     if Get_Kind (Prefix)
                       = Iir_Kind_Signal_Interface_Declaration
                     then
                        declare
                           P : Boolean;
                        begin
                           P := Check_Port_Association_Restriction
                             (Get_Associated_Formal (El), Prefix, El);
                        end;
                     end if;
                  when others =>
                     --  Expression.
                     if Get_Expr_Staticness (Actual) < Globally then
                        Error_Msg_Sem
                          ("actual expression must be globally static",
                           Actual);
                     end if;
               end case;
            end if;
         end loop;
      end if;
   end Sem_Generic_Port_Association_List;

   --  Semantize entity aspect ASPECT and return the entity declaration.
   --  Return NULL_IIR if not found.
   function Sem_Entity_Aspect (Aspect : Iir) return Iir
   is
      Entity : Iir;
      New_Entity : Iir;
      Conf : Iir;
      Arch : Iir;
   begin
      case Get_Kind (Aspect) is
         when Iir_Kind_Entity_Aspect_Entity =>
            Entity := Get_Entity (Aspect);
            New_Entity := Find_Declaration (Entity, Decl_Entity);
            if New_Entity = Null_Iir then
               return Null_Iir;
            end if;
            Set_Entity (Aspect, New_Entity);

            --  Check architecture.
            Arch := Get_Architecture (Aspect);
            if Arch /= Null_Iir then
               --  FIXME: may emit a warning if the architecture does not
               --  exists.
               --  Note: the design needs the architecture.
               Add_Dependence (Aspect);
            end if;
         when Iir_Kind_Entity_Aspect_Configuration =>
            Conf := Get_Configuration (Aspect);
            Conf := Find_Declaration (Conf, Decl_Configuration);
            if Conf = Null_Iir then
               return Null_Iir;
            end if;

            Set_Configuration (Aspect, Conf);

            Libraries.Load_Design_Unit (Conf, Aspect);
            New_Entity := Get_Entity (Get_Library_Unit (Conf));
         when Iir_Kind_Entity_Aspect_Open =>
            return Null_Iir;
         when others =>
            Error_Kind ("sem_entity_aspect", Aspect);
      end case;
      Libraries.Load_Design_Unit (New_Entity, Aspect);
      return Get_Library_Unit (New_Entity);
   end Sem_Entity_Aspect;

   --  Sem the instantiated unit of STMT and return the node constaining
   --  ports and generics (either a entity_declaration or a component
   --  declaration).
   function Sem_Instantiated_Unit
     (Stmt : Iir_Component_Instantiation_Statement)
     return Iir
   is
      Inst : Iir;
   begin
      Inst := Get_Instantiated_Unit (Stmt);
      if Inst = Null_Iir then
         return Null_Iir;
      end if;
      --  Component declarations are already semantized during sem_label.
      if Get_Kind (Inst) = Iir_Kind_Component_Declaration then
         return Inst;
      end if;
      return Sem_Entity_Aspect (Inst);
   end Sem_Instantiated_Unit;

   procedure Sem_Component_Instantiation_Statement
     (Stmt: Iir_Component_Instantiation_Statement; Is_Passive : Boolean)
   is
      Decl : Iir;
      Entity_Unit : Iir_Design_Unit;
      Bind : Iir_Binding_Indication;
   begin
      --  FIXME: move this check in parse ?
      if Is_Passive then
         Error_Msg_Sem ("component instantiation forbidden in entity", Stmt);
      end if;

      -- Check for label.
      --  This cannot be moved in parse since a procedure_call may be revert
      --  into a component instantiation.
      if Get_Label (Stmt) = Null_Identifier then
         Error_Msg_Sem ("a component instantiation needs a label", Stmt);
      end if;

      --  Look for the component.
      --  Note: the component was already semantized during label
      --  semantization (this is required to apply configuration
      --  specification).
      Decl := Sem_Instantiated_Unit (Stmt);
      if Decl = Null_Iir then
         return;
      end if;

      -- The association
      Sem_Generic_Port_Association_List
        (Get_Generic_List (Decl), Get_Generic_Map_Aspect_List (Stmt),
         Get_Port_List (Decl), Get_Port_Map_Aspect_List (Stmt),
         Stmt);

      --  FIXME: add sources for signals, in order to detect multiple sources
      --  to unresolved signals.
      --  What happen if the component is not bound ?

      --  Create a default binding indication if necessary.
      if Get_Component_Configuration (Stmt) = Null_Iir
        and then Get_Kind (Decl) = Iir_Kind_Component_Declaration
      then
         Entity_Unit := Get_Visible_Entity_Declaration (Decl);
         if Entity_Unit = Null_Iir then
            if Flags.Warn_Default_Binding then
               Warning_Msg_Sem
                 ("no default binding for instantiation of "
                  & Disp_Node (Decl), Stmt);
               Explain_No_Visible_Entity (Decl);
            end if;
         else
            --  GHDL: this architecture depends on the entity of the default
            --  binding, since this has checked the binding.
            Add_Dependence (Entity_Unit);

            Bind := Sem_Create_Default_Binding_Indication
              (Decl, Entity_Unit, Stmt, False);
            Set_Default_Binding_Indication (Stmt, Bind);
         end if;
      end if;
   end Sem_Component_Instantiation_Statement;

   procedure Sem_Block_Header (Header : Iir_Block_Header) is
      Generic_List : Iir_Generic_List;
      Port_List : Iir_Port_List;
   begin
      Generic_List := Get_Generic_List (Header);
      Sem_Interface_List (Generic_List, Interface_Generic);
      Port_List := Get_Port_List (Header);
      Sem_Interface_List (Port_List, Interface_Port);

      Sem_Generic_Port_Association_List
        (Generic_List, Get_Generic_Map_Aspect_List (Header),
         Port_List, Get_Port_Map_Aspect_List (Header),
         Header);
   end Sem_Block_Header;

   --  Semantize declarations and concurrent statements of ARCH, which is
   --  either an architecture_declaration or a block_statement.
   procedure Sem_Block (Arch: Iir)
   is
      Stmts : Iir_Concurrent_Statement_List;
      Decls : Iir_Declaration_List;
      Implicit : Implicit_Signal_Declaration_Type;
   begin
      Stmts := Get_Concurrent_Statement_List (Arch);
      Decls := Get_Declaration_List (Arch);
      Push_Declarative_Region_With_Signals (Implicit, Decls);
      Sem_Labels (Stmts);
      if Decls /= null then
         Sem_Declaration_List (Decls, Arch);
      end if;
      Sem_Labels_2 (Stmts);
      if Decls /= null then
         Sem_Specification_List (Decls, Stmts);
      end if;
      Sem_Concurrent_Statement_List (Stmts, False);
      Pop_Declarative_Region_With_Signals (Implicit);
   end Sem_Block;

   procedure Sem_Block_Statement (Stmt: Iir_Block_Statement) is
      Expr: Iir;
      Guard : Iir_Guard_Signal_Declaration;
      Header : Iir_Block_Header;
   begin
      --  LRM93 10.1 (declarative region)
      --  7. A block statement.
      Sem_Name.Open_Declarative_Region;
      Set_Is_Within_Flag (Stmt, True);

      --  LRM93 9.1
      --  If a guard expression appears after the reserved word BLOCK, then a
      --  signal with the simple name GUARD of predefined type BOOLEAN is
      --  implicitly declared at the beginning of the declarative part of the
      --  block, and the guard expression defined the value of that signal at
      --  any given time.
      Guard := Get_Guard_Decl (Stmt);
      if Guard /= Null_Iir then
         --  LRM93 9.1
         --  The type of the guard expression must be type BOOLEAN.
         --  GHDL: guard expression must be semantized before creating the
         --   implicit GUARD signal, since the expression may reference GUARD.
         Set_Expr_Staticness (Guard, None);
         Expr := Get_Guard_Expression (Guard);
         Expr := Sem_Expression (Expr, Boolean_Type_Definition);
         if Expr /= Null_Iir then
            Check_Read (Expr);
            Set_Guard_Expression (Guard, Expr);
         end if;

         -- the guard expression is an implicit definition of a signal named
         -- GUARD.  Create this definition.  This is necessary for the type.
         Set_Base_Name (Guard, Guard);
         Set_Identifier (Guard, Std_Names.Name_Guard);
         Set_Type (Guard, Boolean_Type_Definition);
         Set_Block_Statement (Guard, Stmt);
         Sem_Name.Add_Name (Guard);
         Set_Visible_Flag (Guard, True);
      end if;

      Header := Get_Block_Header (Stmt);
      if Header /= Null_Iir then
         Sem_Block_Header (Header);
      end if;

      Sem_Block (Stmt);
      Set_Is_Within_Flag (Stmt, False);
      Sem_Name.Close_Declarative_Region;
   end Sem_Block_Statement;

   procedure Sem_Generate_Statement (Stmt : Iir_Generate_Statement)
   is
      Scheme : Iir;
   begin
      --  LRM93 10.1 (declarative region)
      --  12. A generate statement.
      Sem_Name.Open_Declarative_Region;

      Scheme := Get_Generation_Scheme (Stmt);
      if Get_Kind (Scheme) = Iir_Kind_Iterator_Declaration then
         Sem_Name.Add_Name (Scheme);
         --  LRM93 7.4.2 (Globally Static Primaries)
         --   4. a generate parameter;
         Sem_Iterator (Scheme, Globally);
         Set_Visible_Flag (Scheme, True);
         --  LRM93 9.7
         --  The discrete range in a generation scheme of the first form must
         --  be a static discrete range;
         if Get_Type (Scheme) /= Null_Iir
           and then Get_Type_Staticness (Get_Type (Scheme)) < Globally
         then
            Error_Msg_Sem ("range must be a static discrete range", Stmt);
         end if;
      else
         Scheme := Sem_Expression (Scheme, Boolean_Type_Definition);
         Check_Read (Scheme);
         --  LRM93 9.7
         --  the condition in a generation scheme of the second form must be
         --  a static expression.
         if Scheme /= Null_Iir
           and then Get_Expr_Staticness (Scheme) < Globally
         then
            Error_Msg_Sem ("condition must be a static expression", Stmt);
         else
            Set_Generation_Scheme (Stmt, Scheme);
         end if;
      end if;

      Sem_Block (Stmt);
      Sem_Name.Close_Declarative_Region;
   end Sem_Generate_Statement;

   procedure Sem_Concurrent_Statement_List
     (List : Iir_Concurrent_Statement_List; Is_Passive : Boolean)
   is
      El: Iir;
      Prev_Concurrent_Statement : Iir;
   begin
      if List = null then
         return;
      end if;

      Prev_Concurrent_Statement := Current_Concurrent_Statement;

      for I in Natural loop
         El := Get_Nth_Element (List, I);
         exit when El = null;

         Current_Concurrent_Statement := El;

         case Get_Kind (El) is
            when Iir_Kind_Concurrent_Conditional_Signal_Assignment =>
               if Is_Passive then
                  Error_Msg_Sem
                    ("signal assignment forbidden in entity statement part",
                     El);
               end if;
               Sem_Concurrent_Conditional_Signal_Assignment (El);
            when Iir_Kind_Concurrent_Selected_Signal_Assignment =>
               if Is_Passive then
                  Error_Msg_Sem ("signal assignment forbidden in entity",
                                 El);
               end if;
               Sem_Concurrent_Selected_Signal_Assignment (El);
            when Iir_Kind_Sensitized_Process_Statement =>
               Set_Passive_Flag (El, Is_Passive);
               Sem_Sensitized_Process_Statement (El);
            when Iir_Kind_Process_Statement =>
               Set_Passive_Flag (El, Is_Passive);
               Sem_Process_Statement (El);
            when Iir_Kind_Component_Instantiation_Statement =>
               Sem_Component_Instantiation_Statement (El, Is_Passive);
            when Iir_Kind_Concurrent_Assertion_Statement =>
               --  FIXME: must check assertion expressions does not contain
               --  non-passive subprograms ??
               Sem_Assertion_Statement (El);
            when Iir_Kind_Block_Statement =>
               if Is_Passive then
                  Error_Msg_Sem ("block forbidden in entity statement part",
                                 El);
               end if;
               Sem_Block_Statement (El);
            when Iir_Kind_Generate_Statement =>
               if Is_Passive then
                  Error_Msg_Sem ("generate statement forbidden in an entity",
                                 El);
               end if;
               Sem_Generate_Statement (El);
            when Iir_Kind_Concurrent_Procedure_Call_Statement =>
               Sem_Procedure_Call (El);
            when others =>
               Error_Kind ("sem_concurrent_statement", El);
         end case;
      end loop;
      Current_Concurrent_Statement := Prev_Concurrent_Statement;
   end Sem_Concurrent_Statement_List;

   procedure Sem_Architecture_Declaration (Arch: Iir_Architecture_Declaration)
   is
      Unit : Iir_Design_Unit;
      Entity_Unit : Iir_Design_Unit;
      Entity_Ident : Name_Id;
      Entity_Library : Iir_Entity_Declaration;
   begin
      -- First, find the entity.
      Entity_Ident := Get_Identifier (Get_Entity (Arch));
      Free_Name (Get_Entity (Arch));
      Entity_Unit := Libraries.Load_Primary_Unit
        (Libraries.Work_Library, Entity_Ident, Arch);
      if Entity_Unit = Null_Iir then
         Error_Msg_Sem ("entity `" & Name_Table.Image (Entity_Ident)
                        & "' was not analysed",
                        Arch);
         return;
      end if;
      Entity_Library := Get_Library_Unit (Entity_Unit);
      if Get_Kind (Entity_Library) /= Iir_Kind_Entity_Declaration then
         Error_Msg_Sem ("primary unit `" & Name_Table.Image (Entity_Ident)
                        & "' is not an entity", Arch);
         return;
      end if;

      --  LRM93 11.4
      --   In each case, the second unit depends on the first unit.
      --  GHDL: an architecture depends on its entity.
      Add_Dependence (Entity_Unit);

      -- Transforms an identifier into an entity_decl.
      --FIXME Free_Iir (Entity_Ident);
      Set_Entity (Arch, Entity_Library);

      Add_Context_Clauses (Entity_Unit);

      Set_Is_Within_Flag (Arch, True);
      Set_Is_Within_Flag (Entity_Library, True);
      Sem_Name.Open_Declarative_Region;
      Sem_Name.Add_Entity_Declarations (Entity_Library);

      --  FIXME: an architecture is not a declaration.
      Unit := Get_Design_Unit (Arch);
      Sem_Name.Add_Name (Unit, True);
      Set_Visible_Flag (Unit, True);

      Sem_Block (Arch);
      Sem_Name.Close_Declarative_Region;
      Set_Is_Within_Flag (Arch, False);
      Set_Is_Within_Flag (Entity_Library, False);
      Check_Full_Declaration (Get_Declaration_List (Entity_Library), Arch);
   end Sem_Architecture_Declaration;

   procedure Sem_Binding_Indication (Bind : Iir_Binding_Indication;
                                     Comp : Iir_Component_Declaration;
                                     Parent : Iir)
   is
      Entity_Aspect : Iir;
      Entity : Iir;
   begin
      if Bind = Null_Iir then
         raise Internal_Error;
         return;
      end if;
      Entity_Aspect := Get_Entity_Aspect (Bind);
      if Entity_Aspect /= Null_Iir then
         Entity := Sem_Entity_Aspect (Entity_Aspect);
      else
         --  LRM93 5.2.1
         --  When a binding indication is used in a configuration
         --  specification, it is an error if the entity aspect is absent.
         --  FIXME.
         raise Internal_Error;
      end if;
      if Entity /= Null_Iir then
         Sem_Generic_Port_Association_List
           (Get_Generic_List (Entity), Get_Generic_Map_Aspect_List (Bind),
            Get_Port_List (Entity), Get_Port_Map_Aspect_List (Bind),
            Bind);
         if Get_Generic_Map_Aspect_List (Bind) = Null_Iir_List then
            Set_Default_Generic_Map_Aspect_List
              (Bind, Create_Default_Map_Aspect (Comp, Entity,
                                                Map_Generic, Parent));
         end if;
         if Get_Port_Map_Aspect_List (Bind) = Null_Iir_List then
            Set_Default_Port_Map_Aspect_List
              (Bind, Create_Default_Map_Aspect (Comp, Entity,
                                                Map_Port, Parent));
         end if;
      end if;
   end Sem_Binding_Indication;

   --  Set configuration_specification or component_configuration SPEC to
   --  component instantiation COMP.
   procedure Apply_Configuration_Specification
     (Comp : Iir_Component_Instantiation_Statement; Spec : Iir)
   is
      Prev_Spec : Iir;
   begin
      if Get_Kind (Spec) = Iir_Kind_Configuration_Specification then
         Prev_Spec := Get_Component_Configuration (Comp);
         if Prev_Spec /= null then
            Error_Msg_Sem ("instantiation was already specified by " &
                           Disp_Node (Prev_Spec), Comp);
         else
            Set_Component_Configuration (Comp, Spec);
         end if;
      else
         Prev_Spec := Get_Component_Configuration (Comp);
         if Prev_Spec /= null then
            Error_Msg_Sem ("instantiation was already specified by " &
                           Disp_Node (Prev_Spec), Comp);
         else
            Set_Component_Configuration (Comp, Spec);
         end if;
      end if;
   end Apply_Configuration_Specification;

   --  Semantize component_configuration or configuration_specification SPEC.
   --  STMTS is the concurrent statement list related to SPEC.
   procedure Sem_Component_Specification
     (Stmts : Iir_Concurrent_Statement_List; Spec : Iir)
   is
      use Sem_Name;
      List : Iir;
      El : Iir;
      Inter : Sem_Name.Name_Interpretation_Type;
      Comp : Iir;
   begin
      Comp := Find_Declaration (Get_Component_Name (Spec), Decl_Component);
      if Comp = Null_Iir then
         return;
      end if;
      Set_Component_Name (Spec, Comp);

      List := Get_Instantiation_List (Spec);
      case Get_Kind (List) is
         when Iir_Kind_Designator_By_All =>
            --  LRM93 5.2
            --  * If the reserved word ALL is supplied, then the configuration
            --    specification applies to all instances of the specified
            --    component declaration whose labels are (implicitly) declared
            --    in the immediately enclosing declarative region part.
            --    This rule applies only to those component instantiation
            --    statements whose corresponding instantiated units name
            --    component.
            for I in Natural loop
               El := Get_Nth_Element (Stmts, I);
               exit when El = null;
               if Get_Kind (El) = Iir_Kind_Component_Instantiation_Statement
                 and then Get_Instantiated_Unit (El) = Comp
               then
                  Apply_Configuration_Specification (El, Spec);
               end if;
            end loop;
         when Iir_Kind_Designator_By_Others =>
            --  LRM93 5.2
            --  * If the reserved word OTHERS is supplied, then the
            --    configuration specification applies to instances of the
            --    specified component declaration whoce labels are (implicitly)
            --    declared in the immediatly enclosing declarative part,
            --    provided that each such component instance is not explicitly
            --    names in the instantiation list of a previous configuration
            --    specification.
            --    This rule applies only to those component instantiation
            --    statements whose corresponding instantiated units name
            --    components.
            for I in Natural loop
               El := Get_Nth_Element (Stmts, I);
               exit when El = null;
               if Get_Kind (El) = Iir_Kind_Component_Instantiation_Statement
                 and then Get_Instantiated_Unit (El) = Comp
                 and then Get_Component_Configuration (El) = Null_Iir
               then
                  Apply_Configuration_Specification (El, Spec);
               end if;
            end loop;
         when Iir_Kind_Designator_List =>
            --  LRM93 5.2
            --  * If a list of instantiation labels is supplied, then the
            --    configuration specification applies to the corresponding
            --    component instances.
            --    Such labels must be (implicitly) declared within the
            --    immediatly enclosing declarative part.
            --    It is an error if these component instances are not instances
            --    of the component declaration named in the component
            --    specification.
            --    It is also an error if any of the labels denote a component
            --    instantiation statement whose corresponding instantiated unit
            --    does not name a component.
            -- FIXME: error message are *really* cryptic.
            for I in Natural loop
               El := Get_Nth_Element (List, I);
               exit when El = null;
               Inter := Sem_Name.Get_Interpretation (Get_Identifier (El));
               if Inter = Sem_Name.No_Name_Interpretation then
                  Error_Msg_Sem ("no component instantation with label `"
                                 & Get_String (El) & ''', El);
               elsif not Sem_Name.Is_In_Current_Declarative_Part (Inter) then
                  --  FIXME.
                  Error_Msg_Sem
                    ("label not in block declarative part", El);
               else
                  Comp := Get_Declaration (Inter);
                  if Get_Kind (Comp)
                    /= Iir_Kind_Component_Instantiation_Statement
                  then
                     Error_Msg_Sem
                       ("label does not denote a component instantiation",
                        El);
                  elsif Get_Instantiated_Unit (Comp)
                    /= Get_Component_Name (Spec)
                  then
                     Error_Msg_Sem ("component names mismatch", El);
                  else
                     Apply_Configuration_Specification (Comp, Spec);
                     Free_Iir (El);
                     Replace_Nth_Element (List, I, Comp);
                  end if;
               end if;
            end loop;
         when others =>
            Error_Kind ("sem_component_specification", List);
      end case;
   end Sem_Component_Specification;

   --  LRM93 5.2.2
   function Get_Visible_Entity_Declaration (Comp: Iir_Component_Declaration)
     return Iir_Design_Unit
   is
      use Sem_Name;

      function Is_Entity_Declaration (Decl : Iir) return Boolean is
      begin
         return Get_Kind (Decl) = Iir_Kind_Design_Unit and then
           Get_Kind (Get_Library_Unit (Decl)) = Iir_Kind_Entity_Declaration;
      end Is_Entity_Declaration;

      Inter : Name_Interpretation_Type;
      Name : Name_Id;
      Decl : Iir;
      Target_Lib : Iir;
   begin
      Name := Get_Identifier (Comp);
      Inter := Get_Interpretation (Name);

      if Inter /= No_Name_Interpretation then
         --  A visible entity declaration is either:
         --
         --  a) An entity declaration that has the same simple name as that of
         --     the instantiated component and that is directly visible
         --     (see 10.3),
         Decl := Get_Declaration (Inter);
         if Is_Entity_Declaration (Decl) then
            return Decl;
         end if;

         --  b)  An entity declaration that has the same simple name that of
         --      the instantiated component and that would be directly
         --      visible in the absence of a directly visible (see 10.3)
         --      component declaration with the same simple name as that
         --      of the entity declaration, or
         if Get_Kind (Decl) = Iir_Kind_Component_Declaration then
            Inter := Get_Under_Interpretation (Name);
            if Inter /= No_Name_Interpretation then
               Decl := Get_Declaration (Inter);
               if Is_Entity_Declaration (Decl) then
                  return Decl;
               end if;
            end if;
         end if;
      end if;

      --  VHDL98:
      --  c) An entity declaration denoted by "L.C", where L is the target
      --     library and C is the simple name of the instantiated component.
      --     The target library is the library logical name of the library
      --     containing the design unit in which the component C is
      --     declared.
      if Flags.Vhdl_Std >= Vhdl_98 or else Flags.Vhdl_Std = Vhdl_93c then
         --  Find target library.
         Target_Lib := Comp;
         while Get_Kind (Target_Lib) /= Iir_Kind_Library_Declaration loop
            Target_Lib := Get_Parent (Target_Lib);
         end loop;

         Decl := Libraries.Find_Primary_Unit (Target_Lib, Name);
         if Decl /= Null_Iir and then Is_Entity_Declaration (Decl) then
            return Decl;
         end if;
      end if;

      return Null_Iir;
   end Get_Visible_Entity_Declaration;

   --  Explain why there is no default binding for COMP.
   procedure Explain_No_Visible_Entity (Comp: Iir_Component_Declaration)
   is
      use Sem_Name;

      Inter : Name_Interpretation_Type;
      Name : Name_Id;
      Decl : Iir;
   begin
      Name := Get_Identifier (Comp);
      Inter := Get_Interpretation (Name);

      if Inter /= No_Name_Interpretation then
         --  A visible entity declaration is either:
         --
         --  a) An entity declaration that has the same simple name as that of
         --     the instantiated component and that is directly visible
         --     (see 10.3),
         Decl := Get_Declaration (Inter);
         Warning_Msg_Sem ("visible declaration for " & Name_Table.Image (Name)
                          & " is " & Disp_Node (Decl), Decl);

         --  b)  An entity declaration that has the same simple name that of
         --      the instantiated component and that would be directly
         --      visible in the absence of a directly visible (see 10.3)
         --      component declaration with the same simple name as that
         --      of the entity declaration, or
         if Get_Kind (Decl) = Iir_Kind_Component_Declaration then
            Inter := Get_Under_Interpretation (Name);
            if Inter /= No_Name_Interpretation then
               Decl := Get_Declaration (Inter);
               Error_Msg_Sem ("interpretation behind the component is "
                              & Disp_Node (Decl), Comp);
            end if;
         end if;
      end if;

      --  VHDL98:
      --  c) An entity declaration denoted by "L.C", where L is the target
      --     library and C is the simple name of the instantiated component.
      --     The target library is the library logical name of the library
      --     containing the design unit in which the component C is
      --     declared.
      if Flags.Vhdl_Std >= Vhdl_98 then
         --  FIXME: to be done
         raise Internal_Error;
      end if;
   end Explain_No_Visible_Entity;

   --  LRM 5.2.2
   --  The default generic map aspect associates each local generic in
   --  the corresponding component instantiation (if any) with a formal
   --  of the same simple name.
   --  It is an error if such a formal does not exist, or if its mode and
   --  type are not appropriate for such an association.
   --  Any remaining unassociated formals are associated with the actual
   --  designator OPEN.

   --  LRM 5.2.2
   --  The default port map aspect associates each local port in the
   --  corresponding component instantiation (if any) with a formal of
   --  the same simple name.
   --  It is an error if such a formal does not exist, or if its mode
   --  and type are not appropriate for such an association.
   --  Any remaining unassociated formals are associated with the actual
   --  designator OPEN.
   type Map_Kind_String_Type is array (Map_Kind_Type) of String_Cst;
   Map_Kind_Name : constant Map_Kind_String_Type :=
     (Map_Generic => new String'("generic"),
      Map_Port => new String'("port"));

   function Create_Default_Map_Aspect
     (Comp : Iir; Entity : Iir; Kind : Map_Kind_Type; Parent : Iir)
     return Iir_Association_List
   is
      Res : Iir_Association_List;
      Comp_El, Ent_El : Iir;
      Assoc : Iir;
      Found : Natural;
      Comp_List : Iir_List;
      Ent_List : Iir_List;
      Error : Boolean;
   begin
      case Kind is
         when Map_Generic =>
            Ent_List := Get_Generic_List (Entity);
            Comp_List := Get_Generic_List (Comp);
         when Map_Port =>
            Ent_List := Get_Port_List (Entity);
            Comp_List := Get_Port_List (Comp);
      end case;

      --  If no formal, then there is no association list.
      --  Just check there is no actuals.
      if Ent_List = Null_Iir_List then
         if Comp_List /= Null_Iir_List then
            Error_Msg_Sem ("no " & Map_Kind_Name (Kind).all & "s of "
                           & Disp_Node (Entity)
                           & " to be associated with "
                           & Map_Kind_Name (Kind).all
                           & "s of " & Disp_Node (Comp), Parent);
         end if;
         return Null_Iir_List;
      end if;

      --  No error found yet.
      Error := False;

      Res := Create_Iir (Iir_Kind_Association_List);
      Found := 0;
      for I in Natural loop
         Ent_El := Get_Nth_Element (Ent_List, I);
         exit when Ent_El = Null_Iir;
         --  Find the component generic/port with the same name.
         if Comp_List /= Null_Iir_List then
            Comp_El := Find_Name_In_List (Comp_List, Get_Identifier (Ent_El));
         else
            Comp_El := Null_Iir;
         end if;
         if Comp_El = Null_Iir then
            Assoc := Create_Iir (Iir_Kind_Association_Element_Open);
         else
            if not Are_Nodes_Compatible (Comp_El, Ent_El) then
               Error_Msg_Sem
                 ("type of "
                  & Disp_Node (Comp_El) & " from " & Disp_Node (Comp)
                  & " and "
                  & Disp_Node (Ent_El) & " from " & Disp_Node (Entity)
                  & " are not compatible for an association",
                  Parent);
               Error := True;
            elsif Kind = Map_Port
              and then
              not Check_Port_Association_Restriction (Ent_El, Comp_El, Parent)
            then
               Error := True;
            end if;
            Assoc := Create_Iir (Iir_Kind_Association_Element_By_Expression);
            Set_Whole_Association_Flag (Assoc, True);
            Set_Actual (Assoc, Comp_El);
            Found := Found + 1;
         end if;
         Set_Formal (Assoc, Ent_El);
         Set_Associated_Formal (Assoc, Ent_El);
         Append_Element (Res, Assoc);
      end loop;
      if not ((Comp_List = Null_Iir and Found = 0)
              or else (Get_Nbr_Elements (Comp_List) = Found))
      then
         --  At least one component generic/port cannot be associated with
         --  the entity one.
         Error_Msg_Sem (Map_Kind_Name (Kind).all & "(s) of " & Disp_Node (Comp)
                        & " have no association in " & Disp_Node (Entity),
                        Parent);
         Error := True;
      end if;
      if Error then
         return Null_Iir_List;
      else
         return Res;
      end if;
   end Create_Default_Map_Aspect;

   function Sem_Create_Default_Binding_Indication
     (Comp : Iir_Component_Declaration;
      Entity_Unit : Iir_Design_Unit;
      Parent : Iir;
      Force : Boolean)
     return Iir_Binding_Indication
   is
      Entity : Iir_Entity_Declaration;
      Aspect : Iir;
      Res : Iir;
      Design_Unit : Iir_Design_Unit;
   begin
      --  LRM 5.2.2
      --  The default binding indication consists of a default entity aspect,
      --  together with a default generic map aspect and a default port map
      --  aspect, as appropriate.

      if Entity_Unit = Null_Iir then
         if not Force then
            return Null_Iir;
         end if;

         --  LRM 5.2.2
         --  If no visible entity declaration has the same simple name as that
         --  of the instantiated component, then the default entity aspect is
         --  OPEN.
         Aspect := Create_Iir (Iir_Kind_Entity_Aspect_Open);
         Location_Copy (Aspect, Comp);
         Res := Create_Iir (Iir_Kind_Binding_Indication);
         Set_Entity_Aspect (Res, Aspect);
         return Res;
      else
         --  LRM 5.2.2
         --  Otherwise, the default entity aspect is of the form:
         --    ENTITY entity_name ( architecture_identifier)
         --  where the entity name is the simple name of the instantiated
         --  component and the architecture identifier is the same as the
         --  simple name of the most recently analyzed architecture body
         --  associated with the entity declaration.
         --
         --  If this rule is applied either to a binding indication contained
         --  within a configuration specification or to a component
         --  configuration that does not contain an explicit inner block
         --  configuration, then the architecture identifier is determined
         --  during elaboration of the design hierarchy containing the binding
         --  indication.
         --
         --  Likewise, if a component instantiation statement contains an
         --  instantiated unit containing the reserved word ENTITY, but does
         --  not contain an explicitly specified architecture identifier, this
         --  rule is applied during the elaboration of the design hierarchy
         --  containing a component instantiation statement.
         --
         --  In all other cases, this rule is applied during analysis of the
         --  binding indication.
         --
         --  It is an error if there is no architecture body associated with
         --  the entity declaration denoted by an entity name that is the
         --  simple name of the instantiated component.
         null;
      end if;

      Design_Unit := Libraries.Load_Primary_Unit
        (Get_Parent (Get_Design_File (Entity_Unit)),
         Get_Identifier (Get_Library_Unit (Entity_Unit)),
         Parent);
      if Design_Unit = Null_Iir then
         --  Found an entity which is not in the library.
         raise Internal_Error;
      end if;

      Entity := Get_Library_Unit (Design_Unit);

      Res := Create_Iir (Iir_Kind_Binding_Indication);
      Aspect := Create_Iir (Iir_Kind_Entity_Aspect_Entity);
      Location_Copy (Aspect, Parent);
      Set_Entity (Aspect, Design_Unit);
      Set_Entity_Aspect (Res, Aspect);

      --  LRM 5.2.2
      --  The default binding indication includes a default generic map aspect
      --  if the design entity implied by the entity aspect contains formal
      --  generics.
      Set_Generic_Map_Aspect_List
        (Res, Create_Default_Map_Aspect (Comp, Entity, Map_Generic, Parent));

      --  LRM 5.2.2
      --  The default binding indication includes a default port map aspect
      --  if the design entity implied by the entity aspect contains formal
      --  ports.
      Set_Port_Map_Aspect_List
        (Res, Create_Default_Map_Aspect (Comp, Entity, Map_Port, Parent));

      return Res;
   end Sem_Create_Default_Binding_Indication;

   --  LRM93 1.3.2
   procedure Sem_Component_Configuration
     (Conf : Iir_Component_Configuration; Father : Iir)
   is
      Block : Iir;
      Configured_Block : Iir;
      Stmts_List : Iir_List;
      Binding : Iir;
      Entity : Iir_Design_Unit;
      Comp : Iir_Component_Declaration;
   begin
      --  LRM93 10.1
      --  11. A component configuration
      Open_Declarative_Region;

      --  LRM93 10.2
      --  If a component configuration appears as a configuration item
      --  immediatly within a block configuration that configures a given
      --  block, and the scope of a given declaration includes the end of the
      --  declarative part of that block, then the scope of the given
      --  declaration extends from the beginning to the end of the
      --  declarative region associated with the given component configuration.
      -- GHDL: this is for labels of component instantiation statements, and
      -- for local ports and generics of the component.
      if Get_Kind (Father) = Iir_Kind_Block_Configuration then
         Configured_Block := Get_Block_Specification (Father);
         if Get_Kind (Configured_Block) = Iir_Kind_Design_Unit then
            raise Internal_Error;
         end if;
         Configured_Block :=
           Get_Block_From_Block_Specification (Configured_Block);
         Sem_Name.Extend_Scope_Of_Block_Declarations (Configured_Block);
         --  FIXME: this is wrong (all declarations should be considered).
         Stmts_List := Get_Concurrent_Statement_List (Configured_Block);
      else
         --  Can a component configuration not be just inside a block
         --  configuration ?
         raise Internal_Error;
      end if;
      Sem_Component_Specification (Stmts_List, Conf);

      --  FIXME: (todo)
      --  If a given component instance is unbound in the corresponding block,
      --  then any explicit component configuration for that instance that does
      --  not contain an explicit binding indication will contain an implicit,
      --  default binding indication (see 5.2.2).  Similarly, if a given
      --  component instance is unbound in the corresponding block, then any
      --  implicit component configuration for that instance will contain an
      --  implicit, default binding indication.
      -- FIXME: ports and generics declared by the component must be
      --   made visible here; create a declarative_region only for this purpose
      Open_Declarative_Region;
      Comp := Get_Component_Name (Conf);
      Sem_Name.Add_Component_Declarations (Comp);
      Binding := Get_Binding_Indication (Conf);
      if Binding /= Null_Iir then
         Sem_Binding_Indication (Binding, Comp, Conf);
      else
         --  Create a default binding indication.
         Entity := Get_Visible_Entity_Declaration (Comp);
         Binding := Sem_Create_Default_Binding_Indication
           (Comp, Entity, Conf, False);

         if Binding /= Null_Iir then
            --  Remap to defaults.
            Set_Default_Entity_Aspect (Binding, Get_Entity_Aspect (Binding));
            Set_Entity_Aspect (Binding, Null_Iir);

            Set_Default_Generic_Map_Aspect_List
              (Binding, Get_Generic_Map_Aspect_List (Binding));
            Set_Generic_Map_Aspect_List (Binding, Null_Iir);

            Set_Default_Port_Map_Aspect_List
              (Binding, Get_Port_Map_Aspect_List (Binding));
            Set_Port_Map_Aspect_List (Binding, Null_Iir);

            Set_Binding_Indication (Conf, Binding);
         end if;
      end if;
      Close_Declarative_Region;
      Block := Get_Block_Configuration (Conf);
      if Block /= Null_Iir then
         Sem_Block_Configuration (Block, Conf);
      end if;
      Close_Declarative_Region;
   end Sem_Component_Configuration;

   --  FATHER is the block_configuration, configuration_declaration,
   --  component_configuration containing the block_configuration BLOCK_CONF.
   procedure Sem_Block_Configuration
     (Block_Conf : Iir_Block_Configuration; Father: Iir)
   is
      List : Iir_Configuration_Item_List;
      El : Iir;
      Block : Iir;
      Stmts_List : Iir_Concurrent_Statement_List;
      Decl_List : Iir_List;
   begin
      case Get_Kind (Father) is
         when Iir_Kind_Configuration_Declaration =>
            --  LRM93 1.3.1
            --  If a block configuration appears immediately within a
            --  configuration declaration, then the block specification of that
            --  block configuration must be an architecture name, and that
            --  architecture name must denote a design entity body whose
            --  interface is defined by the entity declaration denoted by the
            --  entity name of the enclosing configuration declaration.
            declare
               Block_Spec : Iir;
               Arch : Iir_Architecture_Declaration;
               Design: Iir_Design_Unit;
            begin
               Block_Spec := Get_Block_Specification (Block_Conf);
               if Get_Kind (Block_Spec) /= Iir_Kind_Simple_Name then
                  Error_Msg_Sem ("architecture name expected", Block_Spec);
                  return;
               end if;
               --  LRM 10.3 rule b)
               --  For an architecture body associated with a given entity
               --  declaration: at the place of the block specification in a
               --  block configuration for an external block whose interface
               --  is defined by that entity declaration.
               Design := Libraries.Load_Secondary_Unit
                 (Get_Entity (Father), Get_Identifier (Block_Spec),
                  Block_Conf);
               if Design = Null_Iir then
                  Error_Msg_Sem
                    ("no architecture `" & Get_String (Block_Spec) & "'",
                     Block_Conf);
                  return;
               end if;
               Free_Iir (Block_Spec);
               Arch := Get_Library_Unit (Design);
               Set_Block_Specification (Block_Conf, Arch);
               Block := Arch;
               Add_Dependence (Design);
            end;

         when Iir_Kind_Component_Configuration =>
            --  LRM93 1.3.1
            --  If a block configuration appears immediately within a component
            --  configuration, then the corresponding components must be
            --  fully bound, the block specification of that block
            --  configuration must be an architecture name, and that
            --  architecture name must denote the same architecture body as
            --  that to which the corresponding components are bound.
            declare
               Block_Spec : Iir;
               Arch : Iir_Architecture_Declaration;
               Design: Iir_Design_Unit;
               Entity_Aspect : Iir;
               Comp_Arch : Iir;
            begin
               Entity_Aspect :=
                 Get_Entity_Aspect (Get_Binding_Indication (Father));
               if Entity_Aspect = Null_Iir or else
                 Get_Kind (Entity_Aspect) /= Iir_Kind_Entity_Aspect_Entity
               then
                  Error_Msg_Sem ("corresponding component not fully bound",
                                 Block_Conf);
               end if;

               Block_Spec := Get_Block_Specification (Block_Conf);
               if Get_Kind (Block_Spec) /= Iir_Kind_Simple_Name then
                  Error_Msg_Sem ("architecture name expected", Block_Spec);
                  return;
               end if;

               Comp_Arch := Get_Architecture (Entity_Aspect);
               if Comp_Arch /= Null_Iir then
                  if Get_Kind (Comp_Arch) /= Iir_Kind_Simple_Name then
                     raise Internal_Error;
                  end if;
                  if Get_Identifier (Comp_Arch) /= Get_Identifier (Block_Spec)
                  then
                     Error_Msg_Sem
                       ("block specification name is different from "
                        & "component architecture name", Block_Spec);
                     return;
                  end if;
               end if;

               Design := Libraries.Load_Secondary_Unit
                 (Get_Entity (Entity_Aspect), Get_Identifier (Block_Spec),
                  Block_Conf);
               if Design = Null_Iir then
                  Error_Msg_Sem
                    ("no architecture `" & Get_String (Block_Spec) & "'",
                     Block_Conf);
                  return;
               end if;
               Free_Iir (Block_Spec);
               Arch := Get_Library_Unit (Design);
               Set_Block_Specification (Block_Conf, Arch);
               Block := Arch;
            end;

         when Iir_Kind_Block_Configuration =>
            --  LRM93 1.3.1
            --  If a block configuration appears immediately within another
            --  block configuration, then the block specification of the
            --  contained block configuration must be a block statement or
            --  generate statement label, and the label must denote a block
            --  statement or generate statement that is contained immediatly
            --  within the block denoted by the block specification of the
            --  containing block configuration.
            -- FIXME: not yet implemented
            declare
               Block_Spec : Iir;
               Block_Stmts : Iir_Concurrent_Statement_List;
               Block_Spec_Kind : Iir_Kind;
               Prev : Iir_Block_Configuration;
            begin
               Block_Spec := Get_Block_Specification (Block_Conf);
               --  Remember the kind of BLOCK_SPEC, since the node can be free
               --  by find_declaration if it is a simple name.
               Block_Spec_Kind := Get_Kind (Block_Spec);
               case Block_Spec_Kind is
                  when Iir_Kind_Simple_Name =>
                     Block := Block_Spec;
                  when Iir_Kind_Parenthesis_Name =>
                     Block := Get_Prefix (Block_Spec);
                  when Iir_Kind_Slice_Name =>
                     Block := Get_Prefix (Block_Spec);
                  when others =>
                     Error_Msg_Sem ("label expected", Block_Spec);
                     return;
               end case;
               Block := Find_Declaration (Block, Decl_Label);
               if Block = Null_Iir then
                  return;
               end if;
               case Get_Kind (Block) is
                  when Iir_Kind_Block_Statement =>
                     if Block_Spec_Kind /= Iir_Kind_Simple_Name then
                        Error_Msg_Sem
                          ("label does not denote a generate statement",
                           Block_Spec);
                     end if;
                     Prev := Get_Block_Block_Configuration (Block);
                     if Prev /= Null_Iir then
                        Error_Msg_Sem
                          (Disp_Node (Block) & " was already configured at "
                           & Disp_Location (Prev),
                           Block_Conf);
                        return;
                     end if;
                     Set_Block_Block_Configuration (Block, Block_Conf);
                  when Iir_Kind_Generate_Statement =>
                     if Block_Spec_Kind /= Iir_Kind_Simple_Name
                       and then Get_Kind (Get_Generation_Scheme (Block))
                       /= Iir_Kind_Iterator_Declaration
                     then
                        --  LRM93 1.3
                        --  If the block specification of a block configuration
                        --  contains a generate statement label, and if this
                        --  label contains an index specification, then it is
                        --  an error if the generate statement denoted by the
                        --  label does not have a generation scheme including
                        --  the reserved word for.
                        Error_Msg_Sem ("generate statement does not has a for",
                                       Block_Spec);
                        return;
                     end if;
                     Set_Prev_Block_Configuration
                       (Block_Conf, Get_Generate_Block_Configuration (Block));
                     Set_Generate_Block_Configuration (Block, Block_Conf);
                  when others =>
                     Error_Msg_Sem ("block statement label expected",
                                    Block_Conf);
                     return;
               end case;
               Block_Stmts := Get_Concurrent_Statement_List
                 (Get_Block_From_Block_Specification
                  (Get_Block_Specification (Father)));
               if not Is_In_List (Block_Stmts, Block) then
                  Error_Msg_Sem
                    ("label does not denotes an inner block statement",
                     Block_Conf);
                  return;
               end if;

               if Block_Spec_Kind = Iir_Kind_Parenthesis_Name then
                  Block_Spec := Sem_Index_Specification
                    (Block_Spec, Get_Type (Get_Generation_Scheme (Block)));
                  if Block_Spec /= Null_Iir then
                     Set_Prefix (Block_Spec, Block);
                     Set_Block_Specification (Block_Conf, Block_Spec);
                     Block_Spec_Kind := Get_Kind (Block_Spec);
                  end if;
               end if;

               case Block_Spec_Kind is
                  when Iir_Kind_Simple_Name =>
                     Set_Block_Specification (Block_Conf, Block);
                  when Iir_Kind_Indexed_Name
                    | Iir_Kind_Slice_Name =>
                     null;
                  when Iir_Kind_Parenthesis_Name =>
                     null;
                  when others =>
                     raise Internal_Error;
               end case;
            end;

         when others =>
            Error_Kind ("sem_block_configuration", Father);
      end case;

      --  LRM93 10.1
      --  10. A block configuration
      Sem_Name.Open_Scope_Extension;

      --  LRM 10.3
      --  In addition, any declaration that is directly visible at the end of
      --  the declarative part of a given block is directly visible in a block
      --  configuration that configure the given block.  This rule holds unless
      --  a use clause that makes a homograph of the declaration potentially
      --  visible (see 10.4) appears in the corresponding configuration
      --  declaration, and if the scope of that use clause encompasses all or
      --  part of those configuration items.  If such a use clase appears, then
      --  the declaration will be directly visible within the corresponding
      --  configuration items, except at hose places that fall within the scope
      --  of the additional use clause.  At such places, neither name will be
      --  directly visible.
      --  FIXME: handle use clauses.
      Sem_Name.Extend_Scope_Of_Block_Declarations (Block);

      Decl_List := Get_Declaration_List (Block_Conf);
      if Decl_List /= Null_Iir then
         for I in Natural loop
            El := Get_Nth_Element (Decl_List, I);
            exit when El = Null_Iir;
            case Get_Kind (El) is
               when Iir_Kind_Use_Clause =>
                  Sem_Use_Clause (El);
               when others =>
                  --  Parse checks there are only use clauses.
                  raise Internal_Error;
            end case;
         end loop;
      end if;

      Stmts_List := Get_Concurrent_Statement_List (Block);
      Clear_Instantiation_Configuration (Stmts_List);

      List := Get_Configuration_Item_List (Block_Conf);
      if List /= null then
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = null;
            case Get_Kind (El) is
               when Iir_Kind_Block_Configuration =>
                  Sem_Block_Configuration (El, Block_Conf);
               when Iir_Kind_Component_Configuration =>
                  Sem_Component_Configuration (El, Block_Conf);
               when others =>
                  Error_Kind ("sem_block_configuration(2)", El);
            end case;
         end loop;
      end if;
      Sem_Name.Close_Scope_Extension;
   end Sem_Block_Configuration;

   procedure Sem_Configuration_Declaration (Decl: Iir)
   is
      Unit : Iir_Design_Unit;
      Entity_Design: Iir_Design_Unit;
      Name : Iir_Simple_Name;
   begin
      -- LRM93 1.3
      -- The entity name identifies the name of the entity declaration that
      -- defines the design entity at the apex of the design hierarchy.
      -- For a configuration of a given design entity, both the configuration
      -- declaration and the corresponding entity declaration must reside in
      -- the same library.
      Name := Get_Entity (Decl);
      Entity_Design := Libraries.Load_Primary_Unit
        (Libraries.Work_Library, Get_Identifier (Name), Decl);
      if Entity_Design = Null_Iir then
         Error_Msg_Sem ("entity '" & Get_String (Get_Entity (Decl))
                        & "' was not analysed",
                        Decl);
         return;
      end if;
      Free_Name (Name);
      Set_Entity (Decl, Entity_Design);

      --  LRM93 11.4
      --  A primary unit whose name is referenced within a given design unit
      --  must be analyzed prior to the analysis of the given design unit.
      Add_Dependence (Entity_Design);

      Unit := Get_Design_Unit (Decl);
      Sem_Name.Add_Name (Unit);
      Set_Visible_Flag (Unit, True);

      --  LRM93 10.1
      --  2.  A configuration declaration.
      Sem_Name.Open_Declarative_Region;

      --  LRM93 10.2
      --  In addition to the above rules, the scope of any declaration that
      --  includes the end of the declarative part of a given block (wether
      --  it be an external block defined by a design entity or an internal
      --  block defined by a block statement) extends into a configuration
      --  declaration that configures the given block.
      Add_Context_Clauses (Entity_Design);
      Sem_Name.Add_Entity_Declarations (Get_Library_Unit (Entity_Design));

      Sem_Declaration_List (Get_Declaration_List (Decl), Decl);
      Sem_Block_Configuration (Get_Block_Configuration (Decl), Decl);
      Sem_Name.Close_Declarative_Region;
   end Sem_Configuration_Declaration;

   -- Return true if package declaration DECL needs a body.
   -- Ie, it contains subprogram specification or deferred constants.
   function Package_Need_Body_P (Decl: Iir_Package_Declaration)
     return Boolean
   is
      Decl_List: Iir_List;
      El: Iir;
   begin
      Decl_List := Get_Declaration_List (Decl);
      for I in Natural loop
         El := Get_Nth_Element (Decl_List, I);
         exit when El = null;
         case Get_Kind (El) is
            when Iir_Kind_Function_Specification
              | Iir_Kind_Procedure_Specification =>
               return True;
            when Iir_Kind_Constant_Declaration =>
               if Get_Default_Value (El) = null then
                  return True;
               end if;
            when Iir_Kind_Variable_Declaration =>
               null;
            when Iir_Kind_Type_Declaration
              | Iir_Kind_Subtype_Declaration =>
               null;
            when Iir_Kind_Implicit_Function_Declaration
              | Iir_Kind_Implicit_Procedure_Declaration =>
               null;
            when Iir_Kind_File_Declaration =>
               null;
            when Iir_Kind_Signal_Declaration =>
               null;
            when Iir_Kind_Attribute_Declaration =>
               null;
            when Iir_Kind_Attribute_Specification =>
               null;
            when Iir_Kind_Use_Clause =>
               null;
            when Iir_Kind_Component_Declaration =>
               null;
            when others =>
               Error_Kind ("package_need_body_p", El);
         end case;
      end loop;
      return False;
   end Package_Need_Body_P;

   procedure Sem_Package_Declaration (Decl: Iir_Package_Declaration)
   is
      Unit : Iir_Design_Unit;
      Decls : Iir_Declaration_List;
      Implicit : Implicit_Signal_Declaration_Type;
   begin
      Unit := Get_Design_Unit (Decl);
      Sem_Name.Add_Name (Unit);
      Set_Visible_Flag (Unit, True);

      --  LRM93 10.1
      --  4. A package declaration, together with the corresponding
      --     body (if any).
      Sem_Name.Open_Declarative_Region;

      Decls := Get_Declaration_List (Decl);

      Push_Declarative_Region_With_Signals (Implicit, Decls);

      Sem_Declaration_List (Decls, Decl);

      Pop_Declarative_Region_With_Signals (Implicit);
      Sem_Name.Close_Declarative_Region;
      Set_Need_Body (Decl, Package_Need_Body_P (Decl));
   end Sem_Package_Declaration;

   procedure Sem_Package_Body (Decl: Iir)
   is
      Package_Ident: Name_Id;
      Design_Unit: Iir_Design_Unit;
      Package_Decl: Iir;
   begin
      -- First, find the entity.
      Package_Ident := Get_Identifier (Decl);
      Design_Unit := Libraries.Load_Primary_Unit
        (Get_Parent (Get_Design_File (Get_Current_Design_Unit)),
         Package_Ident, Decl);
      if Design_Unit = Null_Iir then
         Error_Msg_Sem ("package `" & Name_Table.Image (Package_Ident)
                        & "' was not analysed",
                        Decl);
         return;
      end if;
      Package_Decl := Get_Library_Unit (Design_Unit);
      if Get_Kind (Package_Decl) /= Iir_Kind_Package_Declaration then
         Error_Msg_Sem
           ("primary unit `" & Name_Table.Image (Package_Ident)
            & "' is not an package", Decl);
         return;
      end if;

      --  Emit a warning is a body is not necessary.
      if not Get_Need_Body (Package_Decl) then
         Warning_Msg_Sem
           (Disp_Node (Package_Decl) & " does not require a body", Decl);
      end if;

      Set_Package (Decl, Package_Decl);
      Add_Dependence (Design_Unit);

      --  Add the context clauses from the primary unit.
      Add_Context_Clauses (Design_Unit);

      Sem_Name.Open_Declarative_Region;

      Sem_Name.Add_Package_Declarations (Package_Decl);

      Sem_Declaration_List (Get_Declaration_List (Decl), Decl);
      Check_Full_Declaration (Get_Declaration_List (Package_Decl), Decl);
      Sem_Name.Close_Declarative_Region;
   end Sem_Package_Body;

   procedure Sem_Library_Clause (Decl: Iir_Library_Clause) is
      use Sem_Name;
      Ident : Name_Id;
      Interpretation: Name_Interpretation_Type;
      Lib: Iir;
   begin
      Ident := Get_Identifier (Decl);
      Interpretation := Get_Interpretation (Ident);
      if Interpretation /= No_Name_Interpretation then
         Lib := Get_Declaration (Interpretation);
         if Get_Kind (Lib) = Iir_Kind_Library_Declaration then
            return;
         end if;
      end if;
      Lib := Libraries.Get_Library (Ident);
      if Lib = Null_Iir then
         Error_Msg_Sem
           ("no resource library `" & Name_Table.Image (Ident) & ''', Decl);
      else
         Set_Library_Declaration (Decl, Lib);
         if Ident /= Std_Names.Name_Work
           and then Lib = Libraries.Work_Library
         then
            Set_Identifier (Lib, Ident);
            Sem_Name.Add_Name (Lib);
            Set_Identifier (Lib, Std_Names.Name_Work);
         else
            Sem_Name.Add_Name (Lib);
         end if;
         Set_Visible_Flag (Lib, True);
      end if;
   end Sem_Library_Clause;

   procedure Sem_Use_Clause (Clause: Iir_Use_Clause)
   is
      Name: Iir;
      Prefix: Iir;
   begin
      --  LRM93 10.4
      --  A use clause achieves direct visibility of declarations that are
      --  visible by selection.
      --  Each selected name is a use clause identifies one or more
      --  declarations that will potentialy become directly visible.

      Name := Get_Selected_Name (Clause);
      case Get_Kind (Name) is
         when Iir_Kind_Selected_By_All_Name
           | Iir_Kind_Selected_Name =>
            Prefix := Get_Prefix (Name);
         when others =>
            Error_Msg_Sem ("use clause allows only selected name", Name);
            return;
      end case;

      Sem_Names (Prefix);
      Prefix := Get_Named_Entity (Prefix);
      if Prefix = Error_Mark then
         return;
      end if;

      case Get_Kind (Prefix) is
         when Iir_Kind_Library_Declaration =>
            null;
         when Iir_Kind_Design_Unit =>
            if Get_Kind (Get_Library_Unit (Prefix))
              /= Iir_Kind_Package_Declaration
            then
               Error_Msg_Sem ("design unit is not a package", Prefix);
               return;
            end if;
            Libraries.Load_Design_Unit_And_Dependences (Prefix, Clause);
         when others =>
            Error_Msg_Sem ("prefix must designate a package or a library",
                           Prefix);
            return;
      end case;

      case Get_Kind (Name) is
         when Iir_Kind_Selected_Name =>
            Sem_Names (Name);
            Name := Get_Named_Entity (Name);
            if Name = Error_Mark then
               return;
            end if;
            Sem_Name.Use_Selected_Name (Name);
         when Iir_Kind_Selected_By_All_Name =>
            Sem_Name.Use_All_Names (Prefix);
         when others =>
            raise Internal_Error;
      end case;
   end Sem_Use_Clause;

   procedure Sem_Context_Clauses (Design_Unit: Iir_Design_Unit) is
      Context_List: Iir_List;
      El: Iir;
   begin
      Context_List := Get_Context_Items (Design_Unit);
      for I in Natural loop
         El := Get_Nth_Element (Context_List, I);
         exit when El = Null_Iir;

         -- Handle declarations.
         case Get_Kind (El) is
            when Iir_Kind_Use_Clause =>
               Sem_Use_Clause (El);
            when Iir_Kind_Library_Clause =>
               Sem_Library_Clause (El);
            when others =>
               Error_Kind ("sem_context_clauses", El);
         end case;
      end loop;
   end Sem_Context_Clauses;

   -- Access to the current design unit.  This is set, saved, restored, cleared
   -- by procedure semantic.
   Current_Design_Unit: Iir_Design_Unit := Null_Iir;

   function Get_Current_Design_Unit return Iir_Design_Unit is
   begin
      return Current_Design_Unit;
   end Get_Current_Design_Unit;

   -- Semantic analysis of a libray unit.
   -- Also add a few node or change some nodes, when for exemple an
   -- identifier is changed into an access to the type.
   procedure Semantic (Design_Unit: Iir_Design_Unit)
   is
      El: Iir;
      Old_Design_Unit: Iir_Design_Unit;
      Implicit : Implicit_Signal_Declaration_Type;
   begin
      --  Sanity check: can analyze either previously analyzed unit or just
      --  parsed unit.
      case Get_Date (Design_Unit) is
         when Date_Parsed
           | Date_Valid =>
            null;
         when others =>
            raise Internal_Error;
      end case;

      -- Save and set current_design_unit.
      Old_Design_Unit := Current_Design_Unit;
      Current_Design_Unit := Design_Unit;
      Push_Declarative_Region_With_Signals (Implicit, null);

      --  Be sure the name table is empty.
      --  It is empty at start-up, or saved before recursing.
      pragma Debug (Name_Table.Assert_No_Infos);
      Open_Declarative_Region;

      -- Set_Dependence_List (Design_Unit,
--                            Create_Iir (Iir_Kind_Design_Unit_List));

      -- LRM93 11.2
      -- Every design unit is assumed to contain the following implicit
      -- context items as part of its context clause:
      --    library STD, WORK; use STD.STANDARD.all;
      Sem_Name.Add_Name (Libraries.Std_Library);
      Sem_Name.Add_Name (Libraries.Work_Library);
      Sem_Name.Use_All_Names (Std_Standard_Unit);
      Add_Dependence (Std_Standard_Unit);

      -- Semantic on context clauses.
      Sem_Context_Clauses (Design_Unit);

      -- semantic on the library unit.
      El := Get_Library_Unit (Design_Unit);
      case Get_Kind (El) is
         when Iir_Kind_Entity_Declaration =>
            Sem_Entity_Declaration (El);
         when Iir_Kind_Architecture_Declaration =>
            Sem_Architecture_Declaration (El);
         when Iir_Kind_Package_Declaration =>
            Sem_Package_Declaration (El);
         when Iir_Kind_Package_Body =>
            Sem_Package_Body (El);
         when Iir_Kind_Configuration_Declaration =>
            Sem_Configuration_Declaration (El);
         when others =>
            Error_Kind ("semantic", El);
      end case;

      Close_Declarative_Region;

      if Get_Date (Design_Unit) = Date_Parsed then
         Set_Date (Design_Unit, Date_Analyzed);
      end if;

      -- Restore current_design_unit.
      Current_Design_Unit := Old_Design_Unit;
      Pop_Declarative_Region_With_Signals (Implicit);
   end Semantic;
end Sem;
