with Ada.Text_IO;
with Types; use Types;
with Std_Package; use Std_Package;
with Errorout; use Errorout;
with Flags;
with Sem_Name; use Sem_Name;
with Sem;
with Libraries;
with Name_Table;
with Iirs_Utils; use Iirs_Utils;
with Std_Names;
with Evaluation; use Evaluation;

package body Sem_Expr is
   -- Semantize an expression (other than a range) with a possible overloading.
   -- Sem_expression_ov (and therefore sem_expression) must be called *once*
   -- for each expression node with A_TYPE1 not null and at most *once* with
   -- A_TYPE1 null.
   --
   -- When A_TYPE1 is null, sem_expression_ov find all possible types
   -- of the expression.  If there is only one possible type (ie, overloading
   -- is non-existant or solved), then the type of the expression is set,
   -- and the node is completly semantized.  Sem_expression_ov must not
   -- be called for such a node.
   -- If there is several possible types (ie overloaded), then the type is
   -- set with a list of overload.  To finishes the semantisation,
   -- sem_expression_ov must be called again with A_TYPE1 set to the
   -- expected type.
   --
   -- If A_TYPE1 is set, sem_expression_ov must finishes the semantisation
   -- of the expression, and set its type, which is not necessary a base type.
   -- A_TYPE1 must be a base type.
   --
   -- In case of error, it displays a message and return null.
   -- In case of success, it returns the semantized expression, which can
   -- be different from EXPR (eg, a character literal is transformed into an
   -- enumeration literal).
   function Sem_Expression_Ov (Expr: Iir; A_Type1: Iir) return Iir;

   -- For a procedure call, A_TYPE must be null.
   function Sem_Subprogram_Call (Expr: Iir; A_Type: Iir)
     return Iir;

   --  If EXPR is a node for an expression, then return EXPR.
   --  Otherwise, emit an error message and return NULL_IIR.
   --  If EXPR is NULL_IIR, NULL_IIR is silently returned.
   function Check_Is_Expression (Expr : Iir) return Iir;

   --  Convert name EXPR to an expression (ie, can create function call).
   --  A_TYPE is the expected type of the expression.
   --  FIXME: it is unclear wether the result must be an expression or not
   --  (ie, it *must* have a type, but may be a range).
   function Name_To_Expression (Name : Iir; A_Type : Iir) return Iir;

   procedure Undeclared (Expr: Iir) is
   begin
      Error_Msg_Sem
        ("identifier `" & Iirs_Utils.Get_String (Expr) & "' not declared",
         Expr);
   end Undeclared;

   procedure Not_Match (Expr: Iir; A_Type: Iir) is
   begin
      Error_Msg_Sem ("can't match " & Disp_Node (Expr) & " with type "
                     & Disp_Node (A_Type), Expr);
   end Not_Match;

   procedure Not_Match (Expr: Iir; Type1: Iir; Type2: Iir) is
   begin
      Error_Msg_Sem
        ("can't match '" & Disp_Node (Expr) & "' with type '"
         & Disp_Node (Type1) & "' or type '" & Disp_Node (Type2) & "'",
         Expr);
   end Not_Match;

   procedure Overloaded (Expr: Iir) is
   begin
      Error_Msg_Sem
        ("cant resolve overloaded identifier '" & Get_String (Expr) & "'",
         Expr);
   end Overloaded;

   procedure Error_Not_Compatible (Expr: Iir) is
   begin
      Error_Msg_Sem ("left and right types of operator are not compatible",
                     Expr);
   end Error_Not_Compatible;

   procedure Error_Overload (Expr: Iir) is
   begin
      Error_Msg_Sem ("can't resolve overload for " & Disp_Node (Expr), Expr);
   end Error_Overload;

   -- Create an overload list.
   -- must be destroyed with free_iir.
   function Get_Overload_List return Iir_Overload_List is
   begin
      return Create_Iir (Iir_Kind_Overload_List);
   end Get_Overload_List;

   function Simplify_Overload_List (List : Iir) return Iir
   is
      Res : Iir;
   begin
      case Get_Nbr_Elements (List) is
         when 0 =>
            Free_Old_Iir (List);
            return Null_Iir;
         when 1 =>
            Res := Get_First_Element (List);
            Free_Old_Iir (List);
            return Res;
         when others =>
            return List;
      end case;
   end Simplify_Overload_List;

   -- Return true if AN_IIR is an overload list.
   function Is_Overload_List (An_Iir: Iir) return Boolean is
   begin
      return Get_Kind (An_Iir) = Iir_Kind_Overload_List;
   end Is_Overload_List;

   -- Replace type of TARGET by A_TYPE.
   -- If TARGET has already a type, it must be an overload list, and in this
   -- case, this list is freed, or it must be A_TYPE.
   -- A_TYPE can't be an overload list.
   --
   -- This procedure can be called in the second pass, when the type is known.
   procedure Replace_Type (Target: Iir; A_Type: Iir) is
      Old_Type: Iir;
   begin
      Old_Type := Get_Type (Target);
      if Old_Type /= null then
         if Is_Overload_List (Old_Type) then
            Free_Iir (Old_Type);
         elsif Old_Type = A_Type then
            return;
         else
            -- Cannot replace a type.
            raise Internal_Error;
         end if;
      end if;
      if A_Type = null then
         return;
      end if;
      if Is_Overload_List (A_Type) then
         raise Internal_Error;
      end if;
      Set_Type (Target, A_Type);
   end Replace_Type;

   -- Return true if ID is overloaded, ie has several meanings.
   function Is_Overloaded (Id: Iir) return Boolean is
   begin
      return Is_Overload_List (Get_Type (Id));
   end Is_Overloaded;

   -- Return the common type of base types LEFT and RIGHT.
   -- LEFT are RIGHT must be really base types (not subtypes).
   -- Roughly speaking, it returns LEFT (= RIGHT) if LEFT = RIGHT (ie, same
   -- type), null otherwise.
   -- However, it handles implicite conversions of universal types.
   function Get_Common_Basetype (Left: Iir; Right: Iir)
     return Iir is
   begin
      if Left = Right then
         return Left;
      end if;
      case Get_Kind (Left) is
         when Iir_Kind_Integer_Type_Definition =>
            if Right = Universal_Integer_Type_Definition then
               return Left;
            elsif Left = Universal_Integer_Type_Definition
              and then Get_Kind (Right) = Iir_Kind_Integer_Type_Definition
            then
               return Right;
            end if;
         when Iir_Kind_Floating_Type_Definition =>
            if Right = Universal_Real_Type_Definition then
               return Left;
            elsif Left = Universal_Real_Type_Definition
              and then Get_Kind (Right) = Iir_Kind_Floating_Type_Definition
            then
               return Right;
            end if;
         when others =>
            null;
      end case;
      return Null_Iir;
   end Get_Common_Basetype;

   -- LEFT are RIGHT must be really a type (not a subtype).
   function Are_Basetypes_Compatible (Left: Iir; Right: Iir)
     return Boolean is
   begin
      return Get_Common_Basetype (Left, Right) /= null;
   end Are_Basetypes_Compatible;

   function Are_Types_Compatible (Left: Iir; Right: Iir)
     return Boolean is
   begin
      return Get_Common_Basetype (Get_Base_Type (Left),
                                  Get_Base_Type (Right)) /= null;
   end Are_Types_Compatible;

   function Are_Nodes_Compatible (Left: Iir; Right: Iir)
     return Boolean is
   begin
      return Are_Types_Compatible (Get_Type (Left), Get_Type (Right));
   end Are_Nodes_Compatible;

   --  Find all named declaration whose identifier is ID in DECL_LIST and
   --  return it.
   --  The result can be NULL (if no such declaration exist),
   --  a declaration, or an overload_list containing all declarations.
   function Find_Declarations_In_List (Decl: Iir; Name : Iir_Selected_Name)
     return Iir
   is
      Res: Iir_List := null;

      --  If indentifier of DECL is ID, then add DECL in the result.
      procedure Handle_Decl (Decl : Iir; Id : Name_Id)
      is
         Nres : Iir;
      begin
         --  Use_clauses may appear in a declaration list.
         case Get_Kind (Decl) is
            when Iir_Kind_Use_Clause =>
               return;
            when others =>
               null;
         end case;
         if Get_Identifier (Decl) /= Id then
            return;
         end if;
         if Res = null then
            Res := Decl;
         elsif Is_Overload_List (Res) then
            Append_Element (Res, Decl);
         else
            Nres := Get_Overload_List;
            Append_Element (Nres, Res);
            Append_Element (Nres, Decl);
            Res := Nres;
         end if;
      end Handle_Decl;

      procedure Iterator_Decl is new Sem_Name.Iterator_Decl
        (Arg_Type => Name_Id, Handle_Decl => Handle_Decl);
      procedure Iterator_Decl_List is new Sem_Name.Iterator_Decl_List
        (Arg_Type => Name_Id, Handle_Decl => Iterator_Decl);

      Id : Name_Id;
   begin
      Id := Get_Identifier (Name);
      case Get_Kind (Decl) is
         when Iir_Kind_Function_Spec_Body
           | Iir_Kind_Function_Specification
           | Iir_Kind_Procedure_Specification
           | Iir_Kind_Procedure_Spec_Body =>
            Iterator_Decl_List (Get_Interface_Declaration (Decl), Id);
         when Iir_Kind_Entity_Declaration =>
            Iterator_Decl_List (Get_Generic_List (Decl), Id);
            Iterator_Decl_List (Get_Port_List (Decl), Id);
         when Iir_Kind_Architecture_Declaration =>
            null;
         when Iir_Kind_Generate_Statement =>
            null;
         when Iir_Kind_Package_Declaration =>
            null;
         when Iir_Kind_Block_Statement =>
            declare
               Header : Iir;
            begin
               Header := Get_Block_Header (Decl);
               if Header /= Null_Iir then
                  Iterator_Decl_List (Get_Generic_List (Header), Id);
                  Iterator_Decl_List (Get_Port_List (Header), Id);
               end if;
            end;
         when Iir_Kind_For_Loop_Statement =>
            Handle_Decl (Get_Iterator_Scheme (Decl), Id);
         when Iir_Kind_Process_Statement
           | Iir_Kind_Sensitized_Process_Statement =>
            null;
         when others =>
            Error_Kind ("find_declarations_in_list", Decl);
      end case;

      case Get_Kind (Decl) is
         when Iir_Kind_Function_Spec_Body
           | Iir_Kind_Function_Specification
           | Iir_Kind_Procedure_Specification
           | Iir_Kind_Procedure_Spec_Body =>
            declare
               Bod : Iir := Iirs_Utils.Get_Subprogram_Body (Decl);
            begin
               Iterator_Decl_List (Get_Subprogram_Declaration (Decl), Id);
               Iterator_Decl_List (Get_Subprogram_Statements (Decl), Id);
            end;
         when Iir_Kind_Architecture_Declaration
           | Iir_Kind_Entity_Declaration
           | Iir_Kind_Generate_Statement
           | Iir_Kind_Block_Statement =>
            Iterator_Decl_List (Get_Declaration_List (Decl), Id);
            Iterator_Decl_List (Get_Concurrent_Statement_List (Decl), Id);
         when Iir_Kind_Package_Declaration =>
            Iterator_Decl_List (Get_Declaration_List (Decl), Id);
         when Iir_Kind_Process_Statement
           | Iir_Kind_Sensitized_Process_Statement =>
            Iterator_Decl_List (Get_Declaration_List (Decl), Id);
            Iterator_Decl_List (Get_Process_Statement_List (Decl), Id);
         when Iir_Kind_For_Loop_Statement =>
            null;
         when others =>
            Error_Kind ("find_declarations_in_list", Decl);
      end case;
      --if Res = Null_Iir then
      --   Error_Msg_Sem ("""" & Name_Table.Image (Id) & """ not defined in "
      --                  & Disp_Node (Decl), Name);
      --end if;
      return Res;
   end Find_Declarations_In_List;

   -- Find a uniq declaration for a name.
   function Find_Declaration (Name: Iir; Kind: Decl_Kind_Type)
     return Iir
   is
      procedure Check_Visible (Res : Iir) is
      begin
         if not Get_Visible_Flag (Res) then
            Error_Msg_Sem (Disp_Node (Res)  & " defined at "
                           & Disp_Location (Res) & " is not yet visible",
                           Name);
         end if;
      end Check_Visible;

      procedure Error (Res : Iir; Str : String)
      is
      begin
         Error_Msg_Sem (Str & " expected, found " & Disp_Node (Res), Name);
      end Error;

      function Check_Kind (Res: Iir; Kind : Iir_Kind; Str: String)
        return Iir
      is
         Res_Kind : Iir_Kind;
      begin
         Res_Kind := Get_Kind (Res);
         if Res_Kind /= Kind then
            Error (Res, Str);
            return Null_Iir;
         else
            Check_Visible (Res);
            return Res;
         end if;
      end Check_Kind;

      function Check_Kind_Unit (Res: Iir; Kind : Iir_Kind; Str: String)
        return Iir
      is
         Res_Kind : Iir_Kind;
      begin
         if Get_Kind (Res) /= Iir_Kind_Design_Unit then
            Error (Res, Str);
            return Null_Iir;
         end if;

         Res_Kind := Get_Kind (Get_Library_Unit (Res));
         if Res_Kind /= Kind then
            Error (Res, Str);
            return Null_Iir;
         else
            Check_Visible (Res);
            return Res;
         end if;
      end Check_Kind_Unit;

      Res: Iir;
   begin
      Sem_Names (Name);
      Res := Get_Named_Entity (Name);
      if Res = Error_Mark then
         --  A message must have been displayed.
         --  FIXME: is it the case for find_selected_declarations ???
         --  Error_Msg_Sem ("identifier is not defined", Name);
         return Null_Iir;
      end if;

      case Kind is
         when Decl_Type
           | Decl_Incomplete_Type =>
            case Get_Kind (Res) is
               when Iir_Kind_Type_Declaration
                 | Iir_Kind_Subtype_Declaration =>
                  Check_Visible (Res);
                  Res := Get_Type (Res);
                  --  Note: RES cannot be NULL_IIR, this is just to be more
                  --  bullet-proof.
                  if Kind /= Decl_Incomplete_Type
                    and then
                    (Res = Null_Iir or else
                     Get_Kind (Res) = Iir_Kind_Incomplete_Type_Definition)
                  then
                     Error_Msg_Sem
                       ("invalid use of an incomplete type definition", Name);
                  end if;
               when others =>
                  Error_Msg_Sem
                    ("type expected, found " & Disp_Node (Res), Name);
                  return Null_Iir;
            end case;
         when Decl_Object =>
            Check_Visible (Res);
         when Decl_Interface =>
            Check_Visible (Res);
         when Decl_Component =>
            Res := Check_Kind (Res, Iir_Kind_Component_Declaration,
                               "component");
         when Decl_Physical_Unit =>
            Check_Visible (Res);
         when Decl_Label =>
            Check_Visible (Res);
         when Decl_Entity =>
            Res := Check_Kind_Unit
              (Res, Iir_Kind_Entity_Declaration, "entity");
         when Decl_Configuration =>
            Res := Check_Kind_Unit (Res, Iir_Kind_Configuration_Declaration,
                                    "configuration");
         when Decl_Group_Template =>
            Res := Check_Kind (Res, Iir_Kind_Group_Template_Declaration,
                               "group template");
         when Decl_Attribute =>
            Res := Check_Kind (Res, Iir_Kind_Attribute_Declaration,
                               "attribute");
         when Decl_Any =>
            Check_Visible (Res);
      end case;
      return Res;
   end Find_Declaration;

   --  Remove procedure specification from LIST.
   procedure Remove_Procedures_From_List (List : Iir_Overload_List)
   is
      El : Iir;
      P : Natural;
   begin
      for I in Natural loop
         El := Get_Nth_Element (List, I);
         if El = Null_Iir then
            return;
         end if;
         if Get_Kind (El) in Iir_Kinds_Procedure_Specification then
            P := I;
            exit;
         end if;
      end loop;
      for I in P + 1 .. Natural'Last loop
         El := Get_Nth_Element (List, I);
         exit when El = Null_Iir;
         if Get_Kind (El) not in Iir_Kinds_Procedure_Specification then
            Replace_Nth_Element (List, P, El);
            P := P + 1;
         end if;
      end loop;
      Set_Nbr_Elements (List, P);
   end Remove_Procedures_From_List;

   --  From the list LIST of function or enumeration literal, extract the
   --  list of (return) types.
   --  If there is only one type, return it.
   --  If there is no types, return NULL.
   --  Otherwise, return the list as an overload list.
   function Create_List_Of_Types (List : Iir_Overload_List)
     return Iir
   is
      Res : Iir_Overload_List;
      Decl : Iir;
   begin
      --  Create the list of possible return types.
      Res := Get_Overload_List;
      for I in Natural loop
         Decl := Get_Nth_Element (List, I);
         exit when Decl = Null_Iir;
         case Get_Kind (Decl) is
            when Iir_Kinds_Function_Specification =>
               Add_Element (Res, Get_Return_Type (Decl));
            when Iir_Kind_Enumeration_Literal
              | Iir_Kind_Function_Call =>
               Add_Element (Res, Get_Type (Decl));
            when others =>
               Error_Kind ("create_list_of_types", Decl);
         end case;
      end loop;
      return Simplify_Overload_List (Res);
   end Create_List_Of_Types;

   --  Create an implicit_dereference node if PREFIX is of type access.
   --  Return PREFIX otherwise.
   --  PARENT is used if an implicit dereference node is created, to copy
   --  location from.
   function Insert_Implicit_Dereference (Prefix : Iir; Parent : Iir)
     return Iir
   is
      Prefix_Type : Iir;
      Res : Iir_Implicit_Dereference;
   begin
      Prefix_Type := Get_Type (Prefix);

      case Get_Kind (Prefix_Type) is
         when Iir_Kind_Access_Type_Definition
           | Iir_Kind_Access_Subtype_Definition =>
            null;
         when others =>
            return Prefix;
      end case;
      Check_Read (Prefix);
      Res := Create_Iir (Iir_Kind_Implicit_Dereference);
      Location_Copy (Res, Parent);
      Set_Type (Res, Get_Designated_Type (Prefix_Type));
      Set_Prefix (Res, Prefix);
      Set_Base_Name (Res, Res);
      return Res;
   end Insert_Implicit_Dereference;

   procedure Sem_Name_Free_Result (List : in out Iir; Keep : Iir)
   is
      procedure Sem_Name_Free (El : in out Iir) is
      begin
         case Get_Kind (El) is
            when Iir_Kind_Function_Call
              | Iir_Kind_Indexed_Name =>
               Free_Iir (El);
            when Iir_Kinds_Function_Specification
              | Iir_Kinds_Procedure_Specification
              | Iir_Kind_Enumeration_Literal =>
               null;
            when others =>
               Error_Kind ("sem_name_free", El);
         end case;
      end Sem_Name_Free;

      El : Iir;
   begin
      if List = Null_Iir then
         return;
      elsif not Is_Overload_List (List) then
         if List /= Keep then
            Sem_Name_Free (List);
         end if;
      else
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            if El /= Keep then
               Sem_Name_Free (El);
            end if;
         end loop;
         Free_Iir (List);
      end if;
   end Sem_Name_Free_Result;

   --  LRM93 6.2
   --  The evaluation of a simple name has no other effect than to determine
   --  the named entity denoted by the name.
   --
   --  NAME may be a string literal too.
   --  GHDL: set interpretation of NAME (possibly an overload list).
   procedure Sem_Simple_Name (Name : Iir)
   is
      use Sem_Name;
      Interpretation: Name_Interpretation_Type;
      Res: Iir;
   begin
      Interpretation := Get_Interpretation (Get_Identifier (Name));

      if Interpretation = No_Name_Interpretation then
         Error_Msg_Sem ("simple name """ & Iirs_Utils.Get_String (Name)
                        & """ not declared", Name);
         Res := Error_Mark;
      elsif Get_Next_Interpretation (Interpretation) = No_Name_Interpretation
      then
         -- not overloaded.
         Res := Get_Declaration (Interpretation);
         if not Get_Visible_Flag (Res) then
            Error_Msg_Sem (Disp_Node (Res) & " is not visible here", Name);
         end if;
      else
         Res := Get_Overload_List;
         while Interpretation /= No_Name_Interpretation loop
            Append_Element (Res, Get_Declaration (Interpretation));
            Interpretation := Get_Next_Interpretation (Interpretation);
         end loop;
      end if;
      Set_Named_Entity (Name, Res);
   end Sem_Simple_Name;

   --  Add new interpretation DECL to RES.
   --  Create an overload_list if necessary.
   procedure Add_Result (Res : in out Iir; Decl : Iir)
   is
      Nres : Iir;
   begin
      if Decl = Null_Iir then
         return;
      end if;
      if Res = Null_Iir then
         Res := Decl;
      elsif Is_Overload_List (Res) then
         Append_Element (Res, Decl);
      else
         Nres := Get_Overload_List;
         Append_Element (Nres, Res);
         Append_Element (Nres, Decl);
         Res := Nres;
      end if;
   end Add_Result;

   --  Move elements of result list LIST to result list RES.
   --  Destroy LIST if necessary.
   procedure Add_Result_List (Res : in out Iir; List : in out Iir)
   is
      El : Iir;
   begin
      if Res = Null_Iir then
         Res := List;
      elsif List = Null_Iir then
         null;
      elsif not Is_Overload_List (List) then
         Add_Result (Res, List);
      else
         if not Is_Overload_List (Res) then
            El := Res;
            Res := Get_Overload_List;
            Append_Element (Res, El);
         end if;
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = Null_Iir;
            Append_Element (Res, El);
         end loop;
         Free_Iir (List);
      end if;
   end Add_Result_List;

   --  If PREFIX is a function specification that cannot be converted to a
   --  function call (because of lack of association), return FALSE.
   function Maybe_Function_Call (Prefix : Iir) return Boolean
   is
      Inter_List : Iir_List;
      Inter : Iir;
   begin
      if Get_Kind (Prefix) not in Iir_Kinds_Function_Specification then
         return True;
      end if;
      Inter_List := Get_Interface_Declaration (Prefix);
      for I in Natural loop
         Inter := Get_Nth_Element (Inter_List, I);
         exit when Inter = Null_Iir;
         if Get_Default_Value (Inter) = Null_Iir then
            return False;
         end if;
      end loop;
      return True;
   end Maybe_Function_Call;

   --  SUB_NAME must be a function specification.
   function Sem_As_Function_Call (Func : Iir; Assoc_List : Iir_List; Loc : Iir)
                                 return Iir_Function_Call
   is
      Call : Iir_Function_Call;
   begin
      Call := Create_Iir (Iir_Kind_Function_Call);
      Location_Copy (Call, Loc);
      Set_Implementation (Call, Func);
      Set_Parameter_Association_List (Call, Assoc_List);
      Set_Type (Call, Get_Return_Type (Func));
      Set_Base_Name (Call, Call);
      return Call;
   end Sem_As_Function_Call;

   --  If PREFIX is a function specification, then return a function call,
   --  else return PREFIX.
   function Maybe_Insert_Function_Call (Prefix : Iir; Loc : Iir) return Iir
   is
   begin
      if Get_Kind (Prefix) in Iir_Kinds_Function_Specification then
         return Sem_As_Function_Call (Prefix, Null_Iir_List, Loc);
      else
         return Prefix;
      end if;
   end Maybe_Insert_Function_Call;

   --  If PTR_TYPE is not NULL_IIR, then return an implciti dereference to
   --  PREFIX, else return PREFIX.
   function Maybe_Insert_Dereference (Prefix : Iir; Ptr_Type : Iir) return Iir
   is
      Id : Iir;
   begin
      if Ptr_Type /= Null_Iir then
         Id := Create_Iir (Iir_Kind_Implicit_Dereference);
         Location_Copy (Id, Prefix);
         Set_Type (Id, Get_Designated_Type (Ptr_Type));
         Set_Prefix (Id, Prefix);
         Set_Base_Name (Id, Id);
         return Id;
      else
         return Prefix;
      end if;
   end Maybe_Insert_Dereference;

   procedure Finish_Sem_Indexed_Name (Expr : Iir)
   is
      Prefix : Iir;
      Prefix_Type : Iir;
      Index_Subtype : Iir;
      Index_List : Iir_List;
      Index : Iir;
      Expr_Staticness : Iir_Staticness;
   begin
      Prefix := Get_Prefix (Expr);
      Prefix_Type := Get_Type (Prefix);
      Expr_Staticness := Locally;

      Index_List := Get_Index_List (Expr);
      -- LRM93 6.4: there must be one such expression for each index
      -- position of the array and each expression must be of the
      -- type of the corresponding index.
      -- Loop on the indexes.
      for I in Natural loop
         Index_Subtype :=
           Get_Nth_Element (Get_Index_List (Prefix_Type), I);
         exit when Index_Subtype = Null_Iir;
         Index := Get_Nth_Element (Index_List, I);
         -- The index_subtype can be an unconstrained index type.
         Index := Check_Is_Expression (Index);
         if Index /= Null_Iir then
            Index := Sem_Expression (Index, Get_Base_Type (Index_Subtype));
         end if;
         if Index /= Null_Iir then
            if Get_Expr_Staticness (Index) = Locally
              and then Get_Type_Staticness (Index_Subtype) = Locally
            then
               Index := Eval_Expr_Check (Index, Index_Subtype);
            end if;
            Replace_Nth_Element (Get_Index_List (Expr), I, Index);
            Expr_Staticness := Min (Expr_Staticness,
                                    Get_Expr_Staticness (Index));
         end if;
      end loop;

      Set_Type (Expr, Get_Element_Subtype (Prefix_Type));

      --  An indexed name cannot be locally static.
      Set_Expr_Staticness
        (Expr, Min (Globally, Min (Expr_Staticness,
                                   Get_Expr_Staticness (Prefix))));

      -- LRM93 6.1:
      -- a name is said to be a static name iff:
      -- The name is an indexed name whose prefix is a static name
      -- and every expression that appears as part of the name is a
      -- static expression.
      --
      -- a name is said to be a locally static name iif:
      -- The name is an indexed name whose prefix is a locally
      -- static name and every expression that appears as part
      -- of the name is a locally static expression.
      Set_Name_Staticness (Expr, Min (Expr_Staticness,
                                      Get_Name_Staticness (Prefix)));

      Set_Base_Name (Expr, Get_Base_Name (Prefix));
   end Finish_Sem_Indexed_Name;

   procedure Finish_Sem_Dereference (Res : Iir)
   is
   begin
      Set_Base_Name (Res, Res);
      Check_Read (Get_Prefix (Res));
      Set_Expr_Staticness (Res, None);
      Set_Name_Staticness (Res, None);
   end Finish_Sem_Dereference;

   procedure Finish_Sem_Slice_Name (Name : Iir_Slice_Name)
   is
      -- The prefix of the slice
      Prefix: Iir;
      Prefix_Type: Iir;
      Prefix_Bt : Iir;
      Index_List: Iir;
      Index_Type: Iir;
      Index_Range : Iir;
      Suffix: Iir;
      Slice_Type : Iir;
      Expr_Type : Iir;
      Staticness : Iir_Staticness;
      Suffix_Rng : Iir;
      Prefix_Rng : Iir;
   begin
      -- Set a type to the prefix.
      Prefix := Get_Prefix (Name);
      Prefix_Type := Get_Type (Prefix);
      Set_Base_Name (Name, Get_Base_Name (Prefix));

      -- LRM93 6.5: the prefix of an indexed name must be appropriate
      -- for an array type.
      Prefix_Bt := Get_Base_Type (Prefix_Type);
      if Get_Kind (Prefix_Bt) /= Iir_Kind_Array_Type_Definition then
         Error_Msg_Sem ("slice can only be applied to an array", Name);
         return;
      end if;

      -- LRM93 6.5:
      -- The prefix of a slice must be appropriate for a
      -- one-dimensionnal array object.
      Index_List := Get_Index_List (Prefix_Type);
      if Get_Nbr_Elements (Index_List) /= 1 then
         Error_Msg_Sem ("slice prefix must be an unidimensional array", Name);
         return;
      end if;

      Index_Type := Get_First_Element (Index_List);
      Index_Range := Get_Range_Constraint (Index_Type);
      Prefix_Rng := Eval_Range (Index_Type);

      --  LRM93 6.5
      --  It is an error if either the bounds of the discrete range does not
      --  belong to the index range of the prefixing array, *unless* the slice
      --  is a null slice.
      --
      --  LRM93 6.5
      --  The slice is a null slice if the discrete range is a null range.

      -- LRM93 6.5:
      -- The bounds of the discrete range [...] must be of the
      -- type of the index of the array.
      Suffix := Sem_Discrete_Range_Expression
        (Get_Suffix (Name), Index_Type);
      if Suffix = Null_Iir then
         return;
      end if;
      Set_Suffix (Name, Suffix);

      -- LRM93 6.5:
      -- It is an error if the direction of the discrete range is not
      -- the same as that of the index range of the array denoted
      -- by the prefix of the slice name.

      -- Check this only if the type is a constrained type.
      Suffix_Rng := Eval_Range (Suffix);
      if Get_Kind (Prefix_Type) = Iir_Kind_Array_Subtype_Definition
        and then Prefix_Rng /= Null_Iir
        and then Suffix_Rng /= Null_Iir
        and then Get_Direction (Suffix_Rng) /= Get_Direction (Prefix_Rng)
      then
         if False and then Flags.Vhdl_Std = Vhdl_87 then
            -- emit a warning for a null slice.
            Warning_Msg_Sem
              ("direction mismatch results in a null slice", Name);
         end if;
         Error_Msg_Sem ("direction of the range mismatch", Name);
      end if;

      --  LRM93 7.4.1
      --  A slice is never a locally static expression.
      Staticness := Get_Discrete_Range_Staticness (Suffix);
      Set_Expr_Staticness
        (Name, Min (Min (Staticness, Get_Expr_Staticness (Prefix)), Globally));
      Set_Name_Staticness
        (Name, Min (Staticness, Get_Name_Staticness (Prefix)));

      --  The type of the slice is a subtype of the base type whose
      --  range contraint is the slice itself.
      if Get_Kind (Suffix) in Iir_Kinds_Discrete_Type_Definition then
         Slice_Type := Suffix;
      else
         case Get_Kind (Get_Base_Type (Index_Type)) is
            when Iir_Kind_Integer_Type_Definition =>
               Slice_Type := Create_Iir (Iir_Kind_Integer_Subtype_Definition);
            when Iir_Kind_Enumeration_Type_Definition =>
               Slice_Type :=
                 Create_Iir (Iir_Kind_Enumeration_Subtype_Definition);
            when others =>
               Error_Kind ("sem_expr: slice_name", Get_Base_Type (Index_Type));
         end case;
         Set_Range_Constraint (Slice_Type, Suffix);
         Set_Type_Staticness (Slice_Type, Staticness);
         Set_Base_Type (Slice_Type, Get_Base_Type (Index_Type));
         Set_Location (Slice_Type, Get_Location (Suffix));
      end if;

      Expr_Type := Create_Iir (Iir_Kind_Array_Subtype_Definition);
      Set_Location (Expr_Type, Get_Location (Suffix));
      Set_Base_Type (Expr_Type, Get_Base_Type (Get_Type (Prefix)));
      Append_Element (Get_Index_List (Expr_Type), Slice_Type);
      Set_Element_Subtype (Expr_Type, Get_Element_Subtype (Prefix_Type));
      if Get_Kind (Prefix_Type) /= Iir_Kind_Array_Type_Definition then
         Set_Resolution_Function
           (Expr_Type, Get_Resolution_Function (Prefix_Type));
      else
         Set_Resolution_Function (Expr_Type, null);
      end if;
      Set_Type_Staticness
        (Expr_Type, Min (Get_Type_Staticness (Prefix_Type),
                         Get_Type_Staticness (Slice_Type)));
      Set_Type (Name, Expr_Type);
   end Finish_Sem_Slice_Name;

   procedure Finish_Sem_Function_Call (Call : Iir)
   is
      Rtype : Iir;
   begin
      Set_Name_Staticness (Call, None);
      --  FIXME: modify sem_subprogram_call to avoid such a type swap.
      Rtype := Get_Type (Call);
      Set_Type (Call, Null_Iir);
      if Sem_Subprogram_Call (Call, Null_Iir) = Null_Iir then
         Set_Type (Call, Rtype);
      end if;
   end Finish_Sem_Function_Call;

   procedure Finish_Sem_Array_Attribute (Attr : Iir; Param : Iir)
   is
      Parameter : Iir;
      Prefix_Type : Iir;
      Index_Type : Iir;
      Prefix : Iir;
   begin
      --  LRM93 14.1
      --  Parameter: A locally static expression of type universal_integer, the
      --  value of which must not exceed the dimensionality of A.  If omitted,
      --  it defaults to 1.
      if Param = Null_Iir then
         Parameter := Universal_Integer_One;
      else
         Parameter := Sem_Expression
           (Param, Universal_Integer_Type_Definition);
         if Parameter = Null_Iir then
            Parameter := Universal_Integer_One;
         else
            if Get_Expr_Staticness (Parameter) /= Locally then
               Error_Msg_Sem ("parameter must be locally static", Parameter);
               Parameter := Universal_Integer_One;
            end if;
         end if;
      end if;
      Prefix := Get_Prefix (Attr);
      Prefix_Type := Get_Type (Prefix);
      declare
         Dim : Iir_Int64;
         Indexes_List : Iir_List;
      begin
         Indexes_List := Get_Index_List (Prefix_Type);
         Dim := Get_Value (Parameter);
         if Dim < 1 or else Dim > Iir_Int64 (Get_Nbr_Elements (Indexes_List))
         then
            Error_Msg_Sem ("parameter value out of bound", Attr);
            Parameter := Universal_Integer_One;
         end if;
         Index_Type := Get_Nth_Element (Indexes_List, Natural (Dim - 1));
      end;

      case Get_Kind (Attr) is
         when Iir_Kind_Left_Array_Attribute
           | Iir_Kind_Right_Array_Attribute
           | Iir_Kind_High_Array_Attribute
           | Iir_Kind_Low_Array_Attribute =>
            Set_Type (Attr, Index_Type);
         when Iir_Kind_Range_Array_Attribute
           | Iir_Kind_Reverse_Range_Array_Attribute =>
            Set_Type (Attr, Index_Type);
         when Iir_Kind_Length_Array_Attribute =>
            Set_Type (Attr, Universal_Integer_Type_Definition);
         when Iir_Kind_Ascending_Array_Attribute =>
            Set_Type (Attr, Boolean_Type_Definition);
         when others =>
            raise Internal_Error;
      end case;

      Set_Parameter (Attr, Parameter);

      Set_Expr_Staticness (Attr, Get_Type_Staticness (Prefix_Type));
   end Finish_Sem_Array_Attribute;

   procedure Finish_Sem_Scalar_Type_Attribute (Attr : Iir; Param : Iir)
   is
      Prefix_Type : Iir;
      Prefix_Bt : Iir;
      Parameter : Iir;
   begin
      if Param = Null_Iir then
         Error_Msg_Sem (Disp_Node (Attr) & " requires a parameter", Attr);
         return;
      end if;

      Prefix_Type := Get_Type (Get_Prefix (Attr));
      Prefix_Bt := Get_Base_Type (Prefix_Type);

      case Get_Kind (Attr) is
         when Iir_Kind_Pos_Attribute =>
            --  LRM93 14.1
            --  Parameter: An expression whose type is the base type of T.
            Parameter := Sem_Expression (Param, Prefix_Bt);
         when Iir_Kind_Val_Attribute =>
            --  LRM93 14.1
            --  Parameter: An expression of any integer type.
            Parameter := Sem_Expression (Param, null);
            --  FIXME: overload.
            if Get_Kind (Get_Base_Type (Get_Type (Parameter)))
              /= Iir_Kind_Integer_Type_Definition
            then
               Error_Msg_Sem ("parameter must be an integer", Attr);
               return;
            end if;
         when Iir_Kind_Succ_Attribute
           | Iir_Kind_Pred_Attribute
           | Iir_Kind_Leftof_Attribute
           | Iir_Kind_Rightof_Attribute =>
            --  LRM93 14.1
            --  Parameter: An expression whose type is the base type of T.
            Parameter := Sem_Expression (Param, Prefix_Bt);
         when Iir_Kind_Image_Attribute =>
            --  LRM93 14.1
            --  Parameter: An expression whose type is the base type of T.
            Parameter := Sem_Expression (Param, Prefix_Bt);
         when Iir_Kind_Value_Attribute =>
            --  Parameter: An expression of type string.
            Parameter := Sem_Expression (Param, String_Type_Definition);
         when others =>
            raise Internal_Error;
      end case;
      if Parameter = Null_Iir then
         return;
      end if;
      Set_Parameter (Attr, Parameter);
      Set_Expr_Staticness (Attr, Min (Get_Type_Staticness (Prefix_Type),
                                      Get_Expr_Staticness (Parameter)));
   end Finish_Sem_Scalar_Type_Attribute;

   procedure Finish_Sem_Signal_Attribute (Attr : Iir; Parameter : Iir)
   is
      Param : Iir;
   begin
      if Parameter = Null_Iir then
         return;
      end if;
      if Get_Kind (Attr)= Iir_Kind_Transaction_Attribute then
         Error_Msg_Sem ("'transaction does not allow a parameter", Attr);
      else
         Param := Sem_Expression (Parameter, Time_Subtype_Definition);
         Set_Parameter (Attr, Param);
      end if;
   end Finish_Sem_Signal_Attribute;

   function Is_Type_Abstract_Numeric (Atype : Iir) return Boolean is
   begin
      case Get_Kind (Atype) is
         when Iir_Kind_Integer_Subtype_Definition
           | Iir_Kind_Integer_Type_Definition
           | Iir_Kind_Floating_Subtype_Definition
           | Iir_Kind_Floating_Type_Definition =>
            return True;
         when others =>
            return False;
      end case;
   end Is_Type_Abstract_Numeric;

   function Are_Types_Closely_Related (Type1, Type2 : Iir) return Boolean
   is
      Ant1, Ant2 : Boolean;
      Index_List1, Index_List2 : Iir_List;
      El1, El2 : Iir;
   begin
      -- LRM93 7.3.5
      -- In particular, a type is closely related to itself.
      if Get_Base_Type (Type1) = Get_Base_Type (Type2) then
         return True;
      end if;

      --  LRM93 7.3.5
      --  a) Abstract Numeric Types: Any abstract numeric type is closely
      --     related to any other abstract numeric type.
      Ant1 := Is_Type_Abstract_Numeric (Type1);
      Ant2 := Is_Type_Abstract_Numeric (Type2);
      if Ant1 and Ant2 then
         return True;
      end if;
      if Ant1 or Ant2 then
         return False;
      end if;

      --  LRM93 7.3.5
      -- b) Array Types: Two array types are closely related if and only if
      --    The types have the same dimensionality; For each index position,
      --    the index types are either the same or are closely related; and
      --    The element types are the same.
      --
      -- No other types are closely related.
      if not (Get_Kind (Type1) in Iir_Kinds_Array_Type_Definition
              and then Get_Kind (Type2) in Iir_Kinds_Array_Type_Definition)
      then
         return False;
      end if;
      Index_List1 := Get_Index_List (Type1);
      Index_List2 := Get_Index_List (Type2);
      if Get_Nbr_Elements (Index_List1) /= Get_Nbr_Elements (Index_List2) then
         return False;
      end if;
      if Get_Base_Type (Get_Element_Subtype (Type1))
        /= Get_Base_Type (Get_Element_Subtype (Type2))
      then
         return False;
      end if;
      for I in Natural loop
         El1 := Get_Nth_Element (Index_List1, I);
         exit when El1 = Null_Iir;
         El2 := Get_Nth_Element (Index_List2, I);
         if not Are_Types_Closely_Related (El1, El2) then
            return False;
         end if;
      end loop;
      return True;
   end Are_Types_Closely_Related;

   procedure Finish_Sem_Type_Conversion (Conv: Iir_Type_Conversion)
   is
      Expr: Iir;
      Staticness : Iir_Staticness;
   begin
      -- LRM93 7.3.5
      -- Furthermore, the operand of a type conversion is not allowed to be
      -- the literal null, an allocator, an aggregate, or a string literal.
      Expr := Get_Expression (Conv);
      case Get_Kind (Expr) is
         when Iir_Kind_Null_Literal
           | Iir_Kind_Aggregate
           | Iir_Kind_String_Literal
           | Iir_Kind_Bit_String_Literal =>
            Error_Msg_Sem
              (Disp_Node (Expr) & " cannot be an type conversion operand",
               Expr);
            return;
         when others =>
            -- LRM93 7.3.5
            -- The type of the operand of a type conversion must be
            -- determinable independent of the context (in particular,
            -- independent of the target type).
            Expr := Sem_Expression (Expr, null);
            if Expr = Null_Iir then
               return;
            end if;
            if Get_Kind (Expr) in Iir_Kinds_Allocator then
               Error_Msg_Sem
                 (Disp_Node (Expr) & " cannot be a type conversion operand",
                  Expr);
            end if;
      end case;

      Check_Read (Expr);
      Set_Expression (Conv, Expr);

      --  LRM93 7.4.1 Locally Static Primaries.
      --  9. a type conversion whose expression is a locally static expression.
      --  LRM93 7.4.2 Globally Static Primaries.
      --  14. a type conversion whose expression is a globally static
      --      expression.
      if Expr /= Null_Iir then
         Staticness := Get_Expr_Staticness (Expr);
         if Flags.Vhdl_Std = Vhdl_87 then
            Staticness := Min (Globally, Staticness);
         end if;
         Set_Expr_Staticness (Conv, Staticness);

         if not Are_Types_Closely_Related (Get_Type (Conv), Get_Type (Expr))
         then
            --  FIXME: should explain why the types are not closely related.
            Error_Msg_Sem
              ("conversion not allowed between not closely related types",
               Expr);
         end if;
      end if;
   end Finish_Sem_Type_Conversion;

   procedure Finish_Sem_Function_Specification (Name : Iir; Spec : Iir)
   is
      Res : Iir;
   begin
      if not Maybe_Function_Call (Spec) then
         Error_Msg_Sem (Disp_Node (Spec) & " requires parameters", Name);
         return;
      end if;
      Res := Maybe_Insert_Function_Call (Spec, Name);
      if Get_Kind (Res) /= Iir_Kind_Function_Call then
         raise Internal_Error;
      end if;
      Finish_Sem_Function_Call (Res);
      Set_Named_Entity (Name, Res);
   end Finish_Sem_Function_Specification;

   procedure Finish_Sem_Name (Name : Iir; Res : Iir);

   procedure Finish_Sem_Implicits (Name : Iir; Pfx : Iir)
   is
      Name_Pfx : Iir;
   begin
      case Get_Kind (Pfx) is
         when Iir_Kinds_Object_Declaration
           | Iir_Kind_Attribute_Value =>
            null;
         when Iir_Kind_Indexed_Name
           | Iir_Kind_Selected_Element
           | Iir_Kind_Slice_Name =>
            Name_Pfx := Get_Prefix (Name);
            if Is_Overload_List (Name_Pfx) then
               Finish_Sem_Name (Name_Pfx, Pfx);
            end if;
         when Iir_Kind_Implicit_Dereference =>
            Finish_Sem_Implicits (Name, Get_Prefix (Pfx));
            Finish_Sem_Dereference (Pfx);
         when Iir_Kind_Function_Call =>
            if Get_Name_Staticness (Pfx) = Unknown then
               Finish_Sem_Function_Call (Pfx);
            else
               Name_Pfx := Get_Prefix (Name);
               if Is_Overload_List (Name_Pfx) then
                  Finish_Sem_Name (Name_Pfx, Pfx);
               end if;
            end if;
         when others =>
            Error_Kind ("finish_sem_implicits", Pfx);
      end case;
   end Finish_Sem_Implicits;

   procedure Finish_Sem_Name (Name : Iir; Res : Iir)
   is
      Pfx : Iir;
   begin
      case Get_Kind (Res) is
         when Iir_Kind_Design_Unit =>
            return;
         when Iir_Kinds_Object_Declaration
           | Iir_Kind_Attribute_Value
           | Iir_Kind_Type_Declaration
           | Iir_Kind_Subtype_Declaration
           | Iir_Kind_Enumeration_Literal
           | Iir_Kind_Component_Declaration =>
            return;
         when Iir_Kind_Type_Conversion =>
            Finish_Sem_Type_Conversion (Res);
            return;
         when Iir_Kind_Indexed_Name
           | Iir_Kind_Selected_Element
           | Iir_Kind_Slice_Name
           | Iir_Kind_Dereference =>
            null;
         when Iir_Kind_Function_Call =>
            Finish_Sem_Function_Call (Res);
            return;
         when Iir_Kinds_Function_Specification
           | Iir_Kinds_Procedure_Specification =>
            --declare
            --   Nres : Iir;
            --begin
            --   Nres := Sem_As_Function_Call (Res, Null_Iir, Name);
            --  Set_Named_Entity (Name, Nres);
            --  Finish_Sem_Function_Call (Nres);
            --end;
            return;
         when Iir_Kind_Length_Array_Attribute
           | Iir_Kind_Range_Array_Attribute =>
            Finish_Sem_Array_Attribute (Res, Null_Iir);
            return;
--          when Iir_Kind_Pos_Attribute =>
--             if Get_Parameter (Res) = Null_Iir then
--                Finish_Sem_Scalar_Type_Attribute (Res, Null_Iir);
--             end if;
--             return;
         when others =>
            Error_Kind ("finish_sem_name", Res);
      end case;

      Pfx := Get_Prefix (Res);
      Finish_Sem_Implicits (Name, Pfx);

      case Get_Kind (Res) is
         when Iir_Kind_Indexed_Name =>
            Finish_Sem_Indexed_Name (Res);
         when Iir_Kind_Slice_Name =>
            Finish_Sem_Slice_Name (Res);
         when Iir_Kind_Selected_Element =>
            Set_Name_Staticness (Res, Get_Name_Staticness (Pfx));
            Set_Expr_Staticness (Res, Get_Expr_Staticness (Pfx));
            Set_Base_Name (Res, Get_Base_Name (Pfx));
         when Iir_Kind_Dereference =>
            Finish_Sem_Dereference (Res);
         when others =>
            Error_Kind ("finish_sem_name(2)", Res);
      end case;
   end Finish_Sem_Name;

   --  LRM93 6.3
   --  Selected Names.
   procedure Sem_Selected_Name (Name: Iir)
   is
      Prefix: Iir;
      Suffix: Name_Id;
      Prefix_Loc : Location_Type;
      Res : Iir;

      procedure Sem_As_Expanded_Name (Sub_Name : Iir)
      is
         Sub_Res : Iir;
      begin
         if Get_Is_Within_Flag (Sub_Name) = True then
            Sub_Res := Find_Declarations_In_List (Sub_Name, Name);
            if Sub_Res /= Null_Iir then
               Add_Result (Res, Sub_Res);
            end if;
         end if;
      end Sem_As_Expanded_Name;

      --  LRM93 6.3
      --  For a selected name that is used to denote a record element,
      --  the suffix must be a simple name denoting an element of a
      --  record object or value.  The prefix must be appropriate for the
      --  type of this object or value.
      procedure Sem_As_Selected_Element (Sub_Name : Iir)
      is
         Base_Type : Iir;
         Ptr_Type : Iir;
         Rec_El : Iir;
         R : Iir;
         Se : Iir;
      begin
         --  FIXME: if not is_expr (sub_name) return.
         Base_Type := Get_Base_Type (Get_Type (Sub_Name));
         if Get_Kind (Base_Type) = Iir_Kind_Access_Type_Definition
         then
            Ptr_Type := Base_Type;
            Base_Type :=
              Get_Base_Type (Get_Designated_Type (Base_Type));
         else
            Ptr_Type := Null_Iir;
         end if;
         if Get_Kind (Base_Type) /= Iir_Kind_Record_Type_Definition then
            return;
         end if;

         Rec_El := Find_Name_In_List
           (Get_Element_Declaration_List (Base_Type), Suffix);
         if Rec_El = Null_Iir then
            return;
         end if;

         if not Maybe_Function_Call (Sub_Name) then
            return;
         end if;

         R := Maybe_Insert_Function_Call (Sub_Name, Name);
         R := Maybe_Insert_Dereference (R, Ptr_Type);

         Se := Create_Iir (Iir_Kind_Selected_Element);
         Location_Copy (Se, Name);
         Set_Prefix (Se, R);
         Set_Type (Se, Get_Type (Rec_El));
         Set_Selected_Element (Se, Rec_El);
         Set_Base_Name (Se, Get_Base_Name (R));
         Add_Result (Res, Se);
      end Sem_As_Selected_Element;

      procedure Error_Selected_Element (Prefix_Type : Iir)
      is
         Base_Type : Iir;
      begin
         Base_Type := Get_Base_Type (Prefix_Type);
         if Get_Kind (Base_Type) = Iir_Kind_Access_Type_Definition then
            Base_Type := Get_Base_Type (Get_Designated_Type (Base_Type));
         end if;
         if Get_Kind (Base_Type) /= Iir_Kind_Record_Type_Definition then
            Error_Msg_Sem
              (Disp_Node (Prefix) & " does not designates a record", Name);
         else
            Error_Msg_Sem
              ("no such element " & Name_Table.Image (Suffix)
               & " in record designated", Name);
         end if;
      end Error_Selected_Element;
   begin
      Prefix := Get_Prefix (Name);
      Prefix_Loc := Get_Location (Prefix);
      Sem_Names (Prefix);
      Prefix := Get_Named_Entity (Prefix);
      if Prefix = Error_Mark then
         Set_Named_Entity (Name, Prefix);
         return;
      end if;
      Suffix := Get_Identifier (Name);

      Res := Null_Iir;

      --  FIXME: do better.
      --
      case Get_Kind (Prefix) is
         when Iir_Kind_Overload_List =>
            --  LRM93 6.3
            --  If, according to the visibility rules, there is at
            --  least one possible interpretation of the prefix of a
            --  selected name as the name of an enclosing entity
            --  interface, architecture, subprogram, block statement,
            --  process statement, generate statement, or loop
            --  statement, then the only interpretations considered are
            --  those of the immediately preceding paragraph.
            --
            --  In this case, the selected name is always interpreted
            --  as an expanded name.  In particular, no interpretations
            --  of the prefix as a function call are considered.
            declare
               El : Iir;
            begin
               for I in Natural loop
                  El := Get_Nth_Element (Prefix, I);
                  exit when El = Null_Iir;
                  Sem_As_Expanded_Name (El);
               end loop;
               if Res /= Null_Iir then
                  for I in Natural loop
                     El := Get_Nth_Element (Prefix, I);
                     exit when El = Null_Iir;
                     Sem_As_Selected_Element (El);
                  end loop;
               end if;
            end;
            if Res = Null_Iir then
               Error_Msg_Sem ("no suffix """ & Name_Table.Image (Suffix)
                              & """ for overloaded selected name", Name);
            end if;
         when Iir_Kind_Library_Declaration =>
            --  LRM93 6.3
            --  An expanded name denotes a primary unit constained in a design
            --  library if the prefix denotes the library and the suffix is the
            --  simple name if a primary unit whose declaration is contained
            --  in that library.
            --  An expanded name is not allowed for a secondary unit,
            --  particularly for an architecture body.
            --  GHDL: FIXME: error message more explicite
            Res := Libraries.Load_Primary_Unit (Prefix, Suffix, Name);
            if Res = Null_Iir then
               Error_Msg_Sem
                 ("primary unit """ & Iirs_Utils.Get_String (Name)
                  & """ not found in " & Disp_Node (Prefix), Name);
            else
               Sem.Add_Dependence (Res);
            end if;
         when Iir_Kind_Process_Statement
           | Iir_Kind_Sensitized_Process_Statement
           | Iir_Kind_Design_Unit
--            | Iir_Kind_Architecture_Declaration
--            | Iir_Kind_Entity_Declaration
--            | Iir_Kind_Package_Declaration
           | Iir_Kind_Generate_Statement
           | Iir_Kind_Block_Statement
           | Iir_Kind_For_Loop_Statement =>
            --  LRM93 6.3
            --  An expanded name denotes a named entity declared immediatly
            --  within a named construct if the prefix that is an entity
            --  interface, an architecture, a subprogram, a block statement,
            --  a process statement, a generate statement, or a loop
            --  statement, and the suffix is the simple name, character
            --  literal, or operator symbol of an named entity whose
            --  declaration occurs immediatly within that construct.
            if Get_Kind (Prefix) = Iir_Kind_Design_Unit then
               if not Get_Loaded_Flag (Prefix) then
                  Libraries.Load_Design_Unit (Prefix, Name);
               end if;
               Sem.Add_Dependence (Prefix);
               Prefix := Get_Library_Unit (Prefix);
            end if;

            Res := Find_Declarations_In_List (Prefix, Name);

            if Res = Null_Iir then
               Error_Msg_Sem
                 ("no declaration for """ & Name_Table.Image (Suffix)
                  & """ in " & Disp_Node (Prefix), Name);
            else
               --  LRM93 6.3
               --  This form of expanded name is only allowed within the
               --  construct itself.
               if Get_Kind (Prefix) /= Iir_Kind_Package_Declaration
                 and then not Get_Is_Within_Flag (Prefix)
               then
                  Error_Msg_Sem
                    ("this expanded name is only allowed within the construct",
                     Prefix_Loc);
                  --  Hum, keep res.
               end if;
            end if;
         when Iir_Kind_Function_Spec_Body
           | Iir_Kind_Function_Specification =>
            Sem_As_Expanded_Name (Prefix);
            if Res = Null_Iir then
               Sem_As_Selected_Element (Prefix);
            end if;
            if Res = Null_Iir then
               Error_Selected_Element (Get_Return_Type (Prefix));
            end if;
         when Iir_Kinds_Object_Declaration
           | Iir_Kind_Indexed_Name
           | Iir_Kind_Selected_Element
           | Iir_Kind_Attribute_Value =>
            Sem_As_Selected_Element (Prefix);
            if Res = Null_Iir then
               Error_Selected_Element (Get_Type (Prefix));
            end if;
         when Iir_Kind_Type_Declaration
           | Iir_Kind_Subtype_Declaration =>
            Error_Msg_Sem
              (Disp_Node (Prefix) & " cannot be selected by name", Prefix_Loc);

         when others =>
            Error_Kind ("sem_selected_name(2)", Prefix);
      end case;
      if Res = Null_Iir then
         Res := Error_Mark;
      elsif not Is_Overload_List (Res) then
         --  Finish sem
         Finish_Sem_Name (Name, Res);
      end if;
      Set_Named_Entity (Name, Res);
   end Sem_Selected_Name;

   --  If ASSOC_LIST has one element, which is an expression without formal,
   --  return the actual, else return NULL_IIR.
   function Get_One_Actual (Assoc_List : Iir_List) return Iir
   is
      Assoc : Iir;
   begin
      if Get_Nbr_Elements (Assoc_List) /= 1 then
         return Null_Iir;
      end if;
      Assoc := Get_Nth_Element (Assoc_List, 0);
      if Get_Kind (Assoc) /= Iir_Kind_Association_Element_By_Expression then
         return Null_Iir;
      end if;
      if Get_Formal (Assoc) /= Null_Iir then
         return Null_Iir;
      end if;
      return Get_Actual (Assoc);
   end Get_One_Actual;

   function Slice_Or_Index (Actual : Iir) return Iir_Kind
   is
   begin
      --  But it may be a slice name.
      case Get_Kind (Actual) is
         when Iir_Kind_Range_Array_Attribute
           | Iir_Kind_Reverse_Range_Array_Attribute
           | Iir_Kind_Subtype_Declaration
           | Iir_Kind_Type_Declaration
           | Iir_Kind_Range_Expression =>
            return Iir_Kind_Slice_Name;
         when others =>
            if Is_Range_Attribute_Name (Actual) then
               return Iir_Kind_Slice_Name;
            end if;
      end case;
      --  By default, this is an indexed name.
      return Iir_Kind_Indexed_Name;
   end Slice_Or_Index;

   function Sem_Index_Specification (Name : Iir_Parenthesis_Name; Itype : Iir)
                                    return Iir
   is
      Actual : Iir;
      Kind : Iir_Kind;
      Res : Iir;
   begin
      Actual := Get_One_Actual (Get_Association_List (Name));
      if Actual = Null_Iir then
         Error_Msg_Sem ("only one index specification is allowed", Name);
         return Null_Iir;
      end if;
      case Get_Kind (Actual) is
         when Iir_Kind_Simple_Name
           | Iir_Kind_Selected_Name =>
            Sem_Names (Actual);
            Actual := Get_Named_Entity (Actual);
            --  FIXME: semantization to be finished.
            --Maybe_Finish_Sem_Name (Actual);
         when others =>
            null;
      end case;
      Kind := Slice_Or_Index (Actual);
      Res := Create_Iir (Kind);
      Location_Copy (Res, Name);
      case Kind is
         when Iir_Kind_Indexed_Name =>
            Actual := Sem_Expression (Actual, Itype);
            if Actual = Null_Iir then
               return Null_Iir;
            end if;
            Check_Read (Actual);
            if Get_Expr_Staticness (Actual) < Globally then
               Error_Msg_Sem ("index must be a static expression", Name);
            end if;
            Append_Element (Get_Index_List (Res), Actual);
         when Iir_Kind_Slice_Name =>
            Actual := Sem_Discrete_Range_Expression (Actual, Itype);
            if Actual = Null_Iir then
               return Null_Iir;
            end if;
            if Get_Discrete_Range_Staticness (Actual) < Globally then
               Error_Msg_Sem ("index must be a static expression", Name);
            end if;
            Set_Suffix (Res, Actual);
         when others =>
            raise Internal_Error;
      end case;
      return Res;
   end Sem_Index_Specification;

   procedure Sem_Parenthesis_Name (Name : Iir_Parenthesis_Name)
   is
      Prefix: Iir;
      Res : Iir;
      Assoc_List : Iir_Association_List;
      Assoc_Len : Integer;

      Slice_Index_Kind : Iir_Kind;

      procedure Index_Or_Not
      is
         El : Iir;
      begin
         Slice_Index_Kind := Iir_Kind_Error;
         for I in Natural loop
            El := Get_Nth_Element (Assoc_List, I);
            exit when El = Null_Iir;
            case Get_Kind (El) is
               when Iir_Kind_Association_Element_By_Expression =>
                  if Get_Formal (El) /= Null_Iir then
                     return;
                  end if;
               when others =>
                  --  Only expression are allowed.
                  return;
            end case;
         end loop;
         Slice_Index_Kind := Iir_Kind_Indexed_Name;
      end Index_Or_Not;

      --  If FINISH is TRUE, then display error message in case of error.
      function Sem_As_Indexed_Or_Slice_Name (Sub_Name : Iir; Finish : Boolean)
        return Iir
      is
         Base_Type : Iir;
         Ptr_Type : Iir;
         P : Iir;
         R : Iir;
      begin
         if Slice_Index_Kind = Iir_Kind_Error then
            if Finish then
               Error_Msg_Sem ("prefix is not a function name", Name);
            end if;
            --  No way.
            return Null_Iir;
         end if;

         Base_Type := Get_Base_Type (Get_Type (Sub_Name));
         if Get_Kind (Base_Type) = Iir_Kind_Access_Type_Definition
         then
            Ptr_Type := Base_Type;
            Base_Type :=
              Get_Base_Type (Get_Designated_Type (Base_Type));
         else
            Ptr_Type := Null_Iir;
         end if;
         if Get_Kind (Base_Type) /= Iir_Kind_Array_Type_Definition then
            if Finish then
               Error_Msg_Sem ("type of prefix is not an array", Name);
            end if;
            return Null_Iir;
         end if;
         if Get_Nbr_Elements (Get_Index_List (Base_Type)) /= Assoc_Len then
            if Finish then
               Error_Msg_Sem
                 ("number of indexes mismatches array dimension", Name);
            end if;
            return Null_Iir;
         end if;

         if not Maybe_Function_Call (Sub_Name) then
            if Finish then
               --  Should not happen.
               raise Internal_Error;
            end if;
            return Null_Iir;
         end if;

         P := Maybe_Insert_Function_Call (Sub_Name, Name);
         P := Maybe_Insert_Dereference (P, Ptr_Type);

         R := Create_Iir (Slice_Index_Kind);
         Location_Copy (R, Sub_Name);
         Set_Prefix (R, P);

         case Slice_Index_Kind is
            when Iir_Kind_Slice_Name =>
               Set_Suffix (R, Get_Actual (Get_Nth_Element (Assoc_List, 0)));
               Set_Type (R, Get_Type (P));
            when Iir_Kind_Indexed_Name =>
               declare
                  Idx_El : Iir;
                  Idx_List : Iir_List;
               begin
                  Idx_List := Get_Index_List (R);
                  for I in Natural loop
                     Idx_El := Get_Nth_Element (Assoc_List, I);
                     exit when Idx_El = Null_Iir;
                     Append_Element (Idx_List, Get_Actual (Idx_El));
                  end loop;
               end;
               Set_Type (R, Get_Element_Subtype (Base_Type));
            when others =>
               raise Internal_Error;
         end case;

         return R;
      end Sem_As_Indexed_Or_Slice_Name;

      Actual : Iir;
      Actual_Expr : Iir;
   begin
      -- The prefix is a function name, a type mark or an array.
      Prefix := Get_Prefix (Name);
      Sem_Names (Prefix);
      Prefix := Get_Named_Entity (Prefix);
      if Prefix = Error_Mark then
         Set_Named_Entity (Name, Prefix);
         return;
      end if;
      Res := Null_Iir;

      Assoc_List := Get_Association_List (Name);
      Assoc_Len := Get_Nbr_Elements (Assoc_List);
      Actual := Get_One_Actual (Assoc_List);

      if Actual /= Null_Iir
        and then Get_Kind (Actual) = Iir_Kind_Range_Expression then
         --  A slice.
         Slice_Index_Kind := Iir_Kind_Slice_Name;
         Add_Result (Res, Sem_As_Indexed_Or_Slice_Name (Prefix, True));
      elsif Actual /= Null_Iir
        and then (Get_Kind (Prefix) = Iir_Kind_Type_Declaration
                  or else Get_Kind (Prefix) = Iir_Kind_Subtype_Declaration)
      then
         --  A type conversion
         Res := Create_Iir (Iir_Kind_Type_Conversion);
         Location_Copy (Res, Name);
         Set_Type (Res, Get_Type (Prefix));
         Set_Expression (Res, Actual);
      else
         if Actual /= Null_Iir
           and then (Get_Kind (Actual) = Iir_Kind_Simple_Name
                     or Get_Kind (Actual) = Iir_Kind_Selected_Name)
         then
            Sem_Names (Actual);
            Actual_Expr := Get_Named_Entity (Actual);
            if Actual_Expr = Error_Mark then
               Set_Named_Entity (Name, Actual_Expr);
               return;
            end if;
            --  Decides between sliced or indexed name to actual.
            Slice_Index_Kind := Slice_Or_Index (Actual_Expr);
         else
            Index_Or_Not;
         end if;

         if Slice_Index_Kind /= Iir_Kind_Slice_Name then
            if Sem_Association_List (Assoc_List) = False then
               null;
            end if;
         end if;

         case Get_Kind (Prefix) is
            when Iir_Kind_Overload_List =>
               declare
                  El : Iir;
                  Used : Boolean;
                  R : Iir;
               begin
                  for I in Natural loop
                     El := Get_Nth_Element (Prefix, I);
                     exit when El = Null_Iir;
                     if Get_Kind (El) in Iir_Kinds_Function_Specification
                       and then
                       Sem_Subprogram_Arguments
                       (Get_Interface_Declaration (El),
                        Assoc_List, False, True, False, Name) /= None
                     then
                        Add_Result
                          (Res, Sem_As_Function_Call (El, Assoc_List, Name));
                        Used := True;
                     else
                        Used := False;
                     end if;
                     if Get_Kind (El) not in Iir_Kinds_Procedure_Specification
                     then
                        R := Sem_As_Indexed_Or_Slice_Name (El, False);
                        if R /= Null_Iir then
                           Add_Result (Res, R);
                           Used := True;
                        end if;
                     end if;
                     if not Used then
                        Sem_Name_Free_Result (El, Null_Iir);
                     end if;
                  end loop;
               end;
               if Res = Null_Iir then
                  Error_Msg_Sem
                    ("neither prefix is a function name "
                     & "nor it can be sliced or indexed", Name);
               end if;
            when Iir_Kinds_Function_Specification =>
               Add_Result
                 (Res, Sem_As_Function_Call (Prefix, Assoc_List, Name));
               Add_Result (Res, Sem_As_Indexed_Or_Slice_Name (Prefix, False));
               if Res = Null_Iir then
                  Error_Msg_Sem
                    ("neither prefix is a function name "
                     & "nor it can be sliced or indexed", Name);
               end if;

            when Iir_Kinds_Object_Declaration
              | Iir_Kind_Indexed_Name
              | Iir_Kind_Slice_Name
              | Iir_Kind_Dereference
              | Iir_Kind_Implicit_Dereference
              | Iir_Kind_Selected_Element
              | Iir_Kind_Attribute_Value
              | Iir_Kind_Function_Call =>
               Add_Result (Res, Sem_As_Indexed_Or_Slice_Name (Prefix, True));

            when Iir_Kinds_Array_Attribute =>
               if Actual /= Null_Iir then
                  Finish_Sem_Array_Attribute (Prefix, Actual);
                  Set_Named_Entity (Name, Prefix);
               else
                  Error_Msg_Sem ("bad attribute parameter", Name);
                  Set_Named_Entity (Name, Error_Mark);
               end if;
               return;

            when Iir_Kinds_Scalar_Type_Attribute =>
               if Actual /= Null_Iir then
                  Finish_Sem_Scalar_Type_Attribute (Prefix, Actual);
                  Set_Named_Entity (Name, Prefix);
               else
                  Error_Msg_Sem ("bad attribute parameter", Name);
                  Set_Named_Entity (Name, Error_Mark);
               end if;
               return;

            when Iir_Kind_Type_Declaration
              | Iir_Kind_Subtype_Declaration =>
               Error_Msg_Sem
                 ("subprogram name is a type mark (missing apostrophe)", Name);

            when Iir_Kind_Stable_Attribute
              | Iir_Kind_Quiet_Attribute
              | Iir_Kind_Delayed_Attribute =>
               if Actual /= Null_Iir then
                  Finish_Sem_Signal_Attribute (Prefix, Actual);
                  Set_Named_Entity (Name, Prefix);
               else
                  Error_Msg_Sem ("bad attribute parameter", Name);
                  Set_Named_Entity (Name, Error_Mark);
               end if;
               return;

            when Iir_Kinds_Procedure_Specification =>
               Error_Msg_Sem ("function name is a procedure", Name);

            when others =>
               Error_Kind ("sem_parenthesis_name", Prefix);
         end case;
      end if;

      if Res = Null_Iir then
         Res := Error_Mark;
      elsif not Is_Overload_List (Res) then
         Finish_Sem_Name (Name, Res);
      end if;
      Set_Named_Entity (Name, Res);
   end Sem_Parenthesis_Name;

   procedure Sem_Selected_By_All_Name (Name : Iir_Selected_By_All_Name)
   is
      Prefix : Iir;
      Res : Iir;

      procedure Sem_As_Selected_By_All_Name (Sub_Name : Iir)
      is
         Base_Type : Iir;
         R : Iir;
      begin
         Base_Type := Get_Base_Type (Get_Type (Sub_Name));
         if Get_Kind (Base_Type) /= Iir_Kind_Access_Type_Definition then
            return;
         end if;

         R := Create_Iir (Iir_Kind_Dereference);
         Location_Copy (R, Name);
         Set_Prefix (R, Sub_Name);
         --  FIXME: access subtype.
         Set_Type (R, Get_Designated_Type (Base_Type));
         Add_Result (Res, R);
      end Sem_As_Selected_By_All_Name;
   begin
      Prefix := Get_Prefix (Name);
      Sem_Names (Prefix);
      Prefix := Get_Named_Entity (Prefix);
      if Prefix = Null_Iir then
         return;
      end if;
      Res := Null_Iir;
      case Get_Kind (Prefix) is
         when Iir_Kinds_Object_Declaration
           | Iir_Kind_Selected_Element
           | Iir_Kind_Dereference
           | Iir_Kind_Indexed_Name =>
            Sem_As_Selected_By_All_Name (Prefix);
         when others =>
            Error_Kind ("sem_selected_by_all_name", Prefix);
      end case;
      if Res = Null_Iir then
         Error_Msg_Sem ("prefix is not an access", Name);
         Res := Error_Mark;
      elsif not Is_Overload_List (Res) then
         Finish_Sem_Name (Name, Res);
      end if;
      Set_Named_Entity (Name, Res);
   end Sem_Selected_By_All_Name;

   function Sem_Base_Attribute (Attr : Iir_Attribute) return Iir
   is
      Prefix : Iir;
      Res : Iir;
      Base_Type : Iir;
   begin
      Prefix := Get_Named_Entity (Get_Prefix (Attr));
      case Get_Kind (Prefix) is
         when Iir_Kind_Type_Declaration
           | Iir_Kind_Subtype_Declaration =>
            null;
         when others =>
            Error_Msg_Sem
              ("prefix of 'base attribute must be a type or a subtype", Attr);
            return Error_Mark;
      end case;
      Res := Create_Iir (Iir_Kind_Base_Attribute);
      Location_Copy (Res, Attr);
      Set_Prefix (Res, Prefix);
      Base_Type := Get_Base_Type (Get_Type (Prefix));
      Base_Type := Get_Type (Get_Type_Declarator (Base_Type));
      Set_Type (Res, Base_Type);
      return Res;
   end Sem_Base_Attribute;

   function Sem_User_Attribute (Attr : Iir_Attribute) return Iir
   is
      Prefix : Iir;
      Value_List : Iir_Attribute_Value_List;
      Value : Iir;
      Attr_Id : Name_Id;
      Spec : Iir_Attribute_Specification;
   begin
      Prefix := Get_Named_Entity (Get_Prefix (Attr));
      Attr_Id := Get_Identifier (Attr);
      --  LRM93 6.6
      --  If the attribute name denotes an alias, then the attribute name
      --  denotes an attribute of the aliased name and not the alias itself,
      --  except when the attribute designator denotes any of the predefined
      --  attributes 'simple_name, 'path_name, or 'instance_name.
      if Get_Kind (Prefix) = Iir_Kind_Alias_Declaration then
         --  GHDL: according to 4.3.3, the name cannot be an alias.
         Prefix := Get_Name (Prefix);
      end if;

      --  LRM93 6.6
      --  If the attribute designator denotes a user-defined attribute, the
      --  prefix cannot denote a subelement or a slice of an object.
      case Get_Kind (Prefix) is
         when Iir_Kind_Selected_By_All_Name
           | Iir_Kind_Selected_Name
           | Iir_Kind_Indexed_Name
           | Iir_Kind_Slice_Name =>
            Error_Msg_Sem ("prefix of user defined attribute cannot be an "
                           & "object subelement", Attr);
            return Null_Iir;
         when Iir_Kinds_Object_Declaration
           | Iir_Kind_Design_Unit
           | Iir_Kind_Type_Declaration
           | Iir_Kind_Subtype_Declaration
           | Iir_Kinds_Function_Specification
           | Iir_Kinds_Procedure_Specification =>
            --  FIXME: to complete
            null;
         when others =>
            Error_Kind ("sem_user_attribute", Prefix);
      end case;

      Value_List := Get_Attribute_Value_List (Prefix);
      if Value_List = Null_Iir then
         Value := Null_Iir;
      else
         for I in Natural loop
            Value := Get_Nth_Element (Value_List, I);
            exit when Value = null;
            Spec := Get_Attribute_Specification (Value);
            exit when Get_Identifier (Get_Attribute_Designator (Spec))
              = Attr_Id;
         end loop;
      end if;
      if Value = Null_Iir then
         Error_Msg_Sem
           (Disp_Node (Prefix) & " was not annotated with attribute `"
            & Name_Table.Image (Attr_Id) & ''', Attr);
         if Attr_Id = Std_Names.Name_First or Attr_Id = Std_Names.Name_Last
         then
            --  Nice (?) message for Ada users.
            Error_Msg_Sem
              ("(you may use 'high, 'low, 'left or 'right attribute)", Attr);
         end if;
         return Error_Mark;
      end if;

      return Value;
   end Sem_User_Attribute;

   function Sem_Scalar_Type_Attribute (Attr : Iir_Attribute)
     return Iir
   is
      use Std_Names;
      Prefix_Name : Iir;
      Prefix : Iir;
      Prefix_Type : Iir;
      Res : Iir;
   begin
      Prefix_Name := Get_Prefix (Attr);
      Prefix := Get_Named_Entity (Prefix_Name);
      --  LRM93 14.1
      --  Prefix: Any discrete or physical type of subtype T.
      case Get_Kind (Prefix) is
         when Iir_Kinds_Type_Declaration
           | Iir_Kind_Base_Attribute =>
            null;
         when others =>
            Error_Msg_Sem
              ("prefix of '" & Get_String (Attr) & " attribute must be a type",
               Attr);
            return Error_Mark;
      end case;
      Prefix_Type := Get_Type (Prefix);
      case Get_Identifier (Attr) is
         when Name_Image
           | Name_Value =>
            if Get_Kind (Prefix_Type) not in Iir_Kinds_Scalar_Type_Definition
            then
               Error_Msg_Sem
                 ("prefix of '" & Get_String (Attr)
                  & " attribute must be a scalar type", Attr);
               Error_Msg_Sem
                 ("found " & Disp_Node (Prefix_Type)
                  & " defined at " & Disp_Location (Prefix_Type), Attr);
               return Error_Mark;
            end if;
         when others =>
            case Get_Kind (Prefix_Type) is
               when Iir_Kinds_Discrete_Type_Definition
                 | Iir_Kind_Physical_Subtype_Definition
                 | Iir_Kind_Physical_Type_Definition =>
                  null;
               when others =>
                  Error_Msg_Sem
                    ("prefix of '" & Get_String (Attr)
                     & " attribute must be discrete or physical type", Attr);
                  Error_Msg_Sem
                    ("found " & Disp_Node (Prefix_Type)
                     & " defined at " & Disp_Location (Prefix_Type), Attr);
                  return Error_Mark;
            end case;
      end case;

      --  Create the resulting node.
      case Get_Identifier (Attr) is
         when Name_Pos =>
            Res := Create_Iir (Iir_Kind_Pos_Attribute);
         when Name_Val =>
            Res := Create_Iir (Iir_Kind_Val_Attribute);
         when Name_Succ =>
            Res := Create_Iir (Iir_Kind_Succ_Attribute);
         when Name_Pred =>
            Res := Create_Iir (Iir_Kind_Pred_Attribute);
         when Name_Leftof =>
            Res := Create_Iir (Iir_Kind_Leftof_Attribute);
         when Name_Rightof =>
            Res := Create_Iir (Iir_Kind_Rightof_Attribute);
         when Name_Image =>
            Res := Create_Iir (Iir_Kind_Image_Attribute);
         when Name_Value =>
            Res := Create_Iir (Iir_Kind_Value_Attribute);
         when others =>
            raise Internal_Error;
      end case;
      Location_Copy (Res, Attr);
      Set_Prefix (Res, Prefix);

      case Get_Identifier (Attr) is
         when Name_Pos =>
            --  LRM93 14.1
            --  Result type: universal_integer.
            Set_Type (Res, Universal_Integer_Type_Definition);
         when Name_Val =>
            --  LRM93 14.1
            --  Result type: the base type of T
            Set_Type (Res, Get_Base_Type (Prefix_Type));
         when Name_Succ
           | Name_Pred
           | Name_Leftof
           | Name_Rightof =>
            --  LRM93 14.1
            --  Result type: the base type of T.
            Set_Type (Res, Get_Base_Type (Prefix_Type));
         when Name_Image =>
            --  LRM93 14.1
            --  Result type: type string
            Set_Type (Res, String_Type_Definition);
         when Name_Value =>
            --  LRM93 14.1
            --  Result type: the base type of T.
            Set_Type (Res, Get_Base_Type (Prefix_Type));
         when others =>
            raise Internal_Error;
      end case;
      return Res;
   end Sem_Scalar_Type_Attribute;

   --  Sem attributes whose prefix is a type or a subtype.
   function Sem_Predefined_Type_Attribute (Attr : Iir_Attribute)
     return Iir
   is
      use Std_Names;
      Res : Iir;
      Prefix : Iir;
      Prefix_Type : Iir;
   begin
      case Get_Identifier (Attr) is
         when Name_Left =>
            Res := Create_Iir (Iir_Kind_Left_Type_Attribute);
         when Name_Right =>
            Res := Create_Iir (Iir_Kind_Right_Type_Attribute);
         when Name_High =>
            Res := Create_Iir (Iir_Kind_High_Type_Attribute);
         when Name_Low =>
            Res := Create_Iir (Iir_Kind_Low_Type_Attribute);
         when Name_Range
           | Name_Reverse_Range =>
            Error_Msg_Sem
              ("prefix of range attribute must be an array type or object",
               Attr);
            return Error_Mark;
         when others =>
            raise Internal_Error;
      end case;
      Location_Copy (Res, Attr);
      Prefix := Get_Named_Entity (Get_Prefix (Attr));
      Set_Prefix (Res, Prefix);
      Prefix_Type := Get_Type (Prefix);
      --  LRM93 14.1
      --  Result Type: Same type as T.
      Set_Type (Res, Prefix_Type);
      Set_Expr_Staticness (Res, Get_Type_Staticness (Prefix_Type));
      return Res;
   end Sem_Predefined_Type_Attribute;

   --  Called for attributes Length, Left, Right, High, Low, Range,
   --  Reverse_Range, Ascending.
   function Sem_Array_Attribute (Attr : Iir_Attribute) return Iir
   is
      procedure Error_Type is
      begin
         Error_Msg_Sem
           ("prefix of '" & Name_Table.Image (Get_Identifier (Attr))
            & " attribute must denote a constrained array subtype",
            Attr);
      end Error_Type;

      use Std_Names;
      Prefix: Iir;
      Prefix_Type : Iir;
      Res : Iir;
   begin
      Prefix := Get_Named_Entity (Get_Prefix (Attr));

      --  LRM93 14.1
      --  Prefix: Any prefix A that is appropriate for an array object, or an
      --  alias thereof, or that denotes a constrained array subtype.
      case Get_Kind (Prefix) is
         when Iir_Kind_Dereference
           | Iir_Kinds_Object_Declaration
           | Iir_Kind_Function_Call
           | Iir_Kind_Selected_Element
           | Iir_Kind_Indexed_Name
           | Iir_Kind_Slice_Name =>
            --  FIXME: list of expr.
            Prefix_Type := Get_Type (Prefix);
            case Get_Kind (Prefix_Type) is
               when Iir_Kind_Access_Type_Definition
                 | Iir_Kind_Access_Subtype_Definition =>
                  declare
                     Designated_Type : Iir;
                  begin
                     Designated_Type :=
                       Get_Designated_Type (Get_Base_Type (Prefix_Type));
                     Prefix := Insert_Implicit_Dereference (Prefix, Attr);
                     Prefix_Type := Designated_Type;
                  end;
               when Iir_Kinds_Array_Type_Definition =>
                  null;
               when others =>
                  Error_Msg_Sem ("object prefix must be an array", Attr);
                  return Error_Mark;
            end case;
--          when Iir_Kind_Array_Subtype_Definition =>
--             Prefix_Type := Prefix;
--          when Iir_Kind_Array_Type_Definition =>
--             Error_Type;
--             return Null_Iir;
         when Iir_Kind_Subtype_Declaration
           | Iir_Kind_Type_Declaration
           | Iir_Kind_Base_Attribute =>
            Prefix_Type := Get_Type (Prefix);
            if Get_Kind (Prefix_Type)
              in Iir_Kinds_Unconstrained_Array_Type_Definition
            then
               Error_Msg_Sem ("prefix type is not constrained", Attr);
               return Error_Mark;
            end if;
         when others =>
            Error_Kind ("sem_array_attribute", Prefix);
      end case;

      case Get_Kind (Prefix_Type) is
         when Iir_Kinds_Scalar_Type_Definition =>
            --  FIXME: check prefix is a scalar type or subtype.
            return Sem_Predefined_Type_Attribute (Attr);
         when Iir_Kinds_Array_Type_Definition =>
            null;
         when others =>
            Error_Type;
            return Error_Mark;
      end case;

      case Get_Identifier (Attr) is
         when Name_Left =>
            Res := Create_Iir (Iir_Kind_Left_Array_Attribute);
         when Name_Right =>
            Res := Create_Iir (Iir_Kind_Right_Array_Attribute);
         when Name_High =>
            Res := Create_Iir (Iir_Kind_High_Array_Attribute);
         when Name_Low =>
            Res := Create_Iir (Iir_Kind_Low_Array_Attribute);
         when Name_Range =>
            Res := Create_Iir (Iir_Kind_Range_Array_Attribute);
         when Name_Reverse_Range =>
            Res := Create_Iir (Iir_Kind_Reverse_Range_Array_Attribute);
         when Name_Length =>
            Res := Create_Iir (Iir_Kind_Length_Array_Attribute);
            Set_Type (Res, Universal_Integer_Type_Definition);
         when Name_Ascending =>
            Res := Create_Iir (Iir_Kind_Ascending_Array_Attribute);
            Set_Type (Res, Boolean_Type_Definition);
         when others =>
            raise Internal_Error;
      end case;
      Location_Copy (Res, Attr);
      Set_Prefix (Res, Prefix);
      return Res;
   end Sem_Array_Attribute;

   function Sem_Signal_Signal_Attribute (Attr : Iir_Attribute; Kind : Iir_Kind)
     return Iir
   is
      Res : Iir;
      Prefix : Iir;
   begin
      Prefix := Get_Named_Entity (Get_Prefix (Attr));
      Res := Create_Iir (Kind);
      if Kind = Iir_Kind_Delayed_Attribute then
         Set_Type (Res, Get_Type (Prefix));
      else
         Set_Type (Res, Boolean_Type_Definition);
      end if;
      Set_Base_Name (Res, Res);
--       Param := Get_Suffix (Attr);
--       if Param /= Null_Iir then
--          if Kind = Iir_Kind_Transaction_Attribute then
--             Error_Msg_Sem ("'transaction does not allow a parameter", Attr);
--             Param := Null_Iir;
--          else
--             Param := Sem_Expression
--               (Param, Time_Subtype_Definition);
--             Set_Parameter (Res, Param);
--          end if;
--       end if;
      if Get_Kind (Prefix) = Iir_Kind_Signal_Interface_Declaration
        and then Get_Kind (Get_Parent (Prefix))
        in Iir_Kinds_Subprogram_Declaration
      then
         Error_Msg_Sem ("'" & Name_Table.Image (Get_Identifier (Attr)) &
                        " is not allowed for a signal parameter", Attr);
      end if;
      Sem.Add_Declaration_For_Implicit_Signal (Res);
      return Res;
   end Sem_Signal_Signal_Attribute;

   function Sem_Signal_Attribute (Attr : Iir_Attribute) return Iir
   is
      use Std_Names;
      Prefix: Iir;
      Res : Iir;
   begin
      Prefix := Get_Named_Entity (Get_Prefix (Attr));
      case Get_Kind (Get_Object_Prefix (Prefix)) is
         when Iir_Kind_Signal_Declaration
           | Iir_Kind_Signal_Interface_Declaration
           | Iir_Kinds_Signal_Attribute =>
            null;
         when others =>
            Error_Msg_Sem
              ("prefix of '" & Name_Table.Image (Get_Identifier (Attr))
               & " attribute must denote a signal", Attr);
            return Error_Mark;
      end case;
      case Get_Identifier (Attr) is
         when Name_Stable =>
            Res := Sem_Signal_Signal_Attribute
              (Attr, Iir_Kind_Stable_Attribute);
         when Name_Quiet =>
            Res := Sem_Signal_Signal_Attribute
              (Attr, Iir_Kind_Quiet_Attribute);
         when Name_Delayed =>
            Res := Sem_Signal_Signal_Attribute
              (Attr, Iir_Kind_Delayed_Attribute);
         when Name_Transaction =>
            Res := Sem_Signal_Signal_Attribute
              (Attr, Iir_Kind_Transaction_Attribute);
         when Name_Event =>
            Res := Create_Iir (Iir_Kind_Event_Attribute);
            Set_Type (Res, Boolean_Type_Definition);
         when Name_Active =>
            Res := Create_Iir (Iir_Kind_Active_Attribute);
            Set_Type (Res, Boolean_Type_Definition);
         when Name_Last_Value =>
            Res := Create_Iir (Iir_Kind_Last_Value_Attribute);
            Set_Type (Res, Get_Type (Prefix));
         when Name_Last_Event =>
            Res := Create_Iir (Iir_Kind_Last_Event_Attribute);
            Set_Type (Res, Time_Type_Definition);
         when Name_Last_Active =>
            Res := Create_Iir (Iir_Kind_Last_Active_Attribute);
            Set_Type (Res, Time_Type_Definition);
         when others =>
            --  Not yet implemented attribute, or really an internal error.
            raise Internal_Error;
      end case;
      --  LRM 4.3.2
      --  The value of an object is said to be read when one of the following
      --  conditions is satisfied:
      --  [...]
      --  * When the object is a signal ans the value of any of its predefined
      --    attributes [...] is read.
      Check_Read (Prefix);
      Location_Copy (Res, Attr);
      Set_Expr_Staticness (Res, None);
      Set_Prefix (Res, Prefix);
      return Res;
   end Sem_Signal_Attribute;

   procedure Sem_Attribute (Attr : Iir_Attribute)
   is
      use Std_Names;
      Prefix : Iir;
      Res : Iir;
   begin
      --  LRM93 6.6
      --  The meaning of the prefix of an attribute name must be determinable
      --  independently of the attribute designator and independently of the
      --  fact that it is the prefix of an attribute.
      Prefix := Get_Prefix (Attr);
      Sem_Names (Prefix);
      Prefix := Get_Named_Entity (Prefix);
      if Prefix = Error_Mark then
         Set_Named_Entity (Attr, Prefix);
         return;
      end if;

      if Get_Kind (Prefix) = Iir_Kind_Overload_List then
         Error_Msg_Sem ("prefix of attribute is overloaded", Attr);
         Set_Named_Entity (Attr, Error_Mark);
         return;
      end if;

      case Get_Identifier (Attr) is
         when Name_Base =>
            Res := Sem_Base_Attribute (Attr);
         when Name_Image
           | Name_Value =>
            if Flags.Vhdl_Std > Vhdl_87 then
               Res := Sem_Scalar_Type_Attribute (Attr);
            else
               Res := Sem_User_Attribute (Attr);
            end if;

         when Name_Pos
           | Name_Val
           | Name_Succ
           | Name_Pred
           | Name_Rightof
           | Name_Leftof =>
            Res := Sem_Scalar_Type_Attribute (Attr);

         when Name_Length
           | Name_Left
           | Name_Right
           | Name_High
           | Name_Low
           | Name_Range
           | Name_Reverse_Range =>
            Res := Sem_Array_Attribute (Attr);

         when Name_Ascending =>
            if Flags.Vhdl_Std > Vhdl_87 then
               Res := Sem_Array_Attribute (Attr);
            else
               Res := Sem_User_Attribute (Attr);
            end if;

         when Name_Stable
           | Name_Event
           | Name_Last_Value
           | Name_Delayed
           | Name_Quiet
           | Name_Transaction
           | Name_Active
           | Name_Last_Active
           | Name_Last_Event =>
            Res := Sem_Signal_Attribute (Attr);

         when Name_Driving
           | Name_Driving_Value =>
            if Flags.Vhdl_Std > Vhdl_87 then
               Res := Sem_Signal_Attribute (Attr);
            else
               Res := Sem_User_Attribute (Attr);
            end if;

         when others =>
            Res := Sem_User_Attribute (Attr);
      end case;

      if Res = Null_Iir then
         Error_Kind ("sem_attribute", Attr);
      end if;
      Set_Named_Entity (Attr, Res);
   end Sem_Attribute;

   --  LRM93 6
   procedure Sem_Names (Name : Iir)
   is
   begin
      if Get_Named_Entity (Name) /= Null_Iir then
         return;
      end if;
      case Get_Kind (Name) is
         when Iir_Kind_Simple_Name
           | Iir_Kind_String_Literal =>
            Sem_Simple_Name (Name);
         when Iir_Kind_Selected_Name =>
            Sem_Selected_Name (Name);
         when Iir_Kind_Parenthesis_Name =>
            Sem_Parenthesis_Name (Name);
         when Iir_Kind_Selected_By_All_Name =>
            Sem_Selected_By_All_Name (Name);
         when Iir_Kind_Attribute =>
            Sem_Attribute (Name);

         --when Iir_Kind_Indexed_Name =>
         --   null;
--          when Iir_Kind_Slice_Name =>
--             return Sem_Slice_Name (Name, Null_Iir);
--          when Iir_Kind_Attribute =>
--             return Sem_Attribute (Name);

--          when Iir_Kind_Design_Unit
--            | Iir_Kind_Library_Declaration =>
--             --  Already semantized name.
--             --  Used by sem_use_clause.
--             return Name;
--          when Iir_Kind_Generate_Statement =>
--             return Name;
--          when Iir_Kind_Procedure_Specification
--            | Iir_Kind_Procedure_Spec_Body =>
--             return Name;
--          when Iir_Kind_Signal_Interface_Declaration
--            | Iir_Kind_Variable_Interface_Declaration
--            | Iir_Kind_Constant_Interface_Declaration
--            | Iir_Kind_File_Interface_Declaration =>
--             return Name;
         when others =>
            Error_Kind ("sem_names", Name);
      end case;
   end Sem_Names;

   --  Finish semantisation of NAME, if necessary.
   procedure Maybe_Finish_Sem_Name (Name : Iir)
   is
      Expr : Iir;
   begin
      Expr := Get_Named_Entity (Name);
      case Get_Kind (Expr) is
         when Iir_Kind_Error =>
            null;
         when Iir_Kinds_Object_Declaration
           | Iir_Kind_Enumeration_Literal
           | Iir_Kind_Physical_Unit =>
            null;
         when Iir_Kind_Type_Declaration
           | Iir_Kind_Subtype_Declaration =>
            null;
         when Iir_Kind_Indexed_Name
           | Iir_Kind_Slice_Name
           | Iir_Kind_Function_Call
           | Iir_Kind_Attribute_Value
           | Iir_Kind_Selected_Element
           | Iir_Kind_Dereference
           | Iir_Kind_Type_Conversion =>
            null;
         when Iir_Kinds_Type_Attribute =>
            null;
         when Iir_Kind_Event_Attribute
           | Iir_Kind_Active_Attribute
           | Iir_Kind_Last_Event_Attribute
           | Iir_Kind_Last_Active_Attribute
           | Iir_Kind_Last_Value_Attribute
           | Iir_Kind_Transaction_Attribute =>
            null;
         when Iir_Kind_Stable_Attribute
           | Iir_Kind_Quiet_Attribute
           | Iir_Kind_Delayed_Attribute =>
            if Get_Parameter (Expr) = Null_Iir then
               Finish_Sem_Signal_Attribute (Expr, Null_Iir);
            end if;
         when Iir_Kinds_Array_Attribute =>
            if Get_Parameter (Expr) = Null_Iir then
               Finish_Sem_Array_Attribute (Expr, Null_Iir);
            end if;
         when Iir_Kinds_Scalar_Type_Attribute =>
            if Get_Parameter (Expr) = Null_Iir then
               Finish_Sem_Scalar_Type_Attribute (Expr, Null_Iir);
            end if;
         when Iir_Kind_Implicit_Dereference =>
            --  Should not happen.
            raise Internal_Error;
         when Iir_Kind_Implicit_Function_Declaration
           | Iir_Kind_Function_Specification
           | Iir_Kind_Function_Spec_Body =>
            Finish_Sem_Function_Specification (Name, Expr);
         when others =>
            Error_Kind ("maybe_finish_sem_name", Expr);
      end case;
   end Maybe_Finish_Sem_Name;

   --  Convert name EXPR to an expression (ie, create function call).
   --  A_TYPE is the expected type of the expression, LOC is the genuine
   --  location of the name.
   function Name_To_Expression (Name : Iir; A_Type : Iir) return Iir
   is
      Ret_Type : Iir;
      Res_Type : Iir;
      Expr : Iir;
      Res : Iir;
      El : Iir;
   begin
      Expr := Get_Named_Entity (Name);
      if Get_Kind (Expr) = Iir_Kind_Error then
         return Null_Iir;
      end if;
      if Check_Is_Expression (Expr) = Null_Iir then
         return Null_Iir;
      end if;
      if Get_Kind (Expr) /= Iir_Kind_Overload_List then
         Maybe_Finish_Sem_Name (Name);
         Expr := Get_Named_Entity (Name);
         if A_Type /= Null_Iir then
            Res_Type := Get_Type (Expr);
            if Res_Type = Null_Iir then
               return Null_Iir;
            end if;
            if not Are_Basetypes_Compatible (Get_Base_Type (Res_Type), A_Type)
            then
               Not_Match (Expr, A_Type);
               return Null_Iir;
            end if;
         end if;
         --Set_Type (Name, Get_Type (Expr));
         --Set_Expr_Staticness (Name, Get_Expr_Staticness (Expr));
         return Expr;
      end if;

      --  Note: RES may contain procedure names...
      Remove_Procedures_From_List (Expr);

      if A_Type /= Null_Iir then
         --  Find the name returning A_TYPE.
         Res := Null_Iir;
         for I in Natural loop
            El := Get_Nth_Element (Expr, I);
            exit when El = Null_Iir;
            if Are_Basetypes_Compatible (Get_Base_Type (Get_Type (El)), A_Type)
            then
               Add_Result (Res, El);
            end if;
         end loop;
         if Res = Null_Iir then
            Not_Match (Name, A_Type);
            return Null_Iir;
         elsif Is_Overload_List (Res) then
            Error_Overload (Name);
            return Null_Iir;
         else
            Sem_Name_Free_Result (Expr, Res);
            Set_Named_Entity (Name, Res);
            Finish_Sem_Name (Name, Res);
            Maybe_Finish_Sem_Name (Name);
            Expr := Get_Named_Entity (Name);
            --Set_Type (Name, Get_Type (Expr));
            --Set_Expr_Staticness (Name, Get_Expr_Staticness (Expr));
            return Expr;
         end if;
      else
         --  Create a list of type.
         Ret_Type := Create_List_Of_Types (Expr);
         if Ret_Type = Null_Iir or else not Is_Overload_List (Ret_Type) then
            --  There is either no types or one type for
            --  several meanings.
            Error_Msg_Sem ("cannot resolve overload", Name);
            return Null_Iir;
         end if;
         Set_Type (Name, Ret_Type);
         return Name;
      end if;
   end Name_To_Expression;

   function Is_Object_Name (Name : Iir) return Boolean
   is
   begin
      case Get_Kind (Name) is
         when Iir_Kind_Alias_Declaration
           | Iir_Kind_Signal_Declaration
           | Iir_Kind_Guard_Signal_Declaration
           | Iir_Kind_Variable_Declaration
           | Iir_Kind_File_Declaration
           | Iir_Kind_Constant_Declaration
           | Iir_Kind_Iterator_Declaration
           | Iir_Kind_Constant_Interface_Declaration
           | Iir_Kind_Variable_Interface_Declaration
           | Iir_Kind_Signal_Interface_Declaration
           | Iir_Kind_File_Interface_Declaration
           | Iir_Kind_Slice_Name
           | Iir_Kind_Indexed_Name
           | Iir_Kind_Selected_Element
           | Iir_Kind_Implicit_Dereference
           | Iir_Kind_Dereference
           | Iir_Kind_Attribute_Value
           | Iir_Kind_Function_Call
           | Iir_Kinds_Signal_Attribute =>
            return True;
         when others =>
            return False;
      end case;
   end Is_Object_Name;

   function Check_Is_Expression (Expr : Iir) return Iir
   is
   begin
      if Expr = Null_Iir then
         return Null_Iir;
      end if;
      case Get_Kind (Expr) is
         when Iir_Kind_Type_Declaration
           | Iir_Kind_Subtype_Declaration
           | Iir_Kinds_Subtype_Definition
           | Iir_Kind_Design_Unit
           | Iir_Kind_Architecture_Declaration
           | Iir_Kind_Entity_Declaration
           | Iir_Kinds_Concurrent_Statement
           | Iir_Kinds_Sequential_Statement
           | Iir_Kind_Library_Declaration
           | Iir_Kind_Component_Declaration
           | Iir_Kinds_Procedure_Declaration =>
            Error_Msg_Sem (Disp_Node (Expr)
                           & " not allowed in an expression", Expr);
            return Null_Iir;
         when Iir_Kinds_Function_Specification =>
            return Expr;
         when Iir_Kind_Overload_List =>
            return Expr;
         when Iir_Kinds_Literal
           | Iir_Kind_Simple_Aggregate
           | Iir_Kind_Physical_Unit
           | Iir_Kind_Enumeration_Literal =>
            return Expr;
         when Iir_Kinds_Object_Declaration
           | Iir_Kind_Aggregate
           | Iir_Kind_Allocator_By_Expression
           | Iir_Kind_Allocator_By_Subtype
           | Iir_Kind_Qualified_Expression =>
            return Expr;
         when Iir_Kinds_Dyadic_Operator
           | Iir_Kinds_Monadic_Operator =>
            return Expr;
         when Iir_Kind_Slice_Name
           | Iir_Kind_Indexed_Name
           | Iir_Kind_Selected_Element
           | Iir_Kind_Dereference
           | Iir_Kind_Implicit_Dereference
           | Iir_Kinds_Expression_Attribute
           | Iir_Kind_Attribute_Value
           | Iir_Kind_Type_Conversion
           | Iir_Kind_Function_Call =>
            return Expr;
         when Iir_Kind_Simple_Name
           | Iir_Kind_Parenthesis_Name
           | Iir_Kind_Attribute
           | Iir_Kind_Selected_Name
           | Iir_Kind_Selected_By_All_Name =>
            return Expr;
         when others =>
            Error_Kind ("check_is_expression", Expr);
            --N := Get_Type (Expr);
            --return Expr;
      end case;
   end Check_Is_Expression;

   -- Find a type compatible with A_TYPE in TYPE_LIST (which can be an
   -- overload list or a simple type) and return it.
   -- In case of failure, return null.
   function Search_Overloaded_Type (Type_List: Iir; A_Type: Iir)
     return Iir
   is
      El: Iir;
   begin
      if not Is_Overload_List (Type_List) then
         if not Are_Types_Compatible (Type_List, A_Type) then
            return Null_Iir;
         else
            return Type_List;
         end if;
      else
         for I in Natural loop
            El := Get_Nth_Element (Type_List, I);
            exit when El = null;
            if Are_Types_Compatible (El, A_Type) then
               return A_Type;
            end if;
         end loop;
         return Null_Iir;
      end if;
   end Search_Overloaded_Type;

   --  LIST1, LIST2 are either a type node or an overload list of types.
   --  Return THE type which is compatible with LIST1 are LIST2.
   --  Return null_iir if there is no such type or if there are several types.
   function Search_Compatible_Type (List1, List2 : Iir) return Iir
   is
      Res : Iir;
      El : Iir;
      Tmp : Iir;
   begin
      if Is_Overload_List (List1) then
         Res := Null_Iir;
         for I in Natural loop
            El := Get_Nth_Element (List1, I);
            exit when El = Null_Iir;
            Tmp := Search_Overloaded_Type (List2, El);
            if Tmp /= Null_Iir then
               if Res = Null_Iir then
                  Res := Tmp;
               else
                  --  Several types match.
                  return Null_Iir;
               end if;
            end if;
         end loop;
         return Res;
      else
         return Search_Overloaded_Type (List2, List1);
      end if;
   end Search_Compatible_Type;

   function Sem_Type_Range (Expr : Iir; A_Type : Iir) return Iir
   is
      Expr_Type : Iir;
   begin
      Expr_Type := Get_Type (Expr);
      if Expr_Type = Null_Iir then
         return A_Type;
      end if;
      if Get_Kind (Expr_Type) in Iir_Kinds_Scalar_Type_Definition then
         return Expr_Type;
      end if;
      Expr_Type := Find_Declaration (Expr_Type, Decl_Type);
      if A_Type /= Null_Iir and then A_Type /= Expr_Type then
         -- This can happend when EXPR is an array subtype index subtype
         -- and A_TYPE is the array index type.
         Error_Msg_Sem ("subtype " & Disp_Node (Expr_Type)
                        & " doesn't match expected type "
                        & Disp_Node (A_Type), Expr);
      end if;
      return Expr_Type;
   end Sem_Type_Range;

   -- Semantize the range expression EXPR.
   -- If A_TYPE is not null_iir, EXPR is expected to be of type A_TYPE.
   -- LRM93 3.2.1.1
   -- FIXME: avoid to run it on an already semantized node, be careful
   --  with range_type_expr.
   function Sem_Range_Expression (Expr: Iir_Range_Expression; A_Type: Iir)
      return Iir_Range_Expression
   is
      Base_Type: Iir;
      Left, Right: Iir;
      Expr_Type : Iir;
   begin
      Expr_Type := Sem_Type_Range (Expr, A_Type);
      if Expr_Type /= Null_Iir then
         Base_Type := Get_Base_Type (Expr_Type);
      else
         Base_Type := Null_Iir;
      end if;

      Left := Get_Left_Limit (Expr);
      Right := Get_Right_Limit (Expr);
      Right := Sem_Expression_Ov (Right, Base_Type);
      Left := Sem_Expression_Ov (Left, Base_Type);
      if Left = Null_Iir or else Right = Null_Iir then
         return Null_Iir;
      end if;
      if Is_Overloaded (Left) or else Is_Overloaded (Right) then
         if Base_Type /= Null_Iir then
            --  Cannot happen, since sem_expression_ov should resolved
            --  ambiguties if a type is given.
            raise Internal_Error;
         end if;

         --  Try to find a common type.
         Base_Type := Search_Compatible_Type
           (Get_Type (Left), Get_Type (Right));
         if Base_Type = Null_Iir then
            Error_Not_Compatible (Expr);
            return Null_Iir;
         else
            Base_Type := Get_Base_Type (Base_Type);
         end if;
         Left := Sem_Expression_Ov (Left, Base_Type);
         Right := Sem_Expression_Ov (Right, Base_Type);
         if Left = Null_Iir or else Right = Null_Iir then
            return Null_Iir;
         end if;
      end if;
      Left := Eval_Expr_If_Static (Left);
      Right := Eval_Expr_If_Static (Right);
      Set_Left_Limit (Expr, Left);
      Set_Right_Limit (Expr, Right);
      Set_Expr_Staticness (Expr, Min (Get_Expr_Staticness (Left),
                                      Get_Expr_Staticness (Right)));
      if Expr_Type /= Null_Iir then
         Set_Type (Expr, Base_Type);
         if Get_Expr_Staticness (Expr) = Locally
           and then Get_Type_Staticness (Expr_Type) = Locally
           and then Get_Kind (Expr_Type) in Iir_Kinds_Subtype_Definition
         then
            Eval_Check_Range (Expr, Expr_Type);
         end if;
      else
         Base_Type := Get_Common_Basetype (Get_Base_Type (Get_Type (Left)),
                                           Get_Base_Type (Get_Type (Right)));
         if Base_Type = Null_Iir then
            Error_Not_Compatible (Expr);
            return Null_Iir;
         end if;
         Set_Type (Expr, Base_Type);
      end if;
      return Expr;
   end Sem_Range_Expression;

   -- Set semantic to expr.
   -- The result can be:
   --  a subtype definition
   --  a range attribute
   --  a range type definition
   -- LRM93 3.2.1.1
   -- FIXME: avoid to run it on an already semantized node, be careful
   --  with range_type_expr.
   function Sem_Discrete_Range_Expression (Expr: Iir; A_Type: Iir)
     return Iir
   is
      Res : Iir;
      Res_Type : Iir;
   begin
      if Get_Kind (Expr) = Iir_Kind_Range_Expression then
         Res := Sem_Range_Expression (Expr, A_Type);
         if Res = Null_Iir then
            return Null_Iir;
         end if;
         Res_Type := Get_Type (Res);
      else
         if Get_Kind (Expr) in Iir_Kinds_Name
           or else Get_Kind (Expr) = Iir_Kind_Attribute
         then
            Sem_Names (Expr);
            Maybe_Finish_Sem_Name (Expr);
            Res := Get_Named_Entity (Expr);
            if Res = Error_Mark then
               return Null_Iir;
            end if;
         else
            Res := Expr;
         end if;

         case Get_Kind (Res) is
            when Iir_Kind_Type_Declaration
              | Iir_Kind_Subtype_Declaration =>
               Res := Get_Type (Res);
               Res_Type := Res;
               if Get_Kind (Res) not in Iir_Kinds_Discrete_Type_Definition
               then
                  Error_Msg_Sem
                    ("type name must denote a discrete type", Expr);
                  return Null_Iir;
               end if;
            when Iir_Kind_Range_Array_Attribute
              | Iir_Kind_Reverse_Range_Array_Attribute =>
               Res_Type := Get_Type (Res);
               Res := Eval_Expr_If_Static (Res);
            when others =>
               Error_Msg_Sem ("name must denote a range", Expr);
               return Null_Iir;
         end case;
         if A_Type /= Null_Iir
           and then Get_Base_Type (Res_Type) /= Get_Base_Type (A_Type)
         then
            Not_Match (Expr, A_Type);
            return Null_Iir;
         end if;
      end if;

      if A_Type /= Null_Iir
        and then Get_Type_Staticness (A_Type) = Locally
        and then Get_Kind (A_Type) in Iir_Kinds_Subtype_Definition
      then
         case Get_Kind (Res) is
            when Iir_Kinds_Type_And_Subtype_Definition =>
               if Get_Type_Staticness (Res) = Locally then
                  Eval_Check_Range (Get_Range_Constraint (Res), A_Type);
               end if;
            when others =>
               if Get_Expr_Staticness (Res) = Locally then
                  Eval_Check_Range (Res, A_Type);
               end if;
         end case;
      end if;
      return Res;
   end Sem_Discrete_Range_Expression;

   function Get_Discrete_Range_Staticness (Expr : Iir) return Iir_Staticness is
   begin
      if Get_Kind (Expr) in Iir_Kinds_Discrete_Type_Definition then
         return Get_Type_Staticness (Expr);
      else
         return Get_Expr_Staticness (Expr);
      end if;
   end Get_Discrete_Range_Staticness;

   --  Return level of compatibility for type nodes LEFT and RIGHT.
   --  Return Via_Conversion if RIGHT is convertible to LEFT.
   function Compatibility (Left_Type, Right_Type : Iir)
     return Compatibility_Level
   is
      Right_Base_Type : Iir;
      Left_Base_Type : Iir;
   begin
      Right_Base_Type := Get_Base_Type (Right_Type);
      Left_Base_Type := Get_Base_Type (Left_Type);
      if Right_Base_Type = Left_Base_Type then
         return Full;
      end if;
      if Get_Kind (Left_Base_Type) = Iir_Kind_Integer_Type_Definition
        and then Right_Base_Type = Universal_Integer_Type_Definition
      then
         return Via_Conversion;
      end if;
      if Get_Kind (Left_Base_Type) = Iir_Kind_Floating_Type_Definition
        and then Right_Base_Type = Universal_Real_Type_Definition
      then
         return Via_Conversion;
      end if;
      return None;
   end Compatibility;

   function Compatibility_Types1 (Left_Type : Iir; Right_Types : Iir)
     return Compatibility_Level
   is
      El : Iir;
      Res : Compatibility_Level;
      Res1 : Compatibility_Level;
   begin
      if Is_Overload_List (Right_Types) then
         Res := None;
         for I in Natural loop
            El := Get_Nth_Element (Right_Types, I);
            exit when El = null;
            Res1 := Compatibility (Left_Type, El);
            case Res1 is
               when Full =>
                  return Full;
               when Via_Conversion =>
                  Res := Via_Conversion;
               when None =>
                  null;
            end case;
         end loop;
         return Res;
      else
         return Compatibility (Left_Type, Right_Types);
      end if;
   end Compatibility_Types1;

   --  Return level of compatibility for nodes LEFT and RIGHT.
   --  LEFT is expected to be an interface of a function definition.
   --  Type of RIGHT can be an overload_list
   --  RIGHT might be implicitly converted to LEFT.
   function Compatibility_Nodes (Left : Iir; Right : Iir)
     return Compatibility_Level
   is
      Left_Type, Right_Type : Iir;
   begin
      Left_Type := Get_Base_Type (Get_Type (Left));
      Right_Type := Get_Type (Right);

      --  Check.
      case Get_Kind (Left_Type) is
         when Iir_Kind_Floating_Type_Definition
           | Iir_Kind_Enumeration_Type_Definition
           | Iir_Kind_Integer_Type_Definition
           | Iir_Kind_Record_Type_Definition
           | Iir_Kind_File_Type_Definition
           | Iir_Kind_Physical_Type_Definition
           | Iir_Kind_Access_Type_Definition
           | Iir_Kind_Array_Type_Definition =>
            null;
         when others =>
            Error_Kind ("are_node_compatible_ov", Left_Type);
      end case;

      return Compatibility_Types1 (Left_Type, Right_Type);
   end Compatibility_Nodes;

   function Compatibility_Types (Left_Types : Iir; Right_Types : Iir)
     return Compatibility_Level
   is
      El : Iir;
      Res : Compatibility_Level;
      Res1 : Compatibility_Level;
   begin
      if Is_Overload_List (Left_Types) then
         Res := None;
         for I in Natural loop
            El := Get_Nth_Element (Left_Types, I);
            exit when El = null;
            Res1 := Compatibility_Types1 (El, Right_Types);
            case Res1 is
               when Full =>
                  return Full;
               when Via_Conversion =>
                  Res := Via_Conversion;
               when None =>
                  null;
            end case;
         end loop;
         return Res;
      else
         return Compatibility_Types1 (Left_Types, Right_Types);
      end if;
   end Compatibility_Types;

   --  Semantize all arguments of ASSOC_LIST (which can be NULL_IIR_LIST).
   --  Return TRUE if no error.
   function Sem_Association_List (Assoc_List : Iir_Association_List)
     return Boolean
   is
      Has_Named : Boolean;
      Ok : Boolean;
      Assoc : Iir;
      Res : Iir;
      Formal : Iir;
   begin
      -- Semantize all arguments
      -- OK is false if there is an error during semantic of one of the
      -- argument, but continue semantisation.
      if Assoc_List = Null_Iir_List then
         return True;
      end if;

      Has_Named := False;
      Ok := True;
      for I in Natural loop
         Assoc := Get_Nth_Element (Assoc_List, I);
         exit when Assoc = Null_Iir;
         Formal := Get_Formal (Assoc);
         if Formal /= Null_Iir then
            Has_Named := True;
            --  FIXME: check FORMAL is well composed.
         elsif Has_Named then
            Error_Msg_Sem ("positional argument after named argument", Assoc);
            Ok := False;
         end if;
         if Get_Kind (Assoc) /= Iir_Kind_Association_Element_Open then
            Res := Sem_Expression_Ov (Get_Actual (Assoc), null);
            if Res = Null_Iir then
               Ok := False;
            else
               Set_Actual (Assoc, Res);
            end if;
         end if;
      end loop;
      return Ok;
   end Sem_Association_List;

   --  Check class of EXPR may be associated with class of INTERFACE.
   procedure Check_Association_Class
     (Interface: Iir;
      Expr : Iir;
      Assoc : Iir_Association_Element_By_Expression)
   is
      Prefix : Iir;
   begin
      case Get_Kind (Interface) is
         when Iir_Kind_Signal_Interface_Declaration =>
            Prefix := Iirs_Utils.Get_Object_Prefix (Expr);
            --  LRM93 2.1.1
            --  In a subprogram call, the actual designator associated with a
            --  formal parameter of class signal must be a signal.
            case Get_Kind (Prefix) is
               when Iir_Kind_Signal_Interface_Declaration
                 | Iir_Kind_Signal_Declaration =>
                  --  LRM93 2.1.1.2
                  --  If an actual signal is associated with a signal parameter
                  --  of any mode, the actual must be denoted by a static
                  --  signal name.
                  if Get_Name_Staticness (Expr) < Globally then
                     Error_Msg_Sem ("actual signal must be static", Expr);
                  end if;
               when Iir_Kind_Guard_Signal_Declaration =>
                  null;
               when others =>
                  Error_Msg_Sem
                    ("signal parameter requires a signal expression", Assoc);
            end case;
         when Iir_Kind_Variable_Interface_Declaration =>
            Prefix := Iirs_Utils.Get_Object_Prefix (Expr);
            --  LRM93 2.1.1
            --  The actual designator associated with a formal of class
            --  variable must be a variable.
            case Get_Kind (Prefix) is
               when Iir_Kind_Variable_Interface_Declaration
                 | Iir_Kind_Variable_Declaration =>
                  null;
               when Iir_Kind_File_Interface_Declaration
                 | Iir_Kind_File_Declaration =>
                  --  LRM87 4.3.1.4
                  --  Such an object is a member of the variable class of
                  --  objects;
                  if Flags.Vhdl_Std >= Vhdl_93 then
                     Error_Msg_Sem
                       ("in vhdl 93, variable parameter cannot be a file",
                        Assoc);
                  end if;
               when others =>
                  Error_Msg_Sem ("variable parameter must be a variable",
                                 Assoc);
            end case;
         when Iir_Kind_File_Interface_Declaration =>
            Prefix := Iirs_Utils.Get_Object_Prefix (Expr);
            --  LRM93 2.1.1
            --  The actual designator associated with a formal of class file
            --  must be a file.
            case Get_Kind (Prefix) is
               when Iir_Kind_File_Interface_Declaration
                 | Iir_Kind_File_Declaration =>
                  null;
               when Iir_Kind_Variable_Declaration
                 | Iir_Kind_Variable_Interface_Declaration =>
                  if Flags.Vhdl_Std >= Vhdl_93 then
                     Error_Msg_Sem
                       ("in vhdl93, file parameter must be a file", Assoc);
                  end if;
               when others =>
                  Error_Msg_Sem ("file parameter must be a file", Assoc);
            end case;
         when Iir_Kind_Constant_Interface_Declaration =>
            --  LRM93 2.1.1
            --  The actual designator associated with a formal of class
            --  constant must be an expression.
            null;
         when others =>
            Error_Kind ("check_association_class", Interface);
      end case;
   end Check_Association_Class;

   procedure Check_Unassociated
     (Interface : Iir; Assoc : Iir; Is_Port : Boolean; Loc : Iir)
   is
      procedure Error (Str : String) is
      begin
         if Assoc = Null_Iir then
            Error_Msg_Sem (Str, Loc);
         else
            Error_Msg_Sem (Str, Assoc);
         end if;
      end Error;
   begin
      if Get_Default_Value (Interface) /= Null_Iir then
         return;
      end if;
      if not Is_Port then
         if Assoc = Null_Iir then
            Error_Msg_Sem ("no parameter for " & Disp_Node (Interface), Loc);
         else
            Error_Msg_Sem ("open association not allowed", Assoc);
         end if;
         return;
      end if;

      -- LRM93 1.1.1.2
      -- A port of mode IN may be unconnected or unassociated only if its
      -- declaration includes a default expression.
      case Get_Mode (Interface) is
         when Iir_In_Mode =>
            Error (Disp_Node (Interface) & " of mode IN must be connected");
         when Iir_Out_Mode
           | Iir_Linkage_Mode
           | Iir_Inout_Mode
           | Iir_Buffer_Mode =>
            if Is_Unconstrained_Type_Definition (Get_Type (Interface)) then
               Error ("unconstrained " & Disp_Node (Interface)
                      & " must be connected");
            end if;
         when Iir_Unknown_Mode =>
            raise Internal_Error;
      end case;
   end Check_Unassociated;

   --  EXPR is a formal or actual expression.
   --  Extract conversion function CONV from EXPR, if:
   --  * argument of the function is of type ARG_TYPE.
   --  * return type of the function is RES_TYPE if RES_TYPE /= Null_Iir
   --    or any type if RES_TYPE = Null_Iir.
   procedure Sem_Conversion
     (Expr : in out Iir; Conv : out Iir; Res_Type : Iir; Arg_Type : Iir)
   is
      Assoc : Iir;
      List : Iir_List;
   begin
      Conv := Null_Iir;
      case Get_Kind (Expr) is
         when Iir_Kind_Parenthesis_Name =>
            raise Internal_Error;
         when Iir_Kind_Function_Call =>
            null;
         when others =>
            return;
      end case;
      Conv := Get_Implementation (Expr);
      List := Get_Parameter_Association_List (Expr);
      Assoc := Get_First_Element (List);
      Expr := Get_Actual (Assoc);
      Free_Iir (Assoc);
      Free_Iir (List);
   end Sem_Conversion;

   --  Finish to semantize association ASSOC associated with INTERFACE.
   --  FORMAL_TYPES is a list of possibles formal types.
   --  FCONV_TYPES is a list of possibles formal conversion types.
   procedure Sem_Association
     (Assoc : Iir;
      Interface : Iir;
      Formal_Types : Iir;
      Fconv_Types : Iir;
      Is_Port : Boolean;
      Loc : Iir)
   is
      Formal, Actual : Iir;
      Expr : Iir;
      Out_Conv, In_Conv : Iir;
      Formal_Type : Iir;
   begin
      Set_Associated_Formal (Assoc, Interface);

      --  Semantize formal.
      Formal := Get_Formal (Assoc);
      if Formal = Null_Iir then
         Formal := Interface;
         Formal_Type := Get_Type (Interface);
      else
         if Fconv_Types /= Null_Iir then
            Sem_Conversion (Formal, Out_Conv, Formal_Types, Fconv_Types);
            Set_Out_Conversion (Assoc, Out_Conv);
         end if;
         Formal_Type := Formal_Types;
      end if;

      if Get_Kind (Assoc) /= Iir_Kind_Association_Element_Open then
         Actual := Get_Actual (Assoc);
      end if;

--       if Get_Kind (Assoc) = Iir_Kind_Association_Element_Open then
--          Formal_Type := Null_Iir;
--       else
--          Actual := Get_Actual (Assoc);
--          if Fconv_Types = Null_Iir then
--             Formal_Type := Formal_Types;
--          else
--             Formal_Type := Fconv_Types;
--          end if;
--          Formal_Type :=
--            Search_Compatible_Type (Formal_Type, Get_Type (Actual));

--          if Formal_Type = Null_Iir then
--             Error_Msg_Sem ("actual and formal types mismatch", Assoc);
--             return;
--          end if;
--       end if;
      Set_Formal (Assoc, Formal);

      --  Semantize actual.
      if Get_Kind (Assoc) = Iir_Kind_Association_Element_Open then
         Check_Unassociated (Interface, Assoc, Is_Port, Loc);
      else
         if Get_Kind (Interface) /= Iir_Kind_Constant_Interface_Declaration
         then
            Sem_Conversion
              (Actual, In_Conv, Get_Type (Interface), Formal_Type);
            Set_In_Conversion (Assoc, In_Conv);
         end if;
         Expr := Sem_Expression (Actual, Formal_Type);
         if Expr /= Null_Iir then
            Expr := Eval_Expr_If_Static (Expr);
            Set_Actual (Assoc, Expr);
            Check_Association_Class (Interface, Expr, Assoc);

            if Expr = Formal then
               --  This case is possible:
               --  Case (1):
               --   B: block
               --     generic         (size : INTEGER);
               --     generic map     (size => size);
               --     port            (P : in STRING(1 to size));
               --     port map        (P => S);
               --   begin
               --     [...]
               --   end block;
               --
               --  Case (2):
               --  procedure recurse (ctxt : integer; val : integer) is
               --  begin
               --    if val > 0 then
               --      recurse (ctxt, val - 1);
               --    end if;
               --  end;
               --
               --  Case (1) is not legal, but case (2) is.
               --  FIXME: find the rules which makes case (1) illegal.
               --  Note: case (1) may be more complex, eg:
               --   generic (a, b: natural);
               --   generic map (a => b, b => a);
               if False then
                  Error_Msg_Sem
                    ("a formal cannot be associed with itself", Expr);
               end if;
            end if;
         end if;
      end if;
   end Sem_Association;

   --  Return a type or a list of types of all possible conversion
   --  functions if the parenthesis name formal of ASSOC is a conversion
   --  function for INTERFACE.
   --  Return NULL_IIR if no match.
   --  If FINISH is true, then performs substitutions (and convert parenthesis
   --   name to function_call)
   function Conversion_Match (Assoc : Iir; Interface : Iir; Finish : Boolean)
     return Iir
   is
      List : Iir;
      El : Iir;
      Act : Iir;
      Res : Iir;
      Call : Iir;
      Inter : Iir;
      Name : Iir;

      function Extract_Return_Type (Func : Iir) return Iir is
      begin
         case Get_Kind (Func) is
            when Iir_Kinds_Function_Specification =>
               return Get_Return_Type (Func);
            when others =>
               return Null_Iir;
         end case;
      end Extract_Return_Type;
   begin
      Name := Get_Formal (Assoc);

      --  LRM93 4.3.2.2
      --  [...] whose single parameter of the function [...]
      List := Get_Association_List (Name);
      if Get_Nbr_Elements (List) /= 1 then
         return Null_Iir;
      end if;
      El := Get_First_Element (List);
      if Get_Kind (El) /= Iir_Kind_Association_Element_By_Expression then
         return Null_Iir;
      end if;

      --  LRM93 4.3.2.2
      --  [...] is the formal designator itself
      Act := Get_Actual (El);
      --  FIXME: the designator can be a slice, a selected or an indexed name.
      if Get_Kind (Act) /= Iir_Kind_Simple_Name
        or else Get_Identifier (Act) /= Get_Identifier (Interface)
      then
         return Null_Iir;
      end if;

      --  Get declarations for the prefix, extract return type.
      Inter := Get_Prefix (Name);
      Sem_Names (Inter);
      Inter := Get_Named_Entity (Inter);
      if Inter = Null_Iir then
         return Null_Iir;
      end if;
      if Is_Overload_List (Inter) then
         Res := Get_Overload_List;
         for I in Natural loop
            El := Get_Nth_Element (Inter, I);
            exit when El = Null_Iir;
            El := Extract_Return_Type (El);
            if El /= Null_Iir then
               Append_Element (Res, El);
            end if;
         end loop;
         Res := Simplify_Overload_List (Res);
      else
         Res := Extract_Return_Type (Inter);
      end if;
      if not Finish then
         if Is_Overload_List (Inter) then
            Free_Iir (Inter);
         end if;
         return Res;
      end if;

      Call := Create_Iir (Iir_Kind_Function_Call);
      Location_Copy (Call, Name);
      Set_Implementation (Call, Inter);
      Set_Base_Name (Call, Call);
      Set_Parameter_Association_List (Call, List);
      Free_Iir (Act);
      Set_Actual (El, Interface);
      Set_Type (Call, Res);
      Set_Formal (Assoc, Call);
      return Res;
   end Conversion_Match;

   --  Return a type or a list of types for a formal expression FORMAL
   --   corresponding to INTERFACE.  Possible cases are:
   --  * FORMAL is the simple name with the same identifier as INTERFACE,
   --    FORMAL_TYPE is set to the type of INTERFACE and CONV_TYPE is set
   --    to NULL_IIR.
   --  * FORMAL is a selected, indexed or slice name whose extreme prefix is
   --    a simple name with the same identifier as INTERFACE, FORMAL_TYPE
   --    is set to the type of the name, and CONV_TYPE is set to NULL_IIR.
   --  * FORMAL is a function call, whose only argument is an
   --    association_element_by_expression, whose actual is a name
   --    whose prefix is the same identifier as INTERFACE (note, since FORMAL
   --    is not semantized, this is parenthesis name), CONV_TYPE is set to
   --    the type or list of type of return type of conversion functions and
   --    FORMAL_TYPE is set to the type of the name.
   --  * otherwise, FORMAL cannot match INTERFACE and both FORMAL_TYPE and
   --    CONV_TYPE are set to NULL_IIR.
   --  If FINISH is true, the simple name is replaced by INTERFACE.
   procedure Sem_Formal (Assoc : Iir;
                         Interface : Iir;
                         Formal_Type : out Iir;
                         Conv_Type : out Iir;
                         Finish : Boolean)
   is
      Formal : Iir;
      Prefix : Iir;
      Suffix : Iir;
      Res : Iir;
      Name : Name_Id;
   begin
      Formal_Type := Null_Iir;
      Conv_Type := Null_Iir;
      Name := Get_Identifier (Interface);
      Formal := Get_Formal (Assoc);
      case Get_Kind (Formal) is
         when Iir_Kind_Simple_Name =>
            if Get_Identifier (Formal) = Name then
               if Finish then
                  Free_Iir (Formal);
                  Set_Formal (Assoc, Interface);
               end if;
               Set_Whole_Association_Flag (Assoc, True);
               Formal_Type := Get_Type (Interface);
            end if;
            return;
         when Iir_Kind_Selected_Name
           | Iir_Kind_Slice_Name =>
            null;
         when Iir_Kind_Parenthesis_Name =>
            if Get_Kind (Interface) /= Iir_Kind_Constant_Interface_Declaration
            then
               Res := Conversion_Match (Assoc, Interface, Finish);
               if Res /= Null_Iir then
                  Set_Whole_Association_Flag (Assoc, True);
                  Conv_Type := Res;
                  Formal_Type := Get_Type (Interface);
                  return;
               end if;
            end if;
            --  Maybe the prefix is a subelement of the interface.
            --  Will be checked just below.
         when others =>
            --  Should have been caught by sem_association_list.
            Error_Kind ("sem_formal", Formal);
      end case;
      --  Check for a sub-element.
      Set_Whole_Association_Flag (Assoc, False);
      Prefix := Get_Prefix (Formal);
      Suffix := Formal;
      loop
         case Get_Kind (Prefix) is
            when Iir_Kind_Simple_Name =>
               if Get_Identifier (Prefix) = Name then
                  if Finish then
                     Set_Named_Entity (Prefix, Interface);
                     Formal := Sem_Expression (Formal, Null_Iir);
                     if Formal /= Null_Iir then
                        Set_Formal (Assoc, Formal);
                        Formal_Type := Get_Type (Formal);
                     end if;
                     return;
                  end if;
               else
                  return;
               end if;
            when Iir_Kind_Selected_Name
              | Iir_Kind_Parenthesis_Name
              | Iir_Kind_Slice_Name =>
               null;
            when others =>
               Error_Kind ("sem_formal(2)", Prefix);
         end case;
         Suffix := Prefix;
         Prefix := Get_Prefix (Prefix);
      end loop;
   end Sem_Formal;

   --  Handle indexed name
   procedure Add_Individual_Assoc_Indexed_Name
     (Sub_Assoc : in out Iir; Formal : Iir)
   is
      Choice_List : Iir_List;
      Choice : Iir;
      Index_List : Iir_List;
      Index : Iir;
      Nbr : Natural;
   begin
      --  Find element.
      Index_List := Get_Index_List (Formal);
      Nbr := Get_Nbr_Elements (Index_List);
      for I in 0 .. Nbr - 1 loop
         Index := Get_Nth_Element (Index_List, I);

         --  Evaluate index.
         Index := Eval_Expr (Index);
         Replace_Nth_Element (Index_List, I, Index);

         --  Find index in choice list.
         Choice_List := Get_Association_List (Sub_Assoc);
         for J in Natural loop
            Choice := Get_Nth_Element (Choice_List, J);
            exit when Choice = Null_Iir;
            case Get_Kind (Choice) is
               when Iir_Kind_Association_Choice_By_Expression =>
                  if Eval_Pos (Get_Expression (Choice)) = Eval_Pos (Index) then
                     goto Found;
                  end if;
               when others =>
                  Error_Kind ("add_individual_assoc_index_name", Choice);
            end case;
         end loop;

         --  If not found, append it.
         Choice := Create_Iir (Iir_Kind_Association_Choice_By_Expression);
         Set_Expression (Choice, Index);
         Location_Copy (Choice, Formal);
         Append_Element (Choice_List, Choice);

         << Found >> null;

         if I < Nbr - 1 then
            Sub_Assoc := Get_Associated (Choice);
            if Sub_Assoc = Null_Iir then
               Sub_Assoc := Create_Iir
                 (Iir_Kind_Association_Element_By_Individual);
               Location_Copy (Sub_Assoc, Index);
               Set_Associated (Choice, Sub_Assoc);
            end if;
         else
            Sub_Assoc := Choice;
         end if;
      end loop;
   end Add_Individual_Assoc_Indexed_Name;

   procedure Add_Individual_Assoc_Slice_Name
     (Sub_Assoc : in out Iir; Formal : Iir)
   is
      Choice : Iir;
      Index : Iir;
   begin
      --  FIXME: handle cases such as param(5 to 6)(5)

      --  Find element.
      Index := Get_Suffix (Formal);

      --  Evaluate index.
      Index := Eval_Expr (Index);
      Set_Suffix (Formal, Index);

      Choice := Create_Iir (Iir_Kind_Association_Choice_By_Range);
      Location_Copy (Choice, Formal);
      Set_Expression (Choice, Index);
      Append_Element (Get_Association_List (Sub_Assoc), Choice);

      Sub_Assoc := Choice;
   end Add_Individual_Assoc_Slice_Name;

   procedure Add_Individual_Assoc_Selected_Name
     (Sub_Assoc : in out Iir; Formal : Iir)
   is
      Choice : Iir;
   begin
      Choice := Create_Iir (Iir_Kind_Association_Choice_By_Name);
      Location_Copy (Choice, Formal);
      Set_Name (Choice, Get_Selected_Element (Formal));
      Append_Element (Get_Association_List (Sub_Assoc), Choice);

      Sub_Assoc := Choice;
   end Add_Individual_Assoc_Selected_Name;

   procedure Add_Individual_Association_1 (Iassoc : in out Iir; Formal : Iir)
   is
      Sub : Iir;
   begin
      case Get_Kind (Formal) is
         when Iir_Kind_Indexed_Name
           | Iir_Kind_Slice_Name
           | Iir_Kind_Selected_Element =>
            Add_Individual_Association_1 (Iassoc, Get_Prefix (Formal));
         when Iir_Kinds_Interface_Declaration =>
            return;
         when others =>
            Error_Kind ("add_individual_association_1", Formal);
      end case;

      case Get_Kind (Iassoc) is
         when Iir_Kind_Association_Element_By_Individual =>
            null;
         when Iir_Kind_Association_Choice_By_Expression =>
            Sub := Get_Associated (Iassoc);
            if Sub = Null_Iir then
               Sub := Create_Iir (Iir_Kind_Association_Element_By_Individual);
               Location_Copy (Sub, Formal);
               Set_Associated_Formal (Sub, Formal);
               Set_Formal (Sub, Iassoc);
               Iassoc := Sub;
            else
               case Get_Kind (Sub) is
                  when Iir_Kind_Association_Element_By_Individual =>
                     Iassoc := Sub;
                  when others =>
                     Error_Msg_Sem
                       ("individual association of "
                        & Disp_Node (Get_Associated_Formal (Iassoc))
                        & " conflicts with that at " & Disp_Location (Sub),
                        Formal);
                     return;
               end case;
            end if;
         when others =>
            Error_Kind ("add_individual_association_1(2)", Iassoc);
      end case;

      case Get_Kind (Formal) is
         when Iir_Kind_Indexed_Name =>
            Add_Individual_Assoc_Indexed_Name (Iassoc, Formal);
         when Iir_Kind_Slice_Name =>
            Add_Individual_Assoc_Slice_Name (Iassoc, Formal);
         when Iir_Kind_Selected_Element =>
            Add_Individual_Assoc_Selected_Name (Iassoc, Formal);
         when others =>
            Error_Kind ("add_individual_association_1(3)", Formal);
      end case;
   end Add_Individual_Association_1;

   procedure Add_Individual_Association (Iassoc : Iir; Assoc : Iir)
   is
      Formal : Iir;
      Iass : Iir;
      Prev : Iir;
   begin
      Formal := Get_Formal (Assoc);
      Iass := Iassoc;
      Add_Individual_Association_1 (Iass, Formal);
      Prev := Get_Associated (Iass);
      if Prev /= null then
         Error_Msg_Sem ("individual association of "
                        & Disp_Node (Get_Associated_Formal (Assoc))
                        & " conflicts with that at " & Disp_Location (Prev),
                        Assoc);
      else
         Set_Associated (Iass, Get_Actual (Assoc));
      end if;
   end Add_Individual_Association;

   procedure Finish_Individual_Assoc_Array (Assoc : Iir; Atype : Iir)
   is
      Index_Tlist : Iir;
      Index_Type : Iir;
      Low, High : Iir;
   begin
      Index_Tlist := Get_Index_List (Atype);
      for I in Natural loop
         Index_Type := Get_Nth_Element (Index_Tlist, I);
         exit when Index_Type = Null_Iir;
         Sem_Choices_Range (Get_Association_List (Assoc), Index_Type, True,
                            Get_Location (Assoc), Low, High);
      end loop;
   end Finish_Individual_Assoc_Array;

   procedure Finish_Individual_Assoc_Record (Assoc : Iir; Atype : Iir)
   is
      El_List : Iir_List;
      Matches : Iir_Array_Acc;
      Assoc_List : Iir_List;
      Ch : Iir;
      Pos : Natural;
      Rec_El : Iir;
   begin
      El_List := Get_Element_Declaration_List (Get_Base_Type (Atype));
      Matches := new Iir_Array (0 .. Get_Nbr_Elements (El_List) -1);
      Matches.all := (others => Null_Iir);
      Assoc_List := Get_Association_List (Assoc);
      for I in Natural loop
         Ch := Get_Nth_Element (Assoc_List, I);
         exit when Ch = Null_Iir;
         Rec_El := Get_Name (Ch);
         Pos := Natural (Get_Element_Position (Rec_El));
         if Matches (Pos) /= Null_Iir then
            Error_Msg_Sem ("individual " & Disp_Node (Rec_El)
                           & " already associated at "
                           & Disp_Location (Matches (Pos)), Ch);
         else
            Matches (Pos) := Ch;
         end if;
      end loop;
      for I in Matches'Range loop
         if Matches (I) = Null_Iir then
            Rec_El := Get_Nth_Element (El_List, I);
            Error_Msg_Sem (Disp_Node (Rec_El) & " not associated", Assoc);
         end if;
      end loop;
   end Finish_Individual_Assoc_Record;

   procedure Finish_Individual_Assoc (Assoc : Iir; Atype : Iir)
   is
   begin
      case Get_Kind (Atype) is
         when Iir_Kind_Array_Subtype_Definition =>
            Finish_Individual_Assoc_Array (Assoc, Atype);
         when Iir_Kind_Record_Type_Definition
           | Iir_Kind_Record_Subtype_Definition =>
            Finish_Individual_Assoc_Record (Assoc, Atype);
         when others =>
            Error_Kind ("finish_individual_assoc", Atype);
      end case;
   end Finish_Individual_Assoc;

   procedure Finish_Individual_Association (Assoc : Iir)
   is
      Formal : Iir;
      Atype : Iir;
   begin
      --  Guard.
      if Assoc = Null_Iir then
         return;
      end if;

      Formal := Get_Associated_Formal (Assoc);
      Atype := Get_Type (Formal);
      Finish_Individual_Assoc (Assoc, Atype);
   end Finish_Individual_Association;

   procedure Sem_Individual_Association (Assocs : Iir_List)
   is
      Assoc : Iir;
      Iassoc : Iir;
      Cur_Iface : Iir;
      Skip_Next : Boolean;
   begin
      Iassoc := Null_Iir;
      Cur_Iface := Null_Iir;
      Skip_Next := False;
      for I in Natural loop
         Assoc := Get_Nth_Element (Assocs, I);
         exit when Assoc = Null_Iir;
         if Get_Associated_Formal (Assoc) /= Cur_Iface then
            Finish_Individual_Association (Iassoc);
            Cur_Iface := Get_Associated_Formal (Assoc);
            Iassoc := Null_Iir;
         end if;
         if Get_Whole_Association_Flag (Assoc) = False then
            if Iassoc = Null_Iir then
               Iassoc :=
                 Create_Iir (Iir_Kind_Association_Element_By_Individual);
               Location_Copy (Iassoc, Assoc);
               Set_Associated_Formal (Iassoc, Cur_Iface);
               Set_Formal (Iassoc, Cur_Iface);
               --  Insert IASSOC at position I.
               declare
                  Nbr : Natural;
               begin
                  Nbr := Get_Nbr_Elements (Assocs);
                  Append_Element (Assocs, Get_Nth_Element (Assocs, Nbr - 1));
                  for J in reverse I + 1 .. Nbr - 1 loop
                     Replace_Nth_Element
                       (Assocs, J, Get_Nth_Element (Assocs, J - 1));
                  end loop;
                  Replace_Nth_Element (Assocs, I, Iassoc);
               end;
               Skip_Next := True;
               Add_Individual_Association (Iassoc, Assoc);
            else
               if Skip_Next then
                  Skip_Next := False;
               else
                  Add_Individual_Association (Iassoc, Assoc);
               end if;
            end if;
         end if;
      end loop;
      Finish_Individual_Association (Iassoc);
   end Sem_Individual_Association;

   --  ASSOC_LIST (if not NULL_IIR_LIST) must already have been semantized
   --   by sem_association_list.
   --  Return the level compatibility between INTERFACE_LIST and ASSOC_LIST.
   --  If FINISH is true, then error messages are displayed (to explain
   --   incompatibility).
   --  If CHECK_MATCH is true, all interface must have an association or
   --   a default value, otherwise, they can be unassociated.
   --  IS_PORT is used for sem_association.
   function Sem_Subprogram_Arguments
     (Interface_List : Iir_List;
      Assoc_List: Iir_List;
      Finish: Boolean;
      Check_Match : Boolean;
      Is_Port : Boolean;
      Loc : Iir)
      return Compatibility_Level
   is
      Res : Compatibility_Level;
      Assoc: Iir;
      Interface: Iir;

      --  Associate ASSOC with interface INTERFACE
      --  This sets RES.
      procedure Try_Association (Formal_Types : Iir; Fconv_Type : Iir)
      is
         Actual: Iir;
         Actual_Type : Iir;
         Aconv_Type : Iir;
         Cassoc_List : Iir_List;
         Cassoc : Iir;
      begin
         if Get_Kind (Assoc) /= Iir_Kind_Association_Element_Open then
            Actual := Get_Actual (Assoc);
            Actual_Type := Get_Type (Actual);
            if Actual_Type = Null_Iir then
               Res := None;
               return;
            end if;
            if Get_Kind (Actual) = Iir_Kind_Function_Call
              and then Get_Kind (Interface)
              /= Iir_Kind_Constant_Interface_Declaration
            then
               Cassoc_List := Get_Parameter_Association_List (Actual);
               if Cassoc_List /= Null_Iir_List
                 and then Get_Nbr_Elements (Cassoc_List) = 1
               then
                  Cassoc := Get_Nth_Element (Cassoc_List, 0);
                  if Get_Kind (Cassoc)
                    = Iir_Kind_Association_Element_By_Expression
                  then
                     Aconv_Type := Actual_Type;
                     Actual_Type := Get_Type (Get_Actual (Cassoc));
                  end if;
               end if;
            end if;

            --  4 cases:
            --    A  => B     type of A = type of B
            --  F(A) => B     type of B = type of F
            --    A  => G(B)  type of A = type of G
            --  F(A) => G(B)  type of B = type of F, type of A = type of G
            if Fconv_Type = Null_Iir and then Aconv_Type = Null_Iir then
               Res := Compatibility_Level'Min
                 (Res, Compatibility_Types (Formal_Types, Actual_Type));
            else
               if Fconv_Type /= Null_Iir then
                  Res := Compatibility_Level'Min
                    (Res, Compatibility_Types (Fconv_Type, Actual_Type));
               end if;
               if Aconv_Type /= Null_Iir then
                  Res := Compatibility_Level'Min
                    (Res, Compatibility_Types (Formal_Types, Aconv_Type));
               end if;
            end if;

            if Res = None then
               if Finish then
                  Not_Match (Actual, Interface);
               end if;
               return;
            end if;
         end if;
         if Finish then
            Sem_Association
              (Assoc, Interface, Formal_Types, Fconv_Type, Is_Port, Loc);
         end if;
      end Try_Association;

      type Param_Assoc_Type is (None, Individual, Whole);
      type Bool_Array is array (Natural range <>) of Param_Assoc_Type;
      Nbr_Arg: constant Natural := Get_Nbr_Elements_Safe (Interface_List);
      Arg_Matched: Bool_Array (0 .. Nbr_Arg - 1) := (others => None);
      Pos: Integer;
      Last_Individual : Iir;
      Formal_Types : Iir;
      Conv_Type : Iir;
      Has_Individual : Boolean;
   begin
      Res := Full;
      Has_Individual := False;

      if Assoc_List /= Null_Iir_List then
         -- Loop on every assoc element, try to match it.
         Pos := 0;
         Last_Individual := Null_Iir;

         for I in Natural loop
            Assoc := Get_Nth_Element (Assoc_List, I);
            exit when Assoc = Null_Iir;

            if Get_Formal (Assoc) = Null_Iir then
               -- Positional argument.
               if Pos = -1 then
                  -- Can't have a positional argument after a named argument.
                  -- This must have been caught by sem_association_list.
                  return None;
               end if;
               -- Try to match actual of ASSOC with the interface.
               Interface := Get_Nth_Element (Interface_List, Pos);
               if Interface = Null_Iir then
                  if Finish then
                     Error_Msg_Sem
                       ("too many arguments for " & Disp_Node (Loc), Assoc);
                  end if;
                  return None;
               end if;
               Try_Association (Get_Type (Interface), Null_Iir);
               if Res = None then
                  return None;
               end if;
               Arg_Matched (Pos) := Whole;
               Set_Whole_Association_Flag (Assoc, True);
               Pos := Pos + 1;
            else
               -- Named association.
               Pos := -1;
               -- FIXME: directly search the formal if finish is true.
               -- Find the Interface.
               for J in 0 .. Nbr_Arg - 1 loop
                  Interface := Get_Nth_Element (Interface_List, J);

                  -- Formal assoc is not necessarily a simple name, it may
                  -- be a conversion function, or even an indexed or
                  -- selected name.
                  Sem_Formal (Assoc, Interface,
                              Formal_Types, Conv_Type, Finish);
                  if Formal_Types /= Null_Iir then
                     Try_Association (Formal_Types, Conv_Type);
                     if Res = None then
                        return None;
                     end if;
                     if Get_Whole_Association_Flag (Assoc) then
                        Last_Individual := Null_Iir;
                        if Arg_Matched (J) = None then
                           Arg_Matched (J) := Whole;
                        else
                           if Finish then
                              Error_Msg_Sem
                                (Disp_Node (Interface) & " already associated",
                                 Assoc);
                              return None;
                           end if;
                        end if;
                     else
                        Has_Individual := True;
                        if Arg_Matched (J) /= Whole then
                           if Finish
                             and then Arg_Matched (J) = Individual
                             and then Last_Individual /= Interface
                           then
                              Error_Msg_Sem
                                ("non consecutive individual association for "
                                 & Disp_Node (Interface),
                                 Assoc);
                              return None;
                           end if;
                           Last_Individual := Interface;
                           Arg_Matched (J) := Individual;
                        else
                           if Finish then
                              Error_Msg_Sem
                                (Disp_Node (Interface) & " already associated",
                                 Assoc);
                           end if;
                           return None;
                        end if;
                     end if;
                     goto Found;
                  end if;
               end loop;

               -- Not found.
               if Finish then
                  Error_Msg_Sem ("no such interface name", Assoc);
               end if;
               return None;

               << Found >> null;
            end if;
         end loop;
      end if;

      if Finish and then Has_Individual then
         Sem_Individual_Association (Assoc_List);
      end if;

      if not Check_Match or else Interface_List = Null_Iir then
         return Res;
      end if;

      for I in Natural loop
         Interface := Get_Nth_Element (Interface_List, I);
         exit when Interface = Null_Iir;
         if Arg_Matched (I) = None
           and then Get_Default_Value (Interface) = Null_Iir
         then
            if Finish then
               Error_Msg_Sem ("no actual for " & Disp_Node (Interface), Loc);
            end if;
            return None;
         end if;
      end loop;
      return Res;
   end Sem_Subprogram_Arguments;

   procedure Set_Function_Call_Staticness (Expr : Iir; Imp : Iir)
   is
      Staticness : Iir_Staticness;
   begin
      --  LRM93 7.4.1 (Locally Static Primaries)
      --  4. a function call whose function name denotes an implicitly
      --     defined operator, and whose actual parameters are each
      --     locally static expressions;
      --
      --  LRM93 7.4.2 (Globally Static Primaries)
      --  9. a function call whose function name denotes a pure function,
      --     and whose actual parameters are each globally static
      --     expressions.
      case Get_Kind (Expr) is
         when Iir_Kinds_Monadic_Operator =>
            Staticness := Get_Expr_Staticness (Get_Operand (Expr));
         when Iir_Kinds_Dyadic_Operator =>
            Staticness := Min (Get_Expr_Staticness (Get_Left (Expr)),
                               Get_Expr_Staticness (Get_Right (Expr)));
         when Iir_Kind_Function_Call =>
            Staticness := Locally;
            declare
               Assoc_List : Iir_Association_List;
               Assoc : Iir;
            begin
               Assoc_List := Get_Parameter_Association_List (Expr);
               if Assoc_List /= null then
                  for I in Natural loop
                     Assoc := Get_Nth_Element (Assoc_List, I);
                     exit when Assoc = null;
                     if Get_Kind (Assoc) /= Iir_Kind_Association_Element_Open
                     then
                        Staticness := Min
                          (Get_Expr_Staticness (Get_Actual (Assoc)),
                           Staticness);
                     end if;
                  end loop;
               end if;
            end;
         when Iir_Kind_Procedure_Call
           | Iir_Kind_Concurrent_Procedure_Call_Statement =>
            return;
         when others =>
            Error_Kind ("set_function_call_staticness (1)", Expr);
      end case;
      case Get_Kind (Imp) is
         when Iir_Kind_Implicit_Function_Declaration =>
            if Get_Implicit_Definition (Imp)
              not in Iir_Predefined_Pure_Functions
            then
               --  Predefined functions such as Now, Endfile are not static.
               Staticness := None;
            end if;
         when Iir_Kinds_User_Function_Specification =>
            if Get_Pure_Flag (Imp) then
               Staticness := Min (Staticness, Globally);
            else
               Staticness := None;
            end if;
         when others =>
            Error_Kind ("set_function_call_staticness (2)", Imp);
      end case;
      Set_Expr_Staticness (Expr, Staticness);
   end Set_Function_Call_Staticness;

   --  Set IMP as the implementation to being called by EXPR.
   --  If the context is a subprogram or a process (ie, if current_subprogram
   --  is not NULL), then mark IMP as callee of current_subprogram, and
   --  update states.
   procedure Sem_Subprogram_Call_Finish (Expr : Iir; Imp : Iir)
   is
      use Sem;
      List : Iir_Callees_List;
      Subprg : Iir := Get_Current_Subprogram;
   begin
      Set_Implementation (Expr, Imp);
      Set_Function_Call_Staticness (Expr, Imp);

      if Subprg = null then
         return;
      end if;
      if Subprg = Imp then
         --  Recursive call.
         return;
      end if;

      case Get_Kind (Imp) is
         when Iir_Kind_Implicit_Procedure_Declaration
           | Iir_Kind_Implicit_Function_Declaration =>
            if Get_Implicit_Definition (Imp)
              in Iir_Predefined_Pure_Functions
            then
               return;
            end if;
         when Iir_Kinds_User_Procedure_Specification
           | Iir_Kinds_User_Function_Specification =>
            List := Get_Callees_List (Subprg);
            if List = null then
               List := Create_Iir (Iir_Kind_Callees_List);
               Set_Callees_List (Subprg, List);
            end if;
            --  FIXME: May use a flag in IMP to speed up the add operation.
            Add_Element (List, Imp);
         when others =>
            raise Internal_Error;
      end case;
      --  Update states.
      if Get_Passive_State (Imp) = False then
         Set_Passive_State (Subprg, False);
         if Get_Kind (Subprg) in Iir_Kinds_Process_Statement
           and then Get_Passive_Flag (Subprg) = True
         then
            Error_Msg_Sem
              (Disp_Node (Subprg) & " must be passive, but calls non-passive "
               & Disp_Node (Imp), Expr);
         end if;
      end if;
      if Get_Wait_State (Imp) = True then
         Set_Wait_State (Subprg, True);
         if Get_Kind (Subprg) = Iir_Kind_Sensitized_Process_Statement then
            Warning_Msg_Sem
              (Disp_Node (Subprg) & " must not contains wait, but calls "
               & Disp_Node (Imp) & " which has a wait", Expr);
         end if;
      end if;
      if Get_Purity_State (Imp) = False then
         Set_Purity_State (Subprg, False);
         if Get_Kind (Subprg) in Iir_Kinds_User_Function_Specification
           and then Get_Pure_Flag (Subprg) = True
         then
            Error_Msg_Sem
              ("call to (impure) " & Disp_Node (Imp) & " makes "
               & Disp_Node (Subprg) & " impure", Expr);
         end if;
      end if;
   end Sem_Subprogram_Call_Finish;

   function Sem_Subprogram_Call_Stage1
     (Expr : Iir; A_Type : Iir; Is_Func_Call : Boolean)
     return Iir
   is
      Nbr_Inter: Natural;
      A_Func: Iir;
      Inter_List: Iir;
      Assoc_List: Iir;
      Res_Type: Iir;
      Inter: Iir;
   begin
      Nbr_Inter := 0;
      Inter_List := Get_Implementation (Expr);
      Assoc_List := Get_Parameter_Association_List (Expr);

      for I in Natural loop
         A_Func := Get_Nth_Element (Inter_List, I);
         exit when A_Func = null;

         --  The identifier of a function call must be a function or an
         --  enumeration literal.
         if Is_Func_Call and then not
           (Get_Kind (A_Func) in Iir_Kinds_Function_Specification
            or else Get_Kind (A_Func) = Iir_Kind_Implicit_Function_Declaration
            or else Get_Kind (A_Func) = Iir_Kind_Enumeration_Literal)
         then
            goto Continue;
         end if;

         --  The identifier of a procedure call must be a procedure.
         if not Is_Func_Call and then not
           (Get_Kind (A_Func) in Iir_Kinds_Procedure_Specification
            or else
            Get_Kind (A_Func) = Iir_Kind_Implicit_Procedure_Declaration)
         then
            goto Continue;
         end if;

         --  Keep this interpretation only if compatible.
         if (A_Type = null or else
           Compatibility_Nodes (A_Type, Get_Return_Type (A_Func)) /= None)
           and then
           Sem_Subprogram_Arguments
           (Get_Interface_Declaration (A_Func),
            Assoc_List, False, True, False, Expr) /= None
         then
            Replace_Nth_Element (Inter_List, Nbr_Inter, A_Func);
            Nbr_Inter := Nbr_Inter + 1;
         end if;

         << Continue >> null;
      end loop;
      Set_Nbr_Elements (Inter_List, Nbr_Inter);

      -- Set_Implementation (Expr, Inter_List);
      -- A set of possible functions to call is in INTER_LIST.
      -- Create a set of possible return type in RES_TYPE.
      case Nbr_Inter is
         when 0 =>
            Error_Msg_Sem ("can't find a subprogram for this overload call",
                           Expr);
            return null;
         when 1 =>
            -- Very simple case: no overloading.
            Inter := Get_First_Element (Inter_List);
            Free_Iir (Inter_List);
            if Is_Func_Call then
               Set_Type (Expr, Get_Return_Type (Inter));
            end if;
            if Sem_Subprogram_Arguments
              (Get_Interface_Declaration (Inter), Assoc_List,
               True, True, False, Expr) = None
            then
               raise Internal_Error;
            end if;
            Sem_Subprogram_Call_Finish (Expr, Inter);
            return Expr;
         when others =>
            if Is_Func_Call then
               if A_Type /= null then
                  -- Cannot find a single interpretation for a given
                  -- type.
                  Error_Overload (Expr);
                  return null;
               end if;
               Res_Type := Get_Overload_List;
               for I in 0 .. Nbr_Inter - 1 loop
                  Add_Element
                    (Res_Type,
                     Get_Return_Type (Get_Nth_Element (Inter_List, I)));
               end loop;
               if Get_Nbr_Elements (Res_Type) = 1 then
                  -- several implementations but one profile.
                  Error_Overload (Expr);
                  return null;
               end if;
               Set_Type (Expr, Res_Type);
            else
               Error_Overload (Expr);
            end if;
            return Expr;
      end case;
   end Sem_Subprogram_Call_Stage1;

   -- For a procedure call, A_TYPE must be null.
   --  Associations must have already been semantized by sem_association_list.
   function Sem_Subprogram_Call (Expr: Iir; A_Type: Iir) return Iir
   is
      Is_Func: Boolean;
      Res_Type: Iir;
      Res: Iir;
      Inter_List: Iir;
      Inter: Iir;
   begin
      Is_Func := Get_Kind (Expr) = Iir_Kind_Function_Call;

      if Is_Func then
         Res_Type := Get_Type (Expr);
      end if;

      if not Is_Func or else Res_Type = Null_Iir then
         -- First call to sem_subprogram_call.
         -- Create the list of possible implementation and possible
         -- return types, according to arguments and A_TYPE.

         -- Select possible interpretations among all interpretations.
         -- NOTE: the list of possible implementations was already created
         --  during the transformation of iir_kind_parenthesis_name to
         --  iir_kind_function_call.
         Inter_List := Get_Implementation (Expr);
         if Is_Overload_List (Inter_List) then
            return Sem_Subprogram_Call_Stage1 (Expr, A_Type, Is_Func);
         else
            if Is_Func then
               if Get_Kind (Inter_List) not in Iir_Kinds_Function_Specification
               then
                  Error_Msg_Sem ("identifier is not a function", Expr);
                  return null;
               end if;
            else
               if Get_Kind (Inter_List) not in
                 Iir_Kinds_Procedure_Specification
                 and then Get_Kind (Inter_List) /=
                 Iir_Kind_Implicit_Procedure_Declaration
               then
                  Error_Msg_Sem ("identifier is not a procedure", Expr);
                  return null;
               end if;
            end if;
            if Sem_Subprogram_Arguments
              (Get_Interface_Declaration (Inter_List),
               Get_Parameter_Association_List (Expr),
               True, True, False, Expr) = None
            then
               --  No need to disp an error message, this is done by
               --  sem_subprogram_arguments.
               return null;
            end if;
            if Is_Func then
               Set_Type (Expr, Get_Return_Type (Inter_List));
            end if;
            Sem_Subprogram_Call_Finish (Expr, Inter_List);
            return Expr;
         end if;
      end if;

      if Is_Func and then A_Type = null then
         -- Impossible case: second call to sem_function_call, without
         -- A_TYPE set.
         raise Internal_Error;
      end if;

      -- The implementation list was set.
      -- The return type was set.
      -- A_TYPE is not null, A_TYPE is *the* return type.

      Inter_List := Get_Implementation (Expr);

      -- Find a single implementation.
      Res := null;
      if Is_Func then
         if Is_Overload_List (Inter_List) then
            -- INTER_LIST is a list of possible declaration to call.
            -- Find one, based on the return type A_TYPE.
            for I in Natural loop
               Inter := Get_Nth_Element (Inter_List, I);
               exit when Inter = null;
               if Are_Basetypes_Compatible
                 (A_Type, Get_Base_Type (Get_Return_Type (Inter)))
               then
                  if Res /= null then
                     Error_Overload (Expr);
                     return null;
                  else
                     Res := Inter;
                  end if;
               end if;
            end loop;
         else
            if Are_Basetypes_Compatible
              (Get_Base_Type (Get_Return_Type (Inter_List)), A_Type)
            then
               Res := Inter_List;
            end if;
         end if;
         if Res = null then
            Not_Match (Expr, A_Type);
            return null;
         end if;

         -- Clean up.
         if Res_Type /= null and then Is_Overload_List (Res_Type) then
            Free_Iir (Res_Type);
         end if;
      else
         -- a procedure call.
         if Is_Overload_List (Inter_List) then
            Error_Overload (Expr);
            return Null_Iir;
         else
            Res := Inter_List;
         end if;
      end if;

      if Is_Overload_List (Inter_List) then
         Free_Iir (Inter_List);
      end if;

      --  Simple case: this is not a call to a function, but an enumeration
      --  literal.
      if Get_Kind (Res) = Iir_Kind_Enumeration_Literal then
         -- Free_Iir (Expr);
         return Res;
      end if;

      -- Set types.
      if Is_Func then
         Set_Type (Expr, Get_Return_Type (Res));
      end if;
      if Sem_Subprogram_Arguments (Get_Interface_Declaration (Res),
                                   Get_Parameter_Association_List (Expr),
                                   True, True, False, Expr) = None
      then
         return Null_Iir;
      end if;
      Sem_Subprogram_Call_Finish (Expr, Res);
      return Expr;
   end Sem_Subprogram_Call;

   procedure Sem_Procedure_Call (Stmt: Iir)
   is
      use Iirs_Utils;

      Imp: Iir;
      Name : Iir;
      Parameters : Iir_List;
      Param : Iir;
      Formal : Iir;
      Prefix : Iir;
   begin
      Name := Get_Implementation (Stmt);
      Sem_Names (Name);
      Imp := Get_Named_Entity (Name);
      if Imp = Null_Iir then
         return;
      end if;
      Set_Implementation (Stmt, Imp);
      Parameters := Get_Parameter_Association_List (Stmt);
      if Sem_Association_List (Parameters) = False then
         return;
      end if;
      if Sem_Subprogram_Call (Stmt, null) /= Stmt then
         return;
      end if;
      if Is_Overload_List (Get_Implementation (Stmt)) then
         --  Failed to resolve overload.
         return;
      end if;

      --  Add a driver for each signal actual parameter associated with a
      --  formal out or inout signal parameter.
      --  This is *NOT* explicitly allowed by LRM, but often used...
      if Parameters = Null_Iir then
         return;
      end if;
      for I in Natural loop
         Param := Get_Nth_Element (Parameters, I);
         exit when Param = Null_Iir;
         Formal := Get_Associated_Formal (Param);
         if Get_Kind (Formal) = Iir_Kind_Signal_Interface_Declaration
           and then Get_Mode (Formal) in Iir_Out_Mode .. Iir_Inout_Mode
         then
            Prefix := Get_Longuest_Static_Prefix (Get_Actual (Param));
            Sem.Sem_Add_Driver (Prefix, Stmt);
         end if;
      end loop;
   end Sem_Procedure_Call;

   --  List must be an overload list containing two functions declarations.
   --  If both have the same profile and only one is an implicit function,
   --  return the explicit function.
   --  Otherwise, return NULL_IIR.
   function Get_Non_Implicit_Function (List : Iir_List) return Iir
   is
      D1, D2 : Iir;
      R : Iir;
   begin
      D1 := Get_Nth_Element (List, 0);
      D2 := Get_Nth_Element (List, 1);
      if Get_Kind (D1) = Iir_Kind_Implicit_Function_Declaration
        and then
        Get_Kind (D2) in Iir_Kinds_User_Function_Specification
      then
         R := D2;
      elsif Get_Kind (D2) = Iir_Kind_Implicit_Function_Declaration
        and then
        Get_Kind (D1) in Iir_Kinds_User_Function_Specification
      then
         R := D1;
      else
         return Null_Iir;
      end if;
      if Get_Base_Type (Get_Return_Type (D1))
        /= Get_Base_Type (Get_Return_Type (D2))
      then
         return Null_Iir;
      end if;
      if Get_Parent (D1) /= Get_Parent (D1) then
         return Null_Iir;
      end if;
      return R;
   end Get_Non_Implicit_Function;

   function Sem_Operator (Expr : Iir; Res_Type : Iir; Arity : Positive)
      return Iir
   is
      Operator : Name_Id;
      Left, Right: Iir;
      Interpretation : Name_Interpretation_Type;
      Decl : Iir;
      Interface_List : Iir_List;
      Overload_List : Iir_List;
      Res_Type_List : Iir_List;
      Compat : Compatibility_Level;
      Nbr_Full_Compat : Natural;
      Full_Compat : Iir;

      -- LEFT and RIGHT must be set.
      function Set_Uniq_Interpretation (Decl : Iir) return Iir
      is
      begin
         Set_Type (Expr, Get_Return_Type (Decl));
         Interface_List := Get_Interface_Declaration (Decl);
         if Is_Overload_List (Get_Type (Left)) then
            Left := Sem_Expression_Ov
              (Left, Get_Type (Get_Nth_Element (Interface_List, 0)));
            if Arity = 1 then
               Set_Operand (Expr, Left);
            else
               Set_Left (Expr, Left);
            end if;
         end if;
         if Arity = 2 and then Is_Overload_List (Get_Type (Right)) then
            Right := Sem_Expression_Ov
              (Right, Get_Type (Get_Nth_Element (Interface_List, 1)));
            Set_Right (Expr, Right);
         end if;
         Sem_Subprogram_Call_Finish (Expr, Decl);
         Free_Iir (Overload_List);
         return Eval_Expr_If_Static (Expr);
         --return Expr;
      end Set_Uniq_Interpretation;

      --  Note: operator and implementation node of expr must be set.
      procedure Error_Operator_Overload (List : Iir_Overload_List)
      is
         El : Iir;
      begin
         Error_Msg_Sem ("operator """ & Name_Table.Image (Operator)
                        & """ is overloaded", Expr);
         Error_Msg_Sem ("possible interpretations are:", Expr);
         for I in Natural loop
            El := Get_Nth_Element (List, I);
            exit when El = null;
            Error_Msg_Sem (Disp_Node (El), El);
         end loop;
      end Error_Operator_Overload;
   begin
      if Arity = 1 then
         Left := Get_Operand (Expr);
         Right := Null_Iir;
      else
         Left := Get_Left (Expr);
         Right := Get_Right (Expr);
      end if;
      Operator := Iirs_Utils.Get_Operator_Name (Expr);

      if Get_Type (Expr) = Null_Iir then
         --  First pass.
         --  Semantize operands.
         --  FIXME: should try to semantize right operand even if semantization
         --  of left operand has failed ??
         if Get_Type (Left) = Null_Iir then
            Left := Sem_Expression_Ov (Left, Null_Iir);
            if Left = Null_Iir then
               return Null_Iir;
            end if;
            if Arity = 1 then
               Set_Operand (Expr, Left);
            else
               Set_Left (Expr, Left);
            end if;
         end if;
         if Arity = 2 and then Get_Type (Right) = Null_Iir then
            Right := Sem_Expression_Ov (Right, Null_Iir);
            if Right = Null_Iir then
               return Null_Iir;
            end if;
            Set_Right (Expr, Right);
         end if;

         Overload_List := Get_Overload_List;
         Nbr_Full_Compat := 0;

         --  Try to find an implementation among user defined function
         Interpretation := Get_Interpretation (Operator);
         while Interpretation /= No_Name_Interpretation loop
            Decl := Get_Declaration (Interpretation);

            --  It is compatible with operand types ?
            if Get_Kind (Decl) not in Iir_Kinds_Function_Specification then
               raise Internal_Error;
            end if;

            --  Check return type.
            if Res_Type /= null then
               Compat := Compatibility (Res_Type, Get_Return_Type (Decl));
               if Compat = None then
                  goto Next;
               else
                  --  Handle case:
                  --    constant a : integer := abs 20;
                  --  ABS can be either for universal integer or for integer.
                  Compat := Full;
               end if;
            else
               Compat := Full;
            end if;

            Interface_List := Get_Interface_Declaration (Decl);

            --  Check arity.
            if Get_Nbr_Elements (Interface_List) /= Arity then
               goto Next;
            end if;

            -- Check operands.
            Compat := Compatibility_Level'Min
              (Compatibility_Nodes (Get_Nth_Element (Interface_List, 0), Left),
               Compat);
            if Compat = None then
               goto Next;
            end if;
            if Arity = 2 then
               Compat := Compatibility_Level'Min
                 (Compatibility_Nodes (Get_Nth_Element (Interface_List, 1),
                                       Right),
                  Compat);
               if Compat = None then
                  goto Next;
               end if;
            end if;

            --  Match.
            if Compat = Full then
               if Nbr_Full_Compat = 0 then
                  Nbr_Full_Compat := 1;
                  Full_Compat := Decl;
               else
                  Nbr_Full_Compat := Nbr_Full_Compat + 1;
               end if;
            end if;

            Append_Element (Overload_List, Decl);

            << Next >> null;
            Interpretation := Get_Next_Interpretation (Interpretation);
         end loop;

         --  The list of possible implementations was computed.
         if Nbr_Full_Compat = 1 then
            return Set_Uniq_Interpretation (Full_Compat);
         end if;
         case Get_Nbr_Elements (Overload_List) is
            when 0 =>
               Error_Msg_Sem
                 ("no function declarations for " & Disp_Node (Expr), Expr);
               Free_Iir (Overload_List);
               return Null_Iir;

            when 1 =>
               return Set_Uniq_Interpretation
                 (Get_First_Element (Overload_List));

            when others =>
               if Get_Nbr_Elements (Overload_List) = 2 then
                  --  There may be an implicit declaration overriden by a user
                  --  declaration.
                  Decl := Get_Non_Implicit_Function (Overload_List);
                  if Decl /= Null_Iir then
                     return Set_Uniq_Interpretation (Decl);
                  end if;
               end if;

               Set_Implementation (Expr, Overload_List);

               --  Create the list of possible return types.
               if Res_Type = Null_Iir then

                  Res_Type_List := Create_List_Of_Types (Overload_List);
                  if not Is_Overload_List (Res_Type_List) then
                     Error_Operator_Overload (Overload_List);
                     return Null_Iir;
                  else
                     Set_Type (Expr, Res_Type_List);
                     return Expr;
                  end if;
               else
                  Error_Operator_Overload (Overload_List);
                  return Null_Iir;
               end if;
         end case;
      else
         declare
            Decl_Compat : Compatibility_Level;
         begin
            --  Second pass
            --  Find the uniq implementation for this call.
            Overload_List := Get_Implementation (Expr);
            Full_Compat := null;
            Compat := None;
            for I in Natural loop
               Decl := Get_Nth_Element (Overload_List, I);
               exit when Decl = Null_Iir;
               --  FIXME: wrong: compatibilty with return type and args.
               Decl_Compat :=
                 Compatibility (Get_Return_Type (Decl), Res_Type);
               case Decl_Compat is
                  when None =>
                     null;
                  when Via_Conversion =>
                     if Compat /= Full then
                        Nbr_Full_Compat := Nbr_Full_Compat + 1;
                        Compat := Via_Conversion;
                     end if;
                  when Full =>
                     if Compat = Full and then Nbr_Full_Compat /= 0 then
                        Error_Operator_Overload (Overload_List);
                        return null;
                     else
                        Full_Compat := Decl;
                        Nbr_Full_Compat := 1;
                        Compat := Full;
                     end if;
               end case;
            end loop;
            --  For a via_conversion.
            if Nbr_Full_Compat /= 1 then
               Error_Operator_Overload (Overload_List);
               return Null_Iir;
            end if;
            Free_Iir (Overload_List);
            Overload_List := Get_Type (Expr);
            Free_Iir (Overload_List);
            return Set_Uniq_Interpretation (Full_Compat);
         end;
      end if;
   end Sem_Operator;

   -- Create a subtype for a string literal.
   -- The literal must have been typed, with a type or a subtype.
   -- FIXME: not general at all!
   function Check_Type_For_String_Literal (A_Type : Iir) return Boolean
   is
      Base_Type : Iir := Get_Base_Type (A_Type);
   begin
      if not Is_Unidim_Array_Type (Base_Type) then
         return False;
      end if;
      case Get_Kind (Get_Element_Subtype (Base_Type)) is
         when Iir_Kind_Enumeration_Subtype_Definition
           | Iir_Kind_Enumeration_Type_Definition =>
            null;
         when others =>
            return False;
      end case;
      return True;
   end Check_Type_For_String_Literal;

   --  Semantize LIT whose elements must be of type EL_TYPE, and return
   --  the length.
   function Sem_String_Literal (Lit: Iir; El_Type : Iir) return Natural
   is
      function Get_Bit_String_Length (Lit : Iir) return Natural
      is
         Str : String := Iirs_Utils.Get_String (Lit);
         Len : Natural := 0;
      begin
         for I in Str'Range loop
            if Str (I) /= '_' then
               Len := Len + 1;
            end if;
         end loop;
         return Len * Get_Bit_String_Base (Lit);
      end Get_Bit_String_Length;

      function Find_Literal (Etype : Iir_Enumeration_Type_Definition;
                             C : Character)
        return Iir_Enumeration_Literal
      is
         Inter : Name_Interpretation_Type;
         Id : Name_Id;
         Decl : Iir;
      begin
         Id := Name_Table.Get_Identifier (C);
         Inter := Get_Interpretation (Id);
         while Inter /= No_Name_Interpretation loop
            Decl := Get_Declaration (Inter);
            if Get_Kind (Decl) = Iir_Kind_Enumeration_Literal
              and then Get_Type (Decl) = Etype
            then
               return Decl;
            end if;
            Inter := Get_Next_Interpretation (Inter);
         end loop;
         --  Character C is not visible...
         if Find_Name_In_List (Get_Enumeration_Literal_List (Etype), Id)
           = Null_Iir
         then
            --  ... because it is not defined.
            Error_Msg_Sem
              ("type " & Disp_Node (Etype) & " does not define character '"
               & C & "'", Lit);
         else
            --  ... because it is not visible.
            Error_Msg_Sem ("character '" & C & "' of type "
                           & Disp_Node (Etype) & " is not visible", Lit);
         end if;
         return Null_Iir;
      end Find_Literal;

      --  For a string_literal, check all characters of the string is a
      --  literal of the type.
      procedure Check_String_Literal (El_Type : Iir; Lit : Iir)
      is
         use Name_Table;
         El : Iir;
      begin
         --  This was wrong, since visibility rules must be checked.
         --          if El_Type = Character_Type_Definition then
         --             --  All characters are defined in character.
         --             --  8 bits characters are caught by scan.
         --             return;
         --          end if;
         Image (Get_Identifier (Lit));
         for I in 1 .. Name_Length loop
            El := Find_Literal (El_Type, Name_Buffer (I));
         end loop;
      end Check_String_Literal;

      Len : Natural;
   begin
      if Get_Kind (Lit) = Iir_Kind_Bit_String_Literal then
         Len := Get_Bit_String_Length (Lit);
         Set_Bit_String_0 (Lit, Find_Literal (El_Type, '0'));
         Set_Bit_String_1 (Lit, Find_Literal (El_Type, '1'));
      else
         Len := Name_Table.Get_Name_Length (Get_Identifier (Lit));
         Check_String_Literal (El_Type, Lit);
      end if;

      Set_Expr_Staticness (Lit, Locally);

      return Len;
   end Sem_String_Literal;

   procedure Sem_String_Literal (Lit: Iir) is
      Lit_Type: Iir;
      Lit_Base_Type : Iir;

      -- The subtype created for the literal.
      N_Type: Iir;
      -- type of the index of the array type.
      Index_Type: Iir;
      Len : Natural;
      El_Type : Iir;
   begin
      Lit_Type := Get_Type (Lit);
      Lit_Base_Type := Get_Base_Type (Lit_Type);

      El_Type := Get_Base_Type (Get_Element_Subtype (Lit_Base_Type));
      Len := Sem_String_Literal (Lit, El_Type);

      case Get_Kind (Lit_Type) is
         when Iir_Kind_Array_Type_Definition
           | Iir_Kind_Unconstrained_Array_Subtype_Definition =>
            -- Set type of the string literal,
            -- according to LRM93 7.3.2.2.
            N_Type := Create_Unidim_Array_By_Length
              (Lit_Base_Type, Iir_Int64 (Len), Lit);
            Set_Type (Lit, N_Type);
         when Iir_Kind_Array_Subtype_Definition =>
            Index_Type := Get_First_Element (Get_Index_List (Lit_Type));
            if Get_Type_Staticness (Index_Type) = Locally then
               if Eval_Discrete_Type_Length (Index_Type) /= Iir_Int64 (Len)
               then
                  Error_Msg_Sem ("string length does not match that of "
                                 & Disp_Node (Index_Type), Lit);
               end if;
            else
               --  FIXME: It this right ?
               --  We really need a locally static type.
               N_Type := Create_Unidim_Array_By_Length
                 (Lit_Base_Type, Iir_Int64 (Len), Lit);
               Set_Type (Lit, N_Type);
            end if;
         when others =>
            Error_Kind ("sem_string_literal_type", Lit_Type);
      end case;
   end Sem_String_Literal;

   -- Perform semantisation on a (sub)aggregate AGGR, which is of type
   -- A_TYPE.
   -- return FALSE is case of failure
   function Sem_Record_Aggregate (Aggr: Iir_Aggregate; A_Type: Iir)
     return boolean
   is
      Element_List: Iir_List;
      El_Type : Iir;
      Matches: Iir_Array_Acc;
      Ok : Boolean;

      procedure Add_Match (El : Iir; Pos : Integer)
      is
         Ass_Type : Iir;
      begin
         if Matches (Pos) /= Null_Iir then
            Error_Msg_Sem
              (Disp_Node (Matches (Pos)) & " was already associated", El);
            Ok := False;
            return;
         end if;
         Matches (Pos) := El;
         Ass_Type := Get_Type (Get_Nth_Element (Element_List, Pos));
         if El_Type = Null_Iir then
            El_Type := Ass_Type;
         elsif not Are_Types_Compatible (El_Type, Ass_Type) then
            Error_Msg_Sem ("elements are not of the same type", El);
            Ok := False;
         end if;
      end Add_Match;

      function Sem_Simple_Choice (Ass : Iir) return Iir
      is
         N_El : Iir;
         Expr : Iir;
         Aggr_El : Iir;
      begin
         Expr := Get_Expression (Ass);
         if Get_Kind (Expr) /= Iir_Kind_Simple_Name then
            Error_Msg_Sem ("element association must be a simple name", Ass);
            return Ass;
         end if;
         Aggr_El := Find_Name_In_List (Element_List, Get_Identifier (Expr));
         if Aggr_El = Null_Iir then
            Error_Msg_Sem
              ("record has no such element " & Disp_Node (Ass), Ass);
            return Ass;
         end if;

         N_El := Create_Iir (Iir_Kind_Association_Choice_By_Name);
         Location_Copy (N_El, Ass);
         Set_Name (N_El, Aggr_El);
         Set_Associated (N_El, Get_Associated (Ass));
         Free_Old_Iir (Ass);
         Add_Match (N_El, Natural (Get_Element_Position (Aggr_El)));
         return N_El;
      end Sem_Simple_Choice;

      Assoc_List: Iir_Association_Choices_List;
      El: Iir;
      Expr: Iir;
      Sub_El : Iir;
      Sub_List : Iir_List;
      Has_Named : Boolean;
   begin
      Ok := True;
      Assoc_List := Get_Association_Choices_List (Aggr);
      Element_List := Get_Element_Declaration_List (Get_Base_Type (A_Type));
      Matches := new Iir_Array (0 .. Get_Nbr_Elements (Element_List) -1);
      Matches.all := (others => Null_Iir);

      Has_Named := False;
      for I in Natural loop
         El := Get_Nth_Element (Assoc_List, I);
         exit when El = Null_Iir;
         El_Type := Null_Iir;
         case Get_Kind (El) is
            when Iir_Kind_Association_Choice_By_None =>
               if Has_Named then
                  Error_Msg_Sem ("positional association after named one", El);
               elsif I > Matches.all'Last then
                  Error_Msg_Sem ("too many elements", El);
                  exit;
               else
                  Add_Match (El, I);
               end if;
            when Iir_Kind_Association_Choice_By_Expression =>
               Has_Named := True;
               El := Sem_Simple_Choice (El);
               Replace_Nth_Element (Assoc_List, I, El);
            when Iir_Kind_Association_Choice_By_List =>
               Has_Named := True;
               Sub_List := Get_Choice_List (El);
               for I in Natural loop
                  Sub_El := Get_Nth_Element (Sub_List, I);
                  exit when Sub_El = Null_Iir;
                  Replace_Nth_Element
                    (Sub_List, I, Sem_Simple_Choice (Sub_El));
               end loop;
            when Iir_Kind_Association_Choice_By_Others =>
               Has_Named := True;
               if I /= Get_Nbr_Elements (Assoc_List) - 1 then
                  Error_Msg_Sem
                    ("choice others must be the last alternative", El);
               end if;
               declare
                  Found : Boolean := False;
               begin
                  for I in Matches.all'Range loop
                     if Matches (I) = Null_Iir then
                        Add_Match (El, I);
                        Found := True;
                     end if;
                  end loop;
                  if not Found then
                     Error_Msg_Sem ("no element for choice others", El);
                  end if;
               end;
            when others =>
               Error_Kind ("sem_record_aggregate", El);
         end case;
         if El_Type /= Null_Iir then
            Expr := Get_Associated (El);
            Expr := Sem_Expression (Expr, El_Type);
            if Expr /= Null_Iir then
               Set_Associated (El, Eval_Expr_If_Static (Expr));
            else
               Ok := False;
            end if;
         else
            Ok := False;
         end if;
      end loop;
      for I in Matches.all'Range loop
         if Matches (I) = Null_Iir then
            Error_Msg_Sem
              ("no value for " & Disp_Node (Get_Nth_Element (Element_List, I)),
               Aggr);
            Ok := False;
         end if;
      end loop;
      Free (Matches);
      return Ok;
   end Sem_Record_Aggregate;

   procedure Sem_String_Choices_Range (Choice_List : Iir_List; Sel : Iir)
   is
      --  True if others choice is present.
      Has_Others : Boolean;

      --  Type of SEL.
      Sel_Type : Iir;

      --  Length of SEL (number of characters in SEL).
      Sel_Length : Iir_Int64;

      El : Iir;
      El1 : Iir;
      Sub_List : Iir_List;

      procedure Sem_Simple_Choice (Choice : Iir)
      is
         Expr : Iir;
      begin
         --  LRM93 8.8
         --  In such case, each choice appearing in any of the case statement
         --  alternative must be a locally static expression whose value is of
         --  the same length as that of the case expression.
         Expr := Sem_Expression (Get_Expression (Choice), Sel_Type);
         if Expr = Null_Iir then
            return;
         end if;
         Set_Expression (Choice, Expr);
         if Get_Expr_Staticness (Expr) > Locally then
            Error_Msg_Sem ("choice must be locally static expression", Expr);
            return;
         end if;
         Expr := Eval_Expr (Expr);
         Set_Expression (Choice, Expr);
         if Eval_Discrete_Type_Length
           (Get_String_Type_Bound_Type (Get_Type (Expr))) /= Sel_Length
         then
            Error_Msg_Sem
              ("value not of the same length of the case expression", Expr);
            return;
         end if;
      end Sem_Simple_Choice;
   begin
      --  LRM93 8.8
      --  If the expression is of one-dimensional character array type, then
      --  the expression must be one of the following:
      --  FIXME: to complete.
      Sel_Type := Get_Type (Sel);
      case Get_Kind (Sel_Type) is
         when Iir_Kind_Array_Subtype_Definition =>
            null;
         when others =>
            Error_Msg_Sem
              ("expression must be discrete or one-dimension array subtype",
               Sel);
            return;
      end case;
      if Get_Nbr_Elements (Get_Index_List (Sel_Type)) /= 1 then
         Error_Msg_Sem ("array type must be one-dimensional", Sel);
         return;
      end if;
      if Get_Type_Staticness (Sel_Type) /= Locally then
         Error_Msg_Sem ("array type must be locally static", Sel);
         return;
      end if;
      Sel_Length := Eval_Discrete_Type_Length
        (Get_String_Type_Bound_Type (Sel_Type));

      Has_Others := False;
      for I in Natural loop
         El := Get_Nth_Element (Choice_List, I);
         exit when El = Null_Iir;
         case Get_Kind (El) is
            when Iir_Kind_Association_Choice_By_None =>
               raise Internal_Error;
            when Iir_Kind_Association_Choice_By_Range =>
               Error_Msg_Sem
                 ("range choice are not allowed for non-discrete type", El);
            when Iir_Kind_Association_Choice_By_Expression =>
               Sem_Simple_Choice (El);
            when Iir_Kind_Association_Choice_By_List =>
               Sub_List := Get_Choice_List (El);
               for I in Natural loop
                  El1 := Get_Nth_Element (Sub_List, I);
                  exit when El1 = Null_Iir;
                  Sem_Simple_Choice (El1);
               end loop;
            when Iir_Kind_Association_Choice_By_Others =>
               if Has_Others then
                  Error_Msg_Sem ("duplicate others choice", El);
               elsif I /= Get_Nbr_Elements (Choice_List) - 1 then
                  Error_Msg_Sem
                    ("choice others must be the last alternative", El);
               end if;
               Has_Others := True;
            when others =>
               Error_Kind ("sem_string_choices_range", El);
         end case;
      end loop;

      -- FIXME:
      -- * check for duplicate choices.
      -- * check for leaking choices.
      -- (should eval strings and bit-strings).
   end Sem_String_Choices_Range;

   procedure Sem_Choices_Range
     (Choice_List : Iir_List;
      Sub_Type : Iir;
      Is_Sub_Range : Boolean;
      Loc : Location_Type;
      Low : out Iir;
      High : out Iir)
   is
      --  Number of positionnal choice.
      Nbr_Pos : Iir_Int64;

      --  Number of named choices.
      Nbr_Named : Natural;

      --  True if others choice is present.
      Has_Others : Boolean;

      Has_Error : Boolean;

      --  True if SUB_TYPE has bounds.
      Type_Has_Bounds : Boolean;

      Arr : Iir_Array_Acc;
      Index : Natural;
      Pos_Max : Iir_Int64;
      El, El1: Iir;
      --Index_Constraint : Iir;
      Staticness : Iir_Staticness;

      --  Semantize a simple (by expression or by range) choice.
      --  Return FALSE in case of error.
      function Sem_Simple_Choice (Choice : Iir; A_Type : Iir) return Boolean
      is
         Expr : Iir;
      begin
         Expr := Get_Expression (Choice);
         if Get_Kind (Choice) = Iir_Kind_Association_Choice_By_Range then
            Expr := Sem_Discrete_Range_Expression (Expr, A_Type);
         else
            Expr := Sem_Expression_Ov (Expr, Get_Base_Type (A_Type));
         end if;
         if Expr = Null_Iir then
            return False;
         end if;
         Expr := Eval_Expr_If_Static (Expr);
         Set_Expression (Choice, Expr);
         Set_Choice_Staticness (Choice, Get_Expr_Staticness (Expr));
         return True;
      end Sem_Simple_Choice;

      --  Get low limit of ASSOC.
      --  First, get the expression of the association, then the low limit.
      --  ASSOC may be either association_by_range (in this case the low limit
      --   is to be fetched), or association_by_expression (and the low limit
      --   is the expression).
      function Get_Low (Assoc : Iir) return Iir
      is
         Expr : Iir;
      begin
         Expr := Get_Expression (Assoc);
         case Get_Kind (Expr) is
            when Iir_Kind_Range_Expression =>
               case Get_Direction (Expr) is
                  when Iir_To =>
                     return Get_Left_Limit (Expr);
                  when Iir_Downto =>
                     return Get_Right_Limit (Expr);
               end case;
            when others =>
               return Expr;
         end case;
      end Get_Low;

      function Get_High (Assoc : Iir) return Iir
      is
         Expr : Iir;
      begin
         Expr := Get_Expression (Assoc);
         case Get_Kind (Expr) is
            when Iir_Kind_Range_Expression =>
               case Get_Direction (Expr) is
                  when Iir_To =>
                     return Get_Right_Limit (Expr);
                  when Iir_Downto =>
                     return Get_Left_Limit (Expr);
               end case;
            when others =>
               return Expr;
         end case;
      end Get_High;

      --  Compare two elements of ARR.
      --  Return true iff OP1 < OP2.
      function Lt (Op1, Op2 : Natural) return Boolean is
      begin
         return
           Eval_Pos (Get_Low (Arr (Op1))) < Eval_Pos (Get_Low (Arr (Op2)));
      end Lt;

      --  Swap two elements of ARR.
      procedure Swap (From : Natural; To : Natural)
      is
         Tmp : Iir;
      begin
         Tmp := Arr (To);
         Arr (To) := Arr (From);
         Arr (From) := Tmp;
      end Swap;

      --  Bubble down element I of a partially ordered heap of length N in
      --  array ARR.
      procedure Bubble_Down (I, N : Natural)
      is
         Child : Natural;
      begin
         Child := 2 * I;
         if Child < N and then Lt (Child, Child + 1) then
            Child := Child + 1;
         end if;
         if Child <= N and then Lt (I, Child) then
            Swap (I, Child);
            Bubble_Down (Child, N);
         end if;
      end Bubble_Down;

      --  Heap sort of ARR.
      procedure Heap_Sort (N : Natural)
      is
      begin
         --  Heapify
         for I in reverse 1 .. N / 2 loop
            Bubble_Down (I, N);
         end loop;

         --  Sort
         for I in reverse 2 .. N loop
            Swap (1, I);
            Bubble_Down (1, I - 1);
         end loop;
      end Heap_Sort;
   begin
      Low := Null_Iir;
      High := Null_Iir;

      --  First:
      --  semantize the choices
      --  compute the range of positionnal choices
      --  compute the number of choice elements (extracted from lists).
      --  check for others presence.
      Nbr_Pos := 0;
      Nbr_Named := 0;
      Has_Others := False;
      Has_Error := False;
      Staticness := Locally;
      for I in Natural loop
         El := Get_Nth_Element (Choice_List, I);
         exit when El = Null_Iir;
         case Get_Kind (El) is
            when Iir_Kind_Association_Choice_By_None =>
               Nbr_Pos := Nbr_Pos + 1;
            when Iir_Kind_Association_Choice_By_Expression
              | Iir_Kind_Association_Choice_By_Range =>
               if Sem_Simple_Choice (El, Sub_Type) then
                  Staticness := Min (Staticness, Get_Choice_Staticness (El));
               else
                  Has_Error := True;
               end if;
               Nbr_Named := Nbr_Named + 1;
            when Iir_Kind_Association_Choice_By_Name =>
               --  It is not possible to have such a choice in an array
               --  aggregate.
               --  Should have been caught previously.
               raise Internal_Error;
            when Iir_Kind_Association_Choice_By_Others =>
               if Has_Others then
                  Error_Msg_Sem ("duplicate others choice", El);
               elsif I /= Get_Nbr_Elements (Choice_List) - 1 then
                  Error_Msg_Sem
                    ("choice others should be the last alternative", El);
               end if;
               Has_Others := True;
            when Iir_Kind_Association_Choice_By_List =>
               declare
                  Sub_List : Iir_Choice_List;
                  El_Staticness : Iir_Staticness;
               begin
                  Sub_List := Get_Choice_List (El);
                  El_Staticness := Locally;
                  for J in Natural loop
                     El1 := Get_Nth_Element (Sub_List, J);
                     exit when El1 = Null_Iir;
                     if Sem_Simple_Choice (El1, Sub_Type) then
                        El_Staticness := Min (El_Staticness,
                                              Get_Choice_Staticness (El1));
                     else
                        Has_Error := True;
                     end if;
                     Nbr_Named := Nbr_Named + 1;
                  end loop;
                  Set_Choice_Staticness (El, El_Staticness);
                  Staticness := Min (Staticness, El_Staticness);
               end;
            when others =>
               Error_Kind ("sem_choices_range", El);
         end case;
      end loop;

      if Has_Error then
         --  Nothing can be done here...
         return;
      end if;
      if Nbr_Pos > 0 and then Nbr_Named > 0 then
         --  LRM93 7.3.2.2
         --  Apart from the final element with the single choice OTHERS, the
         --  rest (if any) of the element associations of an array aggregate
         --  must be either all positionnal or all named.
         Error_Msg_Sem
           ("element associations must be all positional or all named", Loc);
         return;
      end if;

      --  For a positional aggregate.
      if Nbr_Pos > 0 then
         --  Check number of elements match, but only if it is possible.
         if Get_Type_Staticness (Sub_Type) /= Locally then
            return;
         end if;
         Pos_Max := Eval_Discrete_Type_Length (Sub_Type);
         if (not Has_Others and not Is_Sub_Range)
           and then Nbr_Pos < Pos_Max
         then
            Error_Msg_Sem ("not enough elements associated", Loc);
         elsif Nbr_Pos > Pos_Max then
            Error_Msg_Sem ("too many elements associated", Loc);
         end if;
         return;
      end if;

      --  Second:
      --  Create the list of choices
      if Nbr_Named = 0 and then Has_Others then
         --  This is only a others association.
         return;
      end if;
      if Staticness /= Locally then
         --  LRM93 7.3.2.2
         --  A named association of an array aggregate is allowed to have
         --  a choice that is not locally static, or likewise a choice that
         --  is a null range, only if the aggregate includes a single
         --  element association and the element association has a single
         --  choice.
         if Nbr_Named > 1 or Has_Others then
            Error_Msg_Sem ("not static choice exclude others choice", Loc);
         end if;
         return;
      end if;

      --  Set TYPE_HAS_BOUNDS
      case Get_Kind (Sub_Type) is
         when Iir_Kind_Enumeration_Type_Definition
           | Iir_Kind_Enumeration_Subtype_Definition
           | Iir_Kind_Integer_Subtype_Definition =>
            Type_Has_Bounds := True;
         when Iir_Kind_Integer_Type_Definition =>
            Type_Has_Bounds := False;
         when others =>
            Error_Kind ("sem_choice_range(3)", Sub_Type);
      end case;

      Arr := new Iir_Array (1 .. Nbr_Named);
      Index := 0;

      declare
         procedure Add_Choice (Choice : Iir; A_Type : Iir)
         is
            Ok : Boolean;
            Expr : Iir;
         begin
            Expr := Get_Expression (Choice);
            if Type_Has_Bounds
              and then Get_Expr_Staticness (Expr) = Locally
              and then Get_Type_Staticness (A_Type) = Locally
            then
               if Get_Kind (Choice) = Iir_Kind_Association_Choice_By_Range then
                  Ok := Eval_Is_Range_In_Bound (Expr, A_Type);
               else
                  Ok := Eval_Is_In_Bound (Expr, A_Type);
               end if;
               if not Ok then
                  Error_Msg_Sem
                    (Disp_Node (Expr) & " out of index range", Choice);
               end if;
            else
               Ok := True;
            end if;
            if Ok then
               Index := Index + 1;
               Arr (Index) := Choice;
            end if;
         end Add_Choice;
      begin
         --  Fill the array.
         for I in Natural loop
            El := Get_Nth_Element (Choice_List, I);
            exit when El = Null_Iir;
            case Get_Kind (El) is
               when Iir_Kind_Association_Choice_By_None =>
                  --  Only named associations are considered.
                  raise Internal_Error;
               when Iir_Kind_Association_Choice_By_Expression
                 | Iir_Kind_Association_Choice_By_Range =>
                  Add_Choice (El, Sub_Type);
               when Iir_Kind_Association_Choice_By_Others =>
                  null;
               when Iir_Kind_Association_Choice_By_List =>
                  for J in Natural loop
                     El1 := Get_Nth_Element (Get_Choice_List (El), J);
                     exit when El1 = Null_Iir;
                     Add_Choice (El1, Sub_Type);
                  end loop;
               when others =>
                  Error_Kind ("sem_choices_range(2)", El);
            end case;
         end loop;
      end;

      --  Third:
      --  Sort the list
      Heap_Sort (Index);

      --  Set low and high bounds.
      if Index > 0 then
         Low := Get_Low (Arr (1));
         High := Get_High (Arr (Index));
      else
         Low := Null_Iir;
         High := Null_Iir;
      end if;

      --  Fourth:
      --  check for lacking choice (if no others)
      --  check for overlapping choices
      declare
         --  Emit an error message for absence of choices in position L to H
         --  of index type BT at location LOC.
         procedure Error_No_Choice (Bt : Iir;
                                    L, H : Iir_Int64;
                                    Loc : Location_Type)
         is
         begin
            if L = H then
               Error_Msg_Sem ("no choice for " & Disp_Discrete (Bt, L), Loc);
            else
               Error_Msg_Sem
                 ("no choices for " & Disp_Discrete (Bt, L)
                     & " to " & Disp_Discrete (Bt, H), Loc);
            end if;
         end Error_No_Choice;

         --  Lowest and highest bounds.
         Lb, Hb : Iir;
         Pos : Iir_Int64;
         Pos_Max : Iir_Int64;
         E_Pos : Iir_Int64;

         Bt : Iir;
      begin
         Bt := Get_Base_Type (Sub_Type);
         if not Is_Sub_Range
           and then Get_Type_Staticness (Sub_Type) = Locally
           and then Type_Has_Bounds
         then
            Get_Low_High_Limit (Get_Range_Constraint (Sub_Type), Lb, Hb);
         else
            Lb := Low;
            Hb := High;
         end if;
         Pos := Eval_Pos (Lb);
         Pos_Max := Eval_Pos (Hb);
         for I in 1 .. Index loop
            E_Pos := Eval_Pos (Get_Low (Arr (I)));
            if E_Pos > Pos_Max then
               --  Choice out of bound, already handled.
               Error_No_Choice (Bt, Pos, Pos_Max, Get_Location (Arr (I)));
               --  Avoid other errors.
               Pos := Pos_Max + 1;
               exit;
            end if;
            if Pos < E_Pos and then not Has_Others then
               Error_No_Choice (Bt, Pos, E_Pos - 1, Get_Location (Arr (I)));
            elsif Pos > E_Pos then
               if Pos + 1 = E_Pos then
                  Error_Msg_Sem
                    ("duplicate choice for " & Disp_Discrete (Bt, Pos),
                     Arr (I));
               else
                  Error_Msg_Sem
                    ("duplicate choices for" & Disp_Discrete (Bt, E_Pos)
                     & " to " & Disp_Discrete (Bt, Pos), Arr (I));
               end if;
            end if;
            Pos := Eval_Pos (Get_High (Arr (I))) + 1;
         end loop;
         if Pos /= Pos_Max + 1 and then not Has_Others then
            Error_No_Choice (Bt, Pos, Pos_Max, Loc);
         end if;
      end;

      Free (Arr);
   end Sem_Choices_Range;

   -- Find out the MIN and the MAX of an all named association choice list.
   -- It also returns the number of elements associed (counting range).
   procedure Sem_Find_Min_Max_Association_Choice_List
     (List: Iir_Association_Choices_List;
      Min: out Iir;
      Max: out Iir;
      Length: out natural)
   is
      Min_Res: Iir := null;
      Max_Res: Iir := null;
      procedure Update_With_Value (Val: Iir) is
      begin
         if Min_Res = null then
            Min_Res := Val;
            Max_Res := Val;
         elsif Get_Value (Val) < Get_Value (Min_Res) then
            Min_Res := Val;
         elsif Get_Value (Val) > Get_Value (Max_Res) then
            Max_Res := Val;
         end if;
      end Update_With_Value;

      Number_Elements: Natural;

      procedure Update (Choice: Iir) is
         Left, Right: Iir;
         Expr: Iir;
      begin
         case Get_Kind (Choice) is
            when Iir_Kind_Association_Choice_By_Expression =>
               Update_With_Value (Get_Expression (Choice));
               Number_Elements := Number_Elements + 1;
            when Iir_Kind_Association_Choice_By_Range =>
               Expr := Get_Expression (Choice);
               Left := Get_Left_Limit (Expr);
               Right := Get_Right_Limit (Expr);
               Update_With_Value (Left);
               Update_With_Value (Right);
               -- There can't be null range.
               case Get_Direction (Expr) is
                  when Iir_To =>
                     Number_Elements := Number_Elements +
                       Natural (Get_Value (Right) - Get_Value (Left) + 1);
                  when Iir_Downto =>
                     Number_Elements := Number_Elements +
                       Natural (Get_Value (Left) - Get_Value (Right) + 1);
               end case;
            when others =>
               Error_Kind ("sem_find_min_max_association_choice_list", Choice);
         end case;
      end Update;

      El: Iir;
      Sub_List: Iir_Association_Choices_List;
      Sub_El: Iir;
   begin
      Number_Elements := 0;
      for I in Natural loop
         El := Get_Nth_Element (List, I);
         exit when El = null;
         case Get_Kind (El) is
            when Iir_Kind_Association_Choice_By_List =>
               Sub_List := Get_Choice_List (El);
               for J in Natural loop
                  Sub_El := Get_Nth_Element (Sub_List, J);
                  exit when Sub_El = null;
                  Update (Sub_El);
               end loop;
            when others =>
               Update (El);
         end case;
      end loop;
      Min := Min_Res;
      Max := Max_Res;
      Length := Number_Elements;
   end Sem_Find_Min_Max_Association_Choice_List;

   --  Information for each dimension of an aggregate.
   type Array_Aggr_Info is record
      --  False if one sub-aggregate has no others choices.
      --  If FALSE, the dimension is constrained.
      Has_Others : Boolean := True;

      --  True if one sub-aggregate is by named/by position.
      Has_Named : Boolean := False;
      Has_Positional : Boolean := False;

      --  True if one sub-aggregate is dynamic.
      Has_Dynamic : Boolean := False;

      --  LOW and HIGH limits for the dimension.
      Low : Iir := Null_Iir;
      High : Iir := Null_Iir;

      --  Minimum length of the dimension.  This is a minimax.
      Min_Length : Natural := 0;

      --  If not NULL_IIR, this is the bounds of the dimension.
      --  If every dimension has bounds, then the aggregate is constrained.
      Index_Subtype : Iir := Null_Iir;

      --  True if there is an error.
      Error : Boolean := False;
   end record;

   type Array_Aggr_Info_Arr is array (Natural range <>) of Array_Aggr_Info;

   --  Semantize an array aggregate AGGR of *base type* A_TYPE.
   --  The type of the array is computed into A_SUBTYPE.
   --  DIM is the dimension index in A_TYPE.
   --  Return FALSE in case of error.
   procedure Sem_Array_Aggregate_Type_1 (Aggr: Iir;
                                         A_Type: Iir;
                                         Infos : in out Array_Aggr_Info_Arr;
                                         Constrained : Boolean;
                                         Dim: Natural)
   is
      Res: Boolean;
      Assoc_List: Iir_Association_Choices_List;
      Choice: Iir;
      Is_Positional: Tri_State_Type;
      Is_Positional_Choice: Tri_State_Type;
      Low, High : Iir;
      Index_List : Iir_List;
      Has_Others : Boolean;

      Len : Natural;

      --  Type of the index (this is also the type of the choices).
      Index_Type : Iir;

      --Index_Subtype : Iir;
      Index_Subtype_Constraint : Iir_Range_Expression;
      Index_Constraint : Iir_Range_Expression; -- FIXME: 'range.
      Choice_Staticness : Iir_Staticness;

      Info : Array_Aggr_Info renames Infos (Dim);
   begin
      Res := True;
      Index_List := Get_Index_List (A_Type);
      Index_Type := Get_Nth_Element (Index_List, Dim - 1);

      --  Sem choices.
      case Get_Kind (Aggr) is
         when Iir_Kind_Aggregate =>
            Assoc_List := Get_Association_Choices_List (Aggr);
            Sem_Choices_Range (Assoc_List, Index_Type, not Constrained,
                               Get_Location (Aggr), Low, High);

            --  Update infos.
            if Low /= Null_Iir
              and then (Info.Low = Null_Iir
                        or else Eval_Pos (Low) < Eval_Pos (Info.Low))
            then
               Info.Low := Low;
            end if;
            if High /= Null_Iir
              and then (Info.High = Null_Iir
                        or else Eval_Pos (High) > Eval_Pos (Info.High))
            then
               Info.High := High;
            end if;

            --  FIXME: this is wrong: others count for 0, range for
            --  the length, and sub-list are not handled.
            Len := Get_Nbr_Elements (Assoc_List);
            Info.Min_Length := Integer'Max (Info.Min_Length, Len);

            --  Determine if the aggregate is positionnal or named;
            --    and compute choice staticness.
            Is_Positional := Unknown;
            Choice_Staticness := Locally;
            Has_Others := False;
            for I in Natural loop
               Choice := Get_Nth_Element (Assoc_List, I);
               exit when Choice = null;
               case Get_Kind (Choice) is
                  when Iir_Kind_Association_Choice_By_List
                    | Iir_Kind_Association_Choice_By_Range
                    | Iir_Kind_Association_Choice_By_Expression =>
                     Is_Positional_Choice := False;
                     Choice_Staticness :=
                       Iirs.Min (Choice_Staticness,
                                 Get_Choice_Staticness (Choice));
                  when Iir_Kind_Association_Choice_By_None =>
                     Is_Positional_Choice := True;
                  when Iir_Kind_Association_Choice_By_Others =>
                     if not Constrained then
                        Error_Msg_Sem ("`others' choice not allowed for an "
                                       & "aggregate in this context", Aggr);
                        Infos (Dim).Error := True;
                        return;
                     end if;
                     Has_Others := True;
                  when others =>
                     Error_Kind ("sem_array_aggregate_type", Choice);
               end case;
               --  LRM93 7.3.2.2
               --  Apart from the final element with the single choice
               --  OTHERS, the rest (if any) of the element
               --  associations of an array aggregate must be either
               --  all positionnal or all named.
               if Is_Positional /= Unknown
                 and then Is_Positional_Choice /= Is_Positional
               then
                  --  The error has already been emited by sem_choices_range.
                  Infos (Dim).Error := True;
                  return;
               else
                  Is_Positional := Is_Positional_Choice;
               end if;
            end loop;

            if Choice_Staticness = Unknown then
               --  This is possible when a choice is erroneous.
               Infos (Dim).Error := True;
               return;
            end if;

         when Iir_Kind_String_Literal
           | Iir_Kind_Bit_String_Literal =>
            Len := Sem_String_Literal (Aggr, Get_Element_Subtype (A_Type));
            Assoc_List := Null_Iir_List;
            Info.Min_Length := Integer'Max (Info.Min_Length, Len);
            Is_Positional := True;
            Has_Others := False;
            Choice_Staticness := Locally;

         when others =>
            Error_Kind ("sem_array_aggregate_type_1", Aggr);
      end case;

      if Is_Positional = True then
         Info.Has_Positional := True;
      end if;
      if Is_Positional = False then
         Info.Has_Named := True;
      end if;
      if not Has_Others then
         Info.Has_Others := False;
      end if;

      --  LRM93 7.3.2.2
      --  A named association of an array aggregate is allowed to have a choice
      --  that is not locally static, [or likewise a choice that is a null
      --  range], only if the aggregate includes a single element association
      --  and this element association has a single choice.
      if Is_Positional = False and then Choice_Staticness /= Locally then
         Choice := Get_First_Element (Assoc_List);
         if Get_Nbr_Elements (Assoc_List) /= 1 or else
           (Get_Kind (Choice) /= Iir_Kind_Association_Choice_By_Expression
            and then Get_Kind (Choice) /= Iir_Kind_Association_Choice_By_Range)
         then
            Error_Msg_Sem ("non-locally static choice for an aggregate is "
                           & "allowed only if only choice", Aggr);
            Infos (Dim).Error := True;
            return;
         end if;
         Info.Has_Dynamic := True;
      end if;

      --  Compute bounds of the index if there is no index subtype.
      if Info.Index_Subtype = Null_Iir and then Has_Others = False then
         --  LRM93 7.3.2.2
         --  the direction of the index subtype of the aggregate is that of the
         --  index subtype of the base type of the aggregate.

         if Is_Positional = True then
            --  LRM93 7.3.2.2
            --  For a positionnal aggregate, [...] the leftmost bound is given
            --  by S'LEFT where S is the index subtype of the base type of the
            --  array; [...] the rightmost bound is determined by the direction
            --  of the index subtype and the number of element.
            if Get_Type_Staticness (Index_Type) = Locally then
               Info.Index_Subtype := Create_Range_Subtype_By_Length
                 (Index_Type, Iir_Int64 (Len), Get_Location (Aggr));
            end if;
         else
            --  Create an index subtype.
            case Get_Kind (Index_Type) is
               when Iir_Kind_Integer_Subtype_Definition =>
                  Info.Index_Subtype := Create_Iir (Get_Kind (Index_Type));
               when Iir_Kind_Enumeration_Type_Definition
                 | Iir_Kind_Enumeration_Subtype_Definition =>
                  Info.Index_Subtype :=
                    Create_Iir (Iir_Kind_Enumeration_Subtype_Definition);
               when others =>
                  Error_Kind ("sem_array_aggregate_type2", Index_Type);
            end case;
            Location_Copy (Info.Index_Subtype, Aggr);
            Set_Base_Type (Info.Index_Subtype, Get_Base_Type (Index_Type));
            Index_Constraint := Get_Range_Constraint (Index_Type);

            Index_Subtype_Constraint := Create_Iir (Iir_Kind_Range_Expression);
            Location_Copy (Index_Subtype_Constraint, Aggr);
            Set_Range_Constraint
              (Info.Index_Subtype, Index_Subtype_Constraint);
            Set_Type_Staticness (Info.Index_Subtype, Choice_Staticness);
            Set_Direction (Index_Subtype_Constraint,
                           Get_Direction (Index_Constraint));

            --  LRM93 7.3.2.2
            --  For an aggregate that has named associations, the leftmost and
            --  the rightmost bounds are determined by the direction of the
            --  index subtype of the aggregate and the smallest and largest
            --  choice given.
            if Choice_Staticness = Locally then
               case Get_Direction (Index_Constraint) is
                  when Iir_To =>
                     Set_Left_Limit (Index_Subtype_Constraint, Low);
                     Set_Right_Limit (Index_Subtype_Constraint, High);
                  when Iir_Downto =>
                     Set_Left_Limit (Index_Subtype_Constraint, High);
                     Set_Right_Limit (Index_Subtype_Constraint, Low);
               end case;
            else
               --  Dynamic aggregate.
               declare
                  Expr : Iir;
                  Choice : Iir;
               begin
                  Choice := Get_First_Element (Assoc_List);
                  Expr := Get_Expression (Choice);
                  case Get_Kind (Choice) is
                     when Iir_Kind_Association_Choice_By_Expression =>
                        Set_Left_Limit (Index_Subtype_Constraint, Expr);
                        Set_Right_Limit (Index_Subtype_Constraint, Expr);
                     when Iir_Kind_Association_Choice_By_Range =>
                        Set_Range_Constraint (Info.Index_Subtype, Expr);
                        -- FIXME: avoid allocation-free.
                        Free_Iir (Index_Subtype_Constraint);
                     when others =>
                        raise Internal_Error;
                  end case;
               end;
            end if;
         end if;
         --Set_Type_Staticness
         --  (A_Subtype, Iirs.Min (Get_Type_Staticness (A_Subtype),
         --                        Get_Type_Staticness (Index_Subtype)));
         --Append_Element (Get_Index_List (A_Subtype), Index_Subtype);
      elsif Has_Others = False then
         --  Check the subaggregate bounds are the same.
         if Is_Positional = True then
            if Eval_Pos (Eval_Discrete_Range_Left (Get_Range_Constraint
                                                   (Info.Index_Subtype)))
              /= Eval_Pos (Eval_Discrete_Range_Left (Get_Range_Constraint
                                                     (Index_Type)))
            then
               Error_Msg_Sem ("subaggregate bounds mismatch", Aggr);
            else
               if Eval_Discrete_Type_Length (Info.Index_Subtype)
                 /= Iir_Int64 (Len)
               then
                  Error_Msg_Sem ("subaggregate length mismatch", Aggr);
               end if;
            end if;
         else
            declare
               L, H : Iir;
            begin
               Get_Low_High_Limit
                 (Get_Range_Constraint (Info.Index_Subtype), L, H);
               if Eval_Pos (L) /= Eval_Pos (Low)
                 or else Eval_Pos (H) /= Eval_Pos (H)
               then
                  Error_Msg_Sem ("subagregate bounds mismatch", Aggr);
               end if;
            end;
         end if;
      end if;

      --  Semantize aggregate elements.
      if Dim = Get_Nbr_Elements (Index_List) then
         --  A type has been found for AGGR, semantize AGGR as if it was
         --  an aggregate with a subtype.

         if Get_Kind (Aggr) = Iir_Kind_Aggregate then
            -- LRM93 7.3.2.2:
            --   the expression of each element association must be of the
            --   element type.
            declare
               El : Iir;
               Element_Type : Iir;
               Expr : Iir;
            begin
               Element_Type := Get_Element_Subtype (A_Type);
               for I in Natural loop
                  El := Get_Nth_Element (Assoc_List, I);
                  exit when El = null;

                  Expr := Sem_Expression (Get_Associated (El), Element_Type);
                  if Expr /= Null_Iir then
                     Set_Expr_Staticness
                       (Aggr, Min (Get_Expr_Staticness (Aggr),
                                   Get_Expr_Staticness (Expr)));
                     Set_Associated (El, Eval_Expr_If_Static (Expr));
                  else
                     Info.Error := True;
                  end if;
               end loop;
            end;
         end if;
      else
         declare
            Assoc : Iir;
         begin
            for I in Natural loop
               Choice := Get_Nth_Element (Assoc_List, I);
               exit when Choice = Null_Iir;
               Assoc := Get_Associated (Choice);
               case Get_Kind (Assoc) is
                  when Iir_Kind_Aggregate =>
                     Sem_Array_Aggregate_Type_1
                       (Assoc, A_Type, Infos, Constrained, Dim + 1);
                  when Iir_Kind_String_Literal
                    | Iir_Kind_Bit_String_Literal =>
                     if Dim + 1 = Get_Nbr_Elements (Index_List) then
                        Sem_Array_Aggregate_Type_1
                          (Assoc, A_Type, Infos, Constrained, Dim + 1);
                     else
                        Error_Msg_Sem
                          ("string literal not allowed here", Assoc);
                        Infos (Dim + 1).Error := True;
                     end if;
                  when others =>
                     Error_Msg_Sem ("sub-aggregate expected", Assoc);
                     Infos (Dim + 1).Error := True;
               end case;
            end loop;
         end;
      end if;
   end Sem_Array_Aggregate_Type_1;

   --  Semantize an array aggregate whose type is AGGR_TYPE.
   --  If CONSTRAINED is true, then the aggregate appears in one of the
   --  context and can have an 'others' choice.
   --  If CONSTRAINED is false, the aggregate can not have an 'others' choice.
   --  Create a subtype for this aggregate.
   --  Return NULL_IIR in case of error, or AGGR if not.
   function Sem_Array_Aggregate_Type
     (Aggr : Iir; Aggr_Type : Iir; Constrained : Boolean)
     return Iir
   is
      A_Subtype: Iir;
      Base_Type : Iir;
      Index_List : Iir_List := Get_Index_List (Aggr_Type);
      Nbr_Dim : Natural := Get_Nbr_Elements (Index_List);
      Infos : Array_Aggr_Info_Arr (1 .. Nbr_Dim);
      Aggr_Constrained : Boolean;
      Info, Prev_Info : Iir_Aggregate_Info;
   begin
      --  Semantize the aggregate.
      Sem_Array_Aggregate_Type_1 (Aggr, Aggr_Type, Infos, Constrained, 1);

      Aggr_Constrained := True;
      for I in Infos'Range loop
         --  Return now in case of error.
         if Infos (I).Error then
            return Null_Iir;
         end if;
         if Infos (I).Index_Subtype = Null_Iir then
            Aggr_Constrained := False;
         end if;
      end loop;
      Base_Type := Get_Base_Type (Aggr_Type);
      if Aggr_Constrained then
         A_Subtype := Create_Iir (Iir_Kind_Array_Subtype_Definition);
         Location_Copy (A_Subtype, Aggr);
         Set_Base_Type (A_Subtype, Base_Type);
         Set_Element_Subtype (A_Subtype, Get_Element_Subtype (Base_Type));
         Set_Type_Staticness (A_Subtype, Locally);
         for I in Infos'Range loop
            Append_Element (Get_Index_List (A_Subtype),
                            Infos (I).Index_Subtype);
            Set_Type_Staticness
              (A_Subtype,
               Iirs.Min (Get_Type_Staticness (A_Subtype),
                         Get_Type_Staticness (Infos (I).Index_Subtype)));
         end loop;
         Set_Type (Aggr, A_Subtype);
      else
         Set_Type (Aggr, Base_Type);
      end if;

      for I in Infos'Range loop
         --  Create info and link.
         Info := Create_Iir (Iir_Kind_Aggregate_Info);
         if I = 1 then
            Set_Aggregate_Info (Aggr, Info);
         else
            Set_Sub_Aggregate_Info (Prev_Info, Info);
         end if;
         Prev_Info := Info;

         --  Fill info.
         Set_Aggr_Dynamic_Flag (Info, Infos (I).Has_Dynamic);
         Set_Aggr_Named_Flag (Info, Infos (I).Has_Named);
         Set_Aggr_Low_Limit (Info, Infos (I).Low);
         Set_Aggr_High_Limit (Info, Infos (I).High);
         Set_Aggr_Max_Length (Info, Infos (I).Min_Length);
         Set_Aggr_Others_Flag (Info, Infos (I).Has_Others);
      end loop;
      return Aggr;
   end Sem_Array_Aggregate_Type;

   --  Semantize aggregate EXPR whose type is expected to be A_TYPE.
   --  A_TYPE cannot be null_iir (this case is handled in sem_expression_ov)
   function Sem_Aggregate (Expr: Iir_Aggregate; A_Type: Iir)
     return Iir_Aggregate
   is
   begin
      if A_Type = Null_Iir then
         raise Internal_Error;
      end if;

      --  An aggregate is at most globally static.
      Set_Expr_Staticness (Expr, Globally);

      Set_Type (Expr, A_Type); -- FIXME: should free old type
      case Get_Kind (A_Type) is
         when Iir_Kind_Array_Subtype_Definition =>
            return Sem_Array_Aggregate_Type (Expr, A_Type, True);
         when Iir_Kind_Array_Type_Definition =>
            return Sem_Array_Aggregate_Type (Expr, A_Type, False);
         when Iir_Kind_Record_Type_Definition
           | Iir_Kind_Record_Subtype_Definition =>
            if not Sem_Record_Aggregate (Expr, A_Type) then
               return Null_Iir;
            end if;
            return Expr;
         when others =>
            Error_Msg_Sem ("type " & Disp_Node (A_Type) & " is not composite",
                           Expr);
            return Null_Iir;
      end case;
   end Sem_Aggregate;

   -- Transform LIT into a physical_literal.
   -- LIT can be either a not semantized physical literal or
   --  a simple name that is a physical unit.  In the later case, a physical
   --  literal is created.
   function Sem_Physical_Literal (Lit: Iir) return Iir
   is
      Decl: Iir;
      Decl_Type : Iir;
      Res: Iir;
   begin
      case Get_Kind (Lit) is
         when Iir_Kind_Physical_Int_Literal
           | Iir_Kind_Physical_Fp_Literal =>
            Decl := Find_Declaration (Get_Name (Lit), Decl_Physical_Unit);
            Res := Lit;
         when Iir_Kind_Physical_Unit =>
            Res := Create_Iir (Iir_Kind_Physical_Int_Literal);
            Location_Copy (Res, Lit);
            Set_Value (Res, 1);
            Decl := Lit;
         when others =>
            Res := Create_Iir (Iir_Kind_Physical_Int_Literal);
            Location_Copy (Res, Lit);
            Set_Value (Res, 1);
            Decl := Find_Declaration (Lit, Decl_Physical_Unit);
      end case;
      if Decl = Null_Iir then
         return Null_Iir;
      end if;
      Set_Name (Res, Decl);
      Decl_Type := Get_Type (Decl);
      Set_Type (Res, Decl_Type);
      if Decl_Type = Std_Package.Time_Type_Definition
        and then Flags.Vhdl_Std > Vhdl_87
      then
         -- LRM93 7.4.2
         -- 1. a literal of type TIME.
         Set_Expr_Staticness (Res, Globally);
      else
         -- LRM93 7.4.1
         -- 1. a literal of any type other than type TIME;
         Set_Expr_Staticness (Res, Locally);
      end if;
      --Eval_Check_Constraints (Res);
      return Res;
   end Sem_Physical_Literal;


--    function Sem_Attribute (Attr : Iir_Attribute) return Iir
--    is
--       use Std_Names;
--       Prefix : Iir;
--    begin
--       --  LRM93 6.6
--      --  The meaning of the prefix of an attribute name must be determinable
--       --  independently of the attribute designator and independently of the
--       --  fact that it is the prefix of an attribute.
--       --Prefix := Sem_Expression_Ov (Get_Prefix (Attr), null, True);
--       Prefix := Get_Prefix (Attr);
--       Sem_Names (Prefix);
--       Prefix := Get_Named_Entity (Prefix);
--       if Prefix = Null_Iir then
--          return Null_Iir;
--       end if;
--       case Get_Kind (Prefix) is
--          when Iir_Kinds_Nonoverloadable_Declaration =>
--             null;
--          when Iir_Kind_Overload_List =>
--             Overloaded (Prefix);
--             return null;
--          when Iir_Kinds_Function_Specification
--            | Iir_Kinds_Procedure_Specification =>
--             null;
--          when Iir_Kinds_Concurrent_Statement =>
--             null;
--          when Iir_Kinds_Object_Declaration =>
--             null;
--          when Iir_Kind_Design_Unit =>
--             Sem.Add_Dependence (Prefix);
--             Prefix := Get_Library_Unit (Prefix);
--          when Iir_Kind_Selected_By_All_Name
--            | Iir_Kind_Selected_Element
--            | Iir_Kind_Slice_Name
--            | Iir_Kind_Indexed_Name =>
--             if Get_Type (Prefix) = Null_Iir then
--                Prefix := Sem_Expression_Ov (Prefix, Null_Iir, Value_Restr);
--                if Prefix = Null_Iir then
--                   return Null_Iir;
--                elsif Is_Overloaded (Prefix) then
--                   Overloaded (Prefix);
--                   return Null_Iir;
--                end if;
--             end if;
--          when others =>
--             Error_Kind ("sem_attribute", Prefix);
--             return Null_Iir;
--       end case;
--       Set_Prefix (Attr, Prefix);
--    end Sem_Attribute;

   --  Semantize an allocator by expression or an allocator by subtype.
   function Sem_Allocator (Expr : Iir; A_Type : Iir) return Iir
   is
      Arg: Iir;
      Arg_Type : Iir;
   begin
      Arg := Get_Expression (Expr);
      Set_Expr_Staticness (Expr, None);
      if Get_Type (Expr) = Null_Iir then
         if Get_Kind (Expr) = Iir_Kind_Allocator_By_Expression then
            if Get_Kind (Arg) /= Iir_Kind_Qualified_Expression then
               raise Internal_Error;
            end if;
            Arg := Sem_Expression (Arg, Null_Iir);
            if Arg = Null_Iir then
               return Null_Iir;
            end if;
            Check_Read (Arg);
            Arg_Type := Get_Type (Arg);
         else
            Arg := Sem.Sem_Subtype_Indication (Arg);
            if Arg = Null_Iir then
               return Null_Iir;
            end if;
            --  LRM93 7.3.6
            --  If an allocator includes a subtype indication and if the
            --  type of the object created is an array type, then the
            --  subtype indication must either denote a constrained
            --  subtype or include an explicit index constraint.
            if not Sem.Sem_Is_Constrained (Arg) then
               Error_Msg_Sem ("allocator of unconstrained " &
                              Disp_Node (Arg) & " is not allowed", Expr);
            end if;
            --  LRM93 7.3.6
            --  A subtype indication that is part of an allocator must
            --  not include a resolution function.
            if Is_Anonymous_Type_Definition (Arg)
              and then Get_Resolution_Function (Arg) /= Null_Iir
            then
               Error_Msg_Sem ("subtype indication must not include"
                              & " a resolution function", Expr);
            end if;
            Arg_Type := Arg;
         end if;
         Set_Expression (Expr, Arg);
      else
         if Get_Kind (Expr) = Iir_Kind_Allocator_By_Expression then
            Arg_Type := Get_Type (Arg);
         else
            Arg_Type := Arg;
         end if;
      end if;

      if A_Type = Null_Iir then
         --  Pass 1.
         --  LRM 7.3.6 Allocators
         --  The type of the access value returned by an allocator must be
         --  determinable solely from the context, but using the fact that the
         --  value returned is of an access type having the named designated
         --  type.
         declare
            Index : Visible_Type_Index_Type;
            Vtype : Iir;
            Btype : Iir;
            Dtype : Iir;
            List : Iir;
         begin
            List := Get_Overload_List;
            Index := Get_First_Visible_Type;
            while Index /= No_Visible_Type_Index loop
               Vtype := Get_Visible_Type_Decl (Index);
               Btype := Get_Base_Type (Get_Type (Vtype));
               if Get_Kind (Btype) = Iir_Kind_Access_Type_Definition then
                  Dtype := Get_Base_Type (Get_Designated_Type (Btype));
                  if Dtype = Get_Base_Type (Arg_Type) then
                     Add_Element (List, Dtype);
                  end if;
               end if;
               Index := Get_Next_Visible_Type (Index);
            end loop;
            List := Simplify_Overload_List (List);
            Set_Type (Expr, List);
         end;
      else
         if Get_Kind (A_Type) /= Iir_Kind_Access_Type_Definition then
            Error_Msg_Sem ("expected type is not an access type", Expr);
            return Null_Iir;
         end if;
         if not Are_Types_Compatible (Arg_Type, Get_Designated_Type (A_Type))
         then
            Not_Match (Expr, A_Type);
            return Null_Iir;
         end if;
         Free_Old_Iir (Get_Type (Expr));
         Set_Type (Expr, A_Type);
      end if;
      return Expr;
   end Sem_Allocator;

   procedure Check_Read_Aggregate (Aggr : Iir)
   is
   begin
      --  FIXME: todo.
      null;
   end Check_Read_Aggregate;

   --  Check EXPR can be read.
   procedure Check_Read (Expr : Iir)
   is
      Obj : Iir;
   begin
      if Expr = Null_Iir then
         return;
      end if;

      Obj := Expr;
      loop
         case Get_Kind (Obj) is
            when Iir_Kind_Signal_Declaration
              | Iir_Kind_Constant_Declaration
              | Iir_Kind_Constant_Interface_Declaration
              | Iir_Kind_Variable_Declaration
              | Iir_Kind_Attribute_Value
              | Iir_Kind_Iterator_Declaration
              | Iir_Kind_Guard_Signal_Declaration =>
               return;
            when Iir_Kind_Alias_Declaration =>
               Obj := Get_Name (Obj);
            when Iir_Kind_Signal_Interface_Declaration
              | Iir_Kind_Variable_Interface_Declaration =>
               case Get_Mode (Obj) is
                  when Iir_In_Mode
                    | Iir_Inout_Mode
                    | Iir_Buffer_Mode =>
                     return;
                  when Iir_Out_Mode
                    | Iir_Linkage_Mode =>
                     Error_Msg_Sem (Disp_Node (Obj) & " cannot be read", Expr);
                  when Iir_Unknown_Mode =>
                     raise Internal_Error;
               end case;
               return;
            when Iir_Kind_Enumeration_Literal
              | Iir_Kind_Physical_Int_Literal
              | Iir_Kind_Physical_Fp_Literal
              | Iir_Kind_String_Literal
              | Iir_Kind_Bit_String_Literal
              | Iir_Kind_Integer_Literal
              | Iir_Kind_Floating_Point_Literal
              | Iir_Kind_Null_Literal
              | Iir_Kind_Physical_Unit
              | Iir_Kind_Simple_Aggregate =>
               return;
            when Iir_Kinds_Monadic_Operator
              | Iir_Kinds_Dyadic_Operator
              | Iir_Kind_Function_Call =>
               return;
            when Iir_Kind_Qualified_Expression =>
               return;
            when Iir_Kind_Type_Conversion
              | Iir_Kind_Allocator_By_Expression
              | Iir_Kind_Allocator_By_Subtype
              | Iir_Kind_Implicit_Dereference
              | Iir_Kind_Dereference =>
               return;
            when Iir_Kinds_Scalar_Type_Attribute
              | Iir_Kinds_Type_Attribute
              | Iir_Kinds_Array_Attribute
              | Iir_Kind_Image_Attribute
              | Iir_Kind_Value_Attribute
              | Iir_Kinds_Signal_Attribute
              | Iir_Kinds_Signal_Value_Attribute =>
               return;
            when Iir_Kind_Aggregate =>
               Check_Read_Aggregate (Obj);
               return;
            when Iir_Kind_Indexed_Name
              | Iir_Kind_Slice_Name
              | Iir_Kind_Selected_Element =>
               Obj := Get_Base_Name (Obj);
            when others =>
               Error_Kind ("check_read", Obj);
         end case;
      end loop;
   end Check_Read;

   procedure Check_Update (Expr : Iir) is
   begin
      null;
   end Check_Update;

   procedure Check_Constant_Restriction (Expr : Iir)
   is
      Lib : Iir;
      Cur_Lib : Iir;
   begin
      --  LRM93 2.6
      --  Within a package declaration that contains the declaration
      --  of a deferred constant, and within the body of that package,
      --  before the end of the corresponding full declaration, the
      --  use of a name that denotes the deferred constant is only
      --  allowed in the default expression for a local generic,
      --  local port or formal parameter.
      if Get_Deferred_Declaration_Flag (Expr) = False
        or else Get_Deferred_Declaration (Expr) /= Null_Iir
      then
         return;
      end if;

      Lib := Get_Parent (Expr);
      if Get_Kind (Lib) = Iir_Kind_Design_Unit then
         Lib := Get_Library_Unit (Lib);
         raise Internal_Error;
      end if;
      Cur_Lib := Get_Library_Unit (Sem.Get_Current_Design_Unit);
      if (Get_Kind (Cur_Lib) = Iir_Kind_Package_Declaration
          and then Lib = Cur_Lib)
        or else (Get_Kind (Cur_Lib) = Iir_Kind_Package_Body
                 and then Get_Package (Cur_Lib) = Lib)
      then
         Error_Msg_Sem ("invalid use of a deferred constant", Expr);
      end if;
   end Check_Constant_Restriction;

   -- Set semantic to EXPR.
   --  Replace simple_name with the referenced node,
   --  Set type to nodes,
   --  Resolve overloading

   -- If A_TYPE is not null, then EXPR must be of type A_TYPE.
   -- Return null in case of error.
   function Sem_Expression_Ov (Expr: Iir; A_Type1: Iir) return Iir
   is
      A_Type: Iir;
   begin
--     -- Avoid to run sem_expression_ov when a node was already semantized
--     -- except to resolve overload.
--     if Get_Type (Expr) /= Null_Iir then
--        --  EXPR was already semantized.
--        if A_Type1 = null or else not Is_Overload_List (Get_Type (Expr)) then
--           --  This call to sem_expression_ov do not add any informations.
--           Check_Restrictions (Expr, Restriction);
--           return Expr;
--        end if;
--        -- This is an overload list that will be reduced.
--     end if;

      -- A_TYPE must be a type definition and not a subtype.
      if A_Type1 /= Null_Iir then
         A_Type := Get_Base_Type (A_Type1);
         if A_Type /= A_Type1 then
            raise Internal_Error;
         end if;
      else
         A_Type := Null_Iir;
      end if;

      case Get_Kind (Expr) is
         when Iir_Kind_Character_Literal =>
            declare
               Interpretation: Name_Interpretation_Type;
               Decl: Iir;
               List: Iir;
            begin
               Interpretation := Get_Interpretation (Get_Identifier (Expr));

               -- Check the identifier was declared.
               if Interpretation = No_Name_Interpretation then
                  Undeclared (Expr);
                  return Null_Iir;
               end if;

               if Get_Next_Interpretation (Interpretation)
                 = No_Name_Interpretation
               then
                  Decl := Get_Declaration (Interpretation);
                  if A_Type /= null and then A_Type = Get_Type (Decl) then
                     --  Free overload list of expr (if any), and expr.
                     Replace_Type (Expr, null);
                     Iirs_Utils.Free_Name (Expr);
                     return Decl;
                  end if;
               end if;

               -- Character literal can only be an enumeration literal.
               if A_Type /= Null_Iir then
                  while Interpretation /= No_Name_Interpretation loop
                     Decl := Get_Declaration (Interpretation);
                     if Get_Type (Decl) = A_Type then
                        Replace_Type (Expr, null);
                        Iirs_Utils.Free_Name (Expr);
                        return Decl;
                     end if;
                     Interpretation :=
                       Get_Next_Interpretation (Interpretation);
                  end loop;
                  Not_Match (Expr, A_Type);
                  return Null_Iir;
               end if;

               -- Store overloaded interpretation.
               List := Get_Overload_List;
               while Interpretation /= No_Name_Interpretation loop
                  Decl := Get_Declaration (Interpretation);
                  Append_Element (List, Get_Type (Decl));
                  Interpretation := Get_Next_Interpretation (Interpretation);
               end loop;
               Set_Type (Expr, List);
               return Expr;
            end;

         when Iir_Kind_Selected_Name
           | Iir_Kind_Simple_Name
           | Iir_Kind_Parenthesis_Name
           | Iir_Kind_Selected_By_All_Name
           | Iir_Kind_Attribute =>
            declare
               E : Iir;
            begin
               E := Get_Named_Entity (Expr);
               if E = Null_Iir then
                  Sem_Names (Expr);
                  E := Get_Named_Entity (Expr);
                  if E = Null_Iir then
                     raise Internal_Error;
                  end if;
               end if;
               if E = Error_Mark then
                  return Null_Iir;
               end if;
               if Get_Kind (E) = Iir_Kind_Constant_Declaration
                 and then not Deferred_Constant_Allowed
               then
                  Check_Constant_Restriction (E);
               end if;
               E := Name_To_Expression (Expr, A_Type);
               return E;
            end;

--          when Iir_Kind_Selected_Element
--            | Iir_Kind_Indexed_Name =>
--             return Expr;

--          when Iir_Kind_Indexed_Name =>
--             declare
--                -- The prefix of the index.
--                Prefix, Prefix_Type: Iir;

--                El_Type: Iir;
--                Type_Of_Expr: Iir;
--                Index: Iir;
--                Index_Subtype: Iir;
--                Expr_Staticness : Iir_Staticness;
--                Error : Boolean;
--             begin
--                Prefix := Sem_Expression_Ov (Get_Prefix (Expr), null);
--                if Prefix = Null_Iir then
--                   return Null_Iir;
--                end if;

--                Prefix := Insert_Implicit_Dereference (Prefix, Expr);
--                Set_Prefix (Expr, Prefix);
--                Prefix_Type := Get_Type (Prefix);
--                Set_Base_Name (Expr, Get_Base_Name (Prefix));

--             -- LRM93 6.4: the prefix of an indexed name must be appropriate
--                -- for an array type.
--                case Get_Kind (Prefix_Type) is
--                   when Iir_Kind_Array_Subtype_Definition
--                     | Iir_Kind_Array_Type_Definition =>
--                      null;
--                      -- FIXME: functions
--                   when others =>
--                      Error_Msg_Sem ("indexed name is not an array", Expr);
--                      return Null_Iir;
--                end case;

--                El_Type := Get_Element_Subtype (Prefix_Type);
--                Type_Of_Expr := Get_Base_Type (El_Type);

--                if A_Type /= Null_Iir and then Type_Of_Expr /= A_Type then
--                   Not_Match (Expr, A_Type);
--                   return null;
--                end if;

--                if Get_Nbr_Elements (Get_Index_List (Expr))
--                  /= Get_Nbr_Elements (Get_Index_List (Prefix_Type))
--                then
--                   Error_Msg_Sem
--                     ("number of indexes mismatch dimension of array", Expr);
--                   return null;
--                end if;

--                Expr_Staticness := Locally;

--                Error := False;
--              -- LRM93 6.4: there must be one such expression for each index
--                -- position of the array and each expression must be of the
--                -- type of the corresponding index.
--                -- Loop on the indexes.
--                for I in Natural loop
--                   Index_Subtype :=
--                     Get_Nth_Element (Get_Index_List (Prefix_Type), I);
--                   exit when Index_Subtype = null;
--                   Index := Get_Nth_Element (Get_Index_List (Expr), I);
--                   -- The index_subtype can be an unconstrained index type.
--                   Index := Check_Is_Expression (Index);
--                   if Index /= Null_Iir then
--                      Index := Sem_Expression
--                        (Index, Get_Base_Type (Index_Subtype));
--                   end if;
--                   if Index /= Null_Iir then
--                      if Get_Expr_Staticness (Index) = Locally
--                       and then Get_Type_Staticness (Index_Subtype) = Locally
--                      then
--                         Index := Eval_Expr_Check (Index, Index_Subtype);
--                      end if;
--                      Replace_Nth_Element (Get_Index_List (Expr), I, Index);
--                      Expr_Staticness := Min (Expr_Staticness,
--                                              Get_Expr_Staticness (Index));
--                   else
--                      Error := True;
--                   end if;
--                end loop;

--                Set_Type (Expr, Get_Element_Subtype (Prefix_Type));

--                --  An indexed name cannot be locally static.
--                Set_Expr_Staticness
--                  (Expr, Min (Globally, Min (Expr_Staticness,
--                                             Get_Expr_Staticness (Prefix))));

--                -- LRM93 6.1:
--                -- a name is said to be a static name iff:
--                -- The name is an indexed name whose prefix is a static name
--                -- and every expression that appears as part of the name is a
--                -- static expression.
--                --
--                -- a name is said to be a locally static name iif:
--                -- The name is an indexed name whose prefix is a locally
--                -- static name and every expression that appears as part
--                -- of the name is a locally static expression.
--                Set_Name_Staticness (Expr, Min (Expr_Staticness,
--                                              Get_Name_Staticness (Prefix)));
--                if Error then
--                   return Null_Iir;
--                else
--                   return Expr;
--                end if;
--             end;

--          when Iir_Kind_Slice_Name =>
--             return Sem_Slice_Name (Expr, A_Type);

         when Iir_Kinds_Monadic_Operator =>
            return Sem_Operator (Expr, A_Type, 1);

         when Iir_Kinds_Dyadic_Operator =>
            return Sem_Operator (Expr, A_Type, 2);

--          when Iir_Kind_Function_Call =>
--             if Get_Type (Expr) /= Null_Iir
--               and then not Is_Overload_List (Get_Type (Expr))
--             then
--                return Expr;
--             else
--                return Sem_Subprogram_Call (Expr, A_Type);
--             end if;

--          when Iir_Kind_Selected_By_All_Name =>
--             declare
--                Prefix : Iir;
--                Prefix_Type : Iir;
--             begin
--                Prefix := Get_Prefix (Expr);
--                -- FIXME: overload
--                Prefix := Sem_Expression (Prefix, Null_Iir, Value_Restr);
--                if Prefix = Null_Iir then
--                   return Null_Iir;
--                end if;
--                Check_Read (Prefix);
--                Set_Prefix (Expr, Prefix);
--                Prefix_Type := Get_Base_Type (Get_Type (Prefix));
--                if Get_Kind (Prefix_Type) /= Iir_Kind_Access_Type_Definition
--                then
--                   Error_Msg_Sem ("prefix for `.all' must be of type access",
--                                  Expr);
--                   return Null_Iir;
--                end if;
--                Set_Type (Expr, Get_Designated_Type (Prefix_Type));
--                Set_Base_Name (Expr, Expr);
--                Set_Name_Staticness (Expr, None);
--                return Expr;
--             end;

         when Iir_Kind_Enumeration_Literal
           | Iir_Kinds_Object_Declaration =>
            -- All these case have already a type.
            if Get_Type (Expr) = Error_Iir then
               return Error_Iir;
            end if;
            if A_Type /= Null_Iir
              and then not Are_Basetypes_Compatible
              (A_Type, Get_Base_Type (Get_Type (Expr)))
            then
               Not_Match (Expr, A_Type);
               return Error_Iir;
            end if;
            return Expr;

         when Iir_Kind_Integer_Literal =>
            Set_Expr_Staticness (Expr, Locally);
            if A_Type = null then
               Set_Type (Expr, Universal_Integer_Subtype_Definition);
               return Expr;
            elsif Get_Kind (A_Type) = Iir_Kind_Integer_Type_Definition then
               Set_Type (Expr, A_Type);
               return Expr;
            else
               Not_Match (Expr, A_Type);
               return Null_Iir;
            end if;

         when Iir_Kind_Floating_Point_Literal =>
            Set_Expr_Staticness (Expr, Locally);
            if A_Type = Null_Iir then
               Set_Type (Expr, Universal_Real_Subtype_Definition);
               return Expr;
            elsif Get_Kind (A_Type) = Iir_Kind_Floating_Type_Definition then
               Set_Type (Expr, A_Type);
               return Expr;
            else
               Not_Match (Expr, A_Type);
               return null;
            end if;

         when Iir_Kind_Physical_Int_Literal
           | Iir_Kind_Physical_Fp_Literal
           | Iir_Kind_Physical_Unit =>
            declare
               Res: Iir;
            begin
               Res := Sem_Physical_Literal (Expr);
               if Res = Null_Iir then
                  return Null_Iir;
               end if;
               if A_Type /= Null_Iir and then Get_Type (Res) /= A_Type then
                  Not_Match (Res, A_Type);
                  return Null_Iir;
               end if;
               return Res;
            end;

         when Iir_Kind_String_Literal
           | Iir_Kind_Bit_String_Literal =>
            if A_Type /= Null_Iir then
               if not Check_Type_For_String_Literal (A_Type) then
                  Not_Match (Expr, A_Type);
                  return Null_Iir;
               end if;
               -- It is enough ?
               -- FIXME: check against LRM.
               Replace_Type (Expr, A_Type);
               Sem_String_Literal (Expr);
               return Expr;
            end if;

            -- Look on every visible declaration of unidimensional array.
            declare
               Vt: Visible_Type_Index_Type;
               Vt_Type : Iir;
               Decl: Iir;
               List: Iir_List;
            begin
               Vt := Get_First_Visible_Type;
               List := Get_Overload_List;
               while Vt /= No_Visible_Type_Index loop
                  Vt_Type := Get_Type (Get_Visible_Type_Decl (Vt));
                  Decl := Get_Base_Type (Vt_Type);
                  if Get_Kind (Decl) = Iir_Kind_Array_Type_Definition
                    and then Get_Nbr_Elements (Get_Index_List (Decl)) = 1
                    and then Get_Kind
                    (Get_Base_Type (Get_Element_Subtype (Decl)))
                    = Iir_Kind_Enumeration_Type_Definition
                  then
                     -- FIXME: check this is a character type.
                     Append_Element (List, Decl);
                  end if;
                  Vt := Get_Next_Visible_Type (Vt);
               end loop;
               case  Get_Nbr_Elements (List) is
                  when 0 =>
                     Free_Iir (List);
                     Error_Msg_Sem ("no character type for string literal",
                                    Expr);
                     return null;
                  when 1 =>
                     Set_Type (Expr, Get_First_Element (List));
                     Free_Iir (List);
                     Sem_String_Literal (Expr);
                     return Expr;
                  when others =>
                     Set_Type (Expr, List);
                     return Expr;
               end case;
            end;

         when Iir_Kind_Null_Literal =>
            Set_Expr_Staticness (Expr, Locally);
            if A_Type = Null_Iir then
               declare
                  Vt: Visible_Type_Index_Type;
                  Vt_Type : Iir;
                  Decl: Iir;
                  List: Iir_List;
               begin
                  Vt := Get_First_Visible_Type;
                  List := Get_Overload_List;
                  while Vt /= No_Visible_Type_Index loop
                     Vt_Type := Get_Type (Get_Visible_Type_Decl (Vt));
                     Decl := Get_Base_Type (Vt_Type);
                     if Get_Kind (Decl) = Iir_Kind_Access_Type_Definition then
                        Append_Element (List, Decl);
                     end if;
                     Vt := Get_Next_Visible_Type (Vt);
                  end loop;
                  List := Simplify_Overload_List (List);
                  if List = Null_Iir then
                     Error_Msg_Sem ("no visible access type for null literal",
                                    Expr);
                     return Null_Iir;
                  else
                     Set_Type (Expr, List);
                     return Expr;
                  end if;
               end;
            elsif Get_Kind (A_Type) /= Iir_Kind_Access_Type_Definition then
               Error_Msg_Sem ("null literal can only be access type", Expr);
               return Null_Iir;
            else
               Set_Type (Expr, A_Type);
               return Expr;
            end if;

         when Iir_Kind_Qualified_Expression =>
            declare
               N_Type: Iir;
               Res: Iir;
            begin
               N_Type := Sem.Sem_Subtype_Indication (Get_Type_Mark (Expr));
               if N_Type = Null_Iir then
                  return Null_Iir;
               end if;
               Set_Type_Mark (Expr, N_Type);
               Set_Type (Expr, N_Type);
               if A_Type /= Null_Iir
                 and then not Are_Types_Compatible (A_Type, N_Type)
               then
                  Not_Match (Expr, A_Type);
                  return Null_Iir;
               end if;
               Res := Sem_Expression (Get_Expression (Expr), N_Type);
               if Res = Null_Iir then
                  return Null_Iir;
               end if;
               Check_Read (Res);
               Set_Expression (Expr, Res);
               Set_Expr_Staticness (Expr, Min (Get_Expr_Staticness (Res),
                                               Get_Type_Staticness (N_Type)));
               return Expr;
            end;

         when Iir_Kind_Allocator_By_Expression
           | Iir_Kind_Allocator_By_Subtype =>
            return Sem_Allocator (Expr, A_Type);

         when Iir_Kind_Aggregate =>
            if A_Type = Null_Iir then
               declare
                  Vt: Visible_Type_Index_Type;
                  Vt_Type : Iir;
                  Decl: Iir;
                  List: Iir_List;
                  Res : Iir;
               begin
                  Vt := Get_First_Visible_Type;
                  List := Get_Overload_List;
                  while Vt /= No_Visible_Type_Index loop
                     Vt_Type := Get_Type (Get_Visible_Type_Decl (Vt));
                     Decl := Get_Base_Type (Vt_Type);
                     case Get_Kind (Decl) is
                        when Iir_Kind_Array_Type_Definition
                          | Iir_Kind_Record_Type_Definition =>
                           Append_Element (List, Decl);
                        when others =>
                           null;
                     end case;
                     Vt := Get_Next_Visible_Type (Vt);
                  end loop;
                  case Get_Nbr_Elements (List) is
                     when 0 =>
                        Free_Iir (List);
                        Error_Msg_Sem
                          ("no visible composite type for aggregate", Expr);
                        return null;
                     when 1 =>
                        Res := Sem_Aggregate (Expr, Get_First_Element (List));
                        Free_Iir (List);
                        return Res;
                     when others =>
                        Set_Type (Expr, List);
                        return Expr;
                  end case;
               end;
            else
               return Sem_Aggregate (Expr, A_Type);
            end if;

--          when Iir_Kind_Attribute =>
--             declare
--                Res : Iir;
--             begin
--                Res := Sem_Attribute (Expr);
--                if Res = Null_Iir then
--                   return Null_Iir;
--                end if;
--                if A_Type /= Null_Iir
--                  and then Compatibility (A_Type, Get_Type (Res)) = None
--                then
--                   Not_Match (Res, A_Type);
--                   return Null_Iir;
--                end if;
--                return Res;
--             end;

--          when Iir_Kind_Type_Conversion =>
--             return Expr;
--          when Iir_Kind_Attribute_Value =>
--             return Expr;

         when Iir_Kinds_Procedure_Specification =>
            Error_Msg_Sem
              (Disp_Node (Expr) & " cannot be used as an expression", Expr);
            return Null_Iir;

         when others =>
            Error_Kind ("sem_expression_ov", Expr);
            return null;
      end case;
   end Sem_Expression_Ov;

   -- If A_TYPE is not null, then EXPR must be of type A_TYPE.
   -- Return null in case of error.
   function Sem_Expression (Expr: Iir; A_Type: Iir) return Iir
   is
      A_Type1: Iir;
      Res: Iir;
      Expr_Type : Iir;
   begin
      if Check_Is_Expression (Expr) = Null_Iir then
         return Null_Iir;
      end if;

      -- Can't try to run sem_expression_ov when a node was already semantized
      Expr_Type := Get_Type (Expr);
      if Expr_Type /= Null_Iir and then not Is_Overload_List (Expr_Type) then
         return Expr;
      end if;

      -- A_TYPE must be a type definition and not a subtype.
      if A_Type /= null then
         A_Type1 := Get_Base_Type (A_Type);
      else
         A_Type1 := null;
      end if;

      case Get_Kind (Expr) is
         when Iir_Kind_Aggregate =>
            Res := Sem_Aggregate (Expr, A_Type);
         when Iir_Kind_String_Literal
           | Iir_Kind_Bit_String_Literal =>
            if A_Type = Null_Iir then
               Res := Sem_Expression_Ov (Expr, Null_Iir);
            else
               if not Check_Type_For_String_Literal (A_Type) then
                  Not_Match (Expr, A_Type);
                  return Null_Iir;
               end if;
               Set_Type (Expr, A_Type);
               Sem_String_Literal (Expr);
               return Expr;
            end if;
         when others =>
            Res := Sem_Expression_Ov (Expr, A_Type1);
      end case;

      if Res /= null and then Is_Overloaded (Res) then
         Error_Msg_Sem ("cannot resolve overload for " & Disp_Node (Expr),
                        Expr);
         return null;
      end if;
      return Res;
   end Sem_Expression;

end Sem_Expr;
